theory ArcExt
imports Graph SubRel
begin

(* Cette théorie est découpée en trois parties :
   - dans la première, on définit l'extension par ajout d'un arc et on donne les propriétés de base
     liées à cette notion,
   - dans la seconde, on montre comment les ensembles de chemins/sous-chemins évoluent après
     l'extension d'un graphe par ajout d'un arc lorsque les subsomptions ne sont pas prises en
     compte. Cette partie est nécessaire pour montrer que la partie rouge d'un RB graphe est un
     arbre,
   - finalement, on "étend" les théorèmes de la seconde partie au cas avec subsomptions. Les 
     définitions utilisées dans cette troisième partie découlent de la définition d'une relation
     de subsomption. Pour comprendre cette troisième partie, regarder d'abord thy...
 *)

(**************************************************************************************************)
(***************************  EXTENSION PAR AJOUT D'ARCS  *****************************************)
(**************************************************************************************************)
 
(** PARTIE 1 **)

(* On dit qu'un graphe g' est l'extension d'un graphe g par l'ajout d'un arc a si :
   - la source de a est un sommet de g,
   - la cible  de a n'est pas un sommet de g,
   - g' est obtenu de g en lui ajoutant a.

   Le concept d'extension sert surtout englober les différentes contraintes qui le définissent et à 
   alléger les écritures dans la suite.
*)
definition extends :: 
  "('a,'b) rgraph_scheme \<Rightarrow> 
   'a arc                \<Rightarrow> 
   ('a,'b) rgraph_scheme \<Rightarrow> bool" 
where
  "extends g a g' = (src a \<in> Graph.vertices g \<and> tgt a \<notin> Graph.vertices g \<and> g' = (add_arc g a))"

lemma extendsI :
  assumes "src a \<in> Graph.vertices g"
  assumes "tgt a \<notin> Graph.vertices g"
  assumes "g' = add_arc g a"
  shows   "extends g a g'"
using assms unfolding extends_def by (intro conjI)

lemma extendsE1 :
  assumes "extends g a g'"
  shows   "src a \<in> Graph.vertices g"
using assms unfolding extends_def by (elim conjE)

lemma extendsE2 :
  assumes "extends g a g'"
  shows   "tgt a \<notin> Graph.vertices g"
using assms unfolding extends_def by (elim conjE)

lemma extendsE3 :
  assumes "extends g a g'"
  shows   "g' = add_arc g a"
using assms unfolding extends_def by (elim conjE)

lemma extends_root :
  assumes "extends g a g'"
  shows   "root g' = root g"
using assms extendsE3[of g a g'] by (simp add : add_arc_root)

lemma extends_imp_diff_ends :
  assumes "extends g a g'"
  shows   "src a \<noteq> tgt a"
using assms unfolding extends_def by auto

lemma src_in_extends :
  assumes "extends g a g'"
  shows   "src a \<in> Graph.vertices g'"
proof -
  have "g' = add_arc g a" using assms(1) by (rule extendsE3)

  thus ?thesis by (simp add : add_arc_vertices)
qed

lemma tgt_in_extends :
  assumes "extends g a g'"
  shows   "tgt a \<in> Graph.vertices g'"
proof -
  have "g' = add_arc g a" using assms(1) by (rule extendsE3)

  thus ?thesis by (simp add : add_arc_vertices)
qed

lemma extends_vertices :
  assumes "extends g a g'"
  shows   "Graph.vertices g' = Graph.vertices g \<union> {tgt a}"
proof -
  have "g' = add_arc g a"    using assms(1) by (rule extendsE3)

  moreover
  have "src a \<in> Graph.vertices g" using assms(1) by (rule extendsE1)

  ultimately
  show ?thesis by (auto simp add : add_arc_vertices)
qed

lemma extends_tgt_out_arcs :
  assumes "extends g a g'"
  shows   "out_arcs g' (tgt a) = {}"
proof -
  have "src a \<noteq> tgt a" using assms by (rule extends_imp_diff_ends)

  moreover
  have "tgt a \<notin> Graph.vertices g" using assms by (rule extendsE2)

  moreover
  have "g' = add_arc g a" using assms by (rule extendsE3)

  ultimately
  show ?thesis by (simp add : add_arc_tgt_out_arcs)
qed

lemma extends_src_out_arcs :
  assumes "extends g a g'"
  shows   "out_arcs g' (src a) = out_arcs g (src a) \<union> {a}"
using assms extendsE3 add_arc_src_out_arcs by fast

lemma extends_not_end_out_arcs :
  assumes "extends g a g'"
  assumes "v \<noteq> src a \<and> v \<noteq> tgt a"
  shows   "out_arcs g' v = out_arcs g v"
using assms extendsE3[of g a g'] by (auto simp add : out_arcs_def add_arc_arcs)

lemma extends_arcs :
  assumes "extends g a g'"
  shows   "arcs g' = arcs g \<union> {a}"
using assms by (auto simp add : extends_def add_arc_arcs)

lemma (in sub_rel_of) sub_rel_of_extends :
  assumes "extends g a g'"

  shows   "sub_rel_of g' subs"
proof -
  have "g' = add_arc g a" using assms(1) by (rule extendsE3)

  thus ?thesis by (simp add:  sub_rel_of_add_arc)
qed

(*lemma (in wf_sub_rel) extends_wf_sub_rel :
  assumes "extends g a g'"
  shows   "wf_sub_rel subs"
using wf_sub_rel by (simp add : wf_sub_rel_def)*)

lemma (in wf_sub_rel_of) wf_sub_rel_of_extends :
  assumes "extends g a g'"

  shows   "wf_sub_rel_of g' subs"
using assms extendsE3 wf_sub_rel_of_add_arc by fast


(* Si g est bien formé, son extension l'est également *)
lemma (in wf_rgraph) extends_wf_rgraph :
  assumes "extends g a g'"
  shows   "wf_rgraph g'"
using wf_rgraph assms
by (auto simp add : extends_def wf_rgraph_def add_arc_root add_arc_arcs vertices_conv)


(**************************************************************************************************)
(*** Evolution des ensembles de chemins après extends en l'absence de subsomptions ****************)
(**************************************************************************************************)

(** PARTIE 2 **)

(**************************** sous-chemins ********************************************************)

(* Les théorèmes de cette partie permettront de montrer que la partie rouge d'un RB-graphe est un 
   arbre (lorsqu'on met de côté les subsomptions). Cela nous permettra par la suite de connaître,
   pour tout sommet de la partie rouge, l'unique chemin reliant la racine rouge à ce sommet : on 
   pourra donc calculer, à partir d'une configuration initiale, la configuration atteinte en ce 
   sommet (c'est l'exécution symbolique le long de ce chemin depuis la configuration initiale)  *)

(* si g' étend g, alors un sous-chemin de g est un sous-chemin de g' *)
lemma sp_in_extends :
  assumes "extends g a g'"
  assumes "Graph.subpath g  l1 p l2"
  shows   "Graph.subpath g' l1 p l2"
proof -
  have "g' = add_arc g a" using assms(1) by (rule extendsE3)

  thus ?thesis using assms(2) by (simp add : Graph.sp_in_add_arc)
qed

lemma sps_in_extends :
  assumes "extends g a g'"
  shows   "Graph.subpaths g \<subseteq> Graph.subpaths g'"
using assms sp_in_extends by fast

lemma sp_from_in_extends :
  assumes "extends g a g'"
  assumes "Graph.subpath_from g  l as"
  shows   "Graph.subpath_from g' l as"
using assms extendsE3[of g a g'] by (simp add : Graph.sp_from_in_add_arc)

lemma sps_from_in_extends :
  assumes "extends g a g'"
  shows   "Graph.subpaths_from g l \<subseteq> Graph.subpaths_from g' l"
using assms sp_from_in_extends by fast

(* soit un p un sous-chemin de g menant à la source de a, alors p prolongé de a est un sous-chemin 
   de l'extension de g par a *)
lemma sp_end_src_imp_append_in_extends :
  assumes "extends g a g'"
  assumes "Graph.subpath g  l1 p (src a)"
  shows   "Graph.subpath g' l1 (p @ [a]) (tgt a)"
using assms extendsE3[of g a g'] by (simp add : sp_end_src_imp_append_in_add_arc)

(* soit p un sous-chemin de l'extension de g reliant deux sommets de g et ne passant pas 
   par a, alors p est un sous-chemin de g *)
lemma sp_in_extends_minus_arc_imp_p_in_old :
  assumes "extends g a g'"
  assumes "l1 \<in> Graph.vertices g"
  (*assumes "l2 \<in> Graph.vertices g"*)
  assumes "Graph.subpath g' l1 p l2"
  assumes "a \<notin> set p"

  shows   "Graph.subpath g l1 p l2"
using assms extendsE3[of g a g'] Graph.sp_in_add_arc_minus_arc_imp_sp_in_old[of l1 g a p] by simp

(* un sous-chemin de l'extension de g par a reliant deux sommets de g ne passe pas par a *)
lemma sp_from_old_verts_imp_new_arc_not_mem :
  assumes "extends g a g'"
  (*assumes "l1 \<in> Graph.vertices g"*)
  assumes "l2 \<in> Graph.vertices g"
  assumes "Graph.subpath g' l1 p l2"

  shows   "a \<notin> set p"
proof (rule ccontr, simp only : not_not)
  assume "a \<in> set p"

  have "l2 = tgt a"
  proof -
    have  "out_arcs g' (tgt a) = {}" using assms(1) by (rule extends_tgt_out_arcs)
  
    hence "\<exists> p'. p = p' @ [a] \<and> a \<notin> set p'" 
    using assms(3) `a \<in> set p` by (intro sp_through_de_decomp)
  
    then obtain p' where "p = p' @ [a]" "a \<notin> set p'" by blast

    thus ?thesis using assms(3) by (simp add : sp_append_one)
  qed

  moreover
  have "tgt a \<notin> Graph.vertices g" using assms(1) by (rule extendsE2)

  ultimately
  show "False" using assms(2) by simp
qed

(* un sous-chemin de l'extension de g par a reliant deux sommets de g est un sous-chemin de g *)
lemma sp_from_old_verts_imp_sp_in_old :
  assumes "extends g a g'"
  assumes "l1 \<in> Graph.vertices g"
  assumes "l2 \<in> Graph.vertices g"
  assumes "Graph.subpath g' l1 p l2"

  shows   "Graph.subpath g l1 p l2"
proof -
  have "a \<notin> set p" using assms(1,3,4) by (intro sp_from_old_verts_imp_new_arc_not_mem)
  thm sp_from_old_verts_imp_new_arc_not_mem

  moreover
  have "g' = add_arc g a" using assms(1) by (rule extendsE3)
  
  ultimately
  show ?thesis using assms(2,4) by (simp add : Graph.sp_in_add_arc_minus_arc_imp_sp_in_old)               
qed

(* soit p un sous-chemin de l'extension de g par a reliant un sommet de g à la cible de a,
   alors p passe par a  *)
lemma sp_extends_ends_in_tgt_imp_mem :
  assumes "extends g a g'"
  assumes "l \<noteq> tgt a"
  assumes "Graph.subpath g' l p (tgt a)"

  shows   "a \<in> set p"
proof -
  have "tgt a \<notin> Graph.vertices g" using assms(1) by (rule extendsE2)

  moreover
  have "g' = add_arc g a"    using assms(1) by (rule extendsE3)
  
  ultimately
  show ?thesis using assms(2) assms(3) by (simp add : sp_ends_in_tgt_imp_mem)
qed


(* un sous-chemin de l'extension de g par a issu de la cible de a est vide *)
lemma sp_from_tgt_in_extendsion_is_Nil :
  assumes "extends g a g'"
  assumes "Graph.subpath g' (tgt a) p l"
  shows   "p = []"
proof -
  have "out_arcs g' (tgt a) = {}" using assms(1) by (rule extends_tgt_out_arcs)

  thus ?thesis using assms(2) by (rule sp_from_de_empty)
qed


(* soit p un sous-chemin de l'extension de g par a terminant en la cible de a, alors :
   1) p est vide (si p est issu de la cible de a), ou
   2) p est de la forme p'@[a] où p' est un sous-chemin de g menant à la source de a et a 
      ne figure pas dans p'
*)
lemma sp_to_new_arc_tgt_imp :
  assumes "extends g a g'"
  assumes "Graph.subpath g' l p (tgt a)"
  shows   "p = [] \<or> (\<exists> p'. p = p' @ [a] \<and> a \<notin> set p' \<and> Graph.subpath g l p' (src a))"
proof (case_tac "l = tgt a")
  assume "l = tgt a"

  moreover
  have "tgt a \<in> Graph.vertices g'" using assms(1) by (rule tgt_in_extends)

  moreover
  have "out_arcs g' (tgt a) = {}" using assms(1) by (rule extends_tgt_out_arcs)

  ultimately
  show ?thesis using assms(2) by (intro disjI1, simp only : sp_from_de_empty)
  
next

  assume "l \<noteq> tgt a"

  hence  "a \<in> set p" using assms(1) assms(2) by (intro sp_extends_ends_in_tgt_imp_mem)

  moreover
  have "out_arcs g' (tgt a) = {}" using assms(1) by (rule extends_tgt_out_arcs)

  ultimately
  have "\<exists> p'. p = p' @ [a] \<and> a \<notin> set p'" using assms(2) by (elim sp_through_de_decomp)

  then obtain p' where "p = p' @ [a]" and "a \<notin> set p'" by blast

  moreover
  hence "Graph.subpath g' l p' (src a)" using assms(2) by (simp add : sp_append_one)

  have "Graph.subpath g l p' (src a)"
  proof -
    have  "l \<in> Graph.vertices g'" using assms(2) by (rule Graph.fst_of_sp_is_vert)

    hence "l \<in> Graph.vertices g"  using assms(1) `l \<noteq> tgt a` by (simp add : extends_vertices)

    moreover
    have "src a \<in> Graph.vertices g" using assms(1) by (rule extendsE1)

    moreover
    have "g' = add_arc g a" using assms(1) by (rule extendsE3) 
    
    ultimately
    show ?thesis
    using  `Graph.subpath g' l p' (src a)` `a \<notin> set p'` 
    by (simp add : Graph.sp_in_add_arc_minus_arc_imp_sp_in_old)  
  qed

  ultimately
  show ?thesis
  by (intro disjI2 exI conjI)
qed


(* soit p un sous-chemin de l'extension de g par a, alors :
   1) p est vide (si il est issu de la cible de a), ou
   2) p est un sous-chemin de g menant à un sommet quelconque de g, ou
   3) p mène à la cible de a, dans ce cas p est de la forme p' @ [a] où p' est un sous-chemin de 
      g menant à la source de a et a ne figure pas dans p' *)
lemma sp_in_extends_imp :
  assumes "extends g a g'"
  assumes "Graph.subpath g' l1 p l2"

  shows "(p = [] \<or> Graph.subpath g l1 p l2 \<or> (\<exists> p'. p = p'@[a] \<and> a \<notin> set p' \<and> Graph.subpath g l1 p' (src a)))"
proof -

  have "out_arcs g' (tgt a) = {}" using assms(1) by (rule extends_tgt_out_arcs)
  
  have "(l1 \<in> Graph.vertices g \<or> l1 = tgt a) \<and> (l2 \<in> Graph.vertices g \<or> l2 = tgt a)" 
  proof -
    have "l1 \<in> Graph.vertices g'" using assms(2) by (rule Graph.fst_of_sp_is_vert)
    
    moreover
    have "l2 \<in> Graph.vertices g'" using assms(2) by (rule lst_of_sp_is_vert)

    ultimately
    show ?thesis using assms(1) by (auto simp add : extends_vertices)
  qed

  thus ?thesis
  proof (elim conjE disjE)

    assume "l1 \<in> Graph.vertices g"
    and    "l2 \<in> Graph.vertices g"

    hence "Graph.subpath g l1 p l2" using assms  by (elim sp_from_old_verts_imp_sp_in_old)

    thus ?thesis by auto
  next
    assume "l1 \<in> Graph.vertices g"
    and    "l2 = tgt a"

    hence "p = [] \<or> (\<exists> p'. p = p' @ [a] \<and> a \<notin> set p' \<and> Graph.subpath g l1 p' (src a))"
    using assms by (simp add : sp_to_new_arc_tgt_imp)

    thus ?thesis by auto
  next
    assume "l1 = tgt a"

    hence "p = []" using assms by (simp add : sp_from_tgt_in_extendsion_is_Nil)

    thus ?thesis by simp
  next
    assume "l1 = tgt a"

    hence "p = []" using assms by (simp add : sp_from_tgt_in_extendsion_is_Nil)

    thus ?thesis by simp
  qed
qed

lemma
  assumes "extends g a g'"
  shows   "Graph.subpaths_from g' v = Graph.subpaths_from g v \<union> {p @ [a] | p. Graph.subpath g v p (src a)}"
          (is " ?L = ?R")
proof (case_tac "v = tgt a")

  assume "v = tgt a"

  hence "Graph.subpaths_from g' v = {[]}" sorry

  

oops

(* On peut donc caractériser l'ensemble des sous-chemins issu d'un sommet l donné de l'extension de 
   g par a : c'est l'union des ensembles des sous-chemins de g issus de l et de l'ensemble des 
   sous-chemins de g issus de l, menant à la source de a et prolongés par a *)
lemma sp_from_of_extends_eq :
  assumes "extends g a g'"
  assumes "l \<in> Graph.vertices g"
  shows   "Graph.subpaths_from g' l = Graph.subpaths_from g l \<union> {p @ [a] | p. Graph.subpath g l p (src a)}"
          (is " ?L = ?R")
proof (subst set_eq_iff, intro allI iffI)
  fix p

  assume "p \<in> ?L"

  then obtain l' where "Graph.subpath g' l p l'" unfolding Graph.subpath_from_def by blast

  hence "p = [] \<or> Graph.subpath g l p l' \<or> (\<exists> p'. p = p'@[a] \<and> a \<notin> set p' \<and> Graph.subpath g l p' (src a))"
  using assms by (intro sp_in_extends_imp)

  thus "p \<in> ?R"
  proof (elim disjE)
    assume "p = []"

    hence  "Graph.subpath g l p l" using assms(2) by (simp add : Graph.Nil_sp)

    thus ?thesis by (auto simp add : Graph.subpath_from_def)
  next
    assume "Graph.subpath g l p l'" thus ?thesis by (auto simp add : Graph.subpath_from_def)
  next
    assume "\<exists>p'. p = p' @ [a] \<and> a \<notin> set p' \<and> Graph.subpath g l p' (src a)"

    then obtain p' where "p = p' @ [a]"
                   and   "a \<notin> set p'"
                   and   "Graph.subpath g l p' (src a)"
    by blast
    
    thus ?thesis by simp
  qed
next
  fix p

  assume "p \<in> ?R"
  thus   "p \<in> ?L"
  unfolding Un_iff
  proof (elim disjE)
    assume "p \<in> Graph.subpaths_from g l" 
    
    thus ?thesis using assms(1) sp_from_in_extends [of g a g' l] by auto
  next
    assume "p \<in> {p @ [a] |p. Graph.subpath g l p (src a)}"

    then obtain p' where "p = p' @ [a]" 
                   and   "Graph.subpath g l p' (src a)" 
    by blast

    hence "Graph.subpath g' l (p'@[a]) (tgt a)" 
    using assms(1) by (simp add : sp_end_src_imp_append_in_extends)

    thus ?thesis using `p = p' @ [a]` unfolding Graph.subpath_from_def by auto

  qed  
qed


(* l'ensemble des sous-chemins de l'extension de g par a est donc l'union de l'ensemble des sous-
   chemins de g et de l'ensemble des sous-chemins de g menant à la source de a et prolongés par a *)
lemma sp_of_extends_eq :
  assumes "extends g a g'"
  
  shows   "Graph.subpaths g' = Graph.subpaths g \<union> {p @ [a] | p. \<exists> l1. Graph.subpath g l1 p (src a)}"
          (is " ?L = ?R")
proof (subst set_eq_iff, intro allI iffI)
  fix p
  assume "p \<in> ?L"
 
  then obtain l1 l2 where "Graph.subpath g' l1 p l2" by blast

  hence "p = [] \<or> Graph.subpath g l1 p l2 \<or> (\<exists> p'. p = p'@[a] \<and> a \<notin> set p' \<and> Graph.subpath g l1 p' (src a))"
  using assms by (simp add : sp_in_extends_imp)

  thus "p \<in> ?R" (*by (auto simp add : Graph.subpaths_def Graph.subpath_Nil_conv)*)
  apply (elim disjE)
    apply (simp only : Graph.subpath_def Un_iff)
    apply (intro disjI1)
    apply (simp only : mem_Collect_eq)
    apply (rule_tac ?x="root g" in exI)
    apply (rule_tac ?x="root g" in exI)
    apply (simp add : Graph.Nil_sp_from_root vertices_conv)

    apply (force simp add : Graph.subpath_def)
    
    apply (simp only : Un_iff)
    apply (intro disjI2)
    apply fastforce
  done
next
  fix x
  assume "x \<in> ?R"

  thus   "x \<in> ?L"
  unfolding Un_iff
  proof (elim disjE)
    assume "x \<in> Graph.subpaths g" thus ?thesis using assms sps_in_extends by force
  next
    assume "x \<in> {p @ [a] |p. \<exists>l1. Graph.subpath g l1 p (src a)}"

    then obtain p' l1 where "x = p' @ [a]" 
                      and   "Graph.subpath g l1 p' (src a)" 
    by blast

    hence "Graph.subpath g' l1 (p'@[a]) (tgt a)" 
    using assms(1) by (auto simp add : sp_end_src_imp_append_in_extends)

    thus ?thesis using `x = p' @ [a]` unfolding Graph.subpath_def by auto
  qed
qed


(****************** chemins ***********************************************************************)

(* Puisque les chemins d'un graphe sont ses sous-chemins issus de sa racine, certaines des propriétés
   précédentes se "transmettent" aux chemins  *)

lemma path_in_extends :
  assumes "extends g a g'"
  assumes "Graph.path g  p l"
  shows   "Graph.path g' p l"
using assms unfolding Graph.path_def by (simp add : sp_in_extends extends_root)

lemma p_end_src_imp_append_in_extends :
  assumes "extends g a g'"
  assumes "Graph.path g  p (src a)"
  shows   "Graph.path g' (p @ [a]) (tgt a)"
using assms extendsE3[of g a g'] by (simp add : p_end_src_imp_append_in_add_arc)

lemma p_in_extends_minus_arc_imp_p_in_old :
  assumes "extends g a g'"
  (*assumes "l \<in> Graph.vertices g"*)
  assumes "Graph.path g' p l"
  assumes "a \<notin> set p"

  shows   "Graph.path g p l"
using assms unfolding Graph.path_def 
by (simp add : root_is_vert sp_in_extends_minus_arc_imp_p_in_old extends_root)

lemma p_extends_ends_in_tgt_imp_mem :
  assumes "extends g a g'"
  assumes "Graph.path g' p (tgt a)"

  shows   "a \<in> set p"
proof -
  have "root g \<noteq> tgt a"
  proof -
    have "root g \<in> Graph.vertices g" by (rule root_is_vert)

    moreover
    have "tgt a \<notin> Graph.vertices g" using assms(1) by (rule extendsE2)

    ultimately
    show ?thesis by auto
  qed

  moreover
  have "root g' = root g"
  proof -
    have "g' = add_arc g a" using assms(1) by (rule extendsE3)
    thus ?thesis by (simp add : add_arc_root)
  qed

  ultimately
  show ?thesis using assms unfolding Graph.path_def by (simp add : sp_extends_ends_in_tgt_imp_mem)
qed

lemma paths_of_extends_eq :
  assumes "extends g a g'"
  
  shows   "Graph.paths g' = Graph.paths g \<union> {p @ [a] | p. Graph.path g p (src a)}"
          (is " ?L = ?R")
proof -
  
  have "Graph.paths g' = Graph.subpaths_from g (root g) \<union> {p @ [a] | p. Graph.subpath g (root g) p (src a)}"
  using assms  by (simp add : Graph.paths_conv sp_from_of_extends_eq root_is_vert extends_root)

  thus ?thesis by (simp add : Graph.paths_conv Graph.subpath_from_def Graph.path_def)
qed






(**************************************************************************************************)
(*** Evolution des ensembles de chemins après extends en présence de subsomptions *****************)
(**************************************************************************************************)

(** PARTIE 3 **)

(* On prouve les mêmes lemmes que dans la partie précédente, mais en utilisant cette fois une 
   définition des chemins/sous-chemins prenant en compte les subsomptions (cf. thy) *)


(**************************** sous-chemins ********************************************************)

(* si g' étend g, alors un sous-chemin de g est un sous-chemin de g' *)
lemma sp_in_extends_w_subs :
  assumes "extends g a g'"
  assumes "subpath g  l1 p l2 subs"
  shows   "subpath g' l1 p l2 subs"
proof -
  have "g' = add_arc g a" using assms(1) by (rule extendsE3)

  thus ?thesis using assms(2) by (simp add : sp_in_add_arc)
qed


lemma sps_in_extends_w_subs :
  assumes "extends g a g'"
  shows   "subpaths g subs \<subseteq> subpaths g' subs"
using assms sp_in_extends_w_subs by fastforce

lemma sp_from_in_extends_w_subs :
  assumes "extends g a g'"
  assumes "subpath_from g  v as subs"
  shows   "subpath_from g' v as subs"
using assms sp_in_extends_w_subs[of g a g' v as _ subs]
unfolding subpath_from_def by fast


lemma sps_from_in_extends_w_subs :
  assumes "extends g a g'"
  shows   "subpaths_from g l subs \<subseteq> subpaths_from g' l subs"
using assms extendsE3[of g a g'] by (simp add : sps_from_in_add_arc)




lemma (in sub_rel_of) sp_end_src_imp_append_in_extends_w_subs :
  assumes "extends g a g'"
  assumes "subpath g  l1 p (src a) subs"
  shows   "subpath g' l1 (p @ [a]) (tgt a) subs"
using assms extendsE3[of g a g'] by (simp add : sp_end_src_imp_append_in_add_arc)

lemma sp_in_extends_minus_arc_imp_p_in_old_w_subs :
  assumes "extends g a g'"
  assumes "l1 \<in> Graph.vertices g"
  (*assumes "l2 \<in> Graph.vertices g"*)
  assumes "subpath g' l1 p l2 subs"
  assumes "a \<notin> set p"

  shows   "subpath g l1 p l2 subs"
proof -

  (*have "l1 \<in> Graph.vertices g" sorry*)
  
  show ?thesis
  using assms extendsE3[of g a g'] sp_in_add_arc_minus_arc_imp_sp_in_old[of l1 g a p]
  by simp
qed



lemma (in sub_rel_of) extends_imp_tgt_not_in_subrel :
  assumes "extends g a g'"
  shows   "tgt a \<notin> subsumees subs \<and> tgt a \<notin> subsumers subs"
proof (rule ccontr, simp only : de_Morgan_conj not_not)
  assume "tgt a \<in> subsumees subs \<or> tgt a \<in> subsumers subs"

  hence  "tgt a \<in> Graph.vertices g" using sub_rel_of by auto

  thus "False" using assms(1) by (auto simp add : extends_def)
qed

lemma (in sub_rel_of) sp_from_old_verts_imp_new_arc_not_mem_w_subs :
  assumes "extends g a g'"
  assumes "l1 \<in> Graph.vertices g"
  assumes "l2 \<in> Graph.vertices g"
  assumes "subpath g' l1 p l2 subs"

  shows   "a \<notin> set p"
proof (rule ccontr, simp only : not_not)
  assume "a \<in> set p"

  have "l2 = tgt a"
  proof -

    have "sub_rel_of g' subs" using assms(1) by (rule sub_rel_of_extends)

    moreover
    have "tgt a \<notin> subsumees subs" using assms(1) by (simp add : extends_imp_tgt_not_in_subrel)

    moreover
    have  "out_arcs g' (tgt a) = {}" using assms(1) by (rule extends_tgt_out_arcs)
  
    ultimately
    have "\<exists> p'. p = p' @ [a] \<and> a \<notin> set p'" 
    using assms(4) `a \<in> set p` sub_rel_of  by (intro sub_rel_of.sp_through_de_decomp)

    thm sub_rel_of.sp_through_de_decomp[of g' subs a l1 p l2]
  
    then obtain p' where "p = p' @ [a]" "a \<notin> set p'" by blast

    hence "tgt a = l2 \<or> (tgt a, l2) \<in> subs\<^sup>+"
    using assms(4) `sub_rel_of g' subs` by (simp add :  sub_rel_of.sp_append_one)

    thus ?thesis
    proof (elim disjE)
      assume "tgt a = l2" thus ?thesis by simp
    next
      assume "(tgt a,l2) \<in> subs\<^sup>+"
      
      then obtain v where "(tgt a,v) \<in> subs" using tranclD[of "tgt a" l2 subs] by blast
      
      hence "tgt a \<in> subsumees subs" unfolding subsumees_conv by fastforce
      thus ?thesis using `tgt a \<notin> subsumees subs` by simp
    qed
  qed

  moreover
  have "tgt a \<notin> Graph.vertices g" using assms(1) by (rule extendsE2)

  ultimately
  show "False" using assms(3) by simp
qed


lemma (in sub_rel_of) sp_from_old_verts_imp_sp_in_old_w_subs :
  assumes "extends g a g'"
  assumes "l1 \<in> Graph.vertices g"
  assumes "l2 \<in> Graph.vertices g"
  assumes "subpath g' l1 p l2 subs"

  shows   "subpath g  l1 p l2 subs"
proof -
  have "a \<notin> set p" using assms by (rule sp_from_old_verts_imp_new_arc_not_mem_w_subs)

  moreover
  have "g' = add_arc g a" using assms(1) by (rule extendsE3)
  
  ultimately
  show ?thesis using assms by (simp add : sp_in_add_arc_minus_arc_imp_sp_in_old)                     
qed


lemma (in sub_rel_of) sp_extends_ends_in_tgt_imp_mem_w_subs :
  assumes "extends g a g'"
  assumes "l \<noteq> tgt a"
  assumes "subpath g' l p (tgt a) subs"

  shows   "a \<in> set p"
proof -
  have "tgt a \<notin> Graph.vertices g" using assms(1) by (rule extendsE2)

  moreover
  have "g' = add_arc g a"    using assms(1) by (rule extendsE3)
  
  ultimately
  show ?thesis using assms(2) assms(3) by (simp add : sp_ends_in_tgt_imp_mem)
qed      



lemma (in sub_rel_of) sp_from_tgt_in_extendsion_is_Nil_w_subs :
  assumes "extends g a g'"
  assumes "subpath g' (tgt a) p v subs"
  shows   "p = []"
proof -
  have "sub_rel_of g' subs" using assms(1) by (rule sub_rel_of_extends)

  moreover
  have "tgt a \<notin> subsumees subs" using assms(1) by (simp add : extends_imp_tgt_not_in_subrel)
  
  moreover
  have "out_arcs g' (tgt a) = {}" using assms(1) by (rule extends_tgt_out_arcs)

  ultimately
  show ?thesis using assms(2) by (rule sub_rel_of.sp_from_de_empty)
qed



lemma (in sub_rel_of) sp_to_new_arc_tgt_imp_w_subs :
  assumes "extends g a g'"
  assumes "subpath g' l p (tgt a) subs"
  shows   "p = [] \<or> (\<exists> p'. p = p' @ [a] \<and> a \<notin> set p' \<and> subpath g l p' (src a) subs)"
proof (case_tac "l = tgt a")
  assume "l = tgt a"

  moreover
  have "sub_rel_of g' subs" using assms(1) by (rule sub_rel_of_extends)

  moreover
  have "tgt a \<notin> subsumees subs" using assms(1) by (simp add : extends_imp_tgt_not_in_subrel)

  moreover
  have "out_arcs g' (tgt a) = {}" using assms(1) by (rule extends_tgt_out_arcs)

  ultimately
  show ?thesis using assms(2) by (intro disjI1, simp only : sub_rel_of.sp_from_de_empty)
  
next

  assume "l \<noteq> tgt a"

  hence  "a \<in> set p" using assms(1) assms(2) by (elim sp_extends_ends_in_tgt_imp_mem_w_subs)    

  moreover
  have "sub_rel_of g' subs" using assms(1) by (rule sub_rel_of_extends)

  moreover
  have "tgt a \<notin> subsumees subs" using assms(1) by (simp add : extends_imp_tgt_not_in_subrel)
  
  moreover
  have "out_arcs g' (tgt a) = {}" using assms(1) by (rule extends_tgt_out_arcs)

  ultimately
  have "\<exists> p'. p = p' @ [a] \<and> a \<notin> set p'" using assms(2) by (elim sub_rel_of.sp_through_de_decomp)

  then obtain p' where "p = p' @ [a]" and "a \<notin> set p'" by blast

  moreover
  hence "subpath g' l p' (src a) subs" 
  using assms(2) `sub_rel_of g' subs` by (simp add : sub_rel_of.sp_append_one)

  moreover
  have "subpath g l p' (src a) subs"
  proof -
    have  "l \<in> Graph.vertices g'" using assms(2) by (rule fst_of_sp_is_vert)

    hence "l \<in> Graph.vertices g"  using assms(1) `l \<noteq> tgt a` by (simp add : extends_vertices)

    moreover
    have "src a \<in> Graph.vertices g" using assms(1) by (rule extendsE1)

    moreover
    have "g' = add_arc g a" using assms(1) by (rule extendsE3) 
    
    ultimately
    show ?thesis
    using  `subpath g' l p' (src a) subs` `a \<notin> set p'` 
    by (simp add : sp_in_add_arc_minus_arc_imp_sp_in_old)  
  qed

  ultimately
  show ?thesis
  by (intro disjI2 exI conjI)
qed                                  



lemma (in sub_rel_of) sp_in_extends_imp_w_subs :
  assumes "extends g a g'"
  assumes "subpath g' l1 p l2 subs"

  shows   "p = [] \<or> subpath g l1 p l2 subs \<or> 
           (\<exists> p'. p = p'@[a] \<and> a \<notin> set p' \<and> subpath g l1 p' (src a) subs)"
proof -

  have "out_arcs g' (tgt a) = {}" using assms(1) by (rule extends_tgt_out_arcs)
  
  have "(l1 \<in> Graph.vertices g \<or> l1 = tgt a) \<and> (l2 \<in> Graph.vertices g \<or> l2 = tgt a)" 
  proof -
    have "l1 \<in> Graph.vertices g'" using assms(2) by (rule fst_of_sp_is_vert)
    
    moreover
    have "sub_rel_of g' subs" using assms(1) by (rule sub_rel_of_extends)
    
    hence "l2 \<in> Graph.vertices g'" using assms(2) by (rule sub_rel_of.lst_of_sp_is_vert)

    ultimately
    show ?thesis using assms(1) by (auto simp add : extends_vertices)
  qed

  thus ?thesis
  proof (elim conjE disjE)

    assume "l1 \<in> Graph.vertices g"
    and    "l2 \<in> Graph.vertices g"

    hence "subpath g l1 p l2 subs" using assms by (elim sp_from_old_verts_imp_sp_in_old_w_subs)

    thus ?thesis by auto
  next
    assume "l1 \<in> Graph.vertices g"
    and    "l2 = tgt a"

    hence "p = [] \<or> (\<exists> p'. p = p' @ [a] \<and> a \<notin> set p' \<and> subpath g l1 p' (src a) subs)"
    using assms by (simp add : sp_to_new_arc_tgt_imp_w_subs)

    thus ?thesis by auto
  next
    assume "l1 = tgt a"

    hence "p = []" using assms by (simp add : sp_from_tgt_in_extendsion_is_Nil_w_subs)

    thus ?thesis by simp
  next
    assume "l1 = tgt a"

    hence "p = []" using assms by (simp add : sp_from_tgt_in_extendsion_is_Nil_w_subs)

    thus ?thesis by simp
  qed
qed   


lemma (in sub_rel_of) p_in_extends_imp_w_subs :
  assumes "extends g a g'"
  assumes "path g' p l subs"

  shows "p = [] \<or> path g p l subs \<or> 
         (\<exists> p'. p = p'@[a] \<and> a \<notin> set p' \<and> path g p' (src a) subs)"
using assms sp_in_extends_imp_w_subs[of a g' "root g" p l] extends_root
by (force simp add : path_def)



lemma (in sub_rel_of) sp_of_extends_eq_w_subs :
  assumes "extends g a g'"
  
  shows   "subpaths g' subs = subpaths g subs \<union> {p @ [a] | p. \<exists> l1. subpath g l1 p (src a) subs}"
          (is " ?L = ?R")

proof (subst set_eq_iff, intro allI iffI)
  fix p
  assume "p \<in> ?L"
 
  then obtain l1 l2 where "subpath g' l1 p l2 subs" by blast

  hence "p = [] \<or> subpath g l1 p l2 subs \<or> (\<exists> p'. p = p'@[a] \<and> a \<notin> set p' \<and> subpath g l1 p' (src a) subs)"
  using assms by (simp add : sp_in_extends_imp_w_subs)

  thus "p \<in> ?R" (*by (auto simp add : Graph.subpaths_def Graph.subpath_Nil_conv)*)
  apply (elim disjE)
    apply (simp only : subpath_def Un_iff)
    apply (intro disjI1)
    apply (simp only : mem_Collect_eq)
    apply (rule_tac ?x="root g" in exI)
    apply (rule_tac ?x="root g" in exI)
    apply (simp add : Nil_sp_from_root vertices_conv)

    apply (auto simp del : split_paired_Ex)
  done
next
  fix x
  assume "x \<in> ?R"

  thus   "x \<in> ?L"
  unfolding Un_iff
  proof (elim disjE)
    assume "x \<in> subpaths g subs" thus ?thesis using assms sps_in_extends_w_subs by fastforce
  next
    assume "x \<in> {p @ [a] |p. \<exists>l1. subpath g l1 p (src a) subs}"

    then obtain p' l1 where "x = p' @ [a]" 
                      and   "subpath g l1 p' (src a) subs" 
    by blast

    hence "subpath g' l1 (p'@[a]) (tgt a) subs" 
    using assms(1) by (auto simp add : sp_end_src_imp_append_in_extends_w_subs)

    thus ?thesis 
    using `x = p' @ [a]` by (auto simp del : split_paired_Ex)
  qed
qed


thm sub_rel_of.sp_in_extends_imp_w_subs

lemma (in sub_rel_of) tmp :
  assumes "extends g a g'"
  assumes "subpath_from g' v as subs"
  shows   "as = [] \<or>
           subpath_from g  v as subs \<or> 
           (\<exists> as'. as = as'@[a] \<and> a \<notin> set as' \<and> subpath g v as' (src a) subs)"
using assms sp_in_extends_imp_w_subs unfolding subpath_from_def by blast


(*lemma (in sub_rel_of) sp_from_of_extends_eq_w_subs :
  assumes "extends g a g'"
  shows   "subpaths_from g' v subs = subpaths_from g v subs 
                                   \<union> {p @ [a] | p. subpath g v p (src a) subs}"
unfolding set_eq_iff
proof (intro allI iffI)

  case (goal1 as)

  hence  "as = [] \<or>
          subpath_from g  v as subs \<or> 
          (\<exists> as'. as = as'@[a] \<and> a \<notin> set as' \<and> subpath g v as' (src a) subs)"  
  using assms by (simp add : tmp)

  thus ?case
  proof (elim disjE)

    assume "as = []" 

    thus ?thesis using Nil_sp

thus ?case by (auto simp add : Nil_sp_from_root)

oops*)

lemma (in sub_rel_of) sp_from_of_extends_eq_w_subs :
  assumes "extends g a g'"
  assumes "l \<in> Graph.vertices g"
  shows   "subpaths_from g' l subs = subpaths_from g l subs 
                                          \<union> {p @ [a] | p. a \<notin> set p \<and> subpath g l p (src a) subs}"
          (is " ?L = ?R") 
proof (subst set_eq_iff, intro allI iffI)
  fix p

  assume "p \<in> ?L"

  then obtain l' where "subpath g' l p l' subs" 
  unfolding subpath_from_def by blast

  hence "p = [] \<or> subpath g l p l' subs \<or> (\<exists> p'. p = p'@[a] \<and> a \<notin> set p' \<and> subpath g l p' (src a) subs)"
  using assms by (intro sp_in_extends_imp_w_subs)

  thus "p \<in> ?R"
  proof (elim disjE)
    assume "p = []"

    hence  "subpath g l p l subs" using assms(2) by (simp add : Nil_sp)

    thus ?thesis by (auto simp only : subpath_from_def)
  next
    assume "subpath g l p l' subs" thus ?thesis by (auto simp only : subpath_from_def)
  next
    assume "\<exists>p'. p = p' @ [a] \<and> a \<notin> set p' \<and> subpath g l p' (src a) subs"

    then obtain p' where "p = p' @ [a]"
                   and   "a \<notin> set p'"
                   and   "subpath g l p' (src a) subs"
    by blast
    
    thus ?thesis by simp
  qed
next
  fix p

  assume "p \<in> ?R"
  thus   "p \<in> ?L"
  unfolding Un_iff
  proof (elim disjE)
    assume "p \<in> subpaths_from g l subs" 
    
    thus ?thesis using assms(1) sp_from_in_extends_w_subs [of g a g' l] by auto
  next
    assume "p \<in> {p @ [a] |p. a \<notin> set p \<and> subpath g l p (src a) subs}"
 
    then obtain p' where "p = p' @ [a]" 
                   and   "a \<notin> set p'"
                   and   "subpath g l p' (src a) subs" 
    by blast

    hence "subpath g' l (p'@[a]) (tgt a) subs" 
    using assms(1) by (simp add : sp_end_src_imp_append_in_extends_w_subs)

    thus ?thesis 
    using `p = p' @ [a]` unfolding subpath_from_def by (auto simp del : split_paired_Ex)

  qed  
qed





(****************** chemins ***********************************************************************)

lemma path_in_extends_w_subs :
  assumes "extends g a g'"
  assumes "path g  p l subs"
  shows   "path g' p l subs"
using assms unfolding path_def by (simp add : sp_in_extends_w_subs extends_root)

lemma (in sub_rel_of) p_end_src_imp_append_in_extends_w_subs :
  assumes "extends g a g'"
  assumes "path g  p (src a) subs"
  shows   "path g' (p @ [a]) (tgt a) subs"
proof -
  have "g' = add_arc g a" using assms(1) by (rule extendsE3)

  thus ?thesis
  using assms(2)
  unfolding path_def
  using sp_end_src_imp_append_in_add_arc[of "root g" p a] by (auto simp only : add_arc_root)
  
qed

lemma (in sub_rel_of) p_to_new_arc_tgt_imp_w_subs :
  assumes "extends g a g'"
  assumes "path g' p (tgt a) subs"
  shows   "p = [] \<or> (\<exists> p'. p = p' @ [a] \<and> a \<notin> set p' \<and> path g p' (src a) subs)"
using sp_to_new_arc_tgt_imp_w_subs[of a g' "root g" p] assms 
by (simp add : extends_root path_def)


lemma p_in_extends_minus_arc_imp_p_in_old_w_subs :
  assumes "extends g a g'"
  (*assumes "l \<in> Graph.vertices g"*)
  assumes "path g' p l subs"
  assumes "a \<notin> set p"

  shows   "path g p l subs"
using assms unfolding path_def 
by (simp add : root_is_vert sp_in_extends_minus_arc_imp_p_in_old_w_subs extends_root)

lemma (in sub_rel_of) p_extends_ends_in_tgt_imp_mem_w_subs :
  assumes "extends g a g'"
  assumes "path g' p (tgt a) subs"

  shows   "a \<in> set p"
proof -
  have "root g \<noteq> tgt a"
  proof -
    have "root g \<in> Graph.vertices g" by (rule root_is_vert)

    moreover
    have "tgt a \<notin> Graph.vertices g" using assms(1) by (rule extendsE2)

    ultimately
    show ?thesis by auto
  qed

  moreover
  have "root g' = root g"
  proof -
    have "g' = add_arc g a" using assms(1) by (rule extendsE3)
    thus ?thesis by (simp add : add_arc_root)
  qed

  ultimately
  show ?thesis 
  using assms unfolding path_def by (simp add : sp_extends_ends_in_tgt_imp_mem_w_subs)
qed

lemma (in sub_rel_of)
  assumes "extends g a g'"
  assumes "path g' as v subs"
  shows   "path g  as v subs \<or> (\<exists> as'. as = as' @ [a] \<and> a \<notin> set as' \<and> path g as' (src a) subs)"
using assms sp_from_of_extends_eq_w_subs[of a g' v]
proof - 

  thm sp_of_extends_eq_w_subs

oops

lemma (in sub_rel_of) paths_of_extends_eq_w_subs :
  assumes "extends g a g'"
  
  shows   "paths g' subs = paths g subs \<union> {p @ [a] | p. a \<notin> set p \<and> path g p (src a) subs}"
          (is " ?L = ?R")
proof -
  
  have "paths g' subs = subpaths_from g (root g) subs 
                             \<union> {p @ [a] | p. a \<notin> set p \<and> subpath g (root g) p (src a) subs}"
  using assms sp_from_of_extends_eq_w_subs[of a g' "root g"]
        root_is_vert[of g] 
        extends_root[of g a g']
  by (simp only : paths_conv)

  thus ?thesis by (simp add : paths_conv subpath_from_def path_def)
qed






lemma (in sub_rel_of) sp_from_not_tgt_imp :
  assumes "extends g a g'"
  assumes "v1 \<noteq> tgt a"
  assumes "subpath g' v1 as (tgt a) subs"
  shows   "\<exists> as'. as = as' @ [a] \<and> a \<notin> set as' \<and> subpath g v1 as' (src a) subs"
using sp_extends_ends_in_tgt_imp_mem_w_subs[OF assms]
      sp_to_new_arc_tgt_imp_w_subs[OF assms(1,3)]
by auto

end
