theory Bexp
imports Aexp Main
begin

(* Bexp.thy : analogue à Aexp.thy avec valeurs dans bool. 2 fichiers pour pouvoir utiliser les mêmes noms 
   pour les mêmes notions *)
type_synonym 'var_type bexp =  "'var_type state \<Rightarrow> bool"

definition vars :: "'var_type bexp \<Rightarrow> 'var_type set" where
          "vars exp = {v. \<exists> \<sigma> val. exp (\<sigma>(v := val)) \<noteq> exp \<sigma>}"

abbreviation fresh :: "'var_type \<Rightarrow> 'var_type bexp \<Rightarrow> bool" where
            "fresh v exp \<equiv> v \<notin> vars exp"

(* Une expression est sat. si il existe un etat la rendant vraie *)
definition sat :: "'var_type bexp \<Rightarrow> bool" where
          "sat exp = (\<exists> \<sigma>. exp \<sigma>)"

(* calcule la conjonction des expr. booléennes de E *)
abbreviation conjunct :: "'a bexp set \<Rightarrow> 'a bexp" where
          "conjunct E \<equiv> (\<lambda> \<sigma>. \<forall> e \<in> E. e \<sigma>)"

(* Une expression booleene \<phi> entraine \<psi> si les etats rendant vraie \<phi> 
   rendent egalement vraie \<psi> *)
definition entails :: "'a bexp \<Rightarrow> 'a bexp \<Rightarrow> bool" (infixl "\<Turnstile>\<^sub>B" 55) where
          "\<phi> \<Turnstile>\<^sub>B \<psi> = (\<forall> \<sigma>. \<phi> \<sigma> \<longrightarrow> \<psi> \<sigma>)"

(***********************************************************)
(******************* Théroèmes *****************************)
(***********************************************************)

lemma fresh_is :
  "fresh v exp \<Longrightarrow> exp (\<sigma>(v := val)) = exp \<sigma>"
by (simp add : vars_def)

lemma fresh_is2 :
  "\<forall> \<sigma> val. exp (\<sigma>(v := val)) = exp \<sigma> \<Longrightarrow> fresh v exp"
apply (auto simp add: vars_def)
done


(* les variables d'une conjonction sont inclues dans l'union des variables des 
   expressions la constituant *)
lemma vars_of_conj :
  "vars (\<lambda> \<sigma>. e1 \<sigma> \<and> e2 \<sigma>) \<subseteq> vars e1 \<union> vars e2"
unfolding subset_iff
proof (intro allI impI)

  let ?e' = "(\<lambda> \<sigma>. e1 \<sigma> \<and> e2 \<sigma>)"

  fix v

  assume "v \<in> vars ?e'"

  then obtain \<sigma> val where "?e' (\<sigma> (v := val)) \<noteq> ?e' \<sigma>" unfolding vars_def by blast

  hence "e1 (\<sigma> (v := val)) \<noteq> e1 \<sigma> \<or> e2 (\<sigma> (v := val)) \<noteq> e2 \<sigma>" by auto

  thus "v \<in> vars e1 \<union> vars e2" unfolding vars_def by blast
qed

(* idem pour l'egalite (entre expr. booleenes) *)
lemma vars_of_eq_b :
  fixes e1 :: "'a bexp"
  shows  "vars (\<lambda> \<sigma>. e1 \<sigma> = e2 \<sigma>) \<subseteq> vars e1 \<union> vars e2"
unfolding subset_iff
proof (intro allI impI)

  let ?e' = "(\<lambda> \<sigma>. e1 \<sigma> = e2 \<sigma>)"

  fix v

  assume "v \<in> vars ?e'"

  then obtain \<sigma> val where "?e' (\<sigma> (v := val)) \<noteq> ?e' \<sigma>" unfolding vars_def by blast

  hence "e1 (\<sigma> (v := val)) \<noteq> e1 \<sigma> \<or> e2 (\<sigma> (v := val)) \<noteq> e2 \<sigma>" by auto

  thus "v \<in> vars e1 \<union> vars e2" unfolding vars_def by blast
qed


(* idem pour l'egalite entre expr. arithmetiques *)
lemma vars_of_eq_a :
  fixes e1 :: "'a aexp"
 shows  "vars (\<lambda> \<sigma>. e1 \<sigma> = e2 \<sigma>) \<subseteq> Aexp.vars e1 \<union> Aexp.vars e2"
unfolding subset_iff
proof (intro allI impI)

  let ?e' = "(\<lambda> \<sigma>. e1 \<sigma> = e2 \<sigma>)"

  fix v

  assume "v \<in> vars ?e'"

  then obtain \<sigma> val where "?e' (\<sigma> (v := val)) \<noteq> ?e' \<sigma>" unfolding vars_def by blast

  hence "e1 (\<sigma> (v := val)) \<noteq> e1 \<sigma> \<or> e2 (\<sigma> (v := val)) \<noteq> e2 \<sigma>" by auto

  thus "v \<in> Aexp.vars e1 \<union> Aexp.vars e2" unfolding Aexp.vars_def by blast
qed


(* Par consequent, si les ensembles de variables de e1 et e2 sont finis, 
   l'ensemble des variables de l'egalite l'est egalement *)
lemma finite_vars_of_a_eq :
  fixes e1 :: "'a aexp"
  assumes "finite (Aexp.vars e1)"
  assumes "finite (Aexp.vars e2)"
  shows   "finite (vars (\<lambda> \<sigma>. e1 \<sigma> = e2 \<sigma>))"
using assms vars_of_eq_a[of e1 e2]
      rev_finite_subset[of "Aexp.vars e1 \<union> Aexp.vars e2"]
by auto

(* la conjonction de l'union de 2 ensembles est la conjonction des conjonctions
   de ces ensembles *)
lemma conjunct_Un :
  "conjunct (E1 \<union> E2) = (\<lambda> \<sigma>. conjunct E1 \<sigma> \<and> conjunct E2 \<sigma>)"
by auto


lemma conjunct_insert :
  "conjunct (insert e E) = (\<lambda> \<sigma>. e \<sigma> \<and> conjunct E \<sigma>)"
by auto

(* Soit E un ensemble fini d'expressions possedant chacune un nombre fini de 
   variables, la conjonction de ces expressions possede egalement un nombre
   fini de variables. *)
lemma finite_conj :
  assumes "finite E"
  assumes "\<forall> e \<in> E. finite (Bexp.vars e)"
  shows   "finite (Bexp.vars (conjunct E))"
using assms
proof (induct rule : finite_induct)
  
  case goal1 thus ?case by (simp add : Bexp.vars_def)

next

  case (goal2 e E)

  hence "finite (Bexp.vars e)" and "\<forall> e \<in> E. finite (Bexp.vars e)" by auto
  
  hence "finite ((Bexp.vars e) \<union> Bexp.vars (conjunct E))" using goal2 by auto

  thus ?case
  using vars_of_conj[of e "conjunct E"] conjunct_insert[of e E]
  by (simp only : rev_finite_subset)
  
qed




find_unused_assms

end
