theory Graph
imports Main
begin

lemma neq_Nil_conv2 :
  "l \<noteq> [] = (\<exists> x l'. l = l' @ [x])"
by (induct l rule : rev_induct, auto)

(**************************************************************************************************)
(***************************  GRAPHES  ************************************************************)
(**************************************************************************************************)

record 'a arc = 
  src   :: "'a"
  tgt   :: "'a"


abbreviation ui_arc :: 
  "('a \<times> nat) arc \<Rightarrow> 'a arc" 
where
  "ui_arc t \<equiv> \<lparr> src = fst (src t), tgt = fst (tgt t) \<rparr>"

abbreviation ui_as ::
  "('a \<times> nat) arc list \<Rightarrow> 'a arc list"
where
  "ui_as as \<equiv> map ui_arc as"



(* Un graphe possède un sommet initial et un ensemble d'arcs *)
record 'a rgraph =
  root :: "'a"
  arcs :: "'a arc set"


locale rgraph = 
  fixes g :: "('a,'b) rgraph_scheme" (structure)


(* pour l'instant, un graphe est bien formé si :
   - son ensemble d'arcs est vide,
   - sinon, si il existe un arc issu de son sommet initial *)
locale wf_rgraph = rgraph +
  assumes arcs_empty_or_root_vert : "root g \<in> src `arcs g = (arcs g \<noteq> {})"
begin
  lemmas wf_rgraph = arcs_empty_or_root_vert
end




(* le lemme suivant est donc obtenu par définition de wf_rgraph. Les lemmes de ce genre (qui ne font
   que reprendre une partie de la def.) permettent de ne plus avoir à utiliser la définition 
   du concept en question, mais directement ses propriétés (ce qui facilite la compréhension et 
   surtout la rédaction des preuves). *)
lemma empty_graph_wf :
  assumes "arcs g = {}"
  shows   "wf_rgraph g"
using assms by (simp add : wf_rgraph_def)


(* Je pensais avoir besoin de ce concept pour prouver certaines propriétés par la suite, mais 
   pour l'instant il semble qu'on en ait pas besoin. Je le laisse quand même, ainsi que quelques 
   propriétés qui l'utilisent : on pourrait en avoir besoin pour la suite... *)
locale loop_free = rgraph +
  assumes no_loop : "a \<in> arcs g \<Longrightarrow> src a \<noteq> tgt a"



(**************************************************************************************************)
(***************************  SOMMETS  ************************************************************)
(**************************************************************************************************)

(* ensemble des sommets d'un graphe *)
definition vertices :: 
  "('a,'b) rgraph_scheme \<Rightarrow> 'a set"
where
  "vertices g = {root g} \<union> src `arcs g \<union> tgt ` arcs g"


(* quelques lemmes simples sur les sommets *)

lemma vertices_conv (*[simp]*) :
  "vertices g = {root g} \<union> {l. \<exists> t \<in> arcs g. src t = l \<or> tgt t = l}"
by (auto simp add : vertices_def)

lemma root_is_vert :
  "root g \<in> vertices g"
unfolding vertices_def by simp

lemma src_is_vert :
  assumes "t \<in> arcs g"
  shows   "src t \<in> vertices g"
using assms unfolding vertices_def by auto

lemma src_not_vert :
  assumes "src a \<notin> vertices g"
  shows   "a \<notin> arcs g"
using assms unfolding vertices_def by auto

lemma tgt_not_vert : 
  assumes "tgt a \<notin> vertices g"
  shows   "a \<notin> arcs g"
using assms unfolding vertices_def by auto

lemma tgt_is_vert : 
  assumes "t \<in> arcs g"
  shows   "tgt t \<in> vertices g"
using assms unfolding vertices_def by auto

lemma in_vertE : 
  assumes "l \<in> vertices g"
  shows   "l = root g \<or> (\<exists> t \<in> arcs g. l = src t \<or> l = tgt t)"
using assms unfolding vertices_def by auto

lemma empty_graph_vertices :
  assumes "arcs g = {}"
  shows   "vertices g = {root g}"
using assms unfolding vertices_def by simp


lemma (in loop_free) root_only_vert_imp_arcs_empty :
  assumes "vertices g = {root g}"
  shows   "arcs g = {}"
proof (rule ccontr)
  assume "arcs g \<noteq> {}"

  then obtain a where "a \<in> arcs g" by blast

  moreover
  hence "src a = root g" and "tgt a = root g" 
  using assms by (auto simp add : vertices_conv)

  ultimately
  show "False" using no_loop[of a] by simp
qed


lemma (in loop_free) arcs_empty_imp_root_only_vert :
  "arcs g = {} = (vertices g = {root g})"
using empty_graph_vertices root_only_vert_imp_arcs_empty by auto



(**************************************************************************************************)
(***************************  ARCS SORTANTS  ******************************************************)
(**************************************************************************************************)

(* arc issus d'un sommet donné *)
definition out_arcs ::
  "('a,'b) rgraph_scheme \<Rightarrow>
   'a                    \<Rightarrow> 'a arc set" 
where
  "out_arcs g l \<equiv> {t \<in> arcs g. src t = l}"

lemma arcs_conv :
  "arcs g = (\<Union> v \<in> vertices g. out_arcs g v)"
by (auto simp add : vertices_conv out_arcs_def)

lemma out_arcsI:
  assumes "t \<in> arcs lts"
  assumes "src t = l"
  shows   "t \<in> out_arcs lts l"
using assms unfolding out_arcs_def by simp

lemma out_arcsE1 :
  assumes "t \<in> out_arcs g l"
  shows   "t \<in> arcs g"
using assms unfolding out_arcs_def by simp

lemma out_arcsE2 :
  assumes "t \<in> out_arcs lts l"
  shows   "src t = l"
using assms unfolding out_arcs_def by simp

lemma out_arcs_are_arcs :
  "out_arcs g l \<subseteq> arcs g"
unfolding subset_iff by (intro allI impI, rule out_arcsE1)


lemma (in wf_rgraph) arcs_not_empty_imp :
  assumes "arcs g \<noteq> {}"
  shows   "out_arcs g (root g) \<noteq> {}"
using assms wf_rgraph by (auto simp add : out_arcs_def)


(**************************************************************************************************)
(***************************  CHEMINS  ************************************************************)
(**************************************************************************************************)


(* Dans cette partie, on modélise les ensembles de chemins et sous-chemins d'un graphe (en l'absence
   de subsomptions). On donne les propriétés de base concernant les chemins et sous-chemins, 
   notamment lorsqu'un arc est ajouté au graphe.

   Par la suite, ces définitions/propriétés seront utilisés pour montrer l'évolution des ensembles 
   de chemins/sous-chemins dans une extension par ajout d'arcs, puis finalement pour montrer que la
   partie rouge d'un RB-graphe est bien un arbre. 

   Cependant, on est également intéressé par le cas avec subsomptions, donc une "deuxième version" 
   de ces définitions et lemmes prenant en compte les subsomptions sera donnée dans SubRel.thy.

 *)




(**************** Consistance des séquences d'arcs) ****************)

(* une séquence d'arcs allant de l à l' est consistante si :
   - la source de son premier arc est l
   - la cible de son dernier arc est 'l
   - la cible de chaque arc de la séquence est la source de l'arc qui lui succède *)
fun cas :: 
  "'a          \<Rightarrow> 
   'a arc list \<Rightarrow> 
   'a          \<Rightarrow> bool" 
where
  "cas l [] l' = (l = l')"
| "cas l (t#ts) l' = (src t = l \<and> cas (tgt t) ts l')"

lemma cas_Nil : "cas l [] l" by simp

lemma cas_one : 
  "cas l [a] l' = (src a = l \<and> tgt a = l')"
by simp

lemma cas_append_one :
  "cas l (ts@[t]) l' = (cas l ts (src t) \<and> tgt t = l')" (is "?P l")
by (induct ts arbitrary : l, auto) 

lemma cas_append :
  "cas l1 (ts1 @ ts2) l2 = (\<exists> l. cas l1 ts1 l \<and> cas l ts2 l2)" (is "?P l1 ts1")
by (induct ts1 arbitrary : l1, auto)


(**************** Sous-chemins ****************)

(* un sous-chemin est une séquence d'arcs consistante reliant deux sommets du graphe et 
   dont les éléments sont des arcs du graphe (note : la contrainte "l \<in> vertices g" est nécessaire
   pour le cas du chemin vide car autrement on pourrait avoir "cas l [] l \<and> set [] \<subseteq> arcs g" sans
   que l soit un sommet de g) *)
definition subpath :: 
  "('a,'b) rgraph_scheme \<Rightarrow> 
   'a                    \<Rightarrow> 
   'a arc list           \<Rightarrow> 
   'a                    \<Rightarrow> bool" 
where
  "subpath g l p l' \<equiv> cas l p l' \<and> l \<in> vertices g \<and> set p \<subseteq> arcs g"

(*lemma sp_imp_elem_in_arcs :
  assumes "subpath g l p l'"
  shows   "set p \<subseteq> arcs g"
using assms unfolding subpath_def by simp*)

lemma Nil_sp_from_root :
  "subpath g (root g) [] (root g)"
unfolding subpath_def by (auto simp add : root_is_vert)

lemma Nil_sp : 
  "subpath g l [] l' = (l = l' \<and> l \<in> vertices g)"
unfolding subpath_def by auto

lemma sp_one :
  "subpath g l [a] l' = (src a = l \<and> a \<in> arcs g \<and> tgt a = l')"
using src_is_vert unfolding subpath_def by auto

lemma sp_Cons :
  "subpath g l1 (t#p) l2 = (src t = l1 \<and> t \<in> arcs g \<and> subpath g (tgt t) p l2)"
using tgt_is_vert src_is_vert unfolding subpath_def by auto

lemma fst_of_sp_is_vert :
  assumes "subpath g l1 p l2"
  shows   "l1 \<in> vertices g"
using assms unfolding subpath_def by simp

lemma lst_of_sp_is_vert :
  assumes "subpath g l1 p l2"
  shows   "l2 \<in> vertices g"
using assms tgt_is_vert
by (induction p arbitrary : l1, auto simp add: subpath_def)

(*lemma in_sp_imp_src_vert :
  assumes "subpath g l1 p l2"
  assumes "a \<in> set p"
  shows   "src a \<in> vertices g"
proof -
  have "a \<in> arcs g" using assms by (auto simp add : subpath_def)

  thus ?thesis by (auto simp add : vertices_conv)
qed

lemma in_sp_imp_tgt_vert :
  assumes "subpath g l1 p l2"
  assumes "a \<in> set p"
  shows   "tgt a \<in> vertices g"
proof -
  have "a \<in> arcs g" using assms by (auto simp add : subpath_def)

  thus ?thesis by (auto simp add : vertices_conv)
qed*)

(*lemma sp_append_one_tgt :
  assumes "subpath g l (ts@[t]) l'"
  shows   "tgt t = l'"
using assms 
by (induct ts arbitrary : l) 
   (simp add : sp_one, auto simp add : sp_Cons)*)

lemma sp_append_one :
  "subpath g l (ts@[t]) l' = (subpath g l ts (src t) \<and> t \<in> arcs g \<and> tgt t = l')"
by (induct ts arbitrary : l, auto simp add : subpath_def vertices_conv)

lemma sp_append :
  "subpath g l1 (ts1@ts2) l2 = (\<exists> l. subpath g l1 ts1 l \<and> subpath g l ts2 l2)"
by (induct ts1 arbitrary : l1)
   (auto simp add : Nil_sp fst_of_sp_is_vert sp_Cons)

lemma empty_graph_imp_sp_eq_Nil :
  assumes "arcs g = {}"
  assumes "subpath g l p l'"
  shows   "p = []"
using assms unfolding subpath_def by auto



lemma arcs_empty_imp_sp_eq :
  assumes "arcs g = {}"
  shows   "subpath g l p l' = (l = root g \<and> p = [] \<and> l' = root g)"
using assms unfolding subpath_def by (auto simp add : vertices_conv)

lemma arc_sp_eq :
  "subpath g (src a) [a] (tgt a) = (a \<in> arcs g)"
using src_is_vert unfolding subpath_def by auto



(* Tout sous chemin issu d'un sommet dont aucun arc ne part est vide *)
lemma sp_from_de_empty :
  assumes "out_arcs g l = {}"
  assumes "subpath g l p l'"
  
  shows   "p = []"
proof (rule ccontr)
  assume "p \<noteq> []"

  then obtain a p' where "p = a # p'" using neq_Nil_conv[of p] by blast

  hence "src a = l" and "a \<in> arcs g" 
  using assms(2) by (auto simp add : sp_Cons)
  
  hence "a \<in> out_arcs g l" by (intro out_arcsI)

  thus "False" using assms by simp
qed



(* Un sous-chemin p passant par un arc a tel qu'aucun arc n'est issu de sa cible termine par a.
   De plus a n'apparait pas "plus tot" dans p, sinon il existerait un arc issu de la cible de a. *)
lemma sp_through_de_decomp :
  assumes "out_arcs g (tgt a) = {}"
  assumes "subpath g l1 p l2"
  assumes "a \<in> set p"

  shows   "\<exists> p'. p = p' @ [a] \<and> a \<notin> set p'"

using assms(2) assms(3)
proof (induction p arbitrary : l1)
  case Nil thus ?case by simp
next
  case (Cons a' p)
  thus ?case
  proof -
  
    have sp_from_tgt : "subpath g (tgt a') p l2" 
    using Cons(2) by (auto simp only : sp_Cons)

    have "a = a' \<or> (a \<noteq> a' \<and> a \<in> set p)" using Cons(3) by auto

    thus ?thesis
    proof (elim disjE)
      assume "a = a'"

      moreover
      hence "p = []" using sp_from_tgt assms(1) by (simp only : sp_from_de_empty)
      
      ultimately
      show ?thesis by simp
    next
      assume "a \<noteq> a' \<and> a \<in> set p"

      moreover
      then obtain p' where "p = p' @ [a]" and "a \<notin> set p'"
      using sp_from_tgt Cons.IH[of "tgt a'"] by blast

      ultimately
      show ?thesis by auto
    qed
  qed
qed


(* On peut définir l'ensemble des sous-chemins de la manière suivante (le fait que 
   "subpath g l1 p l2" soit vrai garantit que l1 et l2 sont des sommets de g) *)
abbreviation subpaths :: "('a,'b) rgraph_scheme \<Rightarrow> 'a arc list set" where
  "subpaths g \<equiv> {p. \<exists> l1 l2. subpath g l1 p l2}"



lemma empty_graph_subpaths :
  assumes "arcs g = {}"
  shows   "subpaths g = {[]}"
using assms
by (auto simp add : empty_graph_imp_sp_eq_Nil Nil_sp)
   (rule_tac ?x="root g" in exI, rule root_is_vert)

lemma empty_subpaths_imp_empty_graph :
  assumes "subpaths g = {[]}"
  shows   "arcs g = {}"
unfolding set_eq_iff empty_iff
proof (intro allI iffI)
  fix a

  assume "a \<in> arcs g"

  hence  "subpath g (src a) [a] (tgt a)" by (simp only : arc_sp_eq)

  thus   "False" using assms(1) by auto
next
  fix a
  assume "False" thus "a \<in> arcs g" by (elim FalseE)
qed


lemma empty_graph_eq : 
  "(arcs g = {}) = (subpaths g = {[]})"
using empty_graph_subpaths empty_subpaths_imp_empty_graph by auto


(* On définit l'ensemble des sous-chemins issus d'un sommet donné, ce qui nous 
   permettra de caractériser l'ensemble des chemins du graphe : ce sont les sous-chemins issus du 
   sommet initial. *)

definition subpath_from  ::
  "('a,'b) rgraph_scheme \<Rightarrow>
   'a                    \<Rightarrow>
   'a arc list           \<Rightarrow> bool"
where
  "subpath_from g v as \<equiv> \<exists> v'. subpath g v as v'"

lemma Nil_sp_from_root2 :
  assumes "v \<in> vertices g"
  shows   "subpath_from g v []"
using assms Nil_sp by (force simp add : subpath_from_def)

lemma empty_graph_imp_sp_from_eq_Nil :
  assumes "arcs g = {}"
  assumes "subpath_from g v as"
  shows   "as = []"
using assms empty_graph_imp_sp_eq_Nil by (auto simp add : subpath_from_def)

abbreviation subpaths_from :: 
  "('a,'b) rgraph_scheme \<Rightarrow> 
   'a                    \<Rightarrow> 'a arc list set" 
where
  "subpaths_from g v \<equiv> {ras. subpath_from g v ras}"

lemma empty_graph_sps_from_eq :
  assumes "arcs g = {}"
  assumes "v \<in> vertices g"
  shows   "subpaths_from g v = {[]}"
unfolding set_eq_iff singleton_iff
proof (intro allI iffI)
  case goal1 thus ?case using assms by (simp add : empty_graph_imp_sp_from_eq_Nil)
next
  case goal2 thus ?case using assms Nil_sp_from_root2 by auto
qed



lemma empty_graph_imp_sps_from_are_sps :
  assumes "arcs g = {}"
  shows   "subpaths_from g (root g) = subpaths g"
proof (simp only : set_eq_iff, intro allI iffI)
  fix p

  assume "p \<in> subpaths_from g (root g)"
  thus   "p \<in> subpaths g" unfolding subpath_from_def by auto
  
next
  fix p

  assume "p \<in> subpaths g"

  then obtain l1 l2 where "subpath g l1 p l2" by blast

  hence "l1 = root g" 
  using fst_of_sp_is_vert[of g l1 p l2] assms empty_graph_vertices by auto

  thus "p \<in> subpaths_from g (root g)" 
  using `subpath g l1 p l2` by (auto simp add : subpath_from_def)
qed





(**************** Chemins ****************)

(* un chemin d'un graphe est un sous-chemin issu du sommet initial de ce graphe *)
definition path :: 
  "('a,'b) rgraph_scheme \<Rightarrow> 
   'a arc list           \<Rightarrow> 
   'a                    \<Rightarrow> bool" 
where
  "path g ts l \<equiv> subpath g (root g) ts l"

abbreviation paths :: 
  "('a,'b) rgraph_scheme \<Rightarrow> 'a arc list set" 
where
  "paths g \<equiv> {p. \<exists> l. path g p l}"




lemma paths_conv :
  "paths g = subpaths_from g (root g)"
by (simp add : subpath_from_def path_def)


lemma lst_of_path_is_vert :
  assumes "path g p l"
  shows   "l \<in> vertices g"
using assms unfolding path_def by (rule lst_of_sp_is_vert)

(*lemma Nil_path_from_root [intro] :
  "path g [] (root g)"
unfolding path_def by (rule Nil_sp_from_root)*)

lemma Nil_path  :
  "path g [] l = (l = root g)"
unfolding path_def by (auto simp add : Nil_sp root_is_vert)

lemma Nil_in_paths :
  "[] \<in> paths g"
using Nil_path by force

lemma path_one :
  "path g [a] l = (src a = root g \<and> a \<in> arcs g \<and> tgt a = l)"
unfolding path_def by (rule sp_one)

lemma path_Cons : 
  "path g (t#p) l = (src t = root g \<and> t \<in> arcs g \<and> subpath g (tgt t) p l)"
unfolding path_def by (rule sp_Cons)


lemma path_append_one :
  "path g (p @ [t]) l = (path g p (src t) \<and> tgt t = l \<and> t \<in> arcs g)"
unfolding path_def by (auto simp add : sp_append_one)

lemma path_append :
  "path g (p1 @ p2) l = (\<exists> l'. path g p1 l' \<and> subpath g l' p2 l)"
unfolding path_def by (rule sp_append)

lemma path_appendI :
  assumes "path g p1 l1"
  assumes "subpath g l1 p2 l2"
  shows   "path g (p1 @ p2) l2"
unfolding path_append using assms by (intro exI conjI)

lemma path_in_empty_graph_imp_empty :
  assumes "arcs g = {}"
  assumes "path g p l'"
  shows   "p = []"
using assms unfolding path_def by (rule empty_graph_imp_sp_eq_Nil)

lemma arcs_empty_imp_path_eq :
  assumes "arcs g = {}"
  shows   "path g p l' = (p = [] \<and> l' = root g)"
using assms unfolding path_def by (simp add : arcs_empty_imp_sp_eq)

(* Un chemin p passant par un arc a tel qu'aucun arc n'est issu de sa cible termine par a.
   De plus a n'apparait pas "plus tot" dans p, sinon il existerait un arc issu de la cible de a. *)
lemma path_through_de_decomp :
  assumes "out_arcs g (tgt a) = {}"
  assumes "path g p l"
  assumes "a \<in> set p"

  shows   "\<exists> p'. p = p' @ [a] \<and> a \<notin> set p'"
using assms unfolding path_def by (simp add : sp_through_de_decomp)


lemma empty_graph_paths :
  assumes "arcs g = {}"
  shows   "paths g = {[]}"
using assms empty_graph_subpaths[of g] empty_graph_imp_sps_from_are_sps[of g]
by (simp only : paths_conv)










(**************************************************************************************************)
(***************************  AJOUT D'ARCS  *******************************************************)
(**************************************************************************************************)

(* la définition suivante sert principalement à alléger les écritures pour la suite. *)

find_consts name:"rgraph" name:"update"

definition add_arc :: 
  "('a,'b) rgraph_scheme \<Rightarrow> 
   'a arc                \<Rightarrow> ('a,'b) rgraph_scheme" 
where
  "add_arc g a \<equiv> rgraph.arcs_update (\<lambda> arcs. arcs \<union> {a}) g"

lemma add_arc_vertices :
  "vertices (add_arc g a) = (vertices g \<union> {src a, tgt a})"
by (auto simp add : vertices_conv add_arc_def)

lemma add_arc_root :
  "root (add_arc g a) = root g"
by (auto simp add : add_arc_def)

lemma add_arc_arcs :
  "arcs (add_arc g a) = arcs g \<union> {a}"
by (auto simp add : add_arc_def)

lemma sp_in_add_arc :
  assumes "subpath g l1 p l2"
  shows   "subpath (add_arc g a) l1 p l2"
using assms by (auto simp only : subpath_def add_arc_vertices add_arc_arcs)

lemma path_in_add_arc :
  assumes "path g p l"
  shows   "path (add_arc g a) p l"
using assms unfolding path_def by (simp add : sp_in_add_arc add_arc_root)



lemma sps_in_add_arc :
  "subpaths g \<subseteq> subpaths (add_arc g a)"
using sp_in_add_arc by force

lemma sp_from_in_add_arc :
  assumes "subpath_from g v as"
  shows   "subpath_from (add_arc g a) v as"
using assms sp_in_add_arc[of g v as _ a] by (auto simp add : subpath_from_def)

lemma sps_from_in_add_arc :
  "subpaths_from g l \<subseteq> subpaths_from (add_arc g a) l"
using sp_from_in_add_arc by fast

lemma sp_end_src_imp_append_in_add_arc :
  assumes "subpath g l1 p (src a)"
  shows   "subpath (add_arc g a) l1 (p @ [a]) (tgt a)"
using assms by (auto simp add : sp_append_one sp_in_add_arc add_arc_arcs)

lemma p_end_src_imp_append_in_add_arc :
  assumes "path g p (src a)"
  shows   "path (add_arc g a) (p @ [a]) (tgt a)"
using assms unfolding path_def by (simp add : sp_end_src_imp_append_in_add_arc add_arc_root)


(* si on ajoute au graphe un arc dont la cible n'est pas un sommet du graphe, alors l'ensemble 
   des arcs sortants depuis cette cible est vide dans le nouveau graphe.

   On a besoin en plus de l'hypothèse "src a \<noteq> tgt a" sinon cet ensemble est {a} *)

lemma add_arc_tgt_out_arcs :
  assumes "src a \<noteq> tgt a"
  assumes "tgt a \<notin> vertices g"

  shows   "out_arcs (add_arc g a) (tgt a) = {}" (is "?oa = {}")
proof (rule ccontr)
  assume "?oa \<noteq> {}"
  
  then obtain a' where "a' \<in> ?oa" using ex_in_conv[of ?oa] by blast

  hence "a' \<in> out_arcs g (tgt a)" using assms by (auto simp add : out_arcs_def add_arc_arcs)

  hence "a' \<in> arcs g" by (rule out_arcsE1)

  hence "src a' \<in> vertices g" by (rule src_is_vert)

  moreover
  have "src a' = tgt a" using `a' \<in> out_arcs g (tgt a)` by (rule out_arcsE2)

  ultimately
  show "False" using assms by auto
qed

lemma add_arc_src_out_arcs :
  "out_arcs (add_arc g a) (src a) = out_arcs g (src a) \<union> {a}"
by (auto simp add : add_arc_def out_arcs_def)

lemma add_arc_not_end_out_arcs :
  assumes "v \<noteq> src a"
  
  shows   "out_arcs (add_arc g a) v = out_arcs g v"
using assms by (auto simp add : out_arcs_def add_arc_arcs)


lemma sp_in_add_arc_minus_arc_imp_sp_in_old :
  assumes "l1 \<in> vertices g"
  assumes "subpath (add_arc g a) l1 p l2"
  assumes "a \<notin> set p"
  
  shows   "subpath g l1 p l2"
using assms unfolding subpath_def by (auto simp add : add_arc_arcs)
thm add_arc_arcs

lemma p_in_add_arc_minus_arc_imp_p_in_old :
  assumes "path (add_arc g a) p l"
  assumes "a \<notin> set p"
  
  shows   "path g p l"
proof -
  have "root g \<in> vertices g" by (auto simp add : root_is_vert)

  thus ?thesis using assms unfolding path_def add_arc_root
  by (rule sp_in_add_arc_minus_arc_imp_sp_in_old)
qed







lemma sp_ends_in_tgt_imp_mem :
  assumes "tgt a \<notin> vertices g"
  assumes "l \<noteq> tgt a"
  assumes "subpath (add_arc g a) l p (tgt a)"

  shows   "a \<in> set p"
proof -
  have "p \<noteq> []" using assms(2) assms(3) by (auto simp add : Nil_sp)

  then obtain a' p' where "p = p' @ [a']" using neq_Nil_conv2[of p] by blast

  hence "a' \<in> arcs (add_arc g a)" 
  and   "tgt a' = tgt a"
  using assms(3) by (auto simp add : sp_append_one)

  hence "a' = a \<or> (a' \<noteq> a \<and> a' \<in> arcs g)" by (auto simp add : add_arc_arcs)

  thus ?thesis 
  proof (elim disjE)
    assume "a' = a" thus ?thesis using `p = p' @ [a']` by simp
  next
    assume "a' \<noteq> a \<and> a' \<in> arcs g"

    hence  "tgt a' \<in> vertices g" by (simp only : tgt_is_vert)

    thus   ?thesis using `tgt a' = tgt a` assms(1) by simp
  qed
qed


lemma p_ends_in_tgt_imp_mem :
  assumes "tgt a \<notin> vertices g"
  assumes "path (add_arc g a) p (tgt a)"

  shows   "a \<in> set p"
proof -
  have "root g \<noteq> tgt a" using assms(1) by (auto simp add : vertices_conv)

  thus ?thesis using assms unfolding path_def by (simp only : sp_ends_in_tgt_imp_mem add_arc_root)
qed

lemma sp_same_src_imp_same_tgt :
  assumes "subpath g v as v1"
  assumes "subpath g v as v2"
  shows   "v1 = v2"
using assms
by (induct as arbitrary : v) (auto simp add :  Nil_sp sp_Cons)


end