theory PreConf
imports Store Finite_Set Labels (*LTS*)
begin

(* preconf = store + une liste de contraintes sur les variables symboliques *)
record 'var_type pre_conf = 
  store :: "'var_type store"
  \<pi>     :: "'var_type symvar bexp set"

(* le prédicat de cheminement d'une préconf est simplement la conjonction des contraintes 
   contenues dans \<pi> *)
(*definition path_pred :: 
  "'var_type pre_conf \<Rightarrow> 'var_type symvar bexp" 
where
  "conjunct (\<pi> pc) = conjunct (set (\<pi> pc))"*)


(**********************************************)
(*** satisfiabilité d'une pré-configuration ***)
(**********************************************)

(* une préconf est satisfiable si son prédicat de cheminement est satisfiable *)
abbreviation sat :: "'var_type pre_conf \<Rightarrow> bool" where
            "sat pc \<equiv> Bexp.sat (conjunct (\<pi> pc))"


(**********************************************)
(*** états d'une pré-configuration ************)
(**********************************************)

(* ensemble des états représentés par une préconf *)
definition
  states :: "'var_type pre_conf \<Rightarrow> 'var_type state set" 
where
 "states pc = {\<sigma>. \<exists> \<sigma>\<^sub>s\<^sub>y\<^sub>m. consistent \<sigma> \<sigma>\<^sub>s\<^sub>y\<^sub>m (store pc) \<and> conjunct (\<pi> pc) \<sigma>\<^sub>s\<^sub>y\<^sub>m}"

(**********************************************)
(*** sémantique d'une pré-configuration *******)
(**********************************************)

definition 
  sem :: "'var_type pre_conf \<Rightarrow> 'var_type bexp" 
where
 "sem pc = (\<lambda> \<sigma>. \<sigma> \<in> states pc)"

(************************************************)
(*** validation d'une formule par une pré-conf **)
(************************************************)

definition 
  models :: "'var_type pre_conf \<Rightarrow> 'var_type bexp \<Rightarrow> bool" (infixl "\<Turnstile>\<^sub>p\<^sub>c" 55) 
where
  "pc \<Turnstile>\<^sub>p\<^sub>c \<phi> \<equiv> (\<forall> \<sigma>. sem pc \<sigma> \<longrightarrow> \<phi> \<sigma>)"

(************************************************)
(*** subsomption entre pré-configurations *******)
(************************************************)

definition 
  subsums :: "'var_type pre_conf \<Rightarrow> 'var_type pre_conf \<Rightarrow> bool" (infixl "\<sqsubseteq>\<^sub>p\<^sub>c" 55) 
where     
  "pc\<^sub>2 \<sqsubseteq>\<^sub>p\<^sub>c pc\<^sub>1 \<equiv> (states pc\<^sub>2 \<subseteq> states pc\<^sub>1)"

(*********************************************************************)
(******* Abstractions ************************************************)
(*********************************************************************)

definition abstract :: "'var_type symvar bexp \<Rightarrow> 'var_type pre_conf \<Rightarrow> 'var_type pre_conf" where
          "abstract \<phi> pc = \<lparr> store = store pc, \<pi> = \<pi> pc - {\<phi>} \<rparr>"

definition Abstract :: "'var_type symvar bexp list \<Rightarrow> 'var_type pre_conf \<Rightarrow> 'var_type pre_conf" where
          "Abstract bexps pc = foldl (\<lambda> pc \<phi>. abstract \<phi> pc) pc bexps"


(************************************************)
(*** variables symboliques **********************)
(************************************************)

definition symvars :: "'var_type pre_conf \<Rightarrow> 'var_type symvar set" where
          (*"symvars pc = Store.symvars (store pc) \<union> (\<Union> \<phi> \<in> set (\<pi> pc). Bexp.vars \<phi>)"*)
          "symvars pc = Store.symvars (store pc) \<union> Bexp.vars (conjunct (\<pi> pc))"

(* une variable symbolique est fraiche pour une preconf si elle n'appartient pas a l'ensemble des 
   variables symboliques de cette preconf *)
definition fresh_symvar :: "'var_type symvar \<Rightarrow> 'var_type pre_conf \<Rightarrow> bool" where
          "fresh_symvar sv pc = (sv \<notin> symvars pc)"

(* Une pre_conf peut être mise à jour (potentiellement pour toute variable v) si, pour chaque 
   variable v, il existe une variable symbolique fraîche pour cette pre_conf *)
definition updatable :: "'var_type pre_conf \<Rightarrow> bool" where
  "updatable pc \<equiv> \<forall> v. \<exists> sv. fst sv = v \<and> fresh_symvar sv pc"

lemma fresh_symvar_eq :
  "fresh_symvar sv pc = (Store.fresh_symvar sv (store pc) \<and> Bexp.fresh sv (conjunct (\<pi> pc)))"
unfolding fresh_symvar_def symvars_def Store.fresh_symvar_def by simp

(*lemma
  "fresh_symvar sv pc = (snd sv \<noteq> store pc (fst sv) \<and> sv \<notin> Bexp.vars (conjunct (\<pi> pc)))"
by (simp only : tmp1 fresh_eq)*)

(* ensemble des variables symboliques de la forme (v,n) d'une preconf pc pour un v donne *)
definition symvars_of_var :: "'var_type pre_conf \<Rightarrow> 'var_type \<Rightarrow> 'var_type symvar set" where
          "symvars_of_var pc v = {sv \<in> symvars pc. fst sv = v}"



(***************************************************)
(*** Exécution symbolique pour les pré-confs *******)
(***************************************************)


inductive se ::
  "'a pre_conf \<Rightarrow> 'a label \<Rightarrow> 'a pre_conf \<Rightarrow> bool"
where
  "se pc Skip pc"
| "se pc Break pc"
| "se pc (Assume e) \<lparr> store = store pc, \<pi> = \<pi> pc \<union> {adapt_bexp e (store pc)} \<rparr>"
| "fst sv = v         \<Longrightarrow>
   fresh_symvar sv pc \<Longrightarrow>
   se pc (Assign v e) \<lparr> store = (store pc)(v := snd sv), 
                            \<pi> = \<pi> pc \<union> {(\<lambda> \<sigma>. \<sigma> sv = (adapt_aexp e (store pc)) \<sigma>)} \<rparr>"

fun SE ::
  "'a pre_conf \<Rightarrow> 'a label \<Rightarrow> 'a pre_conf"
where
  "SE pc Skip  = pc"
| "SE pc Break = pc"
| "SE pc (Assume e) = \<lparr> store = store pc, \<pi> = \<pi> pc \<union> {adapt_bexp e (store pc)} \<rparr>"
| "SE pc (Assign v e) = (
    let sv = (SOME sv. fst sv = v \<and> fresh_symvar sv pc) 
    in
      \<lparr> store = (store pc)(v := snd sv), \<pi> = insert (\<lambda> \<sigma>. \<sigma> sv = (adapt_aexp e (store pc)) \<sigma>) (\<pi> pc) \<rparr>
   )"



inductive se_star :: "'a pre_conf \<Rightarrow> 'a label list \<Rightarrow> 'a pre_conf \<Rightarrow> bool" where
  "se_star pc [] pc"
| "se pc1 l pc2 \<Longrightarrow> se_star pc2 ls pc3 \<Longrightarrow> se_star pc1 (l # ls) pc3"

find_theorems name:"se_star"

definition feasible :: "'a pre_conf \<Rightarrow> 'a label list \<Rightarrow> bool" where
  "feasible pc labels \<equiv> (\<exists> pc'. se_star pc labels pc' \<and> sat pc')"


                                                          

(***************************************************)
(*** Weakest Preconditions *************************)
(***************************************************)

fun WP :: 
  "'a label \<Rightarrow> 'a bexp \<Rightarrow> 'a bexp"
where
  "WP Skip           post = post"
| "WP Break          post = post"
| "WP (Assume \<phi>)     post = (\<lambda> \<sigma>. (\<phi> \<sigma>) \<longrightarrow> (post \<sigma>))"
| "WP (Assign v exp) post = (\<lambda> \<sigma>. post (\<sigma>(v := exp \<sigma>)))"

definition WP_star :: "'a label list \<Rightarrow> 'a bexp \<Rightarrow> 'a bexp" where
          "WP_star \<omega> post = (foldr WP \<omega> post)"

fun SP ::
  "'var_type bexp \<Rightarrow> 'var_type label \<Rightarrow> 'var_type bexp"
where
  "SP pre Skip           = pre"
| "SP pre Break          = pre"
| "SP pre (Assume \<phi>)     = (\<lambda> \<sigma>. pre \<sigma> \<and> \<phi> \<sigma>)"
| "SP pre (Assign v exp) = (\<lambda> \<sigma>. \<exists> \<sigma>'. \<sigma> = \<sigma>'(v := exp \<sigma>') \<and> pre \<sigma>')"


definition SP_star :: "'var_type bexp \<Rightarrow> 'var_type label list \<Rightarrow> 'var_type bexp" where
          "SP_star pre bexps = (foldl SP pre bexps)"




(******************************************************************)
(*** Théorèmes : avec preuves *************************************)
(******************************************************************)

(* l'ensemble des états d'une configuration est vide ssi cette configuration est insatisfiable 
   
  Idée : les états de pc sont les \<sigma> tels qu'il existe un état \<sigma>\<^sub>s\<^sub>y\<^sub>m tel que \<sigma> \<sigma>\<^sub>s\<^sub>y\<^sub>m soient consistents 
  avec "store pc" et \<sigma>\<^sub>s\<^sub>y\<^sub>m satisfait \<pi> pc. Si \<sigma>\<^sub>s\<^sub>y\<^sub>m ne satisfait \<sigma>\<^sub>s\<^sub>y\<^sub>m, l'ensemble de ces \<sigma>\<^sub> est vide. *)
lemma states_empty_eq :
  "(states pc = {}) = (\<not> sat pc)"
by (simp add : states_def sat_def consistent_def) fast

lemma sat_iff :
  "sat pc \<longleftrightarrow> states pc \<noteq> {}"
using ex_consistent_state1
by (intro iffI) (force simp add : sat_def states_def)+

lemma states_not_empty_eq :
  "states pc \<noteq> {} = sat pc"
by (simp add : states_empty_eq)

(* la sémantique d'une préconfiguration décrit exactement l'ensemble d'états de cette conf. *)
lemma state_sat_sem : 
  "(\<sigma> \<in> states pc) = (sem pc \<sigma>)"
unfolding sem_def by (rule refl)

(* Une configuration valide False ssi elle est insatisfiable *)
lemma models_False_imp_unsat :
  "pc \<Turnstile>\<^sub>p\<^sub>c (\<lambda> \<sigma>. False) = (\<not> sat pc)"
by (simp add: models_def sem_def sat_iff)

lemma subsums_refl [simp] :
  "pc \<sqsubseteq>\<^sub>p\<^sub>c pc"
by (simp only : subsums_def)

lemma subsums_trans :
  "pc1 \<sqsubseteq>\<^sub>p\<^sub>c pc2 \<Longrightarrow> pc2 \<sqsubseteq>\<^sub>p\<^sub>c pc3 \<Longrightarrow> pc1 \<sqsubseteq>\<^sub>p\<^sub>c pc3"
unfolding subsums_def by simp

(* Etablir une subsomption revient à vérifier que la sémantique du subsumée entraîne celle du subsumeur.
   En pratique, cela nous permet de détecter une subsomption sans passer par une énumarétion d'états *)
theorem subsums_eq_entails :
  "(pc\<^sub>2 \<sqsubseteq>\<^sub>p\<^sub>c pc\<^sub>1) = (pc\<^sub>2 \<Turnstile>\<^sub>p\<^sub>c sem pc\<^sub>1)"
unfolding subsums_def models_def by (auto simp add : state_sat_sem)

(* une préconfiguration sat. ne peut etre subsumée que par une préconf. sat. *)
lemma sat_sub_by_sat :
  assumes "sat pc\<^sub>2"
  and     "pc\<^sub>2 \<sqsubseteq>\<^sub>p\<^sub>c pc\<^sub>1"
  shows   "sat pc\<^sub>1"
proof -
  have  "states pc\<^sub>2 \<noteq> {}" using assms by (simp add : sat_iff)
  hence "states pc\<^sub>1 \<noteq> {}" using assms by (auto simp add : subsums_def)
  thus ?thesis by (simp add : sat_iff)
qed

(* quelques propriétés des abstractions *)
lemma abstract_subsums :
  "pc \<sqsubseteq>\<^sub>p\<^sub>c abstract \<phi> pc"
by (auto simp add : abstract_def subsums_def states_def)

lemma Abstract_Nil [simp] :
  "Abstract [] pc = pc"
by (simp add : Abstract_def)

lemma Abstract_one [simp] :
  "Abstract [v] pc = abstract v pc"
by (simp add : Abstract_def)

lemma Abstract_append_one :
  "Abstract (var_list @ [v]) pc = abstract v (Abstract var_list pc)"
by (simp add : Abstract_def)

lemma Abstract_append :
  "Abstract (l1 @ l2) pc = Abstract l2 (Abstract l1 pc)"
by (induction l1, simp_all add : Abstract_def)

lemma Abstract_subsums :
  "pc \<sqsubseteq>\<^sub>p\<^sub>c Abstract bexps pc"
proof (induct bexps rule : rev_induct)
  case Nil show ?case by (simp add : Abstract_def)
next
  fix e exps
  
  assume "pc \<sqsubseteq>\<^sub>p\<^sub>c Abstract exps pc"
  
  thus "pc \<sqsubseteq>\<^sub>p\<^sub>c Abstract (exps @ [e]) pc" using abstract_subsums
  by (rule_tac ?pc2.0="Abstract exps pc" in subsums_trans, 
      simp_all add : Abstract_append_one)
qed

lemma abstract_unsat_imp :
  assumes "\<not> sat (abstract \<phi> pc)"
  shows   "\<not> sat pc"
proof (rule ccontr, simp only : not_not)
  assume "sat pc"
  
  hence  "sat (abstract \<phi> pc)" using abstract_subsums sat_sub_by_sat by auto

  thus "False" using assms by simp
qed

lemma sat_imp_abstract_sat :
  assumes "sat pc"
  shows   "sat (abstract e pc)"
using assms abstract_subsums[of pc e] sat_sub_by_sat[of pc "abstract e pc"]
by simp



(* Exprime le fait que pour tout ensemble fini SV de variables symboliques et pour toute variable v,
   on peut trouver une variable symbolique (pour v) n'appartenant pas à SV et possèdant un indice 
   supérieur ou égal aux indices des variables symboliques de SV pour v.

   Cette variable symbolique est donc "fraîche pour SV".
 *)
lemma finite_symvars_imp_exists_greatest_symvar_gen2 :
  fixes SV :: "'a symvar set"
  assumes "finite SV"
  assumes "\<exists> sv \<in> SV. fst sv = v"
  shows   "\<exists> sv \<in> {sv \<in> SV. fst sv = v}. \<forall> sv' \<in> {sv \<in> SV. fst sv = v}. snd sv' \<le> snd sv"
proof -

  from assms(2) have "SV \<noteq> {}" by auto

  with assms(1) show ?thesis using `\<exists> sv \<in> SV. fst sv = v`
  proof (induct SV rule : finite_ne_induct)

    fix some_sv::"'a symvar"

    assume "\<exists> sv \<in> {some_sv}. fst sv = v"
         
    thus "\<exists> sv \<in> {sv \<in> {some_sv}. fst sv = v}. 
            \<forall> sv' \<in> {sv \<in> {some_sv}. fst sv = v}. snd sv' \<le> snd sv"
    by blast

  next

    fix sv::"'a symvar" 
    fix SV::"'a symvar set"

    assume "finite SV"

    assume "SV \<noteq> {}"

    assume "sv \<notin> SV"

    assume IH : "\<exists> sv \<in> SV. fst sv = v \<Longrightarrow> 
                 \<exists> sv \<in> {sv \<in> SV. fst sv = v}.
                    \<forall> sv'\<in> {sv \<in> SV. fst sv = v}. snd sv' \<le> snd sv"

    assume H : "\<exists> sv' \<in> insert sv SV. fst sv' = v"

    show "\<exists> sv' \<in>{sv' \<in> insert sv SV. fst sv' = v}. 
            \<forall> sv'' \<in> {sv' \<in> insert sv SV. fst sv' = v}. snd sv'' \<le> snd sv'"
    proof (case_tac "fst sv = v")
      assume "fst sv = v" 

      show ?thesis
      proof (case_tac "\<exists> sv \<in> SV. fst sv = v")

        assume "\<exists> sv \<in> SV. fst sv = v"

        then obtain max_sv where "max_sv \<in> {sv \<in> SV. fst sv = v}"
                           and   "\<forall> sv' \<in> {sv \<in> SV. fst sv = v}. snd sv' \<le> snd max_sv"
        using IH by blast
      
        show ?thesis
        proof (case_tac "snd sv \<le> snd max_sv")

          assume "snd sv \<le> snd max_sv"

          show ?thesis
          unfolding Ball_def Bex_def mem_Collect_eq
          proof (rule_tac ?x="max_sv" in exI, intro conjI allI impI)
            show "max_sv \<in> insert sv SV" using `max_sv \<in> {sv \<in> SV. fst sv = v}` by simp
          next
            show "fst max_sv = v" using `max_sv \<in> {sv \<in> SV. fst sv = v}` by simp
          next
            fix sv'

            assume "sv' \<in> insert sv SV \<and> fst sv' = v"

            thus "snd sv' \<le> snd max_sv"
            using `snd sv \<le> snd max_sv` `\<forall> sv' \<in> {sv \<in> SV. fst sv = v}. snd sv' \<le> snd max_sv`
            by (auto simp del : split_paired_All)
          qed
        
        next

          assume "\<not> snd sv \<le> snd max_sv"
        
          show ?thesis
          unfolding Ball_def Bex_def mem_Collect_eq
          proof (rule_tac ?x="sv" in exI, intro conjI allI impI)
            show "sv \<in> insert sv SV" by simp
          next
            show "fst sv = v" by (rule `fst sv = v`)
          next
            fix sv'

            assume "sv' \<in> insert sv SV \<and> fst sv' = v"

            thus   "snd sv' \<le> snd sv"
            unfolding insert_iff
            proof (elim conjE disjE)
              assume "sv' = sv" thus ?thesis by simp
            next
              assume "sv' \<in> SV"
              and    "fst sv' = v"

              hence "snd sv' \<le> snd max_sv" 
              using `\<forall> sv' \<in> {sv \<in> SV. fst sv = v}. snd sv' \<le> snd max_sv` by auto

              thus ?thesis using `\<not> snd sv \<le> snd max_sv` by simp
            qed
          qed
        qed
   
      next
        assume "\<not> (\<exists> sv \<in> SV. fst sv = v)"
        
        thus ?thesis using `fst sv = v` by (auto simp del : split_paired_Ex split_paired_All)
      qed
    
    next
      assume "fst sv \<noteq> v" 
      
      show ?thesis
      proof (case_tac "\<exists> sv \<in> SV. fst sv = v")
        
        assume "\<exists> sv \<in> SV. fst sv = v"
        
        then obtain max_sv where "max_sv \<in> {sv \<in> SV. fst sv = v}"
                           and   "\<forall> sv' \<in> {sv \<in> SV. fst sv = v}. snd sv' \<le> snd max_sv"
        using IH by blast

        thus ?thesis using `fst sv \<noteq> v` by (auto simp del : split_paired_All split_paired_Ex)
      
      next
        assume "\<not> (\<exists> sv \<in> SV. fst sv = v)"

        hence  "\<not> (\<exists> sv' \<in> insert sv SV. fst sv' = v)" using `fst sv \<noteq> v` by simp

        thus ?thesis using H by (elim notE)
      qed
    qed
  qed
qed




(* Le fait que l'ensemble des variables symboliques d'une pre_conf soit fini suffit à rendre cette
   pre_conf updatable. *)
lemma finite_pred_imp_se_updatable :
  assumes "finite (Bexp.vars (conjunct (\<pi> pc)))" (is "finite ?V")
  shows   "updatable pc"
unfolding updatable_def
proof (intro allI)

  fix v

  show "\<exists>sv. fst sv = v \<and> fresh_symvar sv pc"
  proof (case_tac "\<exists> sv \<in> ?V. fst sv = v")
    
    assume "\<exists> sv \<in> ?V. fst sv = v"

    then obtain max_sv where       "max_sv \<in> ?V"
                       and         "fst max_sv = v"
                       and   max : "\<forall>sv'\<in>{sv \<in> ?V. fst sv = v}. snd sv' \<le> snd max_sv"
    using assms finite_symvars_imp_exists_greatest_symvar_gen2[of "Bexp.vars (conjunct (\<pi> pc))" v]
    by blast

    show ?thesis
    proof (case_tac "snd max_sv \<le> store pc v")

      assume "snd max_sv \<le> store pc v"

      thus ?thesis
      using max
      unfolding fresh_symvar_def symvars_def Store.symvars_def symvar_of_def
      by (rule_tac ?x="(v,Suc (store pc v))" in exI)
         (auto simp del : split_paired_All)

    next

      assume "\<not> (snd max_sv \<le> store pc v)"

      thus ?thesis
      using max
      unfolding fresh_symvar_def symvars_def Store.symvars_def symvar_of_def
      by (rule_tac ?x="(v,Suc (snd max_sv))" in exI)
         (auto simp del : split_paired_All)
    qed

  next

    assume "\<not> (\<exists> sv \<in> ?V. fst sv = v)"

    thus ?thesis
    by (rule_tac ?x="(v, Suc (store pc v))" in exI)
       (auto simp add : fresh_symvar_def symvars_def Store.symvars_def symvar_of_def)
  qed
qed


(* Si le prédicat de cheminement d'une preconf contient un nombre fini de contraintes, contenant 
   elles meme un nombre fini de variables symboliques, alors le predicat de cheminement de l'abstraction de 
   cette preconf possede egalement un nombre fini de variables symboliques *)
lemma finite_exps_imp_finite_abstract_pred :
  assumes "finite (\<pi> pc)"
  assumes "\<forall> \<phi> \<in> \<pi> pc. finite (Bexp.vars \<phi>)"
  shows   "finite (Bexp.vars (conjunct (\<pi> (abstract e pc))))"
proof -
  from assms
  have "finite (\<pi> (abstract e pc))"
  and  "\<forall> \<phi> \<in> \<pi> (abstract e pc). finite (Bexp.vars \<phi>)"
  by (auto simp add : abstract_def)

  thus ?thesis using finite_conj by auto
qed




(*********************************************************************)
(******* Propriétés de SE ********************************************)
(*********************************************************************)


lemma se_Skip_eq :
  "se pc Skip pc' = (pc' = pc)"
by (simp add : se.simps)

lemma se_Break_eq :
  "se pc Break pc' = (pc' = pc)"
by (simp add : se.simps)

lemma se_Assume_eq :
  "se pc (Assume e) pc' = (pc' = \<lparr> store = store pc, \<pi> = \<pi> pc \<union> {adapt_bexp e (store pc)} \<rparr>)"
by (simp add : se.simps)

lemma se_Assign_eq :
  "se pc (Assign v e) pc' = 
  (\<exists> sv. fresh_symvar sv pc 
      \<and> fst sv = v 
      \<and>  pc' = \<lparr> store = (store pc)(v := snd sv), 
                 \<pi>     = insert (\<lambda>\<sigma>. \<sigma> sv = adapt_aexp e (store pc) \<sigma>) (\<pi> pc)\<rparr>)"
by (simp only : se.simps, blast)


lemma path_pred_of_se_Assume :
  assumes "se pc (Assume e) pc'"
  shows   "conjunct (\<pi> pc') = (\<lambda> \<sigma>. conjunct (\<pi> pc) \<sigma> \<and> adapt_bexp e (store pc) \<sigma>)"
using assms se_Assume_eq[of pc e pc'] by auto


lemma path_pred_of_se_Assign :
  assumes "se pc (Assign v e) pc'"
  shows   "\<exists> sv. conjunct (\<pi> pc') = (\<lambda> \<sigma>. conjunct (\<pi> pc) \<sigma> \<and> \<sigma> sv = adapt_aexp e (store pc) \<sigma>)"
proof -

  from assms
  obtain sv where "fresh_symvar sv pc"
            and   "fst sv = v"
            and   "pc' = \<lparr>store = (store pc)(v := snd sv), \<pi> = insert (\<lambda>\<sigma>. \<sigma> sv = adapt_aexp e (store pc) \<sigma>) (\<pi> pc) \<rparr>"
  using se_Assign_eq[of pc v e pc'] by blast

  thus ?thesis by (simp del : split_paired_Ex) blast

qed


lemma states_of_se_Skip :
  assumes "se pc Skip pc'"
  shows   "states pc' = states pc"
using assms by (simp add : se_Skip_eq)

lemma states_of_se_Break :
  assumes "se pc Break pc'"
  shows   "states pc' = states pc"
using assms by (simp add : se_Break_eq)

(* Les états de "SE pc (Assume \<phi>)" sont les états de pc validant \<phi> *)
lemma states_of_se_assume :
  assumes "se pc (Assume e) pc'"
  shows   "states pc' = {\<sigma> \<in> states pc. e \<sigma>}"
using assms se_Assume_eq[of pc e pc']
by (auto simp add : adapt_bexp_is_subst states_def)






lemma states_of_se_assign1 :
  assumes "se pc (Assign v e) pc'"
  assumes "\<sigma>' \<in> states pc'"
  shows   "\<exists> \<sigma> \<in> states pc. \<sigma>' = (\<sigma> (v := e \<sigma>))"
proof -

  (* on construit un état symbolique consistant avec \<sigma>' pour le store de ?pc' *)
  obtain \<sigma>\<^sub>s\<^sub>y\<^sub>m
  where 1 : "consistent \<sigma>' \<sigma>\<^sub>s\<^sub>y\<^sub>m (store pc')"
  and   2 : "conjunct (\<pi> pc') \<sigma>\<^sub>s\<^sub>y\<^sub>m"
  using assms(2) unfolding states_def by blast

  (* on construit ensuite un état \<sigma> consistant avec cet état symbolique, pour le store de pc *)
  then obtain \<sigma> 
  where 3 : "consistent \<sigma> \<sigma>\<^sub>s\<^sub>y\<^sub>m (store pc)"
  using ex_consistent_state1 by blast

  (* puisque l'état symbolique satisfait le prédicat de ?pc', qui n'est autre que celui
     de pc auquel on a ajouté une contrainte, il satisfait égalemen le prédicat de pc *)
  moreover
  obtain sv where 4 :"fresh_symvar sv pc"
            and   5 : "fst sv = v"
            and   6 : "pc' = \<lparr> store = (store pc)(v := snd sv), 
                               \<pi>    = insert (\<lambda>\<sigma>. \<sigma> sv = adapt_aexp e (store pc) \<sigma>) (\<pi> pc) \<rparr>"
  using assms(1) se_Assign_eq[of pc v e pc'] by blast

  hence "conjunct (\<pi> pc) \<sigma>\<^sub>s\<^sub>y\<^sub>m" using 2 by (auto simp add : )
  
  (* \<sigma> est donc un état de pc *)
  ultimately
  have "\<sigma> \<in> states pc" by (simp add : states_def) blast

  (* on montre maintenant que \<sigma>' est obtenu de \<sigma> par update de v *)
  moreover
  have "\<sigma>' = \<sigma> (v := e \<sigma>)"
  proof -
    have "\<sigma>' v = e \<sigma>" 
    proof -

      (* par définition de \<sigma>', \<sigma>\<^sub>s\<^sub>y\<^sub>m et de consistent, \<sigma>' associe à v la valeur qu'associe \<sigma>\<^sub>s\<^sub>y\<^sub>m
       à la variable symbolique associée à v en ?pc' *)
      have "\<sigma>' v = \<sigma>\<^sub>s\<^sub>y\<^sub>m (symvar_of v (store pc'))" 
      using 1 by (simp add : consistent_def)

      moreover
      (* qui n'est autre que la valeur de l'adaptation de e au store de pc en \<sigma>\<^sub>s\<^sub>y\<^sub>m *)
      have "\<sigma>\<^sub>s\<^sub>y\<^sub>m (symvar_of v (store pc')) = (adapt_aexp e (store pc)) \<sigma>\<^sub>s\<^sub>y\<^sub>m"
      using 2 5 6 by (force simp add : symvar_of_def )

      moreover
      (* qui est elle même égale à la valeur de e en \<sigma> *)
      have "(adapt_aexp e (store pc)) \<sigma>\<^sub>s\<^sub>y\<^sub>m = e \<sigma>"
      using 3 by (rule adapt_aexp_is_subst)
    
      ultimately
      (* donc \<sigma>' v est bien égal à e \<sigma> *)
      show ?thesis by simp
    qed

    moreover
    have "\<forall> x. x \<noteq> v \<longrightarrow> \<sigma>' x = \<sigma> x" 
    proof (intro allI impI)
      fix x

      assume "x \<noteq> v"

      moreover
      (* par définition de \<sigma>', \<sigma>\<^sub>s\<^sub>y\<^sub>m et de consistent, \<sigma>' associe aux variables autres que v les 
      valeurs qu'associe \<sigma>\<^sub>s\<^sub>y\<^sub>m aux variables symboliques correspondantes en ?pc'. Or ces variables 
      symboliques sont les mêmes qu'en pc. *) 
      hence  "\<sigma>' x = \<sigma>\<^sub>s\<^sub>y\<^sub>m (symvar_of x (store pc))"
      using 1 6 unfolding consistent_def symvar_of_def
      by (drule_tac ?x="x" in spec) simp

      moreover
      (* et sont donc égales aux valeurs associées à ces variables par \<sigma>, puisque \<sigma> et \<sigma>\<^sub>s\<^sub>y\<^sub>m 
         sont consistants pour le store de pc. *) 
      have   "\<sigma>\<^sub>s\<^sub>y\<^sub>m (symvar_of x (store pc)) = \<sigma> x"
      using 3 by (auto simp add : consistent_def)
      
      ultimately
      (* donc \<sigma>' x est bien égal à \<sigma> x pour tout x \<noteq> v *)
      show "\<sigma>' x = \<sigma> x" by simp
    qed

    ultimately
    show ?thesis by auto
  qed

  ultimately
  show ?thesis by (simp add : states_def) blast
qed



lemma states_of_se_assign2 :
  assumes "se pc (Assign v e) pc'"
  assumes "\<exists> \<sigma> \<in> states pc. \<sigma>' = \<sigma> (v := e \<sigma>)"
  shows   "\<sigma>' \<in> states pc'"
proof -

  (* on construit un état \<sigma> de pc tel que \<sigma>' est obtenu de \<sigma> par update de v *)
  obtain \<sigma> where "\<sigma> \<in> states pc" and "\<sigma>' = \<sigma> (v := e \<sigma>)" using assms(2) by blast

  (* puis on construit un état symbolique \<sigma>\<^sub>s\<^sub>y\<^sub>m consistant avec \<sigma> pour le store de pc *)
  then obtain \<sigma>\<^sub>s\<^sub>y\<^sub>m 
  where 1 : "consistent \<sigma> \<sigma>\<^sub>s\<^sub>y\<^sub>m (store pc)"
  and   2 : "conjunct (\<pi> pc) \<sigma>\<^sub>s\<^sub>y\<^sub>m" (* \<sigma>\<^sub>s\<^sub>y\<^sub>m satisfait donc le prédicat de pc *)
  unfolding states_def by blast

  (* puisque pc' est l'ES de pc, il existe une variable symbolique sv fraiche pour pc
     qui relie v à sa nouvelle valeur (dans pc') *)
  obtain sv where 3 : "fresh_symvar sv pc"
            and   4 : "fst sv = v"
            and   5 : "pc' = \<lparr> store = (store pc)(v := snd sv), 
                               \<pi>    = insert (\<lambda>\<sigma>. \<sigma> sv = adapt_aexp e (store pc) \<sigma>) (\<pi> pc) \<rparr>"
  using assms(1) se_Assign_eq[of pc v e pc'] by blast

  (* soit \<sigma>\<^sub>s\<^sub>y\<^sub>m' l'état symbolique obtenu de \<sigma>\<^sub>s\<^sub>y\<^sub>m en updatant sv par e \<sigma> *)
  let ?\<sigma>\<^sub>s\<^sub>y\<^sub>m' = "\<sigma>\<^sub>s\<^sub>y\<^sub>m (sv := e \<sigma>)"
  
  (* \<sigma>\<^sub>s\<^sub>y\<^sub>m' est consistant avec \<sigma>' pour le store de pc' *)
  have "consistent \<sigma>' ?\<sigma>\<^sub>s\<^sub>y\<^sub>m' (store pc')"
  using `\<sigma>' = \<sigma> (v := e \<sigma>)` 1 4 5 by (auto simp add : symvar_of_def consistent_def)
  
  (* de plus, \<sigma>\<^sub>s\<^sub>y\<^sub>m' satisfait le prédicat de pc' *)
  moreover
  have "conjunct (\<pi> pc') ?\<sigma>\<^sub>s\<^sub>y\<^sub>m'"
  proof -

    (* en effet, puisque sv est fraîche pour pc, elle l'est pour son prédicat,
       donc \<sigma>\<^sub>s\<^sub>y\<^sub>m' satisfait le prédicat de pc *)
    have "conjunct (\<pi> pc) ?\<sigma>\<^sub>s\<^sub>y\<^sub>m'" 
     proof -
      have "Bexp.fresh sv (conjunct (\<pi> pc))" 
      using 3 by (simp add : fresh_symvar_def symvars_def)

      thus ?thesis 
      using `conjunct (\<pi> pc) \<sigma>\<^sub>s\<^sub>y\<^sub>m`
            Bexp.fresh_is[of sv "conjunct (\<pi> pc)" \<sigma>\<^sub>s\<^sub>y\<^sub>m "e \<sigma>"]
      by simp
    qed

    (* de plus, \<sigma>\<^sub>s\<^sub>y\<^sub>m' associe à sv la valeur que prend l'adaptation de e au store de pc en \<sigma>\<^sub>s\<^sub>y\<^sub>m' *)
    moreover
    have "?\<sigma>\<^sub>s\<^sub>y\<^sub>m' sv = (adapt_aexp e (store pc)) ?\<sigma>\<^sub>s\<^sub>y\<^sub>m'"
    proof -
      have "Aexp.fresh sv (adapt_aexp e (store pc))" 
      using 3 symvars_of_adapt_aexp[of e "store pc"]
      by (auto simp add : fresh_symvar_def symvars_def)

      thus ?thesis using `consistent \<sigma> \<sigma>\<^sub>s\<^sub>y\<^sub>m (store pc)`
      by (simp add : Aexp.fresh_is adapt_aexp_is_subst)
    qed

    ultimately
    show ?thesis using 5 by (simp add: )
  qed

  (* donc \<sigma>' est bien un état de pc' *)
  ultimately
  show ?thesis unfolding states_def by blast
qed


lemma states_of_se_assign :
  assumes "se pc (Assign v e) pc'"
  shows   "states pc' = {\<sigma> (v := e \<sigma>) | \<sigma>. \<sigma> \<in> states pc}"
          (is "?L = ?R")
using assms states_of_se_assign1 states_of_se_assign2 by fast



(* L'exécution symbolique préserve la subsomption *)
lemma se_mono_for_sub :
  fixes pc1 :: "'var_type pre_conf"
  assumes "se pc1 l pc1'"
  assumes "se pc2 l pc2'"
  assumes "pc2 \<sqsubseteq>\<^sub>p\<^sub>c pc1"
  shows   "pc2' \<sqsubseteq>\<^sub>p\<^sub>c pc1'"
using assms
by (cases l, auto simp add : se_Skip_eq se_Break_eq subsums_def)
   (force simp add : states_of_se_assume, force simp add : states_of_se_assign)


lemma unsat_imp_se_unsat :
  assumes "se pc l pc'"
  assumes "\<not> sat pc"
  shows   "\<not> sat pc'"
using assms by cases (simp add :  sat_def)+


lemma se_sat_imp_sat :
  assumes "se pc l pc'"
  assumes "sat pc'"
  shows   "sat pc"
using assms by cases (auto simp add :  sat_def)+


lemma se_assume_subsumed :
  assumes "se pc (Assume e) pc'"
  shows   "pc' \<sqsubseteq>\<^sub>p\<^sub>c pc"
using assms states_of_se_assume[of pc e pc'] by (auto simp add : subsums_def)





lemma se_mono_for_states_eq :
  assumes "states pc1 = states pc2"
  assumes "se pc1 l pc1'"
  assumes "se pc2 l pc2'"
  shows   "states pc2' = states pc1'"
using assms
by (cases l)
   (simp add : states_of_se_Skip, 
    simp add : states_of_se_Break, 
    simp add : states_of_se_assume,
    simp add : states_of_se_assign)

lemma se_succs_states :
  assumes "se pc l pc1"
  assumes "se pc l pc2"
  shows   "states pc1 = states pc2"
using assms se_mono_for_states_eq[of pc pc l pc1 pc2] by simp



(* Etant donnés une pre_conf pc updatable et un label l, on peut toujours trouver une pre_conf pc'
   résultant de l'exécution symbolique de l depuis pc. *)
lemma updatable_imp_ex_se_suc :
  assumes "updatable pc"
  shows   "\<exists> pc'. se pc l pc'"
using assms
by (cases l, simp_all add : se_Skip_eq  se_Break_eq se_Assume_eq se_Assign_eq updatable_def)


(* Soit pc une configuration dont le predicat de cheminement contient un nombre fini 
   de contraintes possedant elles memes un nombre fini de variables symboliques,
   alors, pour toute abstraction de pc et tout label l, il existe une preconf en 
   relation par SE avec cette abstraction. *)
lemma finite_pred_and_symvars_imp_ex_se_rel_to_abstract :
  assumes "finite (\<pi> pc)"
  assumes "\<forall> \<phi> \<in> \<pi> pc. finite (Bexp.vars \<phi>)"
  shows   "\<exists> apc'. se (abstract e pc) l apc'"
using finite_exps_imp_finite_abstract_pred[OF assms]
      finite_pred_imp_se_updatable
      updatable_imp_ex_se_suc
by fast

(* Si pc est en relation par se (par un label l) avec une preconf pc' et qu'une abstraction 
   de pc est elle meme en relation par se avec apc', alors si pc' est satisfiable,
   apc' l'est egalement. *)
lemma se_sat_and_se_abstract_imp_sat :
  assumes "se pc l pc'"
  assumes "sat pc'"
  assumes "se (abstract e pc) l apc'"
  shows   "sat apc'"
using 
  se_mono_for_sub[OF assms(3) assms(1) abstract_subsums[of pc e]]
  sat_sub_by_sat[OF assms(2)]
by simp


(*lemma ex_sat_se_succ_imp :
  assumes "\<exists> pc'. se pc l pc' \<and> sat pc'"
  shows   "\<forall> pc'. se pc l pc' \<longrightarrow> sat pc'"
proof (intro allI impI)

  fix pc2

  assume "se pc l pc2"

  from assms
  obtain pc1 where "se pc l pc1"
             and   "sat pc1"
  by blast

  thus "sat pc2"
  using se_succs_states[OF `se pc l pc2` `se pc l pc1`] sat_iff by blast

qed*)


(* Soient pc et pc' des configurations en relation par se (par un label l). Si 
   les contraintes du predicat de cheminement de pc ne contiennent qu'un nombre fini
   de variables symboliques, alors les contraintes du predicat de pc' egalement. *)
lemma se_preserves_finiteness1 :
  assumes "finite_label l"
  assumes "se pc l pc'"
  assumes "\<forall> e \<in> \<pi> pc.  finite (Bexp.vars e)"
  shows   "\<forall> e \<in> \<pi> pc'. finite (Bexp.vars e)"
proof (cases l)

  case Skip  thus ?thesis using assms by (simp add : se_Skip_eq)

next

  case Break thus ?thesis using assms by (simp add : se_Break_eq)

next

  case (Assume e) thus ?thesis 
  using assms finite_vars_imp_finite_adapt_b
  by (auto simp add : se_Assume_eq finite_label_def)

next

  case (Assign v e) 
  
  thm se_Assign_eq[of pc v e pc']

  then obtain sv where "fresh_symvar sv pc"
                 and   "fst sv = v"
                 and   "pc' = \<lparr> store = (store pc)(v := snd sv),
                                \<pi>     = insert (\<lambda>\<sigma>. \<sigma> sv = adapt_aexp e (store pc) \<sigma>) (\<pi> pc)\<rparr>"
  using assms(2) se_Assign_eq[of pc v e pc'] by blast

  moreover
  have "finite (Bexp.vars (\<lambda>\<sigma>. \<sigma> sv = adapt_aexp e (store pc) \<sigma>))"
  proof -
    have "finite (Aexp.vars (\<lambda>\<sigma>. \<sigma> sv))" by (auto simp add : Aexp.vars_def)

    moreover
    have "finite (Aexp.vars (adapt_aexp e (store pc)))"
    using assms(1) Assign finite_vars_imp_finite_adapt_a
    by (auto simp add : finite_label_def)

    ultimately
    show ?thesis using finite_vars_of_a_eq by auto
  qed
  
  ultimately
  show ?thesis using assms by auto

qed


(* Soient pc et pc' des preconfs en relation par se (par le label l). Si le predicat 
   de cheminement de pc est fini alors celui de pc' l'est egalement *)
lemma se_preserves_finiteness2 :
  assumes "se pc l pc'"
  assumes "finite (\<pi> pc)"
  shows   "finite (\<pi> pc')"
using assms
apply (cases l)
  apply (simp add : se_Skip_eq)
  apply (simp add : se_Break_eq)
  apply (simp add : se_Assume_eq)
  apply (auto simp add : se_Assign_eq)
done


(* Etant donné un label l dont les expressions (arith. et bool.) potentielles possèdent un nombre 
   fini de variables et une pre_conf pc dont le prédicat de cheminement contient un nombre fini de 
   variables symboliques, toute pre_conf obtenue par l'exécution symbolique de l depuis pc a un 
   prédicat de cheminement contenant un nombre fini de variables symboliques.
*)
lemma se_preserves_finiteness3 :
  assumes "finite_label l"
  assumes "se pc l pc'"
  assumes "finite (Bexp.vars (conjunct (\<pi> pc)))"
  shows   "finite (Bexp.vars (conjunct (\<pi> pc')))"
proof (cases l)
  assume "l = Skip"  thus ?thesis using assms by (simp add : se_Skip_eq)
next
  assume "l = Break" thus ?thesis using assms by (simp add : se_Break_eq)
next
  fix e

  assume "l = Assume e"

  hence "finite (Bexp.vars (adapt_bexp e (store pc)))"
  using assms(1,3) by (simp add : finite_label_def finite_vars_imp_finite_adapt_b)

  hence "finite (Bexp.vars (conjunct (\<pi> pc)) \<union> Bexp.vars (adapt_bexp e (store pc)))"
  using assms(3) by auto

  moreover
  have "Bexp.vars (conjunct (\<pi> pc')) \<subseteq> Bexp.vars (conjunct (\<pi> pc)) \<union> Bexp.vars (adapt_bexp e (store pc))"
  using `l = Assume e` assms(2)
  by (simp only : path_pred_of_se_Assume vars_of_conj)
  
  ultimately
  show ?thesis by (rule rev_finite_subset)

next

  fix v e

  assume "l = Assign v e"

  then obtain sv where "fresh_symvar sv pc"
                 and   "fst sv = v"
                 and   "pc' = \<lparr> store = (store pc)(v := snd sv), 
                                \<pi>     = insert (\<lambda>\<sigma>. \<sigma> sv = adapt_aexp e (store pc) \<sigma>) (\<pi> pc) \<rparr>"
  using assms(2) se_Assign_eq[of pc v e pc'] by blast

  hence "conjunct (\<pi> pc') = (\<lambda> \<sigma>. conjunct (\<pi> pc) \<sigma> \<and> \<sigma> sv = adapt_aexp e (store pc) \<sigma>)"
  by (auto simp add : )

  hence "Bexp.vars (conjunct (\<pi> pc')) \<subseteq> Bexp.vars (conjunct (\<pi> pc)) \<union> Bexp.vars (\<lambda> \<sigma>. \<sigma> sv = adapt_aexp e (store pc) \<sigma>)"
  by (simp add : vars_of_conj)

  moreover
  have "finite (Aexp.vars (adapt_aexp e (store pc)))"
  using `l = Assign v e` assms(1) 
  by (simp add : finite_label_def finite_vars_imp_finite_adapt_a)

  hence  "finite (Aexp.vars (\<lambda>\<sigma>. \<sigma> sv) \<union> Aexp.vars (adapt_aexp e (store pc)))" 
  by (auto simp add : Aexp.vars_def)

  hence "finite (Bexp.vars (\<lambda> \<sigma>. \<sigma> sv = adapt_aexp e (store pc) \<sigma>))"
  using Bexp.vars_of_eq_a[of "\<lambda> \<sigma>. \<sigma> sv" "adapt_aexp e (store pc)"] by (rule rev_finite_subset)
  
  hence "finite (Bexp.vars (conjunct (\<pi> pc)) \<union> Bexp.vars (\<lambda> \<sigma>. \<sigma> sv = adapt_aexp e (store pc) \<sigma>))"
  using assms(3) by auto
 
  ultimately
  show ?thesis by (elim rev_finite_subset)

qed




(*********************************************************************)
(******* Propriétés de SE_star ***************************************)
(*********************************************************************)


lemma  [simp] :
  "se_star pc [] pc' = (pc' = pc)"
by (subst se_star.simps) auto


lemma se_star_Cons :
  "se_star pc1 (l # ls) pc2 = (\<exists> pc. se pc1 l pc \<and> se_star pc ls pc2)"
by (subst (1) se_star.simps) blast


lemma se_star_append :
  "se_star pc1 (ls1 @ ls2) pc2 = (\<exists> pc. se_star pc1 ls1 pc \<and> se_star pc ls2 pc2)"
using assms by (induct ls1 arbitrary : pc1) (simp_all add : se_star_Cons, blast)


lemma se_star_one :
  "se_star pc1 [l] pc2 = se pc1 l pc2"
using se_star_Cons[of pc1 l Nil pc2] by auto


lemma se_star_append_one :
  "se_star pc1 (ls @ [l]) pc2 = (\<exists> pc. se_star pc1 ls pc \<and> se pc l pc2)"
unfolding se_star_append se_star_one by (rule refl)


lemma se_star_mono_for_states_eq : 
  assumes "states pc1 = states pc2"
  assumes "se_star pc1 ls pc1'"
  assumes "se_star pc2 ls pc2'"
  shows   "states pc2' = states pc1'"
using assms
proof (induct ls arbitrary : pc1 pc2)
  case Nil thus ?case by (simp add : )
next

  case (Cons l ls pc1 pc2)
  show ?case
  proof -

    from Cons(3,4)
    obtain pc1'' pc2'' where 1 : "se pc1 l pc1''"
                       and   2 : "se_star pc1'' ls pc1'"
                       and   3 : "se pc2 l pc2''"
                       and   4 : "se_star pc2'' ls pc2'"
    unfolding se_star_Cons by blast

    moreover
    hence "states pc2'' = states pc1''" using Cons(2) by (elim se_mono_for_states_eq)

    ultimately
    show ?thesis using Cons(1) by auto
  qed
qed

lemma se_star_succs_states :
  assumes "se_star pc ls pc1"
  assumes "se_star pc ls pc2"
  shows   "states pc1 = states pc2"
using assms se_star_mono_for_states_eq[of pc pc ls pc1 pc2] by simp


lemma TMP1 :
  assumes "finite (\<pi> pc)"
  assumes "\<forall> e \<in> \<pi> pc. finite (Bexp.vars e)"
  assumes "finite_label l"
  shows   "\<exists> pc'. se pc l pc'"
sorry

lemma TMP2 :
  assumes "finite (\<pi> pc)"
  assumes "\<forall> e \<in> \<pi> pc. finite (Bexp.vars e)"
  assumes "finite_labels ls"
  shows   "\<exists> pc'. se_star pc ls pc'"
sorry



lemma se_star_preserves_finitness1 :
  assumes "finite_labels ls"
  assumes "se_star pc ls pc'"
  assumes "\<forall> e \<in> \<pi> pc.  finite (Bexp.vars e)"
  shows   "\<forall> e \<in> \<pi> pc'. finite (Bexp.vars e)"
using assms se_preserves_finiteness1
by (induct ls arbitrary : pc)
   (simp_all add :  finite_labels_Cons se_star_Cons, blast)

lemma se_star_preserves_finiteness2 :
  assumes "se_star pc ls pc'"
  assumes "finite (\<pi> pc)"
  shows   "finite (\<pi> pc')"
using assms se_preserves_finiteness2
by (induct ls arbitrary : pc) 
   (auto simp add :  se_star_Cons)

lemma se_star_preserves_finitness3 :
  assumes "finite_labels ls"
  assumes "se_star pc ls pc'"
  assumes "finite (Bexp.vars (conjunct (\<pi> pc)))"
  shows   "finite (Bexp.vars (conjunct (\<pi> pc')))"
using assms se_preserves_finiteness3
by (induct ls arbitrary : pc)
   (auto simp add:  se_star_Cons finite_labels_Cons)



                                 
lemma finite_pred_imp_se_star_updatable :
  assumes "finite_labels ls"
  assumes "se_star pc ls pc'"
  assumes "finite (Bexp.vars (conjunct (\<pi> pc)))"
  
  shows   "updatable pc'"
using assms se_star_preserves_finitness3[of ls pc pc']
by (simp add : finite_pred_imp_se_updatable)


(* Etant donné une séquence de labels finis et une pre_conf dont le prédicat de cheminement contient
   un nombre fini de variables symboliques, on peut toujours trouver une configuration pc' résultant
   de l'exécution symbolique de ls depuis pc *)
lemma 
  assumes "finite_labels ls"
  assumes "finite (Bexp.vars (conjunct (\<pi> pc)))"
  shows   "\<exists> pc'. se_star pc ls pc'"
using assms
proof (induction ls arbitrary : pc)
  case Nil show ?case by (simp add : )
next
  
  case (Cons l ls pc)
  show ?case
  proof -

    from Cons.prems(1)
    have "finite_label  l"
    and  "finite_labels ls" by (simp_all add : finite_labels_Cons)

    have "\<exists> pc1. se pc l pc1"
    proof -
      from Cons(3)
      have "updatable pc" by (rule finite_pred_imp_se_updatable)

      thus ?thesis by (rule updatable_imp_ex_se_suc)
    qed

    then obtain pc1 where "se pc l pc1" by blast

    moreover
    hence "finite (Bexp.vars (conjunct (\<pi> pc1)))"
    using `finite_label l` Cons(3) by (elim se_preserves_finiteness3)

    hence "\<exists> pc2. se_star pc1 ls pc2" 
    using `finite_labels ls` by (intro Cons.IH)

    ultimately
    show ?thesis by (simp only : se_star_Cons) blast 
  qed
qed




(* L'exécution symbolique d'une séquence de labels depuis une préconfiguration non-satisfiable 
   donne toujours une configuration non-satisfiable *)

lemma unsat_imp_se_star_unsat :
  assumes "se_star pc ls pc'"
  assumes "\<not> sat pc"
  shows   "\<not> sat pc'"
using assms
by (induct ls arbitrary : pc) 
   (simp add : , force simp add : se_star_Cons unsat_imp_se_unsat)



lemma se_star_sat_imp_sat :
  assumes "se_star pc ls pc'"
  assumes "sat pc'"
  shows   "sat pc"
using assms
by (induct ls arbitrary : pc)
   (simp add : , force simp add : se_star_Cons se_sat_imp_sat)







lemma se_star_mono_for_sub :
  assumes "se_star pc1 ls pc1'"
  assumes "se_star pc2 ls pc2'"
  assumes "pc2  \<sqsubseteq>\<^sub>p\<^sub>c pc1"
  shows   "pc2' \<sqsubseteq>\<^sub>p\<^sub>c pc1'"
using assms 
by (induct ls arbitrary : pc1 pc2) 
   (auto simp add :  se_star_Cons se_mono_for_sub)

lemma tmp :
  assumes "sat pc"
  shows   "states pc \<noteq> {}"
using assms ex_consistent_state1
unfolding sat_def states_def by blast


lemma se_models_eq_models_WP :
  assumes "se pc l pc'"
  shows   "pc' \<Turnstile>\<^sub>p\<^sub>c post \<longleftrightarrow> pc \<Turnstile>\<^sub>p\<^sub>c WP l post"
using assms 
      se_Skip_eq[of pc pc'] 
      se_Break_eq[of pc pc'] 
      states_of_se_assume[of pc _ pc'] 
      states_of_se_assign[of pc _ _ pc']
by (cases l) (simp, simp, (force simp add : models_def sem_def)+)




(*lemma feasible_Cons :
  assumes "feasible pc (l#ls)"
  shows   "\<exists> pc'. se pc l pc' \<and> sat pc' \<and> feasible pc' ls"
using assms se_star_sat_imp_sat
by (simp add : feasible_def se_star_Cons) blast

lemma
  assumes "\<exists> pc'. se pc l pc' \<and> sat pc' \<and> feasible pc' ls"
  shows   "feasible pc (l#ls)"
using assms unfolding feasible_def se_star_Cons by blast*)

lemma feasible_Cons :
  "feasible pc (l#ls) \<longleftrightarrow> (\<exists> pc'. se pc l pc' \<and> sat pc' \<and> feasible pc' ls)"
proof (intro iffI)
  case goal1 thus ?case
  using se_star_sat_imp_sat by (simp add : feasible_def se_star_Cons) blast
next
  case goal2 thus ?case using assms unfolding feasible_def se_star_Cons by blast
qed

lemma feasible_imp_sat :
  assumes "feasible pc ls"
  shows   "sat pc"
using assms se_star_sat_imp_sat by (auto simp add : feasible_def)

lemma feasible_appendE :
  assumes "feasible pc (ls1 @ ls2)"
  shows   "(\<exists> pc'. se_star pc ls1 pc' \<and> sat pc' \<and> feasible pc' ls2)"
using assms
proof (induct ls1 arbitrary : pc)

  case Nil thus ?case by (simp add :  feasible_imp_sat)

next

  case (Cons l ls1 pc)

  hence "feasible pc (l # (ls1 @ ls2))" by simp

  then obtain pc1 where "se pc l pc1"
                  and   "sat pc1"
                  and   "feasible pc1 (ls1 @ ls2)"
  using feasible_Cons by blast

  then obtain pc2 where "se_star pc1 ls1 pc2"
                  and   "sat pc2"
                  and   "feasible pc2 ls2"
  using Cons(1) by blast

  moreover
  have "se_star pc (l # ls1) pc2"
  using `se pc l pc1` `se_star pc1 ls1 pc2` by (auto simp add : se_star_Cons)

  ultimately
  show ?case by blast

qed

lemma feasible_appendI :
  assumes "(\<exists> pc'. se_star pc ls1 pc' \<and> sat pc' \<and> feasible pc' ls2)"
  shows   "feasible pc (ls1 @ ls2)"
using assms
proof (induct ls1 arbitrary : pc)
  
  case Nil thus ?case by (simp add : )

next

  case (Cons l ls1 pc)

  then obtain pc1 pc2 where "se pc l pc1"
                      and   "se_star pc1 ls1 pc2"
                      and   "sat pc2"
                      and   "feasible pc2 ls2"
  using se_star_Cons by blast
  
  moreover
  then obtain pc3 where "se_star pc2 ls2 pc3"
                  and   "sat pc3"
  unfolding feasible_def by blast

  ultimately
  have "se_star pc ((l # ls1) @ ls2) pc3"
  by (auto simp add : se_star_Cons se_star_append)

  thus ?case 
  using `sat pc3` unfolding feasible_def by blast
  
qed

lemma feasible_append :
  "feasible pc (ls1 @ ls2) \<longleftrightarrow> (\<exists> pc'. se_star pc ls1 pc' \<and> sat pc' \<and> feasible pc' ls2)"
using feasible_appendE feasible_appendI by fast

lemma subsums_imp_feasible :
  assumes "finite_labels ls"
  assumes "finite (\<pi> pc1)"
  assumes "finite (\<pi> pc2)"
  assumes "\<forall> e \<in> \<pi> pc1. finite (Bexp.vars e)"
  assumes "\<forall> e \<in> \<pi> pc2. finite (Bexp.vars e)"
  assumes "pc2 \<sqsubseteq>\<^sub>p\<^sub>c pc1"
  assumes "feasible pc2 ls"
  shows   "feasible pc1 ls"
using assms
proof (induct ls arbitrary : pc1 pc2)

  case Nil thus ?case by (simp add : feasible_def  sat_sub_by_sat)

next

  case (Cons l ls pc1 pc2)

  then obtain pc2' where "se pc2 l pc2'"
                   and   "sat pc2'"
                   and   "feasible pc2' ls"
  using feasible_Cons by blast

  obtain pc1' where "se pc1 l pc1'"
  using finite_conj[OF Cons(3,5)]
        finite_pred_imp_se_updatable
        updatable_imp_ex_se_suc
  by blast

  moreover
  hence "sat pc1'" 
  using  se_mono_for_sub[OF _ `se pc2 l pc2'` Cons(7)]
         sat_sub_by_sat[OF `sat pc2'`]
  by fast

  moreover
  have "feasible pc1' ls"
  proof -

    have "finite_label  l" 
    and  "finite_labels ls" using Cons(2) by (simp_all add : finite_labels_Cons) 

    have "finite (\<pi> pc1')" by (rule se_preserves_finiteness2[OF `se pc1 l pc1'` Cons(3)])
     
    moreover
    have "finite (\<pi> pc2')" by (rule se_preserves_finiteness2[OF `se pc2 l pc2'` Cons(4)])

    moreover
    have "\<forall>e\<in>\<pi> pc1'. finite (Bexp.vars e)" 
    by (rule se_preserves_finiteness1[OF `finite_label l` `se pc1 l pc1'` Cons(5)])

    moreover
    have "\<forall>e\<in>\<pi> pc2'. finite (Bexp.vars e)"
    by (rule se_preserves_finiteness1[OF `finite_label l` `se pc2 l pc2'` Cons(6)])
    
    moreover
    have "pc2' \<sqsubseteq>\<^sub>p\<^sub>c pc1'" 
    by (rule se_mono_for_sub[OF `se pc1 l pc1'` `se pc2 l pc2'` Cons(7)])
    
    ultimately
    show ?thesis using Cons(1) `feasible pc2' ls` `finite_labels ls` by fast
  qed

  ultimately
  show ?case by (auto simp add : feasible_Cons)
qed








lemma
  assumes "\<Sigma>2 \<subseteq> \<Sigma>1"
  shows   "{\<sigma>(v := n) | \<sigma>. \<sigma> \<in> \<Sigma>2} \<subseteq> {\<sigma>(v := n) | \<sigma>. \<sigma> \<in> \<Sigma>1}"
using assms by auto

thm states_def

term "states pc"

lemma
  fixes \<Sigma>1 :: "('a \<Rightarrow> integer) set"
  assumes "\<Sigma>1 \<noteq> {}"
  assumes "{\<sigma> (v := n) | \<sigma>. \<sigma> \<in> \<Sigma>1} \<subseteq> \<Sigma>1"
  assumes "\<Sigma>1 \<subseteq> \<Sigma>1a"
  shows   "{\<sigma> (v := n) | \<sigma>. \<sigma> \<in> \<Sigma>1a} \<subseteq> \<Sigma>1a"
using assms
apply auto
oops





find_theorems "feasible _ _"

find_unused_assms
end
