theory Store
imports Aexp Bexp
begin


(***************************)
(*** définitions de base ***)
(***************************)

(* Un store est une fonction associant des naturels aux variables (d'un programme) *)
type_synonym 'a store = "'a \<Rightarrow> nat"


(* On considère que la variable symbolique associée à une variable v par un store store est le
   couple (v,store v)  *)
definition
  symvar_of :: "'a \<Rightarrow> 'a store \<Rightarrow> 'a symvar"
where
  "symvar_of v store \<equiv> (v,store v)"

(*lemma
  "inj (\<lambda> v. symvar_of v store)"
by (simp add : inj_on_def  symvar_of_def)*)


(* On construit l'ensemble des variables symboliques associées par un store aux variables en 
   appliquant symvar_of à l'ensemble des variables. *)
definition 
  symvars :: "'a store \<Rightarrow> 'a symvar set" 
where
 "symvars store = (\<lambda> v. symvar_of v store) ` (UNIV::'a set)"



(*definition symvars2 ::
  "'a store \<Rightarrow> 'a set \<Rightarrow> 'a symvar set"
where
  "symvars2 store V = (\<lambda> v. symvar_of v store) ` V"*)


(* Une variable symbolique est dîte fraîche pour un store si elle ne figure pas dans les 
   l'ensemble image de ce store *)
definition 
  fresh_symvar :: "'var_type symvar \<Rightarrow> 'var_type store \<Rightarrow> bool" 
where
 "fresh_symvar sv store = (sv \<notin> symvars store)"

lemma fresh_is :
  "fresh_symvar sv store \<Longrightarrow> (symvar_of v store \<noteq> sv)"
by (auto simp add : fresh_symvar_def symvars_def)

lemma fresh_eq :
  "fresh_symvar sv store = (snd sv \<noteq> store (fst sv))"
unfolding fresh_symvar_def symvars_def symvar_of_def
by (subst surjective_pairing)  (auto simp del : prod.collapse)

(* On peut toujours trouver une variable symbolique fraiche pour un store *)
lemma ex_fresh_symvar :
  "\<exists> sv. fresh_symvar sv store"
by (rule_tac ?x="(v,Suc (store v))" in exI)
   (auto simp add : fresh_symvar_def symvars_def symvar_of_def)




(***************************************)
(*** Consistence des couples d'états ***)
(***************************************)

(* Un état \<sigma> (des variables) et un état symbolique \<sigma>\<^sub>s\<^sub>y\<^sub>m sont dits consistants avec un store si
   pour chaque variable v, \<sigma> associe à v la même valeur que celle associée par \<sigma>\<^sub>s\<^sub>y\<^sub>m à la variable
   symbolique associée à v par le store  *)
definition consistent :: 
  "'var_type state        \<Rightarrow> 
   'var_type symvar state \<Rightarrow> 
   'var_type store        \<Rightarrow> bool" 
where
  "consistent \<sigma> \<sigma>\<^sub>s\<^sub>y\<^sub>m store \<equiv> (\<forall> v. \<sigma>\<^sub>s\<^sub>y\<^sub>m (symvar_of v store) = \<sigma> v)"

(* Etant donne un etat \<sigma> du programme et un store, l'etat symbolique associant 
   a chaque variable symbolique la valeur associe par \<sigma> a la variable a laquelle est
   associe cette variable symbolique est consistant avec \<sigma> pour le store *)
lemma consistentI1 : 
  "consistent \<sigma> (\<lambda> sv. \<sigma> (fst sv)) store" 
by (simp add : consistent_def symvar_of_def)

(* Etant donne un etat symbolique \<sigma>\<^sub>s\<^sub>y\<^sub>m et un store, l'etat qui associe a chaque 
   variable du programme la valeur associe par \<sigma>\<^sub>s\<^sub>y\<^sub>m a la variable symbolique 
   associe a cette variable est consistant avec \<sigma>\<^sub>s\<^sub>y\<^sub>m pour ce store *)
lemma consistentI2 :
  "consistent (\<lambda> v. \<sigma>\<^sub>s\<^sub>y\<^sub>m (symvar_of v store)) \<sigma>\<^sub>s\<^sub>y\<^sub>m store"
by (simp add : consistent_def)

(* Etant donne un store, on peut toujours trouver un couple d'etats 
   consistants avec ce store *)
lemma ex_consistent_states :
  "\<exists> \<sigma> \<sigma>\<^sub>s\<^sub>y\<^sub>m. consistent \<sigma> \<sigma>\<^sub>s\<^sub>y\<^sub>m store"
by (auto simp add : consistent_def)

(* Etant donne un etat symbolique \<sigma>\<^sub>s\<^sub>y\<^sub>m, on peut toujours trouver un etat
   \<sigma> tel que \<sigma> et \<sigma>\<^sub>s\<^sub>y\<^sub>m soient consistants avec ce store *)
lemma ex_consistent_state1 :
  "\<exists> \<sigma>. consistent \<sigma> \<sigma>\<^sub>s\<^sub>y\<^sub>m store"
by (rule_tac ?x="(\<lambda> v. \<sigma>\<^sub>s\<^sub>y\<^sub>m (symvar_of v store))" in exI) (simp add : consistent_def)

(* Etant donne un etat \<sigma>, on peut toujours trouver un etat symbolique \<sigma>\<^sub>s\<^sub>y\<^sub>m
   tel que \<sigma> et \<sigma>\<^sub>s\<^sub>y\<^sub>m soient consistants avec ce store *)
lemma ex_consistent_state2 :
  "\<exists> \<sigma>\<^sub>s\<^sub>y\<^sub>m. consistent \<sigma> \<sigma>\<^sub>s\<^sub>y\<^sub>m store"
by (rule_tac ?x="\<lambda> sv. \<sigma> (fst sv)" in exI) (simp add : consistent_def symvar_of_def)


corollary ex_consistent_state_update :
  "\<exists> \<sigma>'. consistent \<sigma>' \<sigma>\<^sub>s\<^sub>y\<^sub>m (store(v := exp))"
by (rule ex_consistent_state1)

lemma consistent_is :
  "consistent \<sigma> \<sigma>\<^sub>s\<^sub>y\<^sub>m store = (\<sigma> = (\<lambda> v. \<sigma>\<^sub>s\<^sub>y\<^sub>m (symvar_of v store)))"
by (auto simp add : consistent_def)

lemma consistent_eq1 :
  "consistent \<sigma> \<sigma>\<^sub>s\<^sub>y\<^sub>m store = (\<sigma> = (\<lambda> v. \<sigma>\<^sub>s\<^sub>y\<^sub>m (symvar_of v store)))"
by (auto simp add : consistent_def)


lemma consistent_eq2 :
  "consistent \<sigma> \<sigma>\<^sub>s\<^sub>y\<^sub>m store = (\<forall> sv \<in> symvars store. \<sigma>\<^sub>s\<^sub>y\<^sub>m sv = \<sigma> (fst sv))"
by (auto simp add : consistent_def symvars_def symvar_of_def)






(**********************************)
(*** Substitution pour les Aexp ***)
(**********************************)

definition adapt_aexp :: 
  "'var_type aexp  \<Rightarrow> 
   'var_type store \<Rightarrow> 'var_type symvar aexp" 
where
  "adapt_aexp exp store = (\<lambda> \<sigma>. exp (\<lambda> x. \<sigma> (symvar_of x store)))"

(* adapt_aexp consiste à substituer, dans une expression arithmétique portant sur les variables 
   (du programme), les occurrences de ces variables par les variables symboliques qui leur sont 
   associées par le store. adapt_aexp est donc une substitution, pour des couples d'états 
   consistants *)
lemma adapt_aexp_is_subst :
  "consistent \<sigma> \<sigma>\<^sub>s\<^sub>y\<^sub>m store \<Longrightarrow> (adapt_aexp exp store) \<sigma>\<^sub>s\<^sub>y\<^sub>m = exp \<sigma>"
by (simp add : consistent_is adapt_aexp_def)


(* Les variables symboliques figurant dans l'adaptation d'une aexp à un store proviennent de ce 
   store *)
lemma symvars_of_adapt_aexp :
    "Aexp.vars (adapt_aexp e store) \<subseteq> symvars store"
(is "Aexp.vars ?e' \<subseteq> ?SV")
unfolding subset_iff
proof (intro allI impI)
  fix sv

  assume "sv \<in> Aexp.vars ?e'"

  then obtain \<sigma>\<^sub>s\<^sub>y\<^sub>m val 
  where "?e' (\<sigma>\<^sub>s\<^sub>y\<^sub>m (sv := val)) \<noteq> ?e' \<sigma>\<^sub>s\<^sub>y\<^sub>m"
  by (simp add : Aexp.vars_def, blast)

  hence "(\<lambda> x. (\<sigma>\<^sub>s\<^sub>y\<^sub>m (sv := val)) (symvar_of x store)) \<noteq> (\<lambda> x. \<sigma>\<^sub>s\<^sub>y\<^sub>m (symvar_of x store))"
  proof (intro notI)
    assume "(\<lambda>x. (\<sigma>\<^sub>s\<^sub>y\<^sub>m(sv := val)) (symvar_of x store)) = (\<lambda>x. \<sigma>\<^sub>s\<^sub>y\<^sub>m (symvar_of x store))"
    hence  "?e' (\<sigma>\<^sub>s\<^sub>y\<^sub>m (sv := val)) = ?e' \<sigma>\<^sub>s\<^sub>y\<^sub>m" by (simp add : adapt_aexp_def)
    thus   False using `?e' (\<sigma>\<^sub>s\<^sub>y\<^sub>m (sv := val)) \<noteq> ?e' \<sigma>\<^sub>s\<^sub>y\<^sub>m` by (elim notE)
  qed

  then obtain v 
  where "(\<sigma>\<^sub>s\<^sub>y\<^sub>m (sv := val)) (symvar_of v store) \<noteq> \<sigma>\<^sub>s\<^sub>y\<^sub>m (symvar_of v store)" 
  by blast

  hence "sv = symvar_of v store" by (case_tac "sv = symvar_of v store", simp_all)

  thus "sv \<in> symvars store" by (simp add : symvars_def)
qed


(* Si v est une variable d'une expression arithmétique, alors la variable symbolique qui lui est 
   associée par store est elle même une variable de l'adaptation de cette expression au store. *)
lemma in_vars_imp_symvar_in_vars_a :
  assumes "v \<in> Aexp.vars e"
  shows   "symvar_of v store \<in> Aexp.vars (adapt_aexp e store)"
      (is "?sv \<in> Aexp.vars ?e'")
proof -

  obtain \<sigma> val where "e (\<sigma> (v := val)) \<noteq> e \<sigma>" 
  using assms(1) unfolding Aexp.vars_def by blast  
  
  moreover
  have "(\<lambda>va. ((\<lambda>sv. \<sigma> (fst sv))(?sv := val)) (symvar_of va store)) = (\<sigma>(v := val))"
  by (auto simp add : symvar_of_def)  

  ultimately
  show ?thesis
  unfolding Aexp.vars_def mem_Collect_eq
  using consistentI1[of \<sigma> store] 
        consistentI2[of "(\<lambda>sv. \<sigma> (fst sv))(?sv:= val)" store]
  by (rule_tac ?x="\<lambda>sv. \<sigma> (fst sv)" in exI, rule_tac ?x="val" in exI) 
      (simp add : adapt_aexp_is_subst)
qed

(* Réciproquement, si une variable symbolique appartient aux variables de l'adaptation d'une 
   expression arithmétique à un store, alors la variable à laquelle elle est associée est 
   elle-même une variable de l'expression de départ. *)
lemma symvar_in_vars_a_imp_in_vars : 
  assumes "(symvar_of v store) \<in> Aexp.vars (adapt_aexp e store)" (is "?sv \<in> Aexp.vars ?e'")
  shows   "v \<in> Aexp.vars e"
proof -

  obtain \<sigma>' val where 1 : "?e' (\<sigma>' (?sv := val)) \<noteq> ?e' \<sigma>'" 
  using assms unfolding Aexp.vars_def by blast

  moreover
  have "(\<lambda>va. (\<sigma>'(symvar_of v store := val)) (symvar_of va store)) = (\<lambda>v. \<sigma>' (symvar_of v store))(v := val)"
  by (auto simp add : symvar_of_def)

  ultimately
  show ?thesis
  unfolding Aexp.vars_def mem_Collect_eq
  using consistentI2[of "\<sigma>' (?sv := val)" store] consistentI2[of \<sigma>' store]
  by (simp add : adapt_aexp_is_subst) blast
qed


(* On peut donc calculer, à partir de l'ensemble des variables d'une expression, l'ensemble des 
   variables de son adaptation à un store. *)
lemma adapt_aexp_vars :
  "Aexp.vars (adapt_aexp e store) = (\<lambda> v. symvar_of v store) ` Aexp.vars e"
  (is "Aexp.vars ?e' = ?R")
unfolding set_eq_iff image_def mem_Collect_eq Bex_def
proof (intro allI iffI)
  fix sv

  assume a : "sv \<in> Aexp.vars ?e'"

  hence  c : "sv \<in> symvars store" 
  using symvars_of_adapt_aexp by auto
                       
  show   "\<exists> v. v \<in> Aexp.vars e \<and> sv = symvar_of v store"
  proof (rule_tac ?x="fst sv" in exI, intro conjI)
  
    have "symvar_of (fst sv) store = sv" 
    using c by (auto simp add : symvars_def symvar_of_def)

    thus "fst sv \<in> Aexp.vars e" 
    using symvar_in_vars_a_imp_in_vars[of "fst sv" store e] a by auto
 
  next
    show "sv = symvar_of (fst sv) store"
    using c by (force simp add : symvar_of_def symvars_def)
  qed

next
  fix sv

  assume "\<exists> v. v \<in> Aexp.vars e \<and> sv = symvar_of v store"

  then obtain v where "v \<in> Aexp.vars e" "sv = symvar_of v store" by blast

  thus "sv \<in> Aexp.vars (adapt_aexp e store)" 
  using in_vars_imp_symvar_in_vars_a[of v e store] by simp
qed 

(* On a donc que si une expression arithmetique possede un nombre fini de 
   variables, alors son adaptation a un store possede egalement un nombre fini 
   de variables (symboliques).

   On utilisera ce lemme pour montrer comment construire des preconfs dont le 
   predicat de cheminement contient un nombre fini de variables symboliques, 
   ce qui est une condition suffisante pour qu'il existe une variable symbolique 
   fraiche pour une preconf. 
*)
lemma finite_vars_imp_finite_adapt_a :
  assumes "finite (Aexp.vars e)"
  shows   "finite (Aexp.vars (adapt_aexp e store))"
unfolding adapt_aexp_vars using assms by auto


(**********************************)
(*** Substitution pour les Bexp ***)
(**********************************)

(* idem pour les expressions booleenes *)

definition adapt_bexp :: 
  "'var_type bexp \<Rightarrow> 
   'var_type store \<Rightarrow> 'var_type symvar bexp" 
where
  "adapt_bexp exp store = (\<lambda> \<sigma>. exp (\<lambda> x. \<sigma> (symvar_of x store)))"


lemma adapt_bexp_is_subst :
  "consistent \<sigma> \<sigma>\<^sub>s\<^sub>y\<^sub>m store \<Longrightarrow> (adapt_bexp exp store) \<sigma>\<^sub>s\<^sub>y\<^sub>m = exp \<sigma>"
by (simp add : consistent_is adapt_bexp_def)


lemma symvars_of_adapt_bexp :
    "Bexp.vars (adapt_bexp exp store) \<subseteq> symvars store"
(is "Bexp.vars ?e' \<subseteq> ?SV")
proof
  fix sv
  assume "sv \<in> Bexp.vars ?e'"

  then obtain \<sigma>\<^sub>s\<^sub>y\<^sub>m val 
  where "?e' (\<sigma>\<^sub>s\<^sub>y\<^sub>m (sv := val)) \<noteq> ?e' \<sigma>\<^sub>s\<^sub>y\<^sub>m"
  by (simp add : Bexp.vars_def, blast)

  hence "(\<lambda> x. (\<sigma>\<^sub>s\<^sub>y\<^sub>m (sv := val)) (symvar_of x store)) \<noteq> (\<lambda> x. \<sigma>\<^sub>s\<^sub>y\<^sub>m (symvar_of x store))"
  by (auto simp add : adapt_bexp_def)

  hence "\<exists> v. (\<sigma>\<^sub>s\<^sub>y\<^sub>m (sv := val)) (symvar_of v store) \<noteq> \<sigma>\<^sub>s\<^sub>y\<^sub>m (symvar_of v store)" by force

  then obtain v where "(\<sigma>\<^sub>s\<^sub>y\<^sub>m (sv := val)) (symvar_of v store) \<noteq> \<sigma>\<^sub>s\<^sub>y\<^sub>m (symvar_of v store)" by blast

  hence "sv = symvar_of v store" by (case_tac "sv = symvar_of v store", simp_all)

  thus "sv \<in> symvars store" by (simp add : symvars_def)
qed



lemma in_vars_imp_symvar_in_vars_b :
  assumes "v \<in> Bexp.vars e"
  shows   "symvar_of v store \<in> Bexp.vars (adapt_bexp e store)"
      (is "?sv \<in> Bexp.vars ?e'")
proof -

  obtain \<sigma> val where A : "e (\<sigma> (v := val)) \<noteq> e \<sigma>" 
  using assms(1) unfolding Bexp.vars_def by blast  
  
  moreover
  have H : "(\<lambda>va. ((\<lambda>sv. \<sigma> (fst sv))(?sv := val)) (symvar_of va store)) = (\<sigma>(v := val))"
  by (auto simp add : symvar_of_def)  

  ultimately
  show ?thesis
  unfolding Bexp.vars_def mem_Collect_eq
  using consistentI1[of \<sigma> store] 
        consistentI2[of "(\<lambda>sv. \<sigma> (fst sv))(?sv:= val)" store]
  by (rule_tac ?x="\<lambda>sv. \<sigma> (fst sv)" in exI, rule_tac ?x="val" in exI) 
     (simp add : adapt_bexp_is_subst)
qed



lemma symvar_in_vars_b_imp_in_vars : 
  assumes "(symvar_of v store) \<in> Bexp.vars (adapt_bexp e store)" (is "?sv \<in> Bexp.vars ?e'")
  shows   "v \<in> Bexp.vars e"
proof -

  obtain \<sigma>' val where 1 : "?e' (\<sigma>' (?sv := val)) \<noteq> ?e' \<sigma>'" 
  using assms unfolding Bexp.vars_def by blast

  moreover
  have "(\<lambda>va. (\<sigma>'(?sv := val)) (symvar_of va store)) = (\<lambda>v. \<sigma>' (symvar_of v store))(v := val)"
  by (auto simp add : symvar_of_def)

  ultimately
  show ?thesis
  unfolding Bexp.vars_def mem_Collect_eq
  using consistentI2[of "\<sigma>' (?sv := val)" store] consistentI2[of \<sigma>' store]
  by (simp add : adapt_bexp_is_subst) blast
qed


lemma adapt_bexp_vars :
  "Bexp.vars (adapt_bexp e store) = (\<lambda> v. symvar_of v store) ` Bexp.vars e"
  (is "Bexp.vars ?e' = ?R")
unfolding set_eq_iff image_def mem_Collect_eq Bex_def
proof (intro allI iffI)
  fix sv

  assume a : "sv \<in> Bexp.vars ?e'"

  hence  c : "sv \<in> symvars store" 
  using symvars_of_adapt_bexp by auto
                       
  show   "\<exists> v. v \<in> Bexp.vars e \<and> sv = symvar_of v store"
  proof (rule_tac ?x="fst sv" in exI, intro conjI)
  
    have "symvar_of (fst sv) store = sv" 
    using c by (auto simp add : symvars_def symvar_of_def)

    thus "fst sv \<in> Bexp.vars e" 
    using symvar_in_vars_b_imp_in_vars[of "fst sv" store e] a by auto
 
  next
    show "sv = symvar_of (fst sv) store"
    using c by (force simp add : symvar_of_def symvars_def)
  qed

next
  fix sv

  assume "\<exists> v. v \<in> Bexp.vars e \<and> sv = symvar_of v store"

  then obtain v where "v \<in> Bexp.vars e" "sv = symvar_of v store" by blast

  thus "sv \<in> Bexp.vars ?e'" 
  using in_vars_imp_symvar_in_vars_b[of v e store] by simp
qed 

lemma finite_vars_imp_finite_adapt_b :
  assumes "finite (Bexp.vars e)"
  shows   "finite (Bexp.vars (adapt_bexp e store))"
unfolding adapt_bexp_vars using assms by auto


(*lemma
  "adapt_aexp (\<lambda> \<sigma>. \<sigma> x + \<sigma> y) store = (\<lambda> \<sigma>. \<sigma> (symvar_of x store) + \<sigma> (symvar_of y store))"
apply (simp add : adapt_aexp_def)

oops



lemma symvars_of_update :
  "symvars (store (v := n)) \<subseteq> symvars store \<union> {(v,n)}"
by (auto simp add : symvars_def symvar_of_def)

(* Si l'ensemble de definition d'un store est fini, son ensemble d'images l'est également *)
lemma finite_vars_imp_finite_symvars :
  fixes    store :: "'var_type store"
  assumes "finite (UNIV::'var_type set)"
  shows   "finite (symvars store)"
using assms by (auto simp add : symvars_def)

lemma symvar_of_not_surj :
  "\<not> surj (\<lambda>v. symvar_of v store)"
unfolding surj_def split_paired_All symvar_of_def
apply (intro notI)
apply (drule_tac ?x="v" in spec)
apply (drule_tac ?x="Suc (store v)" in spec)
apply force
done

lemma symvars_psubset_UNIV :
  "symvars store \<subset> (UNIV::'var_type symvar set)"
by (auto simp add : symvars_def symvar_of_not_surj)

    



















lemma
  assumes "v \<notin> Aexp.vars e"
  shows   "symvar_of v store \<notin> Aexp.vars (adapt_aexp e store)"
      (is "?sv \<notin> Aexp.vars ?e'")
proof -
  have "\<forall> \<sigma> val. e (\<sigma> (v := val)) = e \<sigma>" using assms by (auto simp add : Aexp.vars_def)

  show ?thesis
  proof (intro notI, simp only : Aexp.vars_def mem_Collect_eq, elim exE) 

    fix \<sigma> val

    assume "?e' (\<sigma> (?sv := val)) \<noteq> ?e' \<sigma>"

    thm consistentI2[of \<sigma> store]
    thm consistentI2[of "(\<sigma>(symvar_of v store := val))" store]

    moreover
    have H : "(\<lambda>va. (\<sigma>(symvar_of v store := val)) (symvar_of va store)) = (\<lambda>v. \<sigma> (symvar_of v store))(v := val)"
    by (auto simp add : symvar_of_def)

    ultimately
    have "e ((\<lambda>v. \<sigma> (symvar_of v store))(v := val)) \<noteq> e (\<lambda>v. \<sigma> (symvar_of v store))"
    using consistentI2[of \<sigma> store] consistentI2[of "(\<sigma>(symvar_of v store := val))" store]
    by (simp add : adapt_aexp_is_subst)
    
    hence "v \<in> Aexp.vars e" unfolding Aexp.vars_def by blast

    thus False using assms(1) by (elim notE)
  qed
qed







lemma
  assumes "finite (Aexp.vars exp)"
  shows   "finite (Aexp.vars (adapt_aexp exp store))"
using assms by (simp add : tmp3)

lemma symvar_of_inj :
  "inj (\<lambda> v. symvar_of v store)"
unfolding symvar_of_def
by (rule BNF_FP_Base.inj_on_convol_ident)

lemma "finite (symvars (store::'a store)) \<longleftrightarrow> finite (UNIV ::'a set)"
apply (intro iffI)
  defer
  apply (simp only : symvars_def)
  apply (rule finite_imageI, assumption)

  apply (subst (asm) symvars_def)
  using symvar_of_inj[of store] apply (elim finite_imageD, assumption)
done*)



find_unused_assms



end
