theory SubRel
imports Graph
begin 

(**************************************************************************************************)
(***************************  RELATION DE SUBSOMPTION  ********************************************)
(**************************************************************************************************)

(* Définition d'une relation de subsomption : 
   - (descendant,ancestor) \<in> subs <-> "descendant" subsumé par "ancestor",
   - pour l'instant, une relation de subsomption n'est définie que pour des graphes rouges (ie. dont
     les sommets sont indicés) *)
locale sub_rel =
  fixes subs :: "(('a \<times> nat) \<times> ('a \<times> nat)) set" (structure)


(* Le premier élément du couple est le subsumé *)
abbreviation subsumee :: 
  "(('a \<times> nat) \<times> ('a \<times> nat)) \<Rightarrow> ('a \<times> nat)" 
where 
  "subsumee sub \<equiv> fst sub"



(* le second le subsumeur *)
abbreviation subsumer :: 
  "(('a \<times> nat) \<times> ('a \<times> nat)) \<Rightarrow> ('a \<times> nat)" 
where 
  "subsumer sub \<equiv> snd sub"



(* ensemble des subsumés *)
definition subsumees :: 
  "(('a \<times> nat) \<times> ('a \<times> nat)) set \<Rightarrow> ('a \<times> nat) set" 
where 
  "subsumees subss \<equiv> subsumee ` subss"

lemma in_subsumees_conv :
  "d \<in> subsumees subs = (\<exists> a. (d,a) \<in> subs)"
by (force simp add : subsumees_def)

lemma subsumees_conv :
  "subsumees subs = {d. \<exists> a. (d,a) \<in> subs}"
by (force simp add : subsumees_def)

lemma sub_surj_pairing : 
  "sub = (subsumee sub, subsumer sub)"
by auto

lemma subsumee_in_subsumees :
  assumes "sub \<in> subs"
  shows   "subsumee sub \<in> subsumees subs"
using assms sub_surj_pairing[of sub] 
unfolding subsumees_conv mem_Collect_eq
by (rule_tac ?x="subsumer sub" in exI) simp
  



(* ensemble des subsumeurs *)
definition subsumers :: 
  "(('a \<times> nat) \<times> ('a \<times> nat)) set \<Rightarrow> ('a \<times> nat) set" 
where 
  "subsumers subss \<equiv> subsumer ` subss"

lemma subsumers_conv :
  "subsumers subs = {a. \<exists> d. (d,a) \<in> subs}"
by (force simp add : subsumers_def)

lemma subsumer_in_subsumers :
  assumes "sub \<in> subs"
  shows   "subsumer sub \<in> subsumers subs"
using assms sub_surj_pairing[of sub]
unfolding subsumers_conv mem_Collect_eq
by (rule_tac ?x="subsumee sub" in exI) simp




(* ensemble des sommets figurant dans la relation de subsomption *)
abbreviation vertices ::
  "(('a \<times> nat) \<times> ('a \<times> nat)) set \<Rightarrow> ('a \<times> nat) set" 
where
  "vertices subs \<equiv> subsumers subs \<union> subsumees subs"

lemma subsumee_in_vertices :
  assumes "sub \<in> subs"
  shows   "subsumee sub \<in> vertices subs"
using assms subsumee_in_subsumees by fast

lemma subsumer_in_vertices :
  assumes "sub \<in> subs"
  shows   "subsumer sub \<in> vertices subs"
using assms subsumer_in_subsumers by fast





(* Une relation de subsomption est bien formée si :
   - il n'existe aucun sommet étant à la fois subsumé et subsumeur,
   - si deux sommets (indicés) sont en relation, alors leurs versions non-indicées sont égales
     (ie. subsomption uniquement entre même sommets de la partie noire),
   - un sommet subsumé l'est par un unique subsumeur.

   Chacune de ces propriétés sera requise pour la suite.
 *)
locale wf_sub_rel = sub_rel +
  assumes inter_empty       : "subsumers subs \<inter> subsumees subs = {}"
  assumes sub_imp_same_verts : "sub \<in> subs \<Longrightarrow> fst (subsumee sub) = fst (subsumer sub)"
  assumes subsumed_by_one   : "\<forall> d \<in> subsumees subs. \<exists>! a. (d,a) \<in> subs"
begin
  lemmas wf_sub_rel = inter_empty sub_imp_same_verts subsumed_by_one

  lemma (in wf_sub_rel) subsumed_by_two_imp : 
    assumes "(d,a1) \<in> subs"
    assumes "(d,a2) \<in> subs"
  
    shows   "a1 = a2"
  using assms wf_sub_rel unfolding Ball_def subsumees_conv by blast
end

lemma empty_sub_rel_wf :
  "wf_sub_rel {}"
by (auto simp add : wf_sub_rel_def subsumers_conv subsumees_conv)







(* une relation de subsomption "appartient" (peut être que le terme est mal choisi) à un graphe si
   les sommets mis en relation sont tous des sommets du graphe  *)
locale sub_rel_of = rgraph + sub_rel +
  assumes related_are_verts : "vertices subs \<subseteq> Graph.vertices g"
begin

  lemmas sub_rel_of = related_are_verts

  lemma sub_rel_of_imp_subsumee_vert :
    assumes "(d,a) \<in> subs"
    shows   "d \<in> Graph.vertices g"
  using assms related_are_verts unfolding subsumers_conv subsumees_conv by fast
  
  lemma sub_rel_of_imp_subsumer_vert :
    assumes "(d,a) \<in> subs"
    shows   "a \<in> Graph.vertices g"
  using assms related_are_verts unfolding subsumers_conv subsumees_conv by fast

  lemma trancl_sub_rel_of :
     "sub_rel_of g (subs\<^sup>+)"
  unfolding sub_rel_of_def subset_iff
  proof (intro allI impI)
    fix v
    assume "v \<in> vertices (subs\<^sup>+)"

    hence "\<exists> v'. (v,v') \<in> subs\<^sup>+ \<or> (v',v) \<in> subs\<^sup>+" 
    by (auto simp add : vertices_def subsumers_conv subsumees_conv)

    then obtain v' where "(v,v') \<in> subs\<^sup>+ \<or> (v',v) \<in> subs\<^sup>+" by blast

    then obtain v'' where "(v,v'') \<in> subs \<or> (v'',v) \<in> subs"
    using tranclD[of v v' subs] tranclD2[of v' v subs] by blast

    thus "v \<in> Graph.vertices g" 
    using sub_rel_of unfolding subsumees_conv subsumers_conv by (auto simp del : split_paired_Ex)
  qed
end


lemma empty_sub_rel_of :
  "sub_rel_of g {}"
by (auto simp add : sub_rel_of_def subsumers_conv subsumees_conv)







(* De manière générale, on est intéressé par des relations de subsomptions bien formées et 
   "appartenant" au graphe considéré. Etant donné un graphe g, on dira qu'une telle relation est
   bien formée pour g. *)
locale wf_sub_rel_of = rgraph + sub_rel +
  assumes sub_rel_of : "sub_rel_of g subs"
  assumes wf_sub_rel : "wf_sub_rel subs"
begin
  
  lemmas wf_sub_rel_of = sub_rel_of wf_sub_rel

  lemma in_trancl_imp :
    assumes "(d,a) \<in> subs\<^sup>+"
    shows   "(d,a) \<in> subs"
  proof -

    obtain v1 where "(d,v1) \<in> subs" 
              and   "(v1,a) \<in> subs\<^sup>*"
    using assms tranclD[of d a subs] by blast

    hence "v1 = a \<or> v1 \<noteq> a \<and> (v1,a) \<in> subs\<^sup>+" by (elim rtranclD)

    thus ?thesis
    proof (elim disjE)
      assume "v1 = a" thus ?thesis using `(d,v1) \<in> subs` by simp
    next
      assume "v1 \<noteq> a \<and> (v1,a) \<in> subs\<^sup>+"

      then obtain v2 where "(v1,v2) \<in> subs"
                     and   "(v2,a)  \<in> subs\<^sup>*"
      using tranclD[of v1 a subs] by blast

      hence "v1 \<in> subsumers subs \<inter> subsumees subs" using `(d,v1) \<in> subs` 
      unfolding subsumers_conv subsumees_conv by blast

      thus ?thesis using wf_sub_rel wf_sub_rel.inter_empty by auto
    qed
  qed

  lemma arcs_empty_imp_empty :
    assumes "arcs g = {}"
    shows   "subs = {}"
  proof (rule ccontr)
    assume "subs \<noteq> {}"

    then obtain sub where "sub \<in> subs" by blast

    hence "subsumee sub \<in> Graph.vertices g"
    and   "subsumer sub \<in> Graph.vertices g"
    using subsumee_in_subsumees subsumer_in_subsumers sub_rel_of
    unfolding sub_rel_of_def by fast+

    hence "subsumee sub = subsumer sub" using assms(1) by (simp add : vertices_conv)

    hence "subsumers subs \<inter> subsumees subs \<noteq> {}"
    using `sub \<in> subs` unfolding subsumees_conv subsumers_conv
    by (subst (asm) (1 2 3) sub_surj_pairing) (auto simp del : split_paired_Ex)

    thus False using wf_sub_rel unfolding wf_sub_rel_def by simp
  qed
                                                            
end


lemma empty_wf_sub_rel_of :
  "wf_sub_rel_of g {}"
using empty_sub_rel_of empty_sub_rel_wf by (auto simp add : wf_sub_rel_of_def)





(**************************************************************************************************)
(***************************  CHEMINS ET SOUS-CHEMINS AVEC SUBSOMPTIONS  **************************)
(**************************************************************************************************)




(***************************  Séquence d'arcs consistante avec subsomptions  **********************)

(* On définit la consistance d'une séquence d'arcs avec une relation de subsomption.

   L'idée est que, si as1 et as2 sont deux séquences d'arcs consistantes (au sens de Graph.thy, ie 
   sans subsomptions) telles que as1 termine en d et as2 démarre en a, alors si la subsomption 
   (d,a) appartient à la relation subs, as1 @ as2 est consistante avec subs. La subsomption (d,a)
   fait le lien entre as1 et as2.

   En particulier, la séquence vide est un sous chemin allant de v1 à v2 si :
   - v1 = v2, ou
   - v1 est subsumé par v2.

   Note : Par contre, avec cette définition de cas, étant donné une relation de subsomption subs contenant 
   les couples (a,c) et (c,b) mais pas (a,b), on n'a pas cas a [] b subs. 

   En pratique, on avait vu qu'il valait mieux de toute manière ne pas accepter de subsumer a par c 
   si c est lui même subsumé par b, ce qui est respecté directement par la définition de cas. Peut 
   être qu'il vaudrait mieux avoir une définition plus générale, et imposer cette contrainte dans
   wf_sub_rel.

   Pour cela, il faudrait plus de lemmes sur les relations de subsomptions et définir cas autrement
   que par une fonction récursive (par exemple, cas2 ne termine pas nécessairement). Pourrait être 
   un prédicat inductif.

   -- fin de la note


   Dans la suite, on retrouve des lemmes correspondant à ceux de Graph.thy plus des lemmes, en plus 
   des lemmes spécifiques à la présence des subsomptions.

 *)

fun cas :: "'a \<Rightarrow> 'a arc list \<Rightarrow> 'a \<Rightarrow> ('a \<times> 'a) set \<Rightarrow> bool" where
  "cas l [] l' subs = (l = l'  \<or> (l,l') \<in> subs\<^sup>+)"
| "cas l (a#as) l' subs = ((l = src a \<or> (l,src a) \<in> subs\<^sup>+) \<and> cas (tgt a) as l' subs)"


lemma cas_imp_cas_w_subs :
  assumes "Graph.cas v1 as v2"
  shows   "cas v1 as v2 subs"
using assms by (induct as arbitrary : v1, auto)


lemma (in wf_sub_rel) acas_imp_dcas :
  assumes "(d,a) \<in> subs"
  assumes "cas a as v subs"
 
  shows   "cas d as v subs"
proof (cases as)
  
  assume "as = []"

  hence "a = v \<or> (a,v) \<in> subs\<^sup>+" using assms(2) by simp

  thus ?thesis
  proof (elim disjE)
    assume "a = v" thus ?thesis using assms(1) `as = []` by auto
  next
    assume "(a,v) \<in> subs\<^sup>+"

    then obtain x where "(a,x) \<in> subs" using tranclD[of a v subs] by blast

    hence "subsumers subs \<inter> subsumees subs \<noteq> {}" 
    using assms(1) by (auto simp only : subsumers_conv subsumees_conv)

    thus ?thesis using inter_empty by simp
  qed

next

  fix arc as'

  assume 1 : "as = arc # as'"

  hence "(a = src arc \<or> (a,src arc) \<in> subs\<^sup>+)"
  and   2 : "cas (tgt arc) as' v subs"
  using assms(2) by auto

  thus ?thesis
  proof (elim disjE)
    assume "a = src arc" thus ?thesis  using assms(1) 1 2 by auto
  next
    assume "(a,src arc) \<in> subs\<^sup>+"

    then obtain x where "(a,x) \<in> subs" using tranclD[of a "src arc" subs] by blast

    hence "subsumers subs \<inter> subsumees subs \<noteq> {}" 
    using assms(1) by (auto simp only : subsumers_conv subsumees_conv)

    thus ?thesis using inter_empty by simp
  qed
qed




lemma cas_Un :
  assumes "cas v as v'  subs1"
  shows   "cas v as v' (subs1 \<union> subs2)"
using assms 
by (induct as arbitrary : v, auto simp add : trancl_mono)




lemma cas_one :
  "cas v1 [a] v2 subs = ((v1 = src a \<or> (v1,src a) \<in> subs\<^sup>+) \<and> 
                         (tgt a = v2 \<or> (tgt a,v2) \<in> subs\<^sup>+))"
by simp

lemma cas_Cons :
  "cas v1 (a#as) v2 subs = (cas v1 [a] (tgt a) subs \<and> cas (tgt a) as v2 subs)"
by auto

lemma cas_append_one :
  "cas v1 (as @ [a]) v2 subs = (cas v1 as (src a) subs \<and> cas (src a) [a] v2 subs)"
by (induct as arbitrary : v1, auto)

lemma cas_appendE :
  assumes "cas v1 (as1 @ as2) v2 subs"
  shows   "\<exists> v. cas v1 as1 v subs \<and> cas v as2 v2 subs"
using assms 
by (induct as1 arbitrary : v1) 
   (simp_all del : split_paired_Ex split_paired_All, blast)

lemma cas_appendI :
  assumes "\<exists> v. cas v1 as1 v subs \<and> cas v as2 v2 subs"
  shows   "cas v1 (as1 @ as2) v2 subs"
using assms
proof (induct as1 arbitrary : v1) 
  case (Nil v1)
  show ?case
  proof -
    
    obtain v where "cas v1 [] v subs" 
             and   "cas v as2 v2 subs"
    using Nil(1) by blast

    hence "v1 = v \<or> (v1,v) \<in> subs\<^sup>+" by simp

    thus ?thesis
    proof (elim disjE)
      assume "v1 = v" thus ?thesis using `cas v as2 v2 subs` by simp
    next
      assume "(v1,v) \<in> subs\<^sup>+"

      thus ?thesis 
      using `cas v as2 v2 subs`
      by (cases as2, simp_all) (intro disjI2, force)+
    qed
  qed

next

  case (Cons arc as1 v1)
  show ?case
  proof -

    obtain v where "cas v1 (arc # as1) v subs" 
             and   "cas v as2 v2 subs"
    using Cons(2) by blast

    moreover
    hence "v1 = src arc \<or> (v1,src arc) \<in> subs\<^sup>+"
    and   "cas (tgt arc) as1 v subs" by auto

    ultimately
    have "cas (tgt arc) (as1 @ as2) v2 subs" using Cons(1) by auto

    thus ?thesis 
    using `v1 = src arc \<or> (v1,src arc) \<in> subs\<^sup>+` by auto
  qed
qed

lemma cas_append :
  "cas v1 (as1 @ as2) v2 subs = (\<exists> v. cas v1 as1 v subs \<and> cas v as2 v2 subs)"
using cas_appendE cas_appendI by fast
  


lemma (in wf_sub_rel) cas_appendI :

  assumes "(d,a) \<in> subs"
  assumes "cas v1 as1 d  subs"
  assumes "cas a  as2 v2 subs"
    
  shows   "cas v1 (as1 @ as2) v2 subs"
using assms acas_imp_dcas by (subst cas_append, blast+)


lemma (in sub_rel_of) cas_in_sub_rel_of_imp_verts :
  assumes "cas v1 as v2 subs"
  assumes "set as \<subseteq> arcs g"
  assumes "v1 \<in> Graph.vertices g" 
  
  shows   "v2 \<in> Graph.vertices g"
using assms
proof (induct as arbitrary : v1)

  case (Nil v1)
  show ?case
  using Nil(1)
  proof (simp only : cas.simps, elim disjE)
    assume "v1 = v2" thus ?thesis using Nil(3) by simp
  next
    assume "(v1,v2) \<in> subs\<^sup>+"
    thus ?thesis using trancl_sub_rel_of unfolding sub_rel_of_def subsumers_conv by blast
  qed

next

  case (Cons arc as v1)
  thus ?case
  by (subst (asm) cas_Cons, subst (asm) cas_one) 
     (elim conjE disjE, auto simp add : tgt_is_vert)
qed




(***************************  Sous-chemins avec subsomptions  **********************)

(* Un sous-chemin d'un graphe avec une relation de subsomption est une séquence d'arcs de ce graphe
   consistante avec la relation de subsomption.

   Comme dans Graph.thy, on a besoin d'imposer que le sommet de départ (v1) soit bien un sommet de g.
   Autrement, on pourrait avoir "subpath g v1 [] v2 subs" pour des v1 et v2 n'étant pas des sommets 
   de g.

 *)


definition subpath :: 
  "('a,'b) rgraph_scheme \<Rightarrow> 
   'a                    \<Rightarrow> 
   'a arc list           \<Rightarrow> 
   'a                    \<Rightarrow> 
   ('a \<times> 'a) set        \<Rightarrow> bool" 
where
  "subpath g v1 p v2 subs \<equiv>  cas v1 p v2 subs \<and> v1 \<in> Graph.vertices g \<and> set p \<subseteq> arcs g"


lemma Nil_sp :
  "subpath g v1 [] v2 subs = (v1 \<in> Graph.vertices g \<and> (v1 = v2 \<or> (v1,v2) \<in> subs\<^sup>+))"
by (auto simp add : subpath_def)

lemma Nil_spI :
  assumes "v \<in> Graph.vertices g"
  shows   "subpath g v [] v subs"
using assms unfolding subpath_def by auto

lemma Nil_sp_from_root :
  "subpath g (root g) [] (root g) subs"
by (simp add : subpath_def root_is_vert)


lemma empty_graph_imp_sp_empty :
  assumes "arcs g = {}"
  assumes "subpath g v1 p v2 subs"
  shows   "p = []"
using assms unfolding subpath_def by auto

lemma sp_Un :
  assumes "subpath g v1 p v2 subs1"
  shows   "subpath g v1 p v2 (subs1 \<union> subs2)"
using assms unfolding subpath_def by (simp add : cas_Un)


(* un sous-chemin issu de l'ancêtre est également un sous-chemin issu du descendant *)
lemma (in wf_sub_rel_of) asp_imp_dsp :
  assumes "(d,a) \<in> subs"
  assumes "subpath g a p v subs"

  shows   "subpath g d p v subs"
using assms unfolding subpath_def
apply (elim conjE)
apply (intro conjI)
  using wf_sub_rel apply (simp add : wf_sub_rel.acas_imp_dcas)
  using sub_rel_of apply (simp add : sub_rel_of.sub_rel_of_imp_subsumee_vert)
  apply assumption
done


lemma (in sub_rel_of) sp_one :
  shows   "subpath g v1 [a] v2 subs = 
           ((v1 = src a \<or> (v1,src a) \<in> subs\<^sup>+) \<and> a \<in> arcs g \<and> (tgt a = v2 \<or> (tgt a,v2) \<in> subs\<^sup>+))"
  (is "?L = ?R")
proof (intro iffI)
  assume "?L" thus "?R" using cas_one[of v1 a v2 subs] by (simp add : subpath_def)
next
  assume "?R" 
  show   "?L"
  unfolding subpath_def
  proof (intro conjI)
    case goal1 show ?case using `?R` by simp
  next
    case goal2 show ?case using `?R`
    proof (elim conjE disjE)
      case goal1 thus ?case by (simp add : vertices_conv) blast
    next
      case goal2 thus ?case by (simp add : vertices_conv) blast
    next
      case goal3 
      thus ?case using trancl_sub_rel_of by (simp only : sub_rel_of_def subsumees_conv) blast
    next
      case goal4 
      thus ?case using trancl_sub_rel_of by (simp only : sub_rel_of_def subsumees_conv) blast
    qed
  next
    case goal3 show ?case using `?R` by simp
  qed
qed

lemma arc_sp_eq :
  "subpath g (src a) [a] (tgt a) subs = (a \<in> arcs g)"
by (auto simp add : subpath_def vertices_conv)


lemma (in sub_rel_of) sp_Cons :
  shows   "subpath g v1 (a # as) v2 subs = 
          ((v1 = src a  \<or> (v1,src a) \<in> subs\<^sup>+) \<and> a \<in> arcs g \<and> subpath g (tgt a) as v2 subs)"
  (is "?L = ?R")
proof (intro iffI)
  assume "?L" thus "?R" by (auto simp add : subpath_def tgt_is_vert)
next
  assume "?R"
  thus   "?L"
  unfolding subpath_def
  proof (intro conjI)
    case goal1 thus ?case by (auto simp add : subpath_def)
  next
    case goal2 thus ?case
    proof (elim conjE disjE)
      case goal1 thus ?case by (force simp add : vertices_conv)
    next
      case goal2 
      thus ?case using trancl_sub_rel_of by (simp only : sub_rel_of_def subsumees_conv) blast
    qed
  next
    case goal3 thus ?case by simp
  qed
qed

lemma (in wf_sub_rel_of) sp_Cons :
  "subpath g v1 (a # as) v2 subs = 
    ((v1 = src a \<or> (v1,src a) \<in> subs) \<and> a \<in> arcs g \<and> subpath g (tgt a) as v2 subs)"
using sub_rel_of.sp_Cons[OF sub_rel_of]
      wf_sub_rel_of.in_trancl_imp[OF ]
      wf_sub_rel_of
unfolding wf_sub_rel_of_def by fast



lemma fst_of_sp_is_vert :
  assumes "subpath g v1 p v2 subs"
  shows   "v1 \<in> Graph.vertices g"
using assms by (simp add : subpath_def)

(* Si la relation de subsomption appartient bien au graphe considéré, alors cette définition de 
   subpath garantit bien que le sommet d'arrivée (v2) est un sommet du graphe.  *)
lemma (in sub_rel_of) lst_of_sp_is_vert :
  assumes "subpath g v1 p v2 subs"
  shows   "v2 \<in> Graph.vertices g"
using assms
proof (induction p arbitrary : v1)
  case Nil thus ?case using trancl_sub_rel_of
  unfolding subpath_def sub_rel_of_def subsumers_conv cas.simps
  by (elim conjE disjE) (simp, blast)
next
  case (Cons a p v1)
  show ?case
  proof -

    have "subpath g (tgt a) p v2 subs" using Cons(2) by (simp add : sp_Cons)

    thus ?thesis by (rule Cons.IH)
  qed
qed


lemma (in sub_rel_of) sp_append_one :
  "subpath g v1 (as @ [a]) v2 subs =
  (subpath g v1 as (src a) subs \<and> a \<in> arcs g  \<and> (tgt a = v2 \<or> (tgt a, v2) \<in> subs\<^sup>+))"
unfolding subpath_def
by (auto simp add :  cas_append_one src_is_vert tgt_is_vert sub_rel_of_imp_subsumer_vert)

lemma (in wf_sub_rel) sp_appendI1 :
  assumes "subpath g v1 as1 v2 subs"
  assumes "subpath g v2 as2 v3 subs"

  shows   "subpath g v1 (as1 @ as2) v3 subs"
using assms unfolding subpath_def by (simp only : cas_append, fastforce)

lemma (in wf_sub_rel) sp_appendI2 :
  assumes "(d,a) \<in> subs"
  assumes "subpath g v1 as1 d  subs"
  assumes "subpath g a  as2 v2 subs"

  shows   "subpath g v1 (as1 @ as2) v2 subs"
using assms unfolding subpath_def by (simp add : cas_appendI)


lemma (in wf_sub_rel_of) sp_appendE :
  assumes "subpath g v1 (as1 @ as2) v2 subs"

  shows   "\<exists> v. subpath g v1 as1 v subs \<and> subpath g v as2 v2 subs"
proof -
  from assms wf_sub_rel_of
  have   1 : "cas v1 (as1 @ as2) v2 subs"
  and    2 : "v1 \<in> Graph.vertices g"
  and    3 : "set (as1 @ as2) \<subseteq> arcs g"
  unfolding subpath_def by simp_all

  moreover
  then obtain v where 4 : "cas v1 as1 v  subs"
                and   5 : "cas v  as2 v2 subs"
  using assms wf_sub_rel_of cas_append[of v1 as1 as2 v2 subs] by blast

  moreover
  from 2 3 4
  have "v \<in> Graph.vertices g" 
  using assms wf_sub_rel_of by (auto simp add:  sub_rel_of.cas_in_sub_rel_of_imp_verts)

  ultimately
  show ?thesis unfolding subpath_def by fastforce
qed



lemma (in wf_sub_rel_of) sp_append :
  "subpath g v1 (as1 @ as2) v2 subs = (\<exists> v. subpath g v1 as1 v subs \<and> subpath g v as2 v2 subs)"
using wf_sub_rel_of sp_appendE wf_sub_rel.sp_appendI1 by blast


(* Etant donnés :
   - un graphe g,
   - une relation de subsomption bien formée pour g,
   - un élément (d,a) de cette relation tel qu'aucun arc de g n'est issu de d,
   - un sous-chemin de g (avec subsomp\<degree>) p non-vide issu de d,
   alors p est également un sous-chemin de g issu de a.
 *)
lemma (in wf_sub_rel_of) non_empty_subsumee_sp_imp_subsumer_sp :
  assumes "(d,a) \<in> subs"
  assumes "subpath g d p v subs"
  assumes "out_arcs g d = {}"
  assumes "p \<noteq> []"

  shows   "subpath g a p v subs"
proof -

  obtain arc p' where "p = arc # p'" using assms(4) by (subst (asm) neq_Nil_conv, blast)

  hence "d = src arc \<or> (d, src arc) \<in> subs\<^sup>+"
  and   "arc \<in> arcs g"
  and   "subpath g (tgt arc) p' v subs"
  using wf_sub_rel_of assms(2) by (simp_all add : sub_rel_of.sp_Cons)

  thus ?thesis
  proof (elim disjE)
    assume "d = src arc"

    hence  "arc \<in> out_arcs g d" using `arc \<in> arcs g` by (simp add : out_arcs_def)

    thus ?thesis using assms(3) by simp
  next

    assume "(d,src arc) \<in> subs\<^sup>+"

    hence  "(d,src arc) \<in> subs" by (rule in_trancl_imp)

    hence "src arc = a"
    using wf_sub_rel_of assms(1) by (simp add : wf_sub_rel.subsumed_by_two_imp)

    thus ?thesis 
    using `p = arc # p'` `arc \<in> arcs g` `subpath g (tgt arc) p' v subs` wf_sub_rel_of
    by (simp add : sub_rel_of.sp_Cons)
  qed
qed


(* Etant donnés :
   - un graphe g,
   - une relation de subsomptions bien formée pour g,
   - un élément (d,a) de cette relation tel qu'aucun arc de g n'est issu de d,
   - un sous-chemin (avec subsomp\<degree>) p non-vide de g issu de d,
   alors p est vide ou bien p est également un sous-chemin de g issu de a.
 *)
lemma (in wf_sub_rel_of) sp_from_subsumee :
  assumes "(d,a) \<in> subs"
  assumes "subpath g d p v subs"
  assumes "out_arcs g d = {}"               

  shows   "p = [] \<or> subpath g a p v subs"

proof (cases p)

  assume "p = []" thus ?thesis by (intro disjI1)

next
                              
  fix arc p'

  assume "p = arc # p'"

  thus ?thesis using assms by (simp add : non_empty_subsumee_sp_imp_subsumer_sp)
qed



lemma (in sub_rel_of) sp_from_de_empty :
  assumes "v1 \<notin> subsumees subs"
  assumes "out_arcs g v1 = {}"
  assumes "subpath g v1 p v2 subs"

  shows   "p = []"
proof (rule ccontr)

  assume "p \<noteq> []"

  then obtain arc p' where "p = arc # p'" using neq_Nil_conv[of p] by blast

  hence "src arc = v1 \<or> (v1,src arc) \<in> subs\<^sup>+" and "arc \<in> arcs g" 
  using assms(3) by (auto simp add: sp_Cons)

  thus "False"
  proof (elim disjE)
    assume "src arc = v1" hence "arc \<in> out_arcs g v1" using `arc \<in> arcs g` by (intro out_arcsI)

    thus ?thesis using assms(2) by simp
  next
    assume "(v1,src arc) \<in> subs\<^sup>+" 
    thus ?thesis using assms(1) tranclD unfolding subsumees_conv by fastforce
  qed
qed




lemma (in sub_rel_of) sp_through_de_decomp :
  assumes "tgt a \<notin> subsumees subs"
  assumes "out_arcs g (tgt a) = {}"
  assumes "subpath g v1 p v2 subs"
  assumes "a \<in> set p"

  shows   "\<exists> p'. p = p' @ [a] \<and> a \<notin> set p'"
using assms(3) assms(4)
proof (induction p arbitrary : v1)

  case (Nil v1) thus ?case by simp

next

  case (Cons arc p v1)
  show ?case
  proof -

    have "subpath g (tgt arc) p v2 subs" using Cons(2) by (simp_all add : sp_Cons)

    have "a = arc \<or> (a \<noteq> arc \<and> a \<in> set p)" using Cons(3) by auto

    thus ?thesis
    proof (elim disjE)

      assume "a = arc"

      moreover
      hence "p = []" 
      using `subpath g (tgt arc) p v2 subs` assms(1) assms(2) by (simp add : sp_from_de_empty)

      ultimately
      show ?thesis by auto

    next

      assume "a \<noteq> arc \<and> a \<in> set p"

      moreover
      then obtain p' where "p = p' @ [a]"   
                     and   "a \<notin> set p'" 
      using `subpath g (tgt arc) p v2 subs` Cons.IH by blast

      ultimately
      show ?thesis by auto
    qed
  qed
qed


abbreviation subpaths :: 
  "(('a \<times> nat),'b) rgraph_scheme \<Rightarrow> 
   (('a \<times> nat) \<times> ('a \<times> nat)) set \<Rightarrow> ('a \<times> nat) arc list set"
where
  "subpaths g subs \<equiv> {p. \<exists> v1 v2. subpath g v1 p v2 subs}"

lemma empty_graph_subpaths :
  assumes "arcs g = {}"
  shows   "subpaths g subs = {[]}"
unfolding set_eq_iff singleton_iff
proof (intro allI iffI)
  case (goal1 as) 
  thus ?case using assms empty_graph_imp_sp_empty by auto
next
  case (goal2 as)
  thus ?case using Nil_sp_from_root by blast
qed


lemma
  assumes "subpath g v as v1 subs"
  assumes "subpath g v as v2 subs"
  shows   "v1 = v2"
using assms
apply (induction as arbitrary : v)
  apply (simp add : Nil_sp)
oops


definition subpath_from ::
  "(('a \<times> nat),'b) rgraph_scheme \<Rightarrow>
   ('a \<times> nat)                    \<Rightarrow>
   ('a \<times> nat) arc list           \<Rightarrow>
   (('a \<times> nat) \<times> ('a \<times> nat)) set \<Rightarrow> bool"
where
  "subpath_from g v as subs \<equiv> \<exists> v'. subpath g v as v' subs"

lemma empty_graph_sp_from_eq_Nil :
  assumes "arcs g = {}"
  assumes "subpath_from g v as subs"
  shows   "as = []"
using assms empty_graph_imp_sp_empty unfolding subpath_from_def by fast

abbreviation subpaths_from :: 
  "(('a \<times> nat),'b) rgraph_scheme \<Rightarrow> 
   ('a \<times> nat)                    \<Rightarrow> 
   (('a \<times> nat) \<times> ('a \<times> nat)) set \<Rightarrow> ('a \<times> nat) arc list set"
where
  "subpaths_from g v subs \<equiv> {p. subpath_from g v p subs}"

lemma
  assumes "arcs g = {}"
  shows   "subpaths_from g v subs = {[]}"
unfolding set_eq_iff singleton_iff
proof (intro allI iffI)
  case (goal1 as) 
  thus ?case using assms empty_graph_imp_sp_empty by (auto simp add : subpath_from_def)
next
  case (goal2 as)
  moreover
  have "v \<in> Graph.vertices g" sorry
  ultimately
  show ?case using Nil_sp[of g v v subs] 
  unfolding subpath_from_def by blast
qed

(*lemma
  assumes "arcs g = {}"
  assumes "v \<in> Graph.vertices g"
  shows   "subpaths_from g v subs = {[]}"
unfolding set_eq_iff singleton_iff
proof (intro allI iffI)

  case (goal1 as)
  thus ?case using assms empty_graph_imp_sp_empty by (force simp add : subpaths_from_def)

next

  case (goal2 as)
  thus ?case 
  using assms(2) Nil_spI[of v g subs]
  unfolding subpaths_from_def mem_Collect_eq
  by blast

oops*)

(* chemins *)

definition path ::
  "(('a \<times> nat),'b) rgraph_scheme \<Rightarrow>
   ('a \<times> nat) arc list           \<Rightarrow>
   ('a \<times> nat)                    \<Rightarrow>
   (('a \<times> nat) \<times> ('a \<times> nat)) set \<Rightarrow> bool"
where
  "path g p v subs \<equiv> subpath g (root g) p v subs"

lemma empty_graph_imp_p_empty :
  assumes "arcs g = {}"
  assumes "path g p v subs"
  shows   "p = []"
using assms unfolding path_def by (rule empty_graph_imp_sp_empty)

lemma Nil_path :
  "path g [] (root g) subs"
unfolding path_def  by (rule Nil_sp_from_root)



lemma path_conv :
  "path g as v subs = subpath_from g (root g) as subs"
oops

thm path_append

lemma (in wf_sub_rel_of) path_append :
  "path g (as1 @ as2) v subs = (\<exists> v'. path g as1 v' subs \<and> subpath g v' as2 v subs)"
unfolding path_def by (rule sp_append)


abbreviation paths ::
  "(('a \<times> nat),'b) rgraph_scheme \<Rightarrow>
   (('a \<times> nat) \<times> ('a \<times> nat)) set \<Rightarrow> ('a \<times> nat) arc list set"
where
  "paths g subs \<equiv> {p. \<exists> v. path g p v subs}"

lemma Nil_in_paths :
  "[] \<in> paths g subs"
using Nil_path by blast

lemma empty_graph_paths :
  assumes "arcs g = {}"
  shows   "paths g subs = {[]}"
unfolding set_eq_iff singleton_iff
proof (intro allI iffI)

  fix p

  assume "p \<in> paths g subs"

  thus "p = []" using empty_graph_imp_p_empty[OF assms, of p _ subs] by fast

next

  case goal2 thus ?case using Nil_path by blast

qed

lemma paths_conv :
  "paths g subs = subpaths_from g (root g) subs"
unfolding path_def subpath_from_def by (rule refl)

lemma (in sub_rel_of) lst_of_p_is_vert :
  assumes "path g p v subs"
  shows   "v \<in> Graph.vertices g"
using assms lst_of_sp_is_vert by (simp add : path_def)







(************************************** Ajout d'arc ***********************************************)

lemma sp_in_add_arc :
  assumes "subpath g l1 p l2 subs"
  shows   "subpath (add_arc g a) l1 p l2 subs"
using assms by (auto simp add : subpath_def add_arc_vertices add_arc_arcs)

lemma p_in_add_arc :
  assumes "path g p l subs"
  shows   "path (add_arc g a) p l subs"
using assms unfolding path_def add_arc_root by (rule sp_in_add_arc)


lemma sps_from_in_add_arc :
  "subpaths_from g l subs \<subseteq> subpaths_from (add_arc g a) l subs"
unfolding subpath_from_def using sp_in_add_arc by force

lemma (in sub_rel_of) sub_rel_of_add_arc :
  "sub_rel_of (add_arc g a) subs"
using sub_rel_of unfolding sub_rel_of_def subsumees_conv subsumers_conv add_arc_vertices by auto

lemma (in wf_sub_rel_of) wf_sub_rel_of_add_arc :
  "wf_sub_rel_of (add_arc g a) subs"
using wf_sub_rel_of unfolding wf_sub_rel_of_def
by (intro conjI) (rule sub_rel_of.sub_rel_of_add_arc)

lemma (in sub_rel_of) sp_end_src_imp_append_in_add_arc :
  assumes "subpath g l1 p (src a) subs"
  shows   "subpath (add_arc g a) l1 (p @ [a]) (tgt a) subs"
proof -
  have "sub_rel_of (add_arc g a) subs" by (rule sub_rel_of_add_arc)

  thus ?thesis using assms by (auto simp add : sub_rel_of.sp_append_one sp_in_add_arc add_arc_arcs)
qed


lemma sp_in_add_arc_minus_arc_imp_sp_in_old :
  assumes "l1 \<in> Graph.vertices g"
  assumes "subpath (add_arc g a) l1 p l2 subs"
  assumes "a \<notin> set p"
  
  shows   "subpath g l1 p l2 subs"
using assms unfolding subpath_def by (auto simp add : add_arc_arcs)


lemma (in sub_rel_of) sp_ends_in_tgt_imp_mem :
  assumes "tgt a \<notin> Graph.vertices g"
  assumes "l \<noteq> tgt a"
  assumes "subpath (add_arc g a) l p (tgt a) subs"

  shows   "a \<in> set p"
proof -
  have "tgt a \<notin> subsumers subs" using assms(1) sub_rel_of by auto

  have "(l,tgt a) \<notin> subs\<^sup>+" 
  proof (rule ccontr, simp only : not_not)
    assume "(l,tgt a) \<in> subs\<^sup>+"

    then obtain v where "(v,tgt a) \<in> subs" using tranclD2[of l "tgt a" subs] by blast

    thus "False" using `tgt a \<notin> subsumers subs`
    unfolding subsumers_conv by fastforce
  qed

  hence "p \<noteq> []" using assms(2) assms(3) by (auto simp add : Nil_sp)

  then obtain a' p' where "p = p' @ [a']" using neq_Nil_conv2[of p] by blast

  hence "a' \<in> arcs (add_arc g a)" using assms(3) by (auto simp add: subpath_def)
  
  moreover
  have "tgt a' = tgt a"  
  proof -

    have "(tgt a',tgt a) \<notin> subs\<^sup>+"
    using `tgt a \<notin> subsumers subs` tranclD2[of "tgt a'" "tgt a" subs] 
    by (auto simp add : subsumers_conv)

    moreover
    have "sub_rel_of (add_arc g a) subs" by (rule sub_rel_of_add_arc)

    ultimately
    show ?thesis using assms(3) `p = p' @ [a']` by (auto simp add : sub_rel_of.sp_append_one)
  qed
  

  ultimately
  have "a' = a \<or> (a' \<noteq> a \<and> a' \<in> arcs g)" by (auto simp add : add_arc_arcs)

  thus ?thesis 
  proof (elim disjE)
    assume "a' = a" thus ?thesis using `p = p' @ [a']` by simp
  next
    assume "a' \<noteq> a \<and> a' \<in> arcs g"

    hence  "tgt a' \<in> Graph.vertices g" by (simp add : tgt_is_vert)

    thus   ?thesis using `tgt a' = tgt a` assms(1) by simp
  qed
qed



lemma
  assumes "src a \<in> Graph.vertices g"
  assumes "tgt a \<notin> Graph.vertices g"
  assumes "subpath (add_arc g a) v as (tgt a) subs"
  shows   "as = [] \<or> (\<exists> as. as = as' @ [a] \<and> a \<notin> set as' \<and> subpath g v as' (src a) subs)"
proof (case_tac "v = tgt a")
oops


lemma graph_sp_imp_sp_w_subs :
  assumes "Graph.subpath g v1 as v2"
  shows   "subpath g v1 as v2 subs"
using assms cas_imp_cas_w_subs unfolding Graph.subpath_def subpath_def by force

lemma graph_p_imp_p_w_subs :
  assumes "Graph.path g as v"
  shows   "path g as v subs"
using assms graph_sp_imp_sp_w_subs by (simp add : Graph.path_def path_def)




lemma (in wf_sub_rel_of) sp_from_subsumee2 :
  assumes "(d,a) \<in> subs"
  assumes "subpath_from g d as subs"
  assumes "out_arcs g d = {}"               

  shows   "subpath_from g a as subs"
proof (cases as)

  assume "as = []"

  moreover
  have "a \<in> Graph.vertices g"
  using assms(1) sub_rel_of unfolding subsumers_conv sub_rel_of_def
  by (auto simp del : split_paired_Ex)
  
  ultimately
  show ?thesis
  using wf_sub_rel_of unfolding subpath_from_def sub_rel_of_def subsumees_conv
  by (auto simp only : Nil_sp)

next
                              
  fix arc as'

  assume "as = arc # as'"

  moreover
  obtain v where "subpath g d as v subs" 
  using assms(2) unfolding subpath_from_def by blast

  ultimately
  have "arc \<in> arcs g" by (simp add : subpath_def)

  moreover
  have "src arc = a"
  proof -

    have "d = src arc \<or> (d, src arc) \<in> subs\<^sup>+"
    using `as = arc # as'` `subpath g d as v subs` sub_rel_of
    by (simp add : sub_rel_of.sp_Cons)
    
    thus ?thesis
    proof (elim disjE)
      assume "d = src arc"
      hence  "arc \<in> out_arcs g d" using `arc \<in> arcs g` by (simp add : out_arcs_def)
      thus ?thesis using assms(3) by simp
    next
      assume "(d, src arc) \<in> subs\<^sup>+"
      hence  "(d, src arc) \<in> subs" by (rule in_trancl_imp)
      thus  "src arc = a" 
      using `(d,a) \<in> subs` wf_sub_rel by (elim wf_sub_rel.subsumed_by_two_imp)
    qed
  qed

  moreover
  have "subpath g (tgt arc) as' v subs" 
  using `as = arc # as'` `subpath g d as v subs` sub_rel_of
  by (simp add : sub_rel_of.sp_Cons)

  ultimately
  show ?thesis
  using `as = arc # as'` sub_rel_of
  unfolding subpath_from_def
  by (rule_tac ?x="v" in exI, simp add : sub_rel_of.sp_Cons)

qed


lemma
  assumes "B"
  assumes "A \<or> B"
  shows   "B"
oops

lemma sp_from_not_subsumed_not_empty :
  assumes "subpath g v1 as v2 subs"
  assumes "v1 \<noteq> v2"
  assumes "v1 \<notin> subsumees subs"
  shows   "as \<noteq> []"
proof (intro notI)

  assume "as = []"

  hence "(v1,v2) \<in> subs\<^sup>+" using assms(1,2) by (simp add : subpath_def)

  then obtain v3 where "(v1,v3) \<in> subs" using tranclD[of v1 v2 subs] by blast
  
  hence "v1 \<in> subsumees subs" unfolding subsumees_conv by blast

  thus False using assms(3) by (elim notE)
qed

lemma (in sub_rel_of) sp_append_sub :
  assumes "subpath g v1 as v2 subs"
  assumes "(v2,v3) \<in> subs"
  shows   "subpath g v1 as v3 subs"
proof (case_tac "as = []")
  
  assume "as = []"

  hence "v1 \<in> Graph.vertices g" 
  and   "v1 = v2 \<or> (v1,v2) \<in> subs\<^sup>+" using assms(1) by (simp_all add : Nil_sp)

  thus ?thesis
  proof (elim disjE)

    case goal1 thus ?thesis using `as = []` assms(2) by (simp add : Nil_sp r_into_trancl)

  next

    case goal2 

    thus ?thesis using `as = []` assms(2) by  (simp add : Nil_sp, intro disjI2, auto)
  qed

next

  assume "as \<noteq> []"

  then obtain as' a where "as = as' @ [a]" using neq_Nil_conv2[of as] by blast

  thus ?thesis
  using assms by (simp add : sp_append_one, intro disjI2, auto)

qed

end

