theory LTS
imports Graph ArcExt SubExt Labels PreConf Tree
begin


(* Dans les théories précédentes, on a généralisé les lemmes et les définitions concernant les 
   graphes de manière à pouvoir traiter des extensions du record rgraph. On construit un lts
   en étendant les rgraph par un état de contrôle final et une fonction d'étiquetage des arcs *)

(***************************************************)
(******* LTS ***************************************)
(***************************************************)

(* Un lts est un graphe muni d'une fonction d'étiquetage des arcs par des labels *)
record ('a,'b) lts = "'a rgraph" +
  end_loc   :: "'a"
  labelling :: "'a arc \<Rightarrow> 'b label"

locale lts =
  fixes lts :: "('a,'b,'c) lts_scheme" (structure)

abbreviation labels :: "('a,'b,'c) lts_scheme \<Rightarrow> 'b label set" where
  "labels lts \<equiv> labelling lts ` arcs lts"

abbreviation vars :: "('a,'b,'c) lts_scheme \<Rightarrow> 'b set" where
  "vars lts \<equiv> 
    \<Union> x \<in> (\<lambda> t. case t of Skip \<Rightarrow> {} | Assume be \<Rightarrow> Bexp.vars be | Assign v ae \<Rightarrow> {v} \<union> Aexp.vars ae) ` 
    (labels lts). x"

abbreviation graph :: "('a,'b,'c) lts_scheme \<Rightarrow> 'a rgraph" where
  "graph lts \<equiv> rgraph.truncate lts"

abbreviation init :: "('a,'b,'c) lts_scheme \<Rightarrow> 'a" where
  "init lts \<equiv> root lts"
 
abbreviation locations :: "('a,'b,'c) lts_scheme \<Rightarrow> 'a set" where
  "locations lts \<equiv> Graph.vertices lts"

lemma init_is_loc :
  "init lts \<in> locations lts" 
by (rule root_is_vert)

type_synonym 'a trans = "'a arc"

(*abbreviation arcs :: "('a,'b,'c) lts_scheme \<Rightarrow> 'a trans set" where
"arcs lts \<equiv> arcs lts"*)

(*abbreviation out_arcs :: "('a,'b,'c) lts_scheme \<Rightarrow> 'a \<Rightarrow> 'a trans set" where
  "out_arcs lts l \<equiv> out_arcs lts l"*)


(* Un lts est bien formé si :
   - son graphe est bien formé,
   - son état de contrôle final est bien un sommet du graphe sous-jacent,
   - aucune transition n'est issue de cet état de contrôle.
 *)
locale wf_lts = lts + 
  assumes wf_graph                  : "wf_rgraph lts"
  assumes end_loc_loc               : "end_loc lts \<in> locations lts"
  assumes end_loc_out_arcs         : "out_arcs lts (end_loc lts) = {}"
  assumes empty_or_end_loc_not_init : "locations lts \<noteq> {init lts} 
                                       \<longrightarrow> end_loc lts \<noteq> init lts"
  (*assumes end_loc_reachable : "\<exists> p. Graph.subpath lts (init lts) p (end_loc lts)"*)
begin
  lemmas wf_lts = 
    wf_graph 
    end_loc_loc 
    end_loc_out_arcs 
    empty_or_end_loc_not_init 
    (*end_loc_reachable*)

  lemma
    assumes "end_loc lts \<noteq> init lts"
    shows   "locations lts \<noteq> {init lts}"
  using assms end_loc_loc by auto
end


locale finite_labels_lts = lts +
  assumes finite_labels : "l \<in> range (labelling lts) \<Longrightarrow> finite_label l"
begin
  lemmas finite_labels_lts = finite_labels
end

lemma
  "arcs lts = {} \<Longrightarrow> wf_rgraph lts"
using assms by (rule empty_graph_wf)

abbreviation trace :: "'a arc list \<Rightarrow> ('a arc \<Rightarrow> 'b label) \<Rightarrow> 'b label list" where
  "trace as L \<equiv> map L as"

(*lemma  : 
  "trace [a] L = [L a]"
by (simp add : )

lemma  :
  "trace (as1@as2) L = ((trace as1 L) @ (trace as2 L)) "
by (simp add: )*)


(* fsp : Feasible SubPath

   Etant donné un lts pc, une configuration, des états de contrôle de départ et d'arrivée l1 et l2, 
   une séquence de arcs est dîte faisable de l1 vers l2 si cette séquence est un sous-chemin 
   (allant de l1 à l2) du graphe sous-jacent et si l'exécution symbolique de sa trace depuis pc est
   produit une configuration satisfiable *)
definition feasible_subpath ::
  "('a,'b,'c) lts_scheme \<Rightarrow> 
   'b pre_conf           \<Rightarrow> 
   'a                    \<Rightarrow> 
   'a arc list           \<Rightarrow> 
   'a                    \<Rightarrow> bool"
where
  "feasible_subpath lts pc l1 as l2 \<equiv> Graph.subpath lts l1 as l2 \<and> 
                                      feasible pc (trace as (labelling lts))"

lemma fsp_append :
  "feasible_subpath lts pc l1 (as1@as2) l2 = 
    (\<exists> l' pc'. feasible_subpath lts pc l1 as1 l' 
             \<and> se_star pc (trace as1 (labelling lts)) pc'
             \<and> feasible_subpath lts pc' l' as2 l2)"
proof (intro iffI)

  case goal1

  then obtain li where "Graph.subpath lts l1 as1 li"
                 and   "Graph.subpath lts li as2 l2"
  unfolding feasible_subpath_def by (simp add : sp_append) blast

  moreover
  obtain pc' where "se_star pc (trace (as1@as2) (labelling lts)) pc'"
             and   "sat pc'"
  using goal1 unfolding feasible_subpath_def feasible_def by blast

  moreover
  then obtain pci where "se_star pc  (trace as1 (labelling lts)) pci"
                   and  "se_star pci (trace as2 (labelling lts)) pc'"
  by (simp add :  se_star_append) blast
  
  moreover
  hence "sat pci" using se_star_sat_imp_sat `sat pc'` by fast

  ultimately
  show ?case
  by (rule_tac ?x="li" in exI, rule_tac ?x="pci" in exI, intro conjI)
     (auto simp add : feasible_subpath_def feasible_def)

next

  case goal2

  then obtain li pci where "feasible_subpath lts pc l1 as1 li"
                     and   "se_star pc (trace as1 (labelling lts)) pci"
                     and   "feasible_subpath lts pci li as2 l2"
  by blast

  thus ?case 
  by (simp add : feasible_subpath_def feasible_def sp_append se_star_append) blast

qed
  

definition feasible_subpath_from ::
  "('a,'b,'c) lts_scheme \<Rightarrow> 
   'b pre_conf           \<Rightarrow> 
   'a                    \<Rightarrow> 
   'a arc list           \<Rightarrow> bool"
where
  "feasible_subpath_from lts pc l as \<equiv> \<exists> l'. feasible_subpath lts pc l as l'"

lemma fsp_from_append :
  "feasible_subpath_from lts pc l (as1@as2) =  
    (\<exists> l1 pc1. feasible_subpath lts pc l as1 l1
             \<and> se_star pc (trace as1 (labelling lts)) pc1
             \<and> feasible_subpath_from lts pc1 l1 as2)"
using fsp_append[of lts pc l as1 as2]
unfolding feasible_subpath_from_def by blast

abbreviation feasible_subpaths_from ::
  "('a,'b,'c) lts_scheme \<Rightarrow> 
   'b pre_conf           \<Rightarrow> 
   'a                    \<Rightarrow> 'a arc list set"
where
  "feasible_subpaths_from lts pc l \<equiv> {ts. feasible_subpath_from lts pc l ts}"

lemma feasible_subpath_from_conv :
  "feasible_subpath_from lts pc l as = 
    (Graph.subpath_from lts l as \<and> feasible pc (trace as (labelling lts)))"
unfolding feasible_subpath_from_def feasible_subpath_def Graph.subpath_from_def by blast

lemma feasible_subpaths_from_are_subpaths_from :
  "feasible_subpaths_from lts pc l \<subseteq> Graph.subpaths_from lts l"
by (auto simp add : feasible_subpath_from_conv)


definition feasible_path ::
  "('a,'b,'c) lts_scheme \<Rightarrow> 
   'b pre_conf           \<Rightarrow>
   'a arc list           \<Rightarrow> 
   'a                    \<Rightarrow> bool"
where
  "feasible_path lts pc as l \<equiv> feasible_subpath lts pc (init lts) as l"

lemma feasible_path_conv :
  "feasible_path lts pc as l = (Graph.path lts as l \<and> feasible pc (trace as (labelling lts)))"
unfolding feasible_path_def feasible_subpath_def Graph.path_def by (rule refl)

abbreviation feasible_paths ::
  "('a,'b,'c) lts_scheme \<Rightarrow> 
   'b pre_conf           \<Rightarrow> 'a arc list set"
where
  "feasible_paths lts pc \<equiv> {as. \<exists> l. feasible_path lts pc as l}"

lemma feasible_paths_conv :
  "feasible_paths lts pc = feasible_subpaths_from lts pc (init lts)"
unfolding feasible_path_def feasible_subpath_from_def by (rule refl)

lemma feasible_paths_are_paths :
  "feasible_paths lts pc \<subseteq> Graph.paths lts"
by (auto simp add : feasible_path_conv)



lemma no_feasible_arc_from_init_imp :
  fixes lts :: "('a,'b,'c) lts_scheme"
  assumes "sat pc"
  assumes "\<not> (\<exists> ba \<in> out_arcs lts (init lts). \<exists> pc'. se pc (labelling lts ba) pc' \<and> sat pc')"
  shows   "feasible_paths lts pc = {[]}"
unfolding set_eq_iff singleton_iff
proof (intro allI iffI)

  fix bas 
  
  assume "bas \<in> feasible_paths lts pc"

  thus   "bas = []"
  proof (case_tac "bas = []")
    assume "bas = []" thus ?thesis by assumption
  next

    assume "bas \<noteq> []"

    then obtain ba bas' where "bas = ba # bas'" using neq_Nil_conv[of bas] by blast

    then obtain pc1 pc2 where "se pc (labelling lts ba) pc1"
                        and   "se_star pc1 (trace bas' (labelling lts)) pc2"
                        and   "sat pc2"
    using `bas \<in> feasible_paths lts pc`
    unfolding feasible_path_def feasible_subpath_def feasible_def
    by (simp add : se_star_Cons) blast

    moreover
    hence "sat pc1" by (simp add : se_star_sat_imp_sat)

    moreover
    have "ba \<in> out_arcs lts (init lts)"
    using `bas \<in> feasible_paths lts pc` `bas = ba # bas'`
    unfolding feasible_path_def feasible_subpath_def
    by (auto simp add : Graph.sp_Cons out_arcs_def)

    ultimately
    show ?thesis using assms(2) by auto

  qed

next

  fix bas :: "'a arc list"

  assume "bas = []"

  thus "bas \<in> feasible_paths lts pc" 
  using assms(1) Graph.Nil_in_paths
  by (auto simp add : feasible_path_conv feasible_def)

qed




(*(* Feasible SubPath From *)
abbreviation fsp_from ::
  "('a,'b,'c) lts_scheme \<Rightarrow> 
   'b pre_conf           \<Rightarrow> 
   'a                    \<Rightarrow> 
   'a arc list           \<Rightarrow> bool"
where
  "fsp_from lts pc l as \<equiv> \<exists> l'. fsp lts pc l as l'"

(* Feasible SubPathS From *)
abbreviation fsps_from ::
  "('a,'b,'c) lts_scheme \<Rightarrow> 
   'b pre_conf           \<Rightarrow> 
   'a                    \<Rightarrow> 'a arc list set"
where
  "fsps_from lts pc l \<equiv> {as. \<exists> l'. fsp lts pc l as l'}"

(* Feasible Path *)
abbreviation fp ::
  "('a,'b,'c) lts_scheme \<Rightarrow> 
   'b pre_conf           \<Rightarrow>
   'a arc list           \<Rightarrow> bool"
where
  "fp lts pc as \<equiv> fsp_from lts pc (init lts) as"

(* Feasible PathS *)
abbreviation fps ::
  "('a,'b,'c) lts_scheme \<Rightarrow> 
   'b pre_conf           \<Rightarrow> 'a arc list set"
where
  "fps lts pc \<equiv> {p. fp lts pc p}"


lemma
  "fps lts pc \<subseteq> Graph.paths lts"
by (simp only : fsp_def Graph.paths_def Graph.path_def) fast*)




(****************************************************************)
(******* Pre RedBlack LTS ***************************************)
(****************************************************************)

(* Un pre_RB est constitué de :
   - une partie rouge qui est un graphe représentant le dépliage (de la partie noire),
   - une partie noire qui est le lts de départ,
   - une relation de subsomption,
   - une configuration initiale,
   - une fonction associant des configurations aux sommets de la partie rouge. La partie rouge,
     la relation de subsomption et cette fonction constituent le dépliage par exécution symbolique.

   Note : on conserve la configuration initiale en dehors de la fonction "confs" car les images de
   cette fonction sont susceptibles d'être modifiées par abstraction. Cependant, on aura besoin 
   dans certains théorèmes de connaître la configuration initiale originale afin de comparer les 
   ensembles de chemins faisables.
 *)
record ('a,'b) pre_RB =
  red       :: "('a \<times> nat) rgraph"
  black     :: "('a,'b) lts"
  subs      :: "(('a \<times> nat) \<times> ('a \<times> nat)) set"
  init_conf :: "'b pre_conf"
  confs     :: "('a \<times> nat) \<Rightarrow> 'b pre_conf"




locale pre_RB =
  fixes prb :: "('a,'b,'c) pre_RB_scheme" (structure)


(* Un pre_RB est bien formé si :
   - sa partie rouge est un graphe bien formé,
   - sa partie noire est un lts bien formé,
   - la racine de la partie rouge et l'état de contrôle initial de la partie noire coïncident,
   - les arcs rouges sont des versions indicées des arcs de la partie noire.
 *)
locale wf_pre_RB = pre_RB +
  assumes red_wf           : "wf_rgraph (red prb)"
  assumes black_wf         : "wf_lts (black prb)"
  assumes consistent_roots : "fst (root (red prb)) = root (black prb)"
  assumes ui_ra_are_ba     : "a \<in> arcs (red prb) \<Longrightarrow> ui_arc a \<in> arcs (black prb)"
begin
  lemmas wf_pre_RB = red_wf black_wf consistent_roots ui_ra_are_ba
end

locale finite_labels_RB = pre_RB +
  assumes finite_init_pred         : "finite (\<pi> (init_conf prb))"
  assumes finite_init_pred_symvars : "\<forall> e \<in> \<pi> (init_conf prb). finite (Bexp.vars e)"
  assumes finite_labels_lts        : "finite_labels_lts (black prb)"
begin
  lemmas finite_labels_RB = finite_init_pred finite_init_pred_symvars finite_labels_lts
end


abbreviation red_locations :: 
  "('a,'b,'c) pre_RB_scheme \<Rightarrow> ('a \<times> nat) set" 
where
  "red_locations lts \<equiv> Graph.vertices (red lts)"

abbreviation black_locations :: 
  "('a,'b,'c) pre_RB_scheme \<Rightarrow> 'a set" 
where
  "black_locations lts \<equiv> Graph.vertices (black lts)"


(* Indique qu'un pre_RB a été obtenu par notre technique de dépliage par exécution symbolique *)

abbreviation se_extends :: 
  "('a,'b,'c) pre_RB_scheme \<Rightarrow>
   (('a \<times> nat)) rgraph          \<Rightarrow>
   ('a \<times> nat) arc               \<Rightarrow>
   'b pre_conf                  \<Rightarrow>
   ('a,'b,'c) pre_RB_scheme \<Rightarrow> bool"
where
  "se_extends prb rg' ra pc' prb' \<equiv> 
     ui_arc ra \<in> arcs (black prb)
   \<and> src ra \<notin> subsumees (subs prb) 
   \<and> extends (red prb) ra rg' 
   \<and> se  (confs prb (src ra)) (labelling (black prb) (ui_arc ra)) pc'
   \<and> sat pc'                         (* this shocks bu, who suggests an explicit cut rule. *)
   \<and> prb' = \<lparr> red       = rg',
              black     = black prb,
              subs      = subs prb,
              init_conf = init_conf prb,
              confs     = (confs prb) (tgt ra := pc'),
              \<dots>        = more prb \<rparr>"

abbreviation subsum_extends ::
  "('a,'b,'c) pre_RB_scheme   \<Rightarrow>
   (('a \<times> nat) \<times> ('a \<times> nat))     \<Rightarrow>
   (('a \<times> nat) \<times> ('a \<times> nat)) set \<Rightarrow>
   ('a,'b,'c) pre_RB_scheme   \<Rightarrow> bool"
where
  "subsum_extends prb sub subrel prb' \<equiv>
     SubExt.extends (red prb) (subs prb) sub subrel
   (*\<and> sat (confs prb (subsumee sub))*)
   \<and> confs prb (subsumee sub) \<sqsubseteq>\<^sub>p\<^sub>c confs prb (subsumer sub)
   \<and> fst (subsumee sub) \<noteq> end_loc (black prb)
   \<and> prb' = \<lparr> red       = red prb,
              black     = black prb,
              subs      = subrel,
              init_conf = init_conf prb,
              confs     = confs prb,
              \<dots>        = more prb \<rparr>"

abbreviation abstract_extends ::
  "('a,'b,'c) pre_RB_scheme \<Rightarrow> 
   ('a \<times> nat)                   \<Rightarrow>
   'b symvar bexp                      \<Rightarrow>
   ('a,'b,'c) pre_RB_scheme \<Rightarrow> bool"
where
  "abstract_extends prb rl e prb' \<equiv> 
     rl \<in> red_locations prb
   \<and> out_arcs (red prb) rl = {}
   \<and> rl \<notin> subsumees (subs prb)
   \<and> prb' =  \<lparr> red       = red prb, 
               black     = black prb, 
               subs      = subs prb, 
               init_conf = init_conf prb,
               confs     = (confs prb)(rl := abstract e (confs prb rl)), 
               \<dots>        = more prb \<rparr>"

(* La relation bleue: transformations des lts rouge-noirs *)
inductive RB ::
  "('a,'b,'c) pre_RB_scheme \<Rightarrow> bool"
where
  base : 
    "fst (root (red prb)) = init (black prb)      \<Longrightarrow>
     arcs (red prb) = {}                          \<Longrightarrow>
     subs prb       = {}                          \<Longrightarrow>
     (confs prb) (root (red prb)) = init_conf prb \<Longrightarrow>
     sat (init_conf prb)                          \<Longrightarrow> RB prb"

| se_step : 
    "RB prb                                   \<Longrightarrow>
     se_extends prb rg' ra pc' prb'               \<Longrightarrow> RB prb'"
                                                 
| subsum_step :                                  
    "RB prb                                   \<Longrightarrow>
     subsum_extends prb sub subrel prb'           \<Longrightarrow> RB prb'"
                                                 
| abstract_step :                                
    "RB prb                                   \<Longrightarrow> 
     abstract_extends prb rl e prb'               \<Longrightarrow> RB prb'"







(***************************************)
(********** Propriétés des RB **********)
(***************************************)

(* invariants de la relation bleue *)
(* La partie rouge d'un RB est finie *)
lemma red_arcs_finite :
  assumes "RB prb"
  shows   "finite (arcs (red prb))"
using assms by (induct prb) (auto simp add : extends_arcs)



(* Dans un RB, les arcs rouges sont des versions indicées des arcs de la partie noire *)
lemma ui_RA_subset_BA :
  assumes "RB prb"
  shows   "ui_arc ` arcs (red prb) \<subseteq> arcs (black prb)"
using assms by (induct rule : RB.induct) (force simp add : extends_arcs)+



(* version spécifique du théorème précédent *)
lemma ui_ra_is_ba :
  assumes "RB prb"
  assumes "ra \<in> arcs (red prb)"
  shows   "ui_arc ra \<in> arcs (black prb)"
using ui_RA_subset_BA[OF assms(1)] assms(2) by auto


(* Dans un RB, la racine de la partie rouge correspond à l'état de contrôle initial de la partie
   noire. *)
lemma consistent_roots :
  assumes "RB prb"
  shows   "fst (root (red prb)) = init (black prb)"
using assms by (induct prb) (auto simp add : extends_root)



(* Un RB dont la partie noire est un lts bien formé possède une partie rouge bien formée *)
lemma wf_black_imp_wf_red :
  assumes "RB prb"
  assumes "wf_lts (black prb)"
  shows   "wf_rgraph (red prb)"
using assms by (induct prb) (auto simp add : empty_graph_wf wf_rgraph.extends_wf_rgraph)
 


(* Un RB dont la partie noire est bien formée est bien formé *)
lemma wf_black_imp_wf :
  assumes "RB prb"
  assumes "wf_lts (black prb)"
  shows   "wf_pre_RB prb"
using assms wf_black_imp_wf_red consistent_roots ui_ra_is_ba
by (auto simp add : wf_pre_RB_def)


(* La partie rouge d'un RB (sans les subsomptions) est un arbre *)
lemma red_is_tree :
  assumes "RB prb"
  shows   "is_tree (red prb)"
using assms by (induct prb) (auto simp add : empty_graph_is_tree is_tree.extends_is_tree)


(* les sommets de la partie rouge sont des versions indicées des sommets de la partie noire *)
lemma ui_RL_subset_BL :
  assumes "RB prb"
  shows   "fst ` red_locations prb \<subseteq> black_locations prb"
using assms consistent_roots ui_ra_is_ba by (auto simp add : vertices_conv Bex_def) fastforce+



(* version spécifique du lemme précédent *)
lemma ui_rl_is_bl :
  assumes "RB prb"
  assumes "rl \<in> red_locations prb"
  shows   "fst rl \<in> black_locations prb"
using assms ui_RL_subset_BL by auto



lemma rl_sat : (* bu est pret d'abandonner cela *)
  assumes "RB prb"
  assumes "rl \<in> red_locations prb"
  shows   "sat (confs prb rl)"
using assms
proof (induct rule : RB.induct)
  case base         
  thus ?case by (simp add : vertices_conv)
next
  case se_step 
  thus ?case by (force simp add : extends_vertices)
next
  case subsum_step 
  thus ?case by simp
next
  case abstract_step 
  thus ?case by (force simp add : sat_imp_abstract_sat)
qed


(* Dans un RB ben formé, aucun arc n'est issu d'un sommet rouge correspondant à l'état de 
   contrôle final de la partie noire. *)
lemma (in wf_pre_RB) end_loc_imp_out_arcs_empty :
  assumes "RB prb"
  assumes "rl \<in> red_locations prb"
  assumes "fst rl = end_loc (black prb)"
  shows   "out_arcs (red prb) rl = {}"
proof (rule ccontr)

  assume "out_arcs (red prb) rl \<noteq> {}"

  then obtain arc where "arc \<in> arcs (red prb)"
                  and   "src arc = rl"
  unfolding out_arcs_def by blast

  hence "ui_arc arc \<in> out_arcs (black prb) (end_loc (black prb))"
  using assms(1) ui_ra_is_ba assms(3) by (force simp add : out_arcs_def)

  thus "False" using wf_pre_RB by (auto simp add : wf_lts_def)
qed


(* La relation de subsomption d'un RB appartient bien à sa partie rouge. *)
lemma subs_sub_rel_of :
  assumes "RB prb"
  shows   "sub_rel_of (red prb) (subs prb)"
using assms
proof (induct prb)
  case base thus ?case 
    by (simp add : empty_sub_rel_of)
next
  case se_step thus ?case 
    by (auto simp add : sub_rel_of.sub_rel_of_extends)
next
  case subsum_step thus ?case 
    by (auto simp add : SubExt.sub_rel_of.extends_imp_sub_rel_of)
next
  case abstract_step thus ?case 
    by simp
qed


(* La relation de subsomption d'un RB est bien formee. *)
lemma subs_wf_sub_rel :
  assumes "RB prb"
  shows   "wf_sub_rel (subs prb)"
using assms
proof (induct prb)
  case base thus ?case by (simp add : empty_sub_rel_wf)
next
  case se_step thus ?case by simp
next
  case subsum_step thus ?case by (auto simp add : wf_sub_rel.extends_imp_wf_sub_rel)
next
  case abstract_step thus ?case by simp
qed


lemma subs_wf_sub_rel_of :
  assumes "RB prb"
  shows   "wf_sub_rel_of (red prb) (subs prb)"
using subs_sub_rel_of[OF assms] 
      subs_wf_sub_rel[OF assms]
by (simp add : wf_sub_rel_of_def)







lemma subsumees_are_locs :
  assumes "RB prb"
  shows   "subsumees (subs prb) \<subseteq> red_locations prb"
using subs_sub_rel_of[OF assms] by (auto simp add : sub_rel_of_def subsumees_conv)


lemma subs_to_same_BL :
  assumes "RB prb"
  assumes "(d,a) \<in> subs prb"
  shows   "fst d = fst a"
using assms by (induct prb) (auto simp add : SubExt.extends_def)


lemma subs_to_same_BL2 :
  assumes "RB prb"
  assumes "sub \<in> subs prb"
  shows   "fst (subsumee sub) = fst (subsumer sub)"
using assms by (induct prb) (auto simp add : SubExt.extends_def)


lemma rcas_imp_bcas :
  assumes "RB prb"
  assumes "SubRel.cas rl1 ras rl2 (subs prb)"
  shows   "Graph.cas (fst rl1) (ui_as ras) (fst rl2)"
using assms(2)
proof (induct ras arbitrary : rl1)
  
  case (Nil rl1)

  hence "fst rl2 = fst rl1"
  using wf_sub_rel_of.in_trancl_imp[OF subs_wf_sub_rel_of[OF assms(1)], of rl1 rl2]
        subs_to_same_BL[OF assms(1), of rl1 rl2]
  by auto

  thus ?case by simp
  
next

  case (Cons ra ras rl1)

  hence 1 : "rl1 = src ra \<or> (rl1, src ra) \<in> (subs prb)\<^sup>+"
  and   2 : "cas (tgt ra) ras rl2 (subs prb)" by simp_all

  from 1 have "src (ui_arc ra) = fst rl1"
  using wf_sub_rel_of.in_trancl_imp[OF subs_wf_sub_rel_of[OF assms(1)], of rl1 "src ra"]
          subs_to_same_BL[OF assms(1), of rl1 "src ra"]
  by auto
  
  moreover
  have "Graph.cas (tgt (ui_arc ra)) (ui_as ras) (fst rl2)" using 2 Cons(1) by simp

  ultimately
  show ?case by simp
qed


lemma rsp_is_bsp :
  assumes "RB prb"
  assumes "subpath (red prb) rl1 rsp rl2 (subs prb)"
  shows   "Graph.subpath (black prb) (fst rl1) (ui_as rsp) (fst rl2)"
proof -

  have 1 : "cas rl1 rsp rl2 (subs prb)"
  and  2 : "rl1 \<in> red_locations prb"
  and  3 : "set rsp \<subseteq> arcs (red prb)"
  using assms(2) by (simp_all add : subpath_def)

  from 1 have "Graph.cas (fst rl1) (ui_as rsp) (fst rl2)" 
  using rcas_imp_bcas[OF assms(1)] by simp

  moreover
  from 2 have "fst rl1 \<in> black_locations prb"
  using ui_rl_is_bl[OF assms(1)] by fastforce

  moreover
  from 3 have "set (ui_as rsp) \<subseteq> arcs (black prb)" 
  using ui_RA_subset_BA[OF assms(1)] by auto

  ultimately
  show ?thesis by (simp add : Graph.subpath_def)
qed



lemma rspf_is_bspf :
  assumes "RB prb"
  assumes "subpath_from (red prb) rl rsp (subs prb)"
  shows   "Graph.subpath_from (black prb) (fst rl) (ui_as rsp)"
using rsp_is_bsp assms unfolding subpath_from_def Graph.subpath_from_def by fastforce

lemma rp_is_bp :
  assumes "RB prb"
  assumes "wf_lts (black prb)"
  assumes "SubRel.path (red prb) rsp rl (subs prb)"
  shows   "Graph.path (black prb) (ui_as rsp) (fst rl)"
using assms rsp_is_bsp consistent_roots unfolding path_def Graph.path_def by fastforce


lemma rspsf_subset_bspsf :
  assumes "RB prb"
  shows   "ui_as ` subpaths_from (red prb) rl (subs prb) 
           \<subseteq> Graph.subpaths_from (black prb) (fst rl)"
using assms rspf_is_bspf by fast



lemma end_red_loc_not_subsumed :
  assumes "RB prb"
  assumes "rl \<in> red_locations prb"
  assumes "fst rl = end_loc (black prb)"
  shows   "rl \<notin> subsumees (subs prb)"
using assms
proof (induct prb)
  case (base prb) thus ?case unfolding subsumees_conv by auto
next
  case (se_step prb rg' ra pc' prb')

  moreover
  have "rl \<in> red_locations prb \<or> rl = tgt ra" using se_step extends_vertices by auto

  ultimately
  show ?case
  proof (elim conjE disjE)
    case goal1 thus ?case using se_step by simp
  next
    case goal2

    moreover
    hence "rl \<notin> red_locations prb" by (simp add : extendsE2)

    moreover
    have "subsumees (subs prb) \<subseteq> red_locations prb"
    using subs_sub_rel_of[OF se_step(1)] by (simp add : sub_rel_of_def)

    ultimately
    show ?case by auto
  qed
next
  case (subsum_step prb sub subrel)
  thus ?case unfolding subsumees_conv by (auto simp add : SubExt.extends_def)
next
  case (abstract_step prb rl' e) thus ?case by simp
qed


lemma (in finite_labels_RB) finite_pred_constr_symvars :
  assumes "RB prb"
  assumes "rl \<in> red_locations prb"
  shows   "\<forall> e \<in> \<pi> (confs prb rl).  finite (Bexp.vars e)"
using assms finite_init_pred_symvars finite_labels_lts
proof (induct prb arbitrary : rl)
  case base thus ?case by (simp add : vertices_conv)
next
  case (se_step prb rg' ra pc' prb') 

  hence "rl \<in> red_locations prb \<or> rl = tgt ra" by (auto simp add : extends_vertices)
  
  thus ?case 
  proof (elim disjE)
    assume "rl \<in> red_locations prb"
    thus ?thesis using se_step by (auto simp add : extendsE2)
  next
    assume "rl = tgt ra"

    moreover
    have "finite_label (labelling (black prb) (ui_arc ra))"
    using se_step by (simp add : finite_labels_lts_def)

    moreover
    have   "\<forall> e \<in> \<pi> (confs prb (src ra)). finite (Bexp.vars e)"
    proof -
      have "src ra \<in> red_locations prb" using se_step by (auto simp add : extendsE1)
      
      thus ?thesis using se_step by auto
    qed

    moreover
    have "se (confs prb (src ra)) (labelling (black prb) (ui_arc ra)) pc'"
    using se_step by auto

    ultimately
    show ?thesis using se_step se_preserves_finiteness1 by fastforce
  qed
next
  case subsum_step thus ?case by simp
next
  case (abstract_step prb rl e prb') thus ?case by (simp add : abstract_def)
qed




lemma (in finite_labels_RB) finite_pred :
  assumes "RB prb"
  assumes "rl \<in> red_locations prb"
  shows   "finite (\<pi> (confs prb rl))"
using assms finite_init_pred 
proof (induct prb arbitrary : rl)
  case base thus ?case by (simp add : vertices_conv)
next
  case (se_step prb rg' ra pc' prb')

  moreover
  hence "rl \<in> red_locations prb \<and> rl \<noteq> tgt ra \<or> rl = tgt ra" by (auto simp add : extends_vertices)

  ultimately
  show ?case 
  proof (elim disjE)

    case goal1 thus ?thesis using se_step by simp

  next

    case goal2

    moreover
    have "src ra \<in> red_locations prb" using se_step by (auto simp add : extendsE1)

    hence "finite (\<pi> (confs prb (src ra)))" using se_step by auto
  
    ultimately
    show ?thesis using se_step by (auto simp add : se_preserves_finiteness2)
  qed

next
  case subsum_step thus ?case by simp
next
  case abstract_step thus ?case by (simp add : abstract_def)
qed


lemma (in finite_labels_RB) ex_se_succ :
  assumes "RB prb"
  assumes "rl \<in> red_locations prb"
  assumes "finite_label l"
  shows   "\<exists> pc'. se (confs prb rl) l pc'"
by (rule TMP1[OF finite_pred[OF assms(1,2)] finite_pred_constr_symvars[OF assms(1,2)] assms(3)])

lemma (in finite_labels_RB) ex_se_star_succ :
  assumes "RB prb"
  assumes "rl \<in> red_locations prb"
  assumes "finite_labels ls"
  shows   "\<exists> pc'. se_star (confs prb rl) ls pc'"
by (rule TMP2[OF finite_pred[OF assms(1,2)] finite_pred_constr_symvars[OF assms(1,2)] assms(3)])

lemma (in finite_labels_RB) confs_are_updatable :
  assumes "RB prb"
  assumes "rl \<in> red_locations prb"
  shows   "updatable (confs prb rl)"
using finite_conj[OF finite_pred[OF assms(1,2)] finite_pred_constr_symvars[OF assms(1,2)]]
by (rule finite_pred_imp_se_updatable)




lemma sub_subsumed :
  assumes "RB prb"
  assumes "sub \<in> subs prb"
  shows   "confs prb (subsumee sub) \<sqsubseteq>\<^sub>p\<^sub>c confs prb (subsumer sub)"
using assms
proof (induct prb)
  
  case base thus ?case by simp
  
next

  case (se_step prb rg' ra pc' prb')

  moreover
  hence "sub \<in> subs prb" by simp

  hence "subsumee sub \<in> red_locations prb" 
  and   "subsumer sub \<in> red_locations prb"
  using subs_sub_rel_of[OF se_step(1)] subsumee_in_vertices subsumer_in_vertices
  unfolding sub_rel_of_def by fast+

  moreover
  have "tgt ra \<notin> red_locations prb" using se_step by (auto simp add : extendsE2)

  ultimately
  show ?case by auto

next
  
  case (subsum_step prb sub subrel prb') thus ?case 
  using SubExt.extendsE2[of "red prb" "subs prb" sub subrel]
  by (auto simp add : SubExt.extendsE2)

next

  case (abstract_step prb rl e prb')

  hence "rl \<noteq> subsumee sub" by (auto simp add : subsumees_def)

  show ?case 
  proof (case_tac "rl = subsumer sub")
    assume "rl = subsumer sub"

    moreover
    hence  "confs prb (subsumer sub) \<sqsubseteq>\<^sub>p\<^sub>c confs prb' (subsumer sub)" 
    using abstract_step abstract_subsums by auto
  
    ultimately
    show ?thesis 
    using abstract_step 
      subsums_trans[of "confs prb  (subsumee sub)" 
                       "confs prb  (subsumer sub)" 
                       "confs prb' (subsumer sub)"]
    by simp
  next
    assume "rl \<noteq> subsumer sub"
    thus ?thesis using abstract_step `rl \<noteq> subsumee sub` by simp
  qed
qed




lemma (in finite_labels_RB) finite_labels_bsp :
  assumes "RB prb"
  assumes "bas \<in> Graph.subpaths (black prb)"
  shows   "finite_labels (trace bas (labelling (black prb)))"
using assms finite_labels_lts.finite_labels[OF finite_labels_lts]
by (simp add : finite_labels_def)






lemma rb_Nil_sp :
  assumes "RB prb"
  assumes "subpath (red prb) rl1 [] rl2 (subs prb)"
  shows   "rl1 = rl2 \<or> (rl1,rl2) \<in> (subs prb)"
proof -
  from assms(2)
  have "rl1 = rl2 \<or> (rl1,rl2) \<in> (subs prb)\<^sup>+" by (simp add : Nil_sp)

  thus ?thesis using wf_sub_rel_of.in_trancl_imp[OF subs_wf_sub_rel_of[OF assms(1)]] by fast
qed





lemma rb_sp_one :
  assumes "RB prb"
  shows "subpath (red prb) rl1 [ra] rl2 (subs prb) = 
         ( (rl1 = src ra \<or> (rl1, src ra) \<in> (subs prb)) 
         \<and> ra \<in> arcs (red prb) \<and> (tgt ra = rl2 \<or> (tgt ra, rl2) \<in> (subs prb)))"
using sub_rel_of.sp_one[OF subs_sub_rel_of[OF assms]]
      wf_sub_rel_of.in_trancl_imp[OF subs_wf_sub_rel_of[OF assms]] 
by fast


lemma rb_sp_Cons_imp :
  assumes "RB prb"
  assumes "subpath (red prb) rl1 (ra#ras) rl2 (subs prb)"
  shows   "rl1 = src ra \<or> (rl1,src ra) \<in> (subs prb)"
proof -
  from assms(2)
  have "rl1 = src ra \<or> (rl1,src ra) \<in> (subs prb)\<^sup>+"
  using subs_sub_rel_of[OF assms(1)] by (auto simp add : sub_rel_of.sp_Cons)

  thus ?thesis using wf_sub_rel_of.in_trancl_imp[OF subs_wf_sub_rel_of[OF assms(1)]] by fast
qed



lemma rb_sp_Cons :
  assumes "RB prb"
  shows   "subpath (red prb) rl1 (ra # ras) rl2 (subs prb) =
           (  (rl1 = src ra \<or> (rl1, src ra) \<in> subs prb) 
            \<and> ra \<in> arcs (red prb)                       
            \<and> subpath (red prb) (tgt ra) ras rl2 (subs prb))"
using assms sub_rel_of.sp_Cons[OF subs_sub_rel_of] rb_sp_Cons_imp by fast



thm sub_rel_of.sp_Cons[of "red prb" "subs prb" rl1 ra ras rl2]



lemma rb_sp_append_one_imp :
  assumes "RB prb"
  assumes "subpath (red prb) rl1 (ras@[ra]) rl2 (subs prb)"
  shows   "tgt ra = rl2 \<or> (tgt ra,rl2) \<in> subs prb"
proof -
  from assms(2)
  have "tgt ra = rl2 \<or> (tgt ra,rl2) \<in> (subs prb)\<^sup>+"
  using subs_sub_rel_of[OF assms(1)] by (auto simp add : sub_rel_of.sp_append_one)

  thus ?thesis using wf_sub_rel_of.in_trancl_imp[OF subs_wf_sub_rel_of[OF assms(1)]] by fast
qed



lemma rb_sp_append_one :
  assumes "RB prb"
  shows   "subpath (red prb) rl1 (ras @ [ra]) rl2 (subs prb) =
            (  subpath (red prb) rl1 ras (src ra) (subs prb) 
            \<and> ra \<in> arcs (red prb) 
            \<and> (tgt ra = rl2 \<or> (tgt ra, rl2) \<in> subs prb))"
using assms rb_sp_append_one_imp sub_rel_of.sp_append_one[OF subs_sub_rel_of] by fast










lemma (in finite_labels_RB) finite_labels_rsp :
  assumes "RB prb"
  assumes "ras \<in> subpaths (red prb) (subs prb)"
  shows   "finite_labels (trace (ui_as ras) (labelling (black prb)))"
using finite_labels_RB assms by (simp add : finite_labels_lts_def finite_labels_def)




lemma (in finite_labels_RB) sp_imp_ex_se_star_succ :
  assumes "RB prb"
  assumes "subpath (red prb) rl1 ras rl2 (subs prb)"
  shows   "\<exists> pc. se_star (confs prb rl1) (trace (ui_as ras) (labelling (black prb))) pc"
using assms(2)
      ex_se_star_succ[OF assms(1) fst_of_sp_is_vert[OF assms(2)] finite_labels_rsp [OF assms(1)]]
by blast




















(***********************************************)
(********** Propriétés des extensions **********)
(***********************************************)

(****** sub_rel_of *******)

lemma sub_rel_of_seE :
  assumes "RB prb"
  assumes "se_extends prb rg' ra pc' prb'"
  shows   "sub_rel_of (red prb') (subs prb')" 
using se_step[OF assms(1,2)] by (rule subs_sub_rel_of)

lemma sub_rel_of_subsumE :
  assumes "RB prb"
  assumes "subsum_extends prb sub subrel prb'"
  shows   "sub_rel_of (red prb') (subs prb')" 
using subsum_step[OF assms(1,2)] by (rule subs_sub_rel_of)

lemma sub_rel_of_abstractE :
  assumes "RB prb"
  assumes "abstract_extends prb rl e prb'"
  shows   "sub_rel_of (red prb') (subs prb')" 
using abstract_step[OF assms(1,2)] by (rule subs_sub_rel_of)


(****** wf_sub_rel *******)

lemma subs_wf_sub_rel_in_seE :
  assumes "RB prb"
  assumes "se_extends prb rg' ra pc' prb'"
  shows   "wf_sub_rel (subs prb')"
using se_step[OF assms(1,2)] by (rule  subs_wf_sub_rel)

lemma subs_wf_sub_rel_in_subsumE :
  assumes "RB prb"
  assumes "subsum_extends prb sub subrel prb'"
  shows   "wf_sub_rel (subs prb')"
using subsum_step[OF assms(1,2)] by (rule subs_wf_sub_rel)

lemma subs_wf_sub_rel_in_abstractE :
  assumes "RB prb"
  assumes "abstract_extends prb rl e prb'"
  shows   "wf_sub_rel (subs prb')"
using abstract_step[OF assms(1,2)] by (rule  subs_wf_sub_rel)


(****** wf_sub_rel_of *******)

lemma subs_wf_sub_rel_of_seE :
  assumes "RB prb"
  assumes "se_extends prb rg' ra pc' prb'"
  shows   "wf_sub_rel_of (red prb') (subs prb')"
using se_step[OF assms(1,2)] by (rule subs_wf_sub_rel_of)

lemma subs_wf_sub_rel_of_subsumE :
  assumes "RB prb"
  assumes "subsum_extends prb sub subrel prb'"
  shows   "wf_sub_rel_of (red prb') (subs prb')"
using subsum_step[OF assms(1,2)] by (rule subs_wf_sub_rel_of)

lemma subs_wf_sub_rel_of_abstractE :
  assumes "RB prb"
  assumes "abstract_extends prb rl e prb'"
  shows   "wf_sub_rel_of (red prb') (subs prb')"
using abstract_step[OF assms(1,2)] by (rule subs_wf_sub_rel_of)



(* finite_labels_pre_rb_lts *)













lemma (in finite_labels_RB) SE_rel :
  assumes "RB prb"
  assumes "subpath (red prb) rl1 ras rl2 (subs prb)"
  assumes "se_star (confs prb rl1) (trace (ui_as ras) (labelling (black prb))) pc"
  shows   "pc \<sqsubseteq>\<^sub>p\<^sub>c (confs prb rl2)"
using assms finite_labels_RB
proof (induct prb arbitrary : rl1 ras rl2 pc) {

  (* Si la partie rouge est vide, alors ras est vide et rl1 et rl2 sont égaux 
     (à la racine de la partie rouge de prb). On a donc : 
     - confs prb rl2 = confs prb rl1, et
     - pc            = confs prb rl1, or la subsomption est reflexive. *)
  
  case (base prb rl1 ras rl2 pc)

  moreover
  hence "ras = []" by (simp add : empty_graph_imp_sp_empty)

  ultimately
  show ?case by (simp add : Nil_sp)
}

next {

  (* On distingue 4 cas :
     1) rl1 et rl2 sont des sommets rouges du rb_lts de départ,
     2) rl1 est un sommet rouge du rb_lts de départ, rl2 est la cible du nouvel arc,
     3) rl1 est la cible du nouvel arc, rl2 est un sommet rouge du rb_lts de départ,
     4) rl1 et rl2 sont tous deux egaux a la cible du nouvel arc. *)

  case (se_step prb rg' ra pc' prb' rl1 ras rl2 pc)

  have "rl1 \<in> red_locations prb'"
  and  "rl2 \<in> red_locations prb'"
  using fst_of_sp_is_vert[OF se_step(4)]
        sub_rel_of.lst_of_sp_is_vert[OF sub_rel_of_seE[OF se_step(1,3)] se_step(4)]
  by simp_all
           
  hence "rl1 \<in> red_locations prb \<and> rl1 \<noteq> tgt ra \<or> rl1 = tgt ra"
  and   "rl2 \<in> red_locations prb \<and> rl2 \<noteq> tgt ra \<or> rl2 = tgt ra"
  using se_step by (auto simp add : extends_vertices)                             

  thus ?case
  proof (elim disjE conjE) {

    case goal1 (* rl1 et rl2 sont des sommets rouges du rb_lts de départ *)

    moreover
    hence "subpath (red prb) rl1 ras rl2 (subs prb)"
    using se_step
          sub_rel_of.sp_from_old_verts_imp_sp_in_old_w_subs 
          [OF subs_sub_rel_of[OF se_step(1)], of ra rg']
    by auto

    ultimately
    show ?thesis using se_step by auto
  }
  next {

    case goal2 (* rl1 est un sommet rouge du rb_lts de départ, rl2 est la cible du nouvel arc *)

    hence  "ras = [] \<or> (\<exists> p'. ras = p' @ [ra]
                            \<and> ra \<notin> set p'
                            \<and> subpath (red prb) rl1 p' (src ra) (subs prb))"
    using se_step 
          sub_rel_of.sp_to_new_arc_tgt_imp_w_subs 
          [OF subs_sub_rel_of[OF se_step(1)], of ra rg']
    by auto

    thus ?thesis
    proof (elim disjE exE conjE) {

      (* Si ras est vide, alors rl1 = rl2 (car il ne peut exister dans la relation 
         de subsomption de prb' de couple (tgt ra,_)). 
         Ce qui contredit le fait que rl1 n'est pas la cible du nouvel arc. *)

      assume "ras = []"

      have "rl1 = rl2"
      proof -
        have  "(rl1,rl2) \<in> (subs prb')"
        using se_step rb_Nil_sp[OF RB.se_step[OF se_step(1,3)], of rl1 rl2] 
              `ras = []` `rl1 \<noteq> tgt ra` `rl2 = tgt ra`
        by simp

        hence "rl2 \<in> subsumers (subs prb)" 
        using se_step  unfolding subsumers_conv by (simp del : split_paired_Ex) blast
          
        hence "tgt ra \<in> red_locations prb"
        using `rl2 = tgt ra` subs_sub_rel_of[OF se_step(1)] by (auto simp add : sub_rel_of_def)

        thus ?thesis using se_step by (auto simp add : extendsE2)
      qed
        
      thus ?thesis using `rl1 \<noteq> tgt ra` `rl2 = tgt ra` by simp
    }
    next {

      (* Si ras est de la forme ras'@[ra], alors il existe une configuration pc' telle que : 
         - pc' est obtenue par ES de ras' depuis (confs prb rl1),
         - pc' est subsumee par (confs prb (src ra)) (par IH), 
         - pc est obtenue par ES de ra depuis pc'. 
           
         Or (confs prb' rl2) est obtenue par ES de ra depuis (confs prb (src ra)), donc on a 
           bien la subsomption par monotonie de l'ES pour la subsomption. *)
      
      fix ras'

      assume "ras = ras' @ [ra]"
      and    "ra \<notin> set ras'"
      and    "subpath (red prb) rl1 ras' (src ra) (subs prb)"

      moreover
      then obtain pc'
      where "se_star (confs prb rl1) (trace (ui_as ras') (labelling (black prb))) pc'"
      and   "se pc' (labelling (black prb) (ui_arc ra)) pc"
      using se_step `rl1 \<noteq> tgt ra` se_star_append_one
      by simp blast
        
      ultimately
      have "pc' \<sqsubseteq>\<^sub>p\<^sub>c (confs prb (src ra))" using se_step by simp

      thus ?thesis 
      using se_step 
            `rl1 \<noteq> tgt ra` `rl2 = tgt ra` `se pc' (labelling (black prb) (ui_arc ra)) pc`
      by (auto simp add : se_mono_for_sub)
    } qed
  }
  next {
    
    case goal3 (* rl1 est la cible du nouvel arc, rl2 est un sommet rouge du rb_lts de départ *)

    (* Si rl1 est la cible du nouvel arc, alors ras est vide et rl2 est egalement la 
       cible du nouvel arc \<longrightarrow> contradiction *)

    moreover
    have "rl1 = rl2"
    proof -
      have "ras = []"
      using se_step goal3 
            sub_rel_of.sp_from_tgt_in_extendsion_is_Nil_w_subs
            [OF subs_sub_rel_of[OF se_step(1)], of ra rg' ras rl2]
      by auto
      
      hence  "(rl1,rl2) \<in> (subs prb')"
      using se_step goal3 rb_Nil_sp[OF RB.se_step[OF se_step(1,3)], of rl1 rl2] by simp

      hence "rl1 \<in> subsumees (subs prb)" 
      using se_step  unfolding subsumees_conv by (simp del : split_paired_Ex) blast
          
      hence "tgt ra \<in> red_locations prb"
      using `rl1 = tgt ra` subs_sub_rel_of[OF se_step(1)] by (auto simp add  : sub_rel_of_def)

      thus ?thesis using se_step by (auto simp add : extendsE2)
    qed

    ultimately
    show ?thesis by simp
  }
  next {

    case goal4 (* rl1 et rl2 sont tous deux egaux a la cible du nouvel arc *)

    (* Dans ce cas on obtient la subsomption par reflexivite *)

    moreover
    have "ras = []"
    using se_step goal4
          sub_rel_of.sp_from_tgt_in_extendsion_is_Nil_w_subs
          [OF subs_sub_rel_of[OF se_step(1)], of ra rg']
    by auto    

    ultimately
    show ?thesis using se_step by simp
  } qed
}

next

  case (subsum_step prb sub subrel prb' rl1 ras rl2 pc)

  (* On distingue selon que rl1 est le subsumee de la nouvelle subsomption ou non *)

  show ?case
  proof (case_tac "rl1 = subsumee sub")

    (* si rl1 est le subsumee de sub, alors soit ras est vide, soit ras est egalement 
       un chemin de la nouvelle partie rouge issu du subsumer de sub. *)

    assume "rl1 = subsumee sub"

    hence  "ras = [] \<or> subpath (red prb) (subsumer sub) ras rl2 (subs prb')"
    using subsum_step
          wf_sub_rel_of.sp_in_extends_imp2[of "red prb" "subs prb" "subsumee sub" "subsumer sub" "subs prb'" ras]
          subs_wf_sub_rel_of[OF subsum_step(1)]
    by simp

    thus ?thesis
    proof (elim disjE)

      (* si ras est vide, alors rl1 = rl2 ou il existe une subsomption dans la nouvelle partie 
         rouge liant rl1 à rl2 *)
      assume "ras = []"

      hence  "rl1 = rl2 \<or> (rl1,rl2) \<in> (subs prb')" 
      using subsum_step rb_Nil_sp[OF RB.subsum_step[OF subsum_step(1,3)]] by fast

      thus ?thesis 
      proof (elim disjE)

        (* si rl1 = rl2, alors pc = confs prb' rl1 = confs prb' rl2, donc on obtient la subsomption
           par reflexivite *)
        assume "rl1 = rl2" thus ?thesis using subsum_step `ras = []` by simp

      next

        (* si il existe une subsomption de rl1 vers rl2, alors rl2 est le subsumer de sub, 
           donc (confs prb' rl1) est subsumee par (confs prb' rl2), or puisque 
           ras est vide, on a pc = (confs prb' rl1), donc on a bien la subsomption. *)
        assume "(rl1, rl2) \<in> (subs prb')"

        hence  "rl2 = subsumer sub"
        using subsum_step `rl1 = subsumee sub`
              wf_sub_rel.subsumed_by_two_imp[of "subs prb'"]
              subs_wf_sub_rel_in_subsumE[OF subsum_step(1,3)]
              SubExt.extendsE1[of "red prb" "subs prb"]
        by auto

        thus ?thesis using subsum_step `rl1 = subsumee sub` `ras = []` by simp
     qed
      
    next

      (* Supposons que ras soit un chemin de la nouvelle partie rouge issu de (subsumer sub).

         Dans ce cas :
         - soit ras est un chemin de l'ancienne partie rouge, 
         - soit ras possede : 
           - un prefixe etant un chemin de l'ancienne partie rouge menant a (subsumee sub)), et 
           - un suffixe issu de (subsumee sub) etant un chemin menant a rl2 dans la nouvelle 
             partie rouge *)

      assume "subpath (red prb) (subsumer sub) ras rl2 (subs prb')"

      have   "subpath (red prb) (subsumer sub) ras rl2 (subs prb) \<or> 
              (\<exists> ras1 ras2. ras = ras1 @ ras2 
                        \<and> ras1 \<noteq> [] 
                        \<and> subpath (red prb) (subsumer sub) ras1 (subsumee sub) (subs prb) 
                        \<and> subpath (red prb) (subsumee sub) ras2 rl2 (subs prb'))"
      using subsum_step
            `subpath (red prb) (subsumer sub) ras rl2 (subs prb')`
            wf_sub_rel_of.sp_in_extends_imp
            [OF subs_wf_sub_rel_of[OF subsum_step(1)], 
             of "subsumee sub" "subsumer sub"]
            SubExt.extendsE3[of "red prb" "subs prb" "sub" "subs prb'"]
      by auto

      thus ?thesis
      proof (elim disjE exE conjE)

        (* si ras est un chemin menant a rl2 dans l'ancienne partie rouge, alors il existe 
           une configuration pc' telle que : 
           - pc' est obtenue par ES de ras depuis (subsumer sub),
           - pc' est subsumee par (confs prb' rl2) (par IH) 

           or puisque : 
           - pc est obtenue par ES de ras depuis (subsumee sub), 
           - (confs prb' (subsumee sub)) est subsumee par (confs prb' (subsumer sub)),

           on obtient pc \<sqsubseteq>\<^sub>p\<^sub>c pc' par motononie de l'ES pour la subsomption, et finalement 
           pc \<sqsubseteq>\<^sub>p\<^sub>c (confs prb' rl2) par transitivite.  *)

        assume "subpath (red prb) (subsumer sub) ras rl2 (subs prb)"

        moreover
        have "\<exists> pc'. se_star (confs prb (subsumer sub)) (trace (ui_as ras) (labelling (black prb'))) pc'"
        proof -

          have "subsumer sub \<in> red_locations prb"
          using subsum_step by (simp add : SubExt.extends_def)          
         
          moreover
          have "finite_labels (trace (ui_as ras) (labelling (black prb)))"
          using subsum_step
                finite_labels_RB.finite_labels_rsp
                [OF _ RB.subsum_step[OF subsum_step(1,3)], of ras]
          by (simp only : finite_labels_RB_def pre_RB.simps)  fast

          ultimately
          show ?thesis using subsum_step finite_labels_RB.ex_se_star_succ[of prb "subsumer sub" ] 
          by (simp add : finite_labels_RB_def)
        qed

        then obtain pc' 
        where "se_star (confs prb (subsumer sub)) (trace (ui_as ras) (labelling (black prb'))) pc'"
        by blast

        ultimately
        have "pc' \<sqsubseteq>\<^sub>p\<^sub>c confs prb' rl2" using subsum_step by simp

        thus ?thesis 
        using subsum_step
              `rl1 = subsumee sub`
              se_star_mono_for_sub[OF `se_star (confs prb (subsumer sub)) (trace (ui_as ras) (labelling (black prb'))) pc'` subsum_step(5)]
              sub_subsumed[of prb sub]
              subsums_trans[of pc pc']
        by auto

      next 

        (* Soient :
           - un préfixe de ras, ras1, etant un chemin de l'ancienne partie rouge menant de 
             subsumer sub a subsumee sub,
           - un suffixe de ras, ras2, menant de subsumee sub a rl2 dans la nouvelle partie rouge.

           On montre la subsomption par induction (par l'arriere) sur ras2. *)

        fix ras1 ras2

        assume "ras = ras1 @ ras2"
        and    "ras1 \<noteq> []"
        and    "subpath (red prb) (subsumer sub) ras1 (subsumee sub) (subs prb)"
        and    "subpath (red prb) (subsumee sub) ras2 rl2 (subs prb')"

        thus ?thesis 
        using subsum_step(5)
        proof (induct ras2 arbitrary : ras rl2 pc rule : rev_induct)
          
          (* Supposons que ras2 soit vide et soit pc' la configuration obtenue par ES de ras
             depuis (subsumer sub). *)

          case (Nil ras rl2 pc) 

          then obtain pc' 
          where "se_star (confs prb (subsumer sub)) (trace (ui_as ras) (labelling (black prb))) pc'"
          using subsum_step
                finite_labels_RB.sp_imp_ex_se_star_succ
                [of prb "subsumer sub" ras "subsumee sub"]
          by (simp add : finite_labels_RB_def)  blast

          (* Puisque pc est obtenue par ES de ras depuis (subsumee sub), pc est subsumee 
             par pc' par monotonie de l'ES pour la subsomption. *)
          hence "pc \<sqsubseteq>\<^sub>p\<^sub>c pc'"
          using subsum_step Nil `rl1 = subsumee sub`
             se_star_mono_for_sub
             [OF `se_star (confs prb (subsumer sub)) (trace (ui_as ras) (labelling (black prb))) pc'`]
          by auto          

          (* de plus, puisque ras2 est vide, ras est un chemin de l'ancienne partie rouge 
             menant à (subsumee sub). Donc pc' est subsumee par (confs prb (subsumee sub)) 
             par l'IH de subsum_step. *)

          (* ICI : on peut utiliser l'IH de subsum_step car le fait de n''abstraire que des feuilles 
             de la partie rouge permet de propager directement les abstractions et donc de 
             n'accepter que des subsomptions qui sont respectées par les abstractions. 
             C'était un gros problème en hiver/printemps 2015. *)
          have "pc' \<sqsubseteq>\<^sub>p\<^sub>c confs prb (subsumee sub)" 
          using subsum_step Nil
                `se_star (confs prb (subsumer sub)) (trace (ui_as ras) (labelling (black prb))) pc'`
          by simp
          
          (* de plus, puisque [] est un chemin allant de (subsumee sub) à rl2,
             (subsumee sub) = rl2 ou il existe une subsomption de (subsumee sub) à rl2.  *)
          have "subsumee sub = rl2 \<or> (subsumee sub,rl2) \<in> (subs prb')" 
          using subsum_step Nil rb_Nil_sp[OF RB.subsum_step[OF subsum_step(1,3)]] by auto

          thus ?case 
          proof (elim disjE)

            (* si subsumee sub = rl2, alors ras est un sous chemin de l'ancienne partie
               rouge allant de rl1 à rl2, alors   (confs prb' rl2) = (conf prb' (subsumee sub)).

               or pc' est subsumee par (conf prb' (subsumee sub)), et pc est subsumee par pc', 
               on obtient pc \<sqsubseteq>\<^sub>p\<^sub>c (confs prb' rl2)  par transitivité de la subsomption. *)
            assume "subsumee sub = rl2"

            thus ?thesis 
            using subsum_step 
                  `pc \<sqsubseteq>\<^sub>p\<^sub>c pc'` `pc' \<sqsubseteq>\<^sub>p\<^sub>c confs prb (subsumee sub)` 
                  subsums_trans 
            by simp

          next

            (* si il existe une subsomption de (subsumee sub) à rl2, alors rl2 = (subsumer sub).

               On a : 
               - pc subsumee par pc',
               - pc' subsumee par (confs prb (subsumee sub)) 
               - (confs prb (subsumee sub)) subsumee par (confs prb (subsumer sub)),

               donc on obtient pc \<sqsubseteq>\<^sub>p\<^sub>c (confs prb' rl2)  par (2x) transitivité de la subsomption.*)
            assume "(subsumee sub,rl2) \<in> (subs prb')"

            hence  "rl2 = subsumer sub"
            using subsum_step
                  SubExt.extendsE1[of "red prb" "subs prb"]
                  wf_sub_rel.subsumed_by_two_imp
                  [OF subs_wf_sub_rel_in_subsumE[OF subsum_step(1,3)]]
            by auto

            thus ?thesis 
            using subsum_step
                  `pc \<sqsubseteq>\<^sub>p\<^sub>c pc'` `pc' \<sqsubseteq>\<^sub>p\<^sub>c confs prb (subsumee sub)`
                  subsums_trans[of pc pc' "confs prb (subsumee sub)"]
                  subsums_trans[of pc "confs prb (subsumee sub)" "confs prb (subsumer sub)"]
            by simp
          qed
        
        next

          (* Supposons que le suffixe soit de la forme (ras2 @ [ra]), alors : 
           - ras2 est un chemin de la nouvelle partie rouge menant en la source de ra, 
           - il existe une configuration pc' obtenue de (confs prb' rl1) par ES de ras2 et
             telle que pc est obtenue depuis pc' par ES de ra. *)

          case (goal2 ra ras2 ras rl2 pc)

          moreover
          hence "subpath (red prb) (subsumee sub) ras2 (src ra) (subs prb')"
          using subsum_step rb_sp_append_one[OF RB.subsum_step[OF subsum_step(1,3)]]
          by simp

          moreover
          def t_ras12 \<equiv> "trace (ui_as (ras1@ras2)) (labelling (black prb'))"

          from goal2(2,6)
          have "se_star (confs prb' rl1) (t_ras12 @ [(labelling (black prb')) (ui_arc ra)]) pc"
          by (simp add : t_ras12_def)

          then obtain pc' 
          where "se_star (confs prb' rl1) t_ras12 pc'"
          and   "se pc' (labelling (black prb) (ui_arc ra)) pc"
          using subsum_step se_star_append_one by simp blast

          (* On a donc que pc' est subsumee par (confs prb (src ra)) par l'IH "courante" *)
          ultimately
          have "pc' \<sqsubseteq>\<^sub>p\<^sub>c (confs prb (src ra))" using subsum_step by (auto simp add : t_ras12_def)

          (* De plus, on nomme pc'' la configuration obtenue depuis (confs prb (src ra)) 
           par ES de ra. *)
          have  "src ra \<in> red_locations prb" 
          using subsum_step
                `subpath (red prb) (subsumee sub) ras2 (src ra) (subs prb')`
                sub_rel_of_subsumE[OF subsum_step(1,3)]
          by (auto simp add : sub_rel_of.lst_of_sp_is_vert)

          then obtain pc''
          where "se (confs prb (src ra)) (labelling (black prb) (ui_arc ra)) pc''"
          using subsum_step
             finite_labels_RB.confs_are_updatable[of prb "src ra"]
             updatable_imp_ex_se_suc[of "confs prb (src ra)" "labelling (black prb) (ui_arc ra)"]
          unfolding finite_labels_RB_def
          by simp blast
         
          (* montrons que pc'' est subsumee par (confs prb' rl2) *)
          have "pc'' \<sqsubseteq>\<^sub>p\<^sub>c confs prb' rl2"
          proof -
          
            (* En effet, [ra] est un sous-chemin de la partie rouge de départ terminant en rl2 et
             donc soit sa cible est rl2, soit il existe une subsomption allant de cette cible 
             à rl2. *)
            have "subpath (red prb) (src ra) [ra] (tgt ra) (subs prb)"
            and  "tgt ra = rl2 \<or> (tgt ra, rl2) \<in> (subs prb')"
            using subsum_step goal2
                  `src ra \<in> red_locations prb`
                  rb_sp_append_one[OF RB.subsum_step[OF subsum_step(1,3)], of "subsumee sub" ras2 ra rl2]
            by (auto simp add : arc_sp_eq)
             
            thus ?thesis 
            proof (elim disjE)

              (* si la cible de ra est rl2, alors on peut utiliser l'IH de subsum_step sur [ra] *)
              assume "tgt ra = rl2" 
              thus ?thesis 
              using subsum_step 
                    `subpath (red prb) (src ra) [ra] (tgt ra) (subs prb)`
                   `se (confs prb (src ra)) (labelling (black prb) (ui_arc ra)) pc''`
              by (auto simp add :  se_star_one)

            next
          
              (* si il existe une subsomption de la cible de ra vers rl2 *)
              assume "(tgt ra, rl2) \<in> (subs prb')"

              (* alors la configuration en tgt ra est subsumee par celle de rl2 *)
              hence  "confs prb (tgt ra) \<sqsubseteq>\<^sub>p\<^sub>c confs prb rl2"
              using subsum_step
                    sub_subsumed[OF RB.subsum_step[OF subsum_step(1,3)], of "(tgt ra, rl2)"]
              by simp

              (* or pc'' est subsumee par la conf en rl2, par l'IH de subsum_step *)
              moreover
              have "pc'' \<sqsubseteq>\<^sub>p\<^sub>c confs prb (tgt ra)" 
              using subsum_step
                    `subpath (red prb) (src ra) [ra] (tgt ra) (subs prb)`
                    `se (confs prb (src ra)) (labelling (black prb) (ui_arc ra)) pc''`
              by (auto simp add :  se_star_one)

              (* donc pc'' est subsumee par la conf en rl2 par transitivite *)
              ultimately
              show ?thesis using subsum_step subsums_trans by auto
            qed
          qed

          (* de plus, puisque 
           - pc   est obtenue par ES de ra depuis pc',
           - pc'' est obtenue par ES de ra depuis (confs prb (src ra)),
           - pc' est subsumee par (confs prb (src ra)),
           on a que pc est subsumee par pc'', par monotonie de l'ES pour la subsomption. *)
        moreover
        have "pc \<sqsubseteq>\<^sub>p\<^sub>c pc''" 
        using se_mono_for_sub[OF `se (confs prb (src ra)) (labelling (black prb) (ui_arc ra)) pc''`
                                 `se pc' (labelling (black prb) (ui_arc ra)) pc`
                                 `pc' \<sqsubseteq>\<^sub>p\<^sub>c (confs prb (src ra))`]
        by assumption

        (* On obtient donc la subsomption par transitivite. *)
        ultimately
        show ?case using subsums_trans by fast
        qed
      qed
    qed
 
  next

    (* Si rl1 n'est pas le nouveau subsumee, alors on distingue les cas ou :
       - ras n'utilise pas la nouvelle subsomption et est donc un chemin de la partie rouge de 
         départ,
       - ras utilise la nouvelle subsomption et se decompose en un prefixe de la partie rouge de 
         depart s'arrêtant en le nouveau subsumee, et un suffixe de la nouvelle debutant en ce 
         subsumee et terminant en rl2. *)

    assume "rl1 \<noteq> subsumee sub"

    hence  "subpath (red prb) rl1 ras rl2 (subs prb) \<or> 
            (\<exists> ras1 ras2. ras = ras1 @ ras2 
                        \<and> ras1 \<noteq> [] 
                        \<and> subpath (red prb) rl1 ras1 (subsumee sub) (subs prb) 
                        \<and> subpath (red prb) (subsumee sub) ras2 rl2 (subs prb'))"
    using subsum_step 
          wf_sub_rel_of.sp_in_extends_imp
          [OF subs_wf_sub_rel_of[OF subsum_step(1)], of "subsumee sub" "subsumer sub"]
    by auto
    
    thus ?thesis
    proof (elim disjE exE conjE) {

      (* si ras est un chemin de la partie rouge de depart, on obtient la subsomption par IH *)
      assume "subpath (red prb) rl1 ras rl2 (subs prb)" thus ?thesis using subsum_step by simp
    }
    next { 

      (* Sinon : par induction (par l'arrière) sur le suffixe *)

      fix ras1 ras2
       
      assume "ras = ras1 @ ras2"
      and    "ras1 \<noteq> []"
      and    "subpath (red prb) rl1 ras1 (subsumee sub) (subs prb)"
      and    "subpath (red prb) (subsumee sub) ras2 rl2 (subs prb')"

      thus ?thesis
      using `se_star (confs prb' rl1) (trace (ui_as ras) (labelling (black prb'))) pc`
      proof (induct ras2 arbitrary : ras rl2 pc rule : rev_induct) {

        case (Nil ras rl2 pc)

        (* Si le suffixe est vide, alors rl2 peut etre soit le subsumee de la nouvelle 
           subsomption, soit son subsumer. *)
        have  "rl2 = subsumee sub \<or> rl2 = subsumer sub" 
        proof -

          have "subsumee sub = rl2 \<or> (subsumee sub, rl2) \<in> (subs prb')\<^sup>+"
          using Nil by (auto simp add : Nil_sp)

          thus ?thesis 
          proof (elim disjE)
            assume "subsumee sub = rl2" thus ?thesis by simp
          next
            assume "(subsumee sub, rl2) \<in> (subs prb')\<^sup>+"

            hence  "(subsumee sub, rl2) \<in> (subs prb')"
            using wf_sub_rel_of.in_trancl_imp
                  [OF subs_wf_sub_rel_of_subsumE[OF subsum_step(1,3)]]
            by simp

            moreover
            have   "(subsumee sub, subsumer sub) \<in> (subs prb')"
            using subsum_step by (auto simp add : SubExt.extendsE1)
            
            ultimately
            show ?thesis 
            using wf_sub_rel.subsumed_by_two_imp
                  [OF subs_wf_sub_rel_in_subsumE[OF subsum_step(1,3)]]
            by auto
          qed
        qed

        (* de plus, ras est donc un chemin de la partie rouge de depart terminant en le 
           nouveau subsumee, donc pc est subsume par (confs prb (subsumee sub)) par IH. *)
        have "pc \<sqsubseteq>\<^sub>p\<^sub>c confs prb (subsumee sub)" using subsum_step Nil by auto

        show ?case 
        using `rl2 = subsumee sub \<or> rl2 = subsumer sub`
        proof (elim disjE) {

          (* si rl2 est le subsumee de sub, alors on a bien la subsomption *)
          case goal1 thus ?thesis using subsum_step `pc \<sqsubseteq>\<^sub>p\<^sub>c confs prb (subsumee sub)` by simp
        }
        next {

          (* si rl2 est le subsumer de sub, alors (confs prb rl2) subsume (confs prb (subsumee sub))
             par def. de l'extension par ajout de sub. On obtient donc la subsomption par 
             transitivité. *)

          case goal2

          moreover
          have  "confs prb (subsumee sub) \<sqsubseteq>\<^sub>p\<^sub>c confs prb (subsumer sub)"
          using subsum_step by simp

          ultimately
          show ?case using subsum_step `pc \<sqsubseteq>\<^sub>p\<^sub>c confs prb (subsumee sub)` subsums_trans by fastforce
        } qed
      }

      next {

        (* Supposons que le suffixe soit de la forme (ras2 @ [ra]), alors : 
           - ras2 est un chemin de la nouvelle partie rouge menant en la source de ra, 
           - il existe une configuration pc' obtenue de (confs prb' rl1) par ES de ras2 et
             telle que pc est obtenue depuis pc' par ES de ra. *)

        case (goal2 ra ras2 ras rl2 pc)

        moreover
        hence "subpath (red prb) (subsumee sub) ras2 (src ra) (subs prb')"
        using subsum_step rb_sp_append_one[OF RB.subsum_step[OF subsum_step(1,3)]]
        by simp

        moreover
        def t_ras12 \<equiv> "trace (ui_as (ras1@ras2)) (labelling (black prb'))"

        from goal2(2,6)
        have "se_star (confs prb' rl1) (t_ras12 @ [(labelling (black prb')) (ui_arc ra)]) pc"
        by (simp add : t_ras12_def)

        then obtain pc' 
        where "se_star (confs prb' rl1) t_ras12 pc'"
        and   "se pc' (labelling (black prb) (ui_arc ra)) pc"
        using subsum_step se_star_append_one by simp blast

        (* On a donc que pc' est subsumee par (confs prb (src ra)) par l'IH "courante" *)
        ultimately
        have "pc' \<sqsubseteq>\<^sub>p\<^sub>c (confs prb (src ra))" using subsum_step by (auto simp add : t_ras12_def)

        (* De plus, on nomme pc'' la configuration obtenue depuis (confs prb (src ra)) 
           par ES de ra. *)
        have  "src ra \<in> red_locations prb" 
        using subsum_step
              `subpath (red prb) (subsumee sub) ras2 (src ra) (subs prb')`
              sub_rel_of_subsumE[OF subsum_step(1,3)]
        by (auto simp add : sub_rel_of.lst_of_sp_is_vert)

       
        then obtain pc''
        where "se (confs prb (src ra)) (labelling (black prb) (ui_arc ra)) pc''"
        using subsum_step
              finite_labels_RB.confs_are_updatable[of prb "src ra"]
              updatable_imp_ex_se_suc[of "confs prb (src ra)" "labelling (black prb) (ui_arc ra)"]
        unfolding finite_labels_RB_def
        by simp blast

        (* montrons que pc'' est subsumee par (confs prb' rl2) *)
        have "pc'' \<sqsubseteq>\<^sub>p\<^sub>c confs prb' rl2"
        proof -

          (* En effet, [ra] est un sous-chemin de la partie rouge de départ terminant en rl2 et
             donc soit sa cible est rl2, soit il existe une subsomption allant de cette cible 
             à rl2. *)
          have "subpath (red prb) (src ra) [ra] (tgt ra) (subs prb)"
          and  "tgt ra = rl2 \<or> (tgt ra, rl2) \<in> (subs prb')"
          using subsum_step goal2
                `src ra \<in> red_locations prb`
                rb_sp_append_one[OF RB.subsum_step[OF subsum_step(1,3)]]
          by (auto simp add : arc_sp_eq)

          thus ?thesis 
          proof (elim disjE)

            (* si la cible de ra est rl2, alors on peut utiliser l'IH de subsum_step sur [ra] *)
            assume "tgt ra = rl2" 
            thus ?thesis 
            using subsum_step 
                  `subpath (red prb) (src ra) [ra] (tgt ra) (subs prb)`
                  `se (confs prb (src ra)) (labelling (black prb) (ui_arc ra)) pc''`
            by (auto simp add :  se_star_one)

          next
          
            (* si il existe une subsomption de la cible de ra vers rl2 *)
            assume "(tgt ra, rl2) \<in> (subs prb')"

            (* alors la configuration en tgt ra est subsumee par celle de rl2 *)
            hence  "confs prb (tgt ra) \<sqsubseteq>\<^sub>p\<^sub>c confs prb rl2"
            using subsum_step
                  sub_subsumed[OF RB.subsum_step[OF subsum_step(1,3)], of "(tgt ra, rl2)"]
            by simp

            (* or pc'' est subsumee par la conf en rl2, par l'IH de subsum_step *)
            moreover
            have "pc'' \<sqsubseteq>\<^sub>p\<^sub>c confs prb (tgt ra)" 
            using subsum_step
                  `subpath (red prb) (src ra) [ra] (tgt ra) (subs prb)`
                  `se (confs prb (src ra)) (labelling (black prb) (ui_arc ra)) pc''`
            by (auto simp add :  se_star_one)

            (* donc pc'' est subsumee par la conf en rl2 par transitivite *)
            ultimately
            show ?thesis using subsum_step subsums_trans by auto
          qed
        qed

        (* de plus, puisque 
           - pc   est obtenue par ES de ra depuis pc',
           - pc'' est obtenue par ES de ra depuis (confs prb (src ra)),
           - pc' est subsumee par (confs prb (src ra)),
           on a que pc est subsumee par pc'', par monotonie de l'ES pour la subsomption. *)
        moreover
        have "pc \<sqsubseteq>\<^sub>p\<^sub>c pc''" 
        using se_mono_for_sub[OF `se (confs prb (src ra)) (labelling (black prb) (ui_arc ra)) pc''`
                                 `se pc' (labelling (black prb) (ui_arc ra)) pc`
                                 `pc' \<sqsubseteq>\<^sub>p\<^sub>c (confs prb (src ra))`]
        by assumption

        (* On obtient donc la subsomption par transitivite. *)
        ultimately
        show ?case using subsums_trans by fast
      } qed
    } qed
  qed

next

  case (abstract_step prb rl e prb' rl1 ras rl2 pc)

  show ?case
  proof (case_tac "rl1 = rl")

    (* si rl1 est le sommet rouge ou a eu lieu l'abstraction, alors ras est vide et dans ce 
       cas on a : 
       - rl1 = rl2, et 
       - pc = confs prb rl2,
       Or (confs prb' rl2) est une abstraction de (confs prb rl2), donc on a bien la subsomption.*)
    
    assume "rl1 = rl"

    moreover
    hence   "ras = []" 
    using abstract_step
          sub_rel_of.sp_from_de_empty[OF subs_sub_rel_of[OF abstract_step(1)], of rl1 ras rl2]
    by simp

    moreover
    have  "rl2 = rl" 
    proof -

      have "rl1 = rl2 \<or> (rl1, rl2) \<in> (subs prb)\<^sup>+"
      using abstract_step(3,4) `ras = []` by (simp add : Nil_sp)

      moreover
      have "(rl1, rl2) \<notin> (subs prb)\<^sup>+"
      proof (intro notI)
        assume "(rl1, rl2) \<in> (subs prb)\<^sup>+"

        hence  "(rl1, rl2) \<in> (subs prb)"
        by (rule wf_sub_rel_of.in_trancl_imp[OF subs_wf_sub_rel_of[OF abstract_step(1)]])

        thus False 
        using abstract_step `rl1 = rl`
        unfolding subsumees_conv by blast
      qed
      
      ultimately
      show ?thesis using `rl1 = rl` by simp
    qed

    ultimately
    show ?thesis 
    using abstract_step 
          abstract_subsums[of "confs prb rl2" e]
          subsums_trans[of pc "confs prb rl2" "confs prb' rl2"]
    by simp
  
  next

    (* Supposons que rl1 ne soit pas le sommet ou a eu lieu la subsomption. 
    
       Si rl2 = rl, alors pc est subsume par (confs prb rl2) par IH et (confs prb rl2) est 
       subsume par (confs prb' rl2) par def de l'abstraction, donc on a bien la subsomption par 
       transitivité. 
       
       Si rl2 n'est pas rl, alors on a la subsomption par IH. *) 

    assume "rl1 \<noteq> rl"

    show ?thesis
    proof (case_tac "rl2 = rl")
      assume "rl2 = rl"

      hence  "confs prb rl2 \<sqsubseteq>\<^sub>p\<^sub>c confs prb' rl2" 
      using abstract_step by (auto simp add : abstract_subsums)

      moreover
      have  "pc \<sqsubseteq>\<^sub>p\<^sub>c confs prb rl2" using abstract_step `rl1 \<noteq> rl` by auto

      ultimately
      show ?thesis by (simp add : subsums_trans[of pc "confs prb rl2" "confs prb' rl2"])
    next
      assume "rl2 \<noteq> rl" thus ?thesis using abstract_step `rl1 \<noteq> rl` by simp 
    qed
  qed
qed









(*******************************)
(********** free_from **********)
(*******************************)

definition free_from :: 
  "('a, 'b, 'c) pre_RB_scheme \<Rightarrow>
   ('a \<times> nat)                     \<Rightarrow>
   'a arc                         \<Rightarrow> bool"
where
  "free_from prb rl ba \<equiv> 
      ba \<in> out_arcs (black prb) (fst rl) 
    \<and> (\<forall> ra \<in> out_arcs (red prb) rl. ui_arc ra \<noteq> ba) 
    \<and> (\<exists> pc'. se (confs prb rl) (labelling (black prb) ba) pc' \<and> sat pc')"


(********** se_extends **********)

lemma free_in_seE_imp_free :
  assumes "se_extends prb rg' ra pc' prb'"
  assumes "rl \<in> red_locations prb"
  assumes "free_from prb' rl ba"
  shows   "free_from prb  rl ba"
proof -
  
  have "rl \<noteq> tgt ra" using assms(1,2) by (auto simp add : extendsE2)

  have "ba \<in> out_arcs (black prb) (fst rl) "
  using assms by (simp add : free_from_def)

  moreover
  have "\<forall> ra \<in> out_arcs (red prb) rl. ui_arc ra \<noteq> ba"
  using assms `rl \<noteq> tgt ra`
  by (case_tac "rl = src ra")
     (auto simp add : extends_src_out_arcs extends_not_end_out_arcs free_from_def)

  moreover
  have "\<exists> pc'. se (confs prb rl) (labelling (black prb) ba) pc' \<and> sat pc'"
  using assms `rl \<noteq> tgt ra` by (auto simp add : free_from_def extends_def)

  ultimately
  show ?thesis by (simp add : free_from_def)
qed



lemma free_not_src_imp_free_in_seE :
  assumes "se_extends prb rg' ra pc' prb'"
  assumes "rl \<in> red_locations prb"
  assumes "rl \<noteq> src ra"
  assumes "free_from prb  rl ba"
  shows   "free_from prb' rl ba"
proof -

  from assms(1,2) have "rl \<noteq> tgt ra" by (auto simp add : extendsE2)

  have "ba \<in> out_arcs (black prb') (fst rl)"
  using assms(1,4) by (auto simp add : free_from_def)

  moreover
  have "\<forall> ra \<in> out_arcs (red prb') rl. ui_arc ra \<noteq> ba"
  using assms(1,3,4) `rl \<noteq> tgt ra` 
  by (auto simp add : free_from_def extends_not_end_out_arcs)
  
  moreover
  have "\<exists> pc'. se (confs prb' rl) (labelling (black prb') ba) pc' \<and> sat pc'" 
  using assms(1,4) `rl \<noteq> tgt ra` by (simp add : free_from_def)

  ultimately
  show ?thesis by (simp add : free_from_def)
qed


lemma free_not_ra_imp_free_in_seE :
  assumes "se_extends prb rg' ra pc' prb'"
  assumes "ba \<noteq> ui_arc ra"
  assumes "free_from prb  (src ra) ba"
  shows   "free_from prb' (src ra) ba"
proof -

  have "ba \<in> out_arcs (black prb') (fst (src ra))"
  using assms(1,3) by (auto simp add : free_from_def)

  moreover
  have "\<forall> ra \<in> out_arcs (red prb') (src ra). ui_arc ra \<noteq> ba"
  using assms by (auto simp add : free_from_def out_arcs_def extends_arcs)
  
  moreover
  have "\<exists> pc'. se (confs prb' (src ra)) (labelling (black prb') ba) pc' \<and> sat pc'" 
  using assms(1,3) extends_imp_diff_ends by (auto simp add : free_from_def)

  ultimately
  show ?thesis by (simp add : free_from_def)
  
qed



(********** subsum_extends **********)

lemma free_in_subsumE_imp_free :
  assumes "subsum_extends prb sub subrel prb'"
  assumes "free_from prb' rl ba"
  shows   "free_from prb  rl ba"
using assms by (auto simp add : free_from_def)


lemma free_imp_free_in_subsumE :
  assumes "subsum_extends prb sub subrel prb'"
  assumes "free_from prb  rl ba"
  shows   "free_from prb' rl ba"
using assms by (auto simp add : free_from_def)




(********** abstract_extends **********)

lemma (in finite_labels_RB) free_imp_free_in_abstractE :
  assumes "RB prb"
  assumes "abstract_extends prb rl e prb'"
  assumes "free_from prb  rl' ba"
  shows   "free_from prb' rl' ba"
proof -

  have "ba \<in> out_arcs (black prb') (fst rl')"
  and  "\<forall>ra\<in>out_arcs (red prb') rl'. ui_arc ra \<noteq> ba"
  using assms by (auto simp add : free_from_def)

  moreover
  have "\<exists>pc'. se (confs prb' rl') (labelling (black prb') ba) pc' \<and> sat pc'"
  proof (case_tac "rl' = rl")

    case goal1
 
    have "rl \<in> red_locations prb" using assms(2) by simp

    hence "finite (\<pi> (confs prb rl))" 
    and   "\<forall> e \<in> \<pi> (confs prb rl). finite (Bexp.vars e)"
    using finite_pred[OF assms(1)] 
          finite_pred_constr_symvars[OF assms(1)]
    by simp_all

    then obtain apc' where "se (abstract e (confs prb rl)) (labelling (black prb) ba) apc'"
    using finite_exps_imp_finite_abstract_pred
          finite_pred_imp_se_updatable
          updatable_imp_ex_se_suc
    by blast

    moreover
    have "sat apc'"
    proof -

      from assms(3)
      obtain pc' where "se (confs prb rl') (labelling (black prb) ba) pc'"
                 and   "sat pc'"
      unfolding free_from_def by blast

      thus ?thesis
      using assms(2) goal1 
            `se (abstract e (confs prb rl)) (labelling (black prb) ba) apc'`
            rl_sat[OF assms(1) `rl \<in> red_locations prb`]
            se_sat_and_se_abstract_imp_sat
      by auto
    qed

    ultimately
    show ?thesis using assms(2) goal1 by simp blast
  next
    case goal2 thus ?thesis using assms by (auto simp add : free_from_def)
  qed

  ultimately
  show ?thesis by (simp add : free_from_def)
qed



lemma free_in_abstractE_imp_free :
  assumes "abstract_extends prb rl e prb'"
  assumes "rl' \<noteq> rl"
  assumes "free_from prb' rl' ba"
  shows   "free_from prb  rl' ba"
using assms by (auto simp add : free_from_def)









(****************************)
(********** fringe **********)
(****************************)

definition fringe :: 
  "('a, 'b, 'c) pre_RB_scheme \<Rightarrow> ('a \<times> nat) set"
where
  "fringe prb \<equiv> 
    {rl \<in> red_locations prb. rl \<notin> subsumees (subs prb) 
                           \<and> (out_arcs (black prb) (fst rl) = {} \<or> (\<exists> ba. free_from prb rl ba))}"


(*********** base ***********)

lemma fringe_of_empty_red_1 :
  assumes "fst (root (red prb)) = init (black prb)"
  assumes "arcs (red prb) = {}"
  assumes "subs prb = {}"
  assumes "out_arcs (black prb) (init (black prb)) = {}"
  shows   "fringe prb = {root (red prb)}"
using assms
unfolding set_eq_iff singleton_iff
proof (intro allI iffI)
  case goal1 thus ?case by (simp add : fringe_def vertices_conv)
next
  case goal2 thus ?case by (simp add : fringe_def vertices_conv subsumees_conv)
qed

lemma fringe_of_empty_red_2 :
  assumes "arcs (red prb) = {}"
  assumes "subs prb = {}"
  assumes "\<exists> ba. free_from prb (root (red prb)) ba"
  shows   "fringe prb = {root (red prb)}"
using assms
unfolding set_eq_iff singleton_iff 
proof (intro allI iffI)
  case goal1 thus ?case by (simp add : fringe_def vertices_conv)
next
  case goal2 thus ?case by (simp add : fringe_def vertices_conv subsumees_conv)
qed

lemma fringe_of_empty_red_3 :
  assumes "arcs (red prb) = {}"
  assumes "subs prb = {}"
  assumes "out_arcs (black prb) (init (black prb)) \<noteq> {}"
  assumes "fst (root (red prb)) = init (black prb)"
  assumes "\<not> (\<exists> ba. free_from prb (root (red prb)) ba)"
  shows   "fringe prb = {}"
unfolding set_eq_iff
using assms
proof (subst empty_iff, intro allI iffI)
  case goal1 thus ?case using assms by (auto simp add : fringe_def vertices_conv)
next
  case goal2 thus ?case using assms by (simp add : fringe_def vertices_conv subsumees_conv)
qed



(*********** se_extends ***********)

lemma seE_fringe1 :
  assumes "RB prb"
  assumes "se_extends prb rg' ra pc' prb'"
  assumes "\<exists> ba. free_from prb' (src ra) ba"
  assumes "out_arcs (black prb) (fst (tgt ra)) = {} \<or> (\<exists> ba. free_from prb' (tgt ra) ba)"
  shows   "fringe prb' = fringe prb \<union> {tgt ra}"
unfolding set_eq_iff Un_iff singleton_iff
proof (intro allI iffI)

  fix rl

  assume L : "rl \<in> fringe prb'"

  hence  "rl \<in> red_locations prb \<or> rl = tgt ra" 
  using assms(2) by (auto simp add : fringe_def extends_vertices)

  thus   "rl \<in> fringe prb \<or> rl = tgt ra" 
  proof (elim disjE)

    assume "rl \<in> red_locations prb"

    moreover
    have   "rl \<notin> subsumees (subs prb)" using assms(2) L by (simp add : fringe_def)

    moreover
    have "rl \<noteq> tgt ra" using assms(2) `rl \<in> red_locations prb`by (auto simp add : extendsE2)
    
    hence "out_arcs (black prb) (fst rl) = {} \<or> (\<exists>ba. free_from prb rl ba)" 
    using assms(2) L `rl \<in> red_locations prb` free_in_seE_imp_free by (simp add : fringe_def) blast

    ultimately
    show ?thesis by (simp add : fringe_def)

  next

    assume "rl = tgt ra" thus ?thesis by simp

  qed

next

  fix rl

  assume R : "rl \<in> fringe prb \<or> rl = tgt ra"

  hence  "rl \<in> red_locations prb'"
  using assms(2) by (auto simp add : fringe_def extends_vertices)

  moreover
  have "rl \<notin> subsumees (subs prb')"
  using R
  proof (elim disjE)
    assume "rl \<in> fringe prb" thus ?thesis using assms(2) by (auto simp add : fringe_def)
  next
    assume "rl = tgt ra"
    
    show ?thesis
    proof (intro notI)
      assume "rl \<in> subsumees (subs prb')"

      hence  "rl \<in> subsumees (subs prb)" using assms(2) by auto

      hence  "tgt ra \<in> red_locations prb"
      using assms(2) `rl = tgt ra` subs_sub_rel_of[OF assms(1)]
      by (auto simp add : sub_rel_of_def subsumees_def)

      thus False using assms(2) by (auto simp add : extendsE2)
    qed
  qed

  moreover
  have "out_arcs (black prb') (fst rl) = {} \<or> (\<exists> ba. free_from prb' rl ba)"
  using R
  proof (elim disjE)
    assume "rl \<in> fringe prb" 
    
    thus ?thesis
    proof (case_tac "rl = src ra")
      assume "rl = src ra" thus ?thesis using assms(3) by simp
    next
      assume "rl \<noteq> src ra"

      moreover
      have   "rl \<noteq> tgt ra" and "rl \<in> red_locations prb" 
      using assms(2) `rl \<in> fringe prb` by (auto simp add : fringe_def extendsE2)

      ultimately
      show ?thesis 
      using assms(2) `rl \<in> fringe prb` free_not_src_imp_free_in_seE
      by (simp add : fringe_def extends_not_end_out_arcs) blast
    qed
  next
    assume "rl = tgt ra" thus ?thesis using assms(2,4) by simp
  qed

  ultimately
  show   "rl \<in> fringe prb'" by (simp add : fringe_def)

qed







lemma seE_fringe2 :
  assumes "RB prb"
  assumes "se_extends prb rg' ra pc' prb'"
  assumes "\<exists> ba. free_from prb' (src ra) ba"
  assumes "\<not> (out_arcs (black prb) (fst (tgt ra)) = {} \<or> (\<exists> ba. free_from prb' (tgt ra) ba))"
  shows   "fringe prb' = fringe prb"
unfolding set_eq_iff
proof (intro allI iffI)

  fix rl

  assume L : "rl \<in> fringe prb'"

  moreover
  have "tgt ra \<notin> fringe prb'" using assms(2,4) by (simp add : fringe_def)

  ultimately
  have   "rl \<in> red_locations prb" using assms(2) by (auto simp add : fringe_def extends_vertices)

  moreover
  have   "rl \<notin> subsumees (subs prb)" using assms(2) L by (auto simp add : fringe_def)

  moreover
  have   "out_arcs (black prb) (fst rl) = {} \<or> (\<exists>ba. free_from prb rl ba)"
  using assms(2) L `rl \<in> red_locations prb` free_in_seE_imp_free by (simp add : fringe_def) blast

  ultimately
  show "rl \<in> fringe prb" by (simp add : fringe_def)

next

  fix rl

  assume "rl \<in> fringe prb"

  hence "rl \<in> red_locations prb'" 
  using assms(2) by (auto simp add : fringe_def extends_vertices)

  moreover
  have "rl \<notin> subsumees (subs prb')"
  proof (intro notI)
    assume "rl \<in> subsumees (subs prb')"

    hence  "rl \<in> subsumees (subs prb)" using assms(2) by auto

    thus False using `rl \<in> fringe prb` by (auto simp add : fringe_def)
  qed

  moreover
  have "out_arcs (black prb') (fst rl) = {} \<or> (\<exists> ba. free_from prb' rl ba)"
  proof (case_tac "rl = src ra")
    assume "rl = src ra" thus ?thesis using assms(3) by simp
  next
    assume "rl \<noteq> src ra"

    moreover
    have   "rl \<noteq> tgt ra" and "rl \<in> red_locations prb" 
    using assms(2) `rl \<in> fringe prb` by (auto simp add : fringe_def extendsE2)

    ultimately
    show ?thesis 
    using assms(2) `rl \<in> fringe prb` free_not_src_imp_free_in_seE
    by (simp add : fringe_def extends_not_end_out_arcs) blast
  qed

  ultimately
  show "rl \<in> fringe prb'" by (simp add : fringe_def)

qed






lemma seE_fringe3 :
  assumes "RB prb"
  assumes "se_extends prb rg' ra pc' prb'"
  assumes "\<not> (\<exists> ba. free_from prb' (src ra) ba)"
  assumes "out_arcs (black prb) (fst (tgt ra)) = {} \<or> (\<exists> ba. free_from prb' (tgt ra) ba)"
  shows   "fringe prb' = fringe prb - {src ra} \<union> {tgt ra}"
unfolding set_eq_iff Diff_iff Un_iff singleton_iff
proof (intro allI iffI)

  fix rl 
  
  assume L : "rl \<in> fringe prb'"

  hence  "rl \<in> red_locations prb \<or> rl = tgt ra" 
  using assms(2) by (auto simp add : fringe_def extends_vertices)

  thus   "rl \<in> fringe prb \<and> rl \<noteq> src ra \<or> rl = tgt ra"
  proof (elim disjE)

    assume "rl \<in> red_locations prb"

    moreover
    have "rl \<notin> subsumees (subs prb)" using assms(2) L by (simp add : fringe_def)

    moreover 
    have "rl \<noteq> src ra" using assms(2,3) L by (auto simp add : out_arcs_def fringe_def)

    moreover
    hence "out_arcs (black prb) (fst rl) = {} \<or> (\<exists>ba. free_from prb rl ba)" 
    using assms(2) L `rl \<in> red_locations prb` free_in_seE_imp_free
    by (simp add : fringe_def) blast

    ultimately
    show ?thesis by (simp add : fringe_def)
  next                                                                            
    assume "rl = tgt ra" thus ?thesis by simp
  qed

next

  fix rl 
  
  assume R : "rl \<in> fringe prb \<and> rl \<noteq> src ra \<or> rl = tgt ra" 

  hence  "rl \<in> red_locations prb'"
  using assms(2) by (auto simp add : fringe_def extends_vertices)

  moreover
  have   "rl \<notin> subsumees (subs prb')"
  using  R
  proof (elim disjE conjE)
    assume "rl \<in> fringe prb" thus ?thesis using assms(2) by (simp add  : fringe_def)
  next
    assume "rl = tgt ra"
    
    hence  "rl \<notin> red_locations prb" using assms(2) by (auto simp add : extendsE2)
    
    thus   ?thesis using assms(2) subs_sub_rel_of[OF assms(1)]
    by (auto simp add : sub_rel_of_def)
  qed

  moreover
  have   "out_arcs (black prb') (fst rl) = {} \<or> (\<exists> ba. free_from prb' rl ba)"
  using  R
  proof (elim disjE)
    assume "rl \<in> fringe prb \<and> rl \<noteq> src ra"
    thus ?thesis using assms(2) free_not_src_imp_free_in_seE by (simp add : fringe_def) blast
  next
    assume "rl = tgt ra" thus ?thesis using assms(2,4) by simp
  qed

  ultimately
  show   "rl \<in> fringe prb'" by (simp add : fringe_def)
qed






lemma seE_fringe4 :
  assumes "RB prb"
  assumes "se_extends prb rg' ra pc' prb'"
  assumes "\<not> (\<exists> ba. free_from prb' (src ra) ba)"
  assumes "\<not> (out_arcs (black prb) (fst (tgt ra)) = {} \<or> (\<exists> ba. free_from prb' (tgt ra) ba))"
  shows   "fringe prb' = fringe prb - {src ra}"
unfolding set_eq_iff Diff_iff singleton_iff
proof (intro allI iffI)
  fix rl

  assume L : "rl \<in> fringe prb'"

  hence   "rl \<in> red_locations prb"
  using assms(2,4) by (auto simp add : fringe_def extends_vertices)

  moreover
  have   "rl \<notin> subsumees (subs prb)" using assms(2) L by (simp add : fringe_def)

  moreover
  have   "rl \<noteq> src ra" using assms(2,3) L by (auto simp add : fringe_def out_arcs_def)

  moreover
  hence   "out_arcs (black prb) (fst rl) = {} \<or> (\<exists> ba. free_from prb rl ba)" 
  using assms(2) L `rl \<in> red_locations prb` free_in_seE_imp_free by (simp add : fringe_def) blast
  
  ultimately
  show   "rl \<in> fringe prb \<and> rl \<noteq> src ra" by (simp add : fringe_def)
next
  fix rl

  assume R : "rl \<in> fringe prb \<and> rl \<noteq> src ra"

  hence  "rl \<in> red_locations prb'" using assms(2) by (auto simp add : fringe_def extends_vertices)

  moreover
  have   "rl \<notin> subsumees (subs prb')" using assms(2) R by (simp add :  fringe_def)

  moreover
  have   "out_arcs (black prb') (fst rl) = {} \<or> (\<exists> ba. free_from prb' rl ba)" 
  using assms(2) R free_not_src_imp_free_in_seE by (simp add : fringe_def) blast

  ultimately
  show   "rl \<in> fringe prb'" by (simp add : fringe_def)
qed






(*********** subsum_extends ***********)

lemma subsumE_fringe :
  assumes "subsum_extends prb sub subrel prb'"
  shows   "fringe prb' = fringe prb - {subsumee sub}"
unfolding set_eq_iff Diff_iff singleton_iff
proof (intro allI iffI)

  fix rl

  assume H1 : "rl \<in> fringe prb'"

  have   "rl \<noteq> subsumee sub"
  proof (intro notI)

    assume "rl = subsumee sub"

    hence  "rl \<in> subsumees (subs prb')"
    using assms SubExt.extendsE2[of "red prb" "subs prb" sub subrel]
    unfolding subsumees_conv mem_Collect_eq
    by (rule_tac ?x="subsumer sub" in exI) auto

    thus False using H1 by (simp add : fringe_def)
  qed

  moreover
  have   "rl \<in> red_locations prb" using assms H1 by (simp add : fringe_def)

  moreover
  have   "rl \<notin> subsumees (subs prb)"
  using assms H1 SubExt.extendsE2[of "red prb" "subs prb" sub subrel]
  by (simp add : fringe_def subsumees_conv)

  moreover
  have   "out_arcs (black prb) (fst rl) = {} \<or> (\<exists> ba. free_from prb rl ba)" 
  using assms H1 free_in_subsumE_imp_free[OF assms]
  unfolding fringe_def mem_Collect_eq
  by (elim conjE disjE) (simp, blast)

  ultimately
  show "rl \<in> fringe prb \<and> rl \<noteq> subsumee sub" by (simp add : fringe_def)

next

  fix rl

  assume H2 : "rl \<in> fringe prb \<and> rl \<noteq> subsumee sub"

  hence "rl \<in> red_locations prb'" using assms(1) by (simp add : fringe_def)

  moreover
  have  "rl \<notin> subsumees (subs prb')"
  using assms H2 SubExt.extendsE2[of "red prb" "subs prb" sub subrel]
  by (auto simp add : fringe_def subsumees_conv)

  moreover
  have  "out_arcs (black prb') (fst rl) = {} \<or> (\<exists> ba. free_from prb' rl ba)" 
  using assms H2 free_imp_free_in_subsumE[OF assms(1)]
  unfolding fringe_def mem_Collect_eq
  by (elim conjE disjE) (simp, blast)

  ultimately
  show  "rl \<in> fringe prb'" by (simp add : fringe_def)

qed


(*********** abstract_extends ***********)

lemma (in finite_labels_RB) abstractE_fringe1 :
  assumes "RB prb"
  assumes "abstract_extends prb rl e prb'"
  assumes "rl \<in> fringe prb"
  shows   "fringe prb' = fringe prb"
unfolding set_eq_iff
proof (intro allI iffI)
  fix rl'

  assume L : "rl' \<in> fringe prb'"

  hence  "rl' \<in> red_locations prb" 
  and    "rl' \<notin> subsumees (subs prb)" using assms(2) by (simp_all add : fringe_def)

  moreover
  have   "out_arcs (black prb) (fst rl') = {} \<or> (\<exists> ba. free_from prb rl' ba)" 
  proof (case_tac "rl' = rl")
    assume "rl' = rl" thus ?thesis using assms(3) by (simp add : fringe_def)
  next
    assume "rl' \<noteq> rl" 
    thus ?thesis using assms(2) L free_in_abstractE_imp_free by (simp add : fringe_def) blast
  qed

  ultimately
  show   "rl' \<in> fringe prb" by (simp add : fringe_def)
next
  fix rl'

  assume R : "rl' \<in> fringe prb"

  hence  "rl' \<in> red_locations prb'"
  and    "rl' \<notin> subsumees (subs prb')" using assms(2) by (simp_all add : fringe_def)

  moreover
  have   "out_arcs (black prb') (fst rl') = {} \<or> (\<exists> ba. free_from prb' rl' ba)"
  using assms(2) R free_imp_free_in_abstractE[OF assms(1,2)]
  by (simp add : fringe_def) blast

  ultimately
  show   "rl' \<in> fringe prb'" by (simp add : fringe_def)
qed


lemma (in finite_labels_RB) abstractE_fringe2 :
  assumes "RB prb"
  assumes "abstract_extends prb rl e prb'"
  assumes "rl \<notin> fringe prb"
  assumes "\<exists> ba. free_from prb' rl ba"
  shows   "fringe prb' = fringe prb \<union> {rl}"
unfolding set_eq_iff Un_iff singleton_iff
proof (intro allI iffI)

  fix rl'

  assume L : "rl' \<in> fringe prb'"

  hence  "rl' \<in> red_locations prb" 
  and    "rl' \<notin> subsumees (subs prb)" using assms(2) by (simp_all add : fringe_def)

  show   "rl' \<in> fringe prb \<or> rl' = rl"
  proof (case_tac "rl' = rl")
    assume "rl' = rl" thus ?thesis by simp
  next
    assume "rl' \<noteq> rl"

    hence  "out_arcs (black prb) (fst rl') = {} \<or> (\<exists> ba. free_from prb rl' ba)"
    using assms(2) L free_in_abstractE_imp_free by (simp add : fringe_def) blast 
    
    thus ?thesis 
    using `rl' \<in> red_locations prb` `rl' \<notin> subsumees (subs prb)` by (simp add : fringe_def)
  qed

next

  fix rl'

  assume R : "rl' \<in> fringe prb \<or> rl' = rl"

  hence "rl' \<in> red_locations prb'" 
  and   "rl' \<notin> subsumees (subs prb') "using assms(2) by (auto simp add : fringe_def)

  moreover
  have "out_arcs (black prb') (fst rl') = {} \<or> (\<exists> ba. free_from prb' rl' ba)" 
  using R
  proof (elim disjE)
    assume "rl' \<in> fringe prb"
    
    thus ?thesis  using assms(1,2) free_imp_free_in_abstractE by (simp add : fringe_def) blast
  next
    assume "rl' = rl" thus ?thesis using assms(4)  by simp
  qed

  ultimately
  show   "rl' \<in> fringe prb'" by (simp add : fringe_def)
qed



lemma (in finite_labels_RB) abstractE_fringe3 :
  assumes "RB prb"
  assumes "abstract_extends prb rl e prb'"
  assumes "rl \<notin> fringe prb"
  assumes "\<not> (\<exists> ba. free_from prb' rl ba)"
  shows   "fringe prb' = fringe prb"
unfolding set_eq_iff
proof (intro allI iffI)
  fix rl'

  assume L : "rl' \<in> fringe prb'"
  
  hence "rl' \<in> red_locations prb"
  and   "rl' \<notin> subsumees (subs prb)" using assms(2) by (simp_all add : fringe_def)

  moreover
  have "out_arcs (black prb) (fst rl') = {} \<or> (\<exists> ba. free_from prb rl' ba)"
  proof -
    have "rl' \<noteq> rl" using assms(2,3,4) L by (auto simp add : fringe_def)

    thus ?thesis using assms(2,3) L free_in_abstractE_imp_free by (simp add : fringe_def) blast
  qed

  ultimately
  show   "rl' \<in> fringe prb" by (simp add : fringe_def)
next
  fix rl'

  assume R : "rl' \<in> fringe prb"

  hence  "rl' \<in> red_locations prb'"
  and    "rl' \<notin> subsumees (subs prb')" using assms(2) by (simp_all add : fringe_def)

  moreover
  have "out_arcs (black prb') (fst rl') = {} \<or> (\<exists> ba. free_from prb' rl' ba)"
  using assms(1,2) R free_imp_free_in_abstractE by (simp add : fringe_def)  blast

  ultimately
  show   "rl' \<in> fringe prb'" by (simp add : fringe_def)
qed






(*********************************)
(********** P & SP_from **********)
(*********************************)

definition P :: 
  "('a, 'b, 'c) pre_RB_scheme \<Rightarrow> 'a arc list set"
where
  "P prb \<equiv> 
     ui_as ` paths (red prb) (subs prb)
   \<union> {ui_as ras1 @ bas2 
      | ras1 bas2. \<exists> rl \<in> fringe prb. path (red prb) ras1 rl (subs prb)
                                    \<and> \<not> (\<exists> ras21 bas22. bas2 = ui_as ras21 @ bas22 
                                                      \<and> ras21 \<noteq> [] 
                                                      \<and> subpath_from (red prb) rl ras21 (subs prb))
                                    \<and> Graph.subpath_from (black prb) (fst rl) bas2}"

definition SP_from ::
  "('a, 'b, 'c) pre_RB_scheme \<Rightarrow> ('a \<times> nat) \<Rightarrow> 'a arc list set"
where
  "SP_from prb rl \<equiv> 
     ui_as ` subpaths_from (red prb) rl (subs prb)
   \<union> {ui_as ras1 @ bas2 
      | ras1 bas2. \<exists> rl' \<in> fringe prb. subpath (red prb) rl ras1 rl' (subs prb)
                                     \<and> \<not> (\<exists> ras21 bas22. bas2 = ui_as ras21 @ bas22 
                                                       \<and> ras21 \<noteq> [] 
                                                       \<and> subpath_from (red prb) rl' ras21 (subs prb))
                                     \<and> Graph.subpath_from (black prb) (fst rl') bas2}"


lemma P_is_SP_from_root :
  "P prb = SP_from prb (root (red prb))"
unfolding SP_from_def subpath_from_def P_def path_def by simp



lemma (in finite_labels_RB) base_SP_from1 :
  assumes "fst (root (red prb)) = init (black prb)"
  assumes "arcs (red prb) = {}"
  assumes "subs prb = {}"
  assumes "confs prb (root (red prb)) = init_conf prb"
  assumes "sat (init_conf prb)"
  assumes "out_arcs (black prb) (fst (root (red prb))) = {} \<or>
           (\<exists> ba. free_from prb (root (red prb)) ba)"
  shows   "SP_from prb (root (red prb)) = Graph.paths (black prb)"
proof -

  have "fringe prb = {root (red prb)}"
  using assms fringe_of_empty_red_1[of prb] fringe_of_empty_red_2[of prb]
  by auto

  show ?thesis
  unfolding set_eq_iff
  proof (intro allI iffI)

    fix bas

    assume "bas \<in> SP_from prb (root (red prb))"

    thus "bas \<in> Graph.paths (black prb)" 
    apply (subst (asm) SP_from_def)
    apply (subst (asm) Un_iff)
    proof (elim disjE exE conjE)
    
      case goal1
      
      hence "bas = []" using assms by (auto simp add : empty_graph_sp_from_eq_Nil)
      
      thus ?thesis using Graph.Nil_in_paths by simp

    next

      case goal2

      then obtain ras1 bas2 rl where "bas = ui_as ras1 @ bas2"
                               and   "rl \<in> fringe prb"
                               and   "subpath (red prb) (root (red prb)) ras1 rl (subs prb)"
                               and   "Graph.subpath_from (black prb) (fst rl) bas2"
      by blast

      moreover
      hence "ras1 = []" using assms by (simp add : empty_graph_imp_sp_empty)

      moreover
      have "rl = root (red prb)" 
      using assms `rl \<in> fringe prb` by (simp add : fringe_def vertices_conv)

      ultimately
      show ?thesis using assms by (simp add : Graph.paths_conv)
    qed

  next

    fix bas

    assume "bas \<in> Graph.paths (black prb)"

    show "bas \<in> SP_from prb (root (red prb))" 
    unfolding SP_from_def Un_iff mem_Collect_eq
    proof (intro disjI2, rule_tac ?x="[]" in exI, rule_tac ?x="bas" in exI, intro conjI)

      case goal1 show ?case by simp

    next

      case goal2

      show ?case
      unfolding Bex_def
      proof (rule_tac ?x="root (red prb)" in exI, intro conjI)
      
        show "root (red prb) \<in> fringe prb" using `fringe prb = {root (red prb)}` by simp
      
      next

        show "subpath (red prb) (root (red prb)) [] (root (red prb)) (subs prb)"
        by (rule Nil_sp_from_root)

      next

        case goal3
        show ?case
        proof (intro notI, elim exE conjE)

          fix ras21 bas22
          
          assume "bas = ui_as ras21 @ bas22"
          and    "ras21 \<noteq> []" 
          and    "subpath_from (red prb) (root (red prb)) ras21 (subs prb)"

          moreover
          hence "ras21 = []" 
          using assms by (simp add : empty_graph_sp_from_eq_Nil)
          
          ultimately
          show False by (elim notE)

        qed

      next

        show "Graph.subpath_from (black prb) (fst (root (red prb))) bas"
        using assms `bas \<in> Graph.paths (black prb)` by (simp add : Graph.paths_conv)
      
      qed
    qed
  qed
qed


lemma (in finite_labels_RB) base_P1 :
  assumes "fst (root (red prb)) = init (black prb)"
  assumes "arcs (red prb) = {}"
  assumes "subs prb = {}"
  assumes "confs prb (root (red prb)) = init_conf prb"
  assumes "sat (init_conf prb)"
  assumes "out_arcs (black prb) (fst (root (red prb))) = {} \<or>
           (\<exists> ba. free_from prb (root (red prb)) ba)"
  shows   "P prb = Graph.paths (black prb)"
proof -

  have "fringe prb = {root (red prb)}"
  using assms fringe_of_empty_red_1[of prb] fringe_of_empty_red_2[of prb]
  by auto

  show ?thesis
  unfolding set_eq_iff
  proof (intro allI iffI)

    fix bas

    assume "bas \<in> P prb"

    hence  "bas = [] \<or> (\<exists> p1 p2. bas = ui_as p1 @ p2 
                               \<and> path (red prb) p1 (root (red prb)) (subs prb)
                               \<and> Graph.subpath_from (black prb) (fst (root (red prb))) p2)"
    using assms(2) `fringe prb = {root (red prb)}` empty_graph_paths[OF assms(2)]
    unfolding P_def by auto

    thus   "bas \<in> Graph.paths (black prb)" 
    proof (elim disjE)

      case goal1 thus ?thesis by (auto simp add : Graph.Nil_path)

    next

      case goal2 

      then obtain p1 p2 where "bas = ui_as p1 @ p2"
                        and   "path (red prb) p1 (root (red prb)) (subs prb)"
                        and   "Graph.subpath_from (black prb) (fst (root (red prb))) p2"
      by blast

      moreover
      hence "p1 = []" using empty_graph_imp_p_empty[OF assms(2)] by fast

      ultimately
      show ?thesis 
      using assms(1) by (simp add : Graph.paths_conv)
      
    qed

  next

    fix bas
      
    assume "bas \<in> Graph.paths (black prb)"

    show   "bas \<in> P prb" 
    unfolding P_def Un_iff mem_Collect_eq
    proof (intro disjI2, rule_tac ?x="[]" in exI, rule_tac ?x="bas" in exI, intro conjI)
    
      case goal1 show ?case by simp

    next

      case goal2 

      show ?case 
      unfolding Bex_def
      proof (rule_tac ?x="root (red prb)" in exI, intro conjI)

        case goal1 show ?case using `fringe prb = {root (red prb)}` by simp

      next

        case goal2 show ?case by (rule Nil_path)

      next

        case goal3

        show ?case
        proof (intro notI, elim exE conjE)

          fix p3 p4

          assume "bas = ui_as p3 @ p4"
          assume "p3 \<noteq> []"

          then obtain ra p3' where "p3 = ra # p3'" using neq_Nil_conv[of p3] by blast

          moreover
          assume "subpath_from (red prb) (root (red prb)) p3 (subs prb)"

          ultimately
          have "ra \<in> arcs (red prb)" by (simp add : subpath_from_def subpath_def)
          
          thus False using assms(2) by auto
        qed

      next

        case goal4

        show ?case 
        using assms(1) `bas \<in> Graph.paths (black prb)` 
        by (simp add : Graph.paths_conv)
      
      qed
    qed
  qed
qed



lemma (in finite_labels_RB) base_SP_from2 :
  assumes "fst (root (red prb)) = init (black prb)"
  assumes "arcs (red prb) = {}"
  assumes "subs prb = {}"
  assumes "confs prb (root (red prb)) = init_conf prb"
  assumes "sat (init_conf prb)"
  assumes "out_arcs (black prb) (fst (root (red prb))) \<noteq> {} \<and> 
           \<not> (\<exists> ba. free_from prb (root (red prb)) ba)"
  shows   "SP_from prb (root (red prb)) = {[]}"
unfolding set_eq_iff singleton_iff
proof (intro allI iffI)

  fix bas

  assume "bas \<in> SP_from prb (root (red prb))"

  thus "bas = []" 
  unfolding SP_from_def Un_iff Bex_def mem_Collect_eq
  proof (elim disjE exE conjE)
    assume "bas \<in> ui_as ` subpaths_from (red prb) (root (red prb)) (subs prb)"
    thus ?thesis using assms(2) empty_graph_sp_from_eq_Nil by fast
  next
    case (goal2 ras1 bas2 rl) 
  
    moreover
    have "fringe prb = {}" using assms fringe_of_empty_red_3[of prb] by auto
      
    ultimately
    show  ?thesis using `fringe prb = {}` by auto
  qed

next

  fix bas::"'a arc list"

  assume "bas = []"

  thus "bas \<in> SP_from prb (root (red prb))" using Nil_in_paths[of "red prb" "subs prb"] 
  by (simp only : SP_from_def paths_conv) blast

qed



lemma (in finite_labels_RB) base_P2 :
  assumes "fst (root (red prb)) = init (black prb)"
  assumes "arcs (red prb) = {}"
  assumes "subs prb = {}"
  assumes "confs prb (root (red prb)) = init_conf prb"
  assumes "sat (init_conf prb)"
  assumes "out_arcs (black prb) (fst (root (red prb))) \<noteq> {} \<and> \<not> (\<exists> ba. free_from prb (root (red prb)) ba)"
  shows   "P prb = {[]}"
proof -

  have "fringe prb = {}" using assms fringe_of_empty_red_3[of prb] by auto

  show ?thesis
  unfolding set_eq_iff singleton_iff
  proof (intro allI iffI)

    fix bas

    assume "bas \<in> P prb"

    thus "bas = []" 
    unfolding P_def Un_iff Bex_def mem_Collect_eq
    proof (elim disjE exE conjE)
      assume "bas \<in> ui_as ` paths (red prb) (subs prb)"
      thus ?thesis using assms(2) by (auto simp add : empty_graph_imp_p_empty)
    next
      case (goal2 ras1 bas2 rl) thus ?thesis using `fringe prb = {}` by auto
    qed

  next

    fix bas::"'a arc list"

    assume "bas = []"

    thus "bas \<in> P prb" using Nil_in_paths[of "red prb" "subs prb"] by (auto simp add : P_def)

  qed

qed























































lemma SP_from_append_oneE :
  assumes "RB prb"
  assumes "bas @ [ba] \<in> SP_from prb rl"
  shows   "bas \<in> SP_from prb rl"
using assms
apply (subst (asm) SP_from_def)
unfolding Un_iff image_def Bex_def mem_Collect_eq
proof (elim disjE exE conjE)

  case (goal1 ras)

  moreover
  then obtain ras' ra where "ras = ras'@[ra]" by (induct ras rule : rev_induct) simp_all

  ultimately
  have "subpath_from (red prb) rl ras' (subs prb)"
  using assms(1) by (simp only : subpath_from_def rb_sp_append_one) blast

  moreover
  have "bas = ui_as ras'" using `bas @ [ba] = ui_as ras` `ras = ras'@[ra]` by simp

  ultimately
  show ?thesis by (simp add : SP_from_def)

next

  case (goal2 ras1 bas2 rl')

  show ?thesis
  proof (case_tac "bas2 = []")

    assume "bas2 = []" 

    then obtain ras' ra where "ras1 = ras'@[ra]" 
    using goal2 by (induct ras1 rule : rev_induct) simp_all

    hence "subpath_from (red prb) rl ras' (subs prb)"
    using assms(1) goal2 by (simp only : subpath_from_def rb_sp_append_one) blast
   
    moreover
    have "bas = ui_as ras'" 
    using `bas @ [ba] = ui_as ras1 @ bas2` `bas2 = []` `ras1 = ras'@[ra]` by simp
 
    ultimately
    show ?thesis by (simp add : SP_from_def)

  next

    assume "bas2 \<noteq> []"

    then obtain ba' bas2' where "bas2 = bas2' @ [ba']" using neq_Nil_conv2[of bas2] by blast

    hence "bas = ui_as ras1 @ bas2'" using `bas @ [ba] = ui_as ras1 @ bas2` by auto
    
    moreover
    have "\<not> (\<exists>ras21 bas22. bas2' = ui_as ras21 @ bas22 
                         \<and> ras21 \<noteq> [] 
                         \<and> subpath_from (red prb) rl' ras21 (subs prb))"
    proof (intro notI, elim exE conjE)

      fix ras21 bas22
       
      assume "bas2' = ui_as ras21 @ bas22"
      and    "ras21 \<noteq> []"
      and    "subpath_from (red prb) rl' ras21 (subs prb)"

      hence  "\<exists>ras21 bas22. bas2 = ui_as ras21 @ bas22 
                          \<and> ras21 \<noteq> [] 
                          \<and> subpath_from (red prb) rl' ras21 (subs prb)"
      apply (rule_tac ?x="ras21" in exI)
      apply (rule_tac ?x="bas22@[ba']" in exI)
      apply (intro conjI)
        apply (simp add: `bas2 = bas2' @ [ba']`)
        apply assumption
        apply assumption
      done

      thus False using goal2 by (elim notE)
    qed

    moreover
    have "Graph.subpath_from (black prb) (fst rl') bas2'" 
    using goal2(6) `bas2 = bas2' @ [ba']`
    by (auto simp add : Graph.subpath_from_def Graph.sp_append_one)
    
    ultimately
    show ?thesis using goal2 unfolding SP_from_def by blast

  qed
qed














lemma (in finite_labels_RB)
  assumes "RB prb"
  assumes "rl \<in> red_locations prb"
  shows   "feasible_subpaths_from (black prb) (confs prb rl) (fst rl) \<subseteq> SP_from prb rl"
using assms finite_labels_RB
proof (induct prb arbitrary : rl)

  (* Cas de base : la partie rouge est vide.

     Dans ce cas, rl est la racine de la partie rouge, et l'ensemble des sous-chemins faisables 
     noirs issus de (fst rl) est l'ensemble des chemins faisables noirs.

     On distingue alors deux cas : 
     
     1) l'ensemble des arcs issus de l'état initial de la partie noire est vide ou bien il existe 
        un arc libre depuis la racine de la partie rouge. Dans ce cas, SP_from est l'ensemble des 
        chemins de la partie noire, ce qui donne bien l'inclusion. 
     
     2) l'ensemble des arcs issus de l'état initial de la partie noire est non-vide et il n'existe 
        aucun arbre libre issu de la racine rouge. Dans ce cas, l'ensemble des chemins faisables 
        noirs contient uniquement []. En effet, la conf. initial est sat., donc [] est faisable, 
        mais aucun arc ne mene à une conf. sat., puisqu'aucun arc n'a encore été exécuté mais 
        qu'aucun n'est libre. De plus, SP_from est également égal à {[]}. *)

  case (base prb) 

  hence 1 : "rl = root (red prb)" by (simp add : vertices_conv)

  hence 2 : "feasible_subpaths_from (black prb) (confs prb rl) (fst rl) 
             = feasible_paths (black prb) (confs prb (root (red prb)))"
  using base feasible_paths_conv[of "black prb"] by simp

  show ?case 
  proof (case_tac "out_arcs (black prb) (fst (root (red prb))) = {} \<or>
                   (\<exists> ba. free_from prb (root (red prb)) ba)")

    case goal1

    thus ?thesis 
    using base 1 2
          finite_labels_RB.base_SP_from1[of prb]
          feasible_paths_are_paths[of "black prb" "confs prb rl"] 
    by (simp only : finite_labels_RB_def)
   
  next 

    case goal2 

    moreover
    hence "\<not> (\<exists>ba\<in>out_arcs (black prb) (init (black prb)).
              \<exists>pc'. se (confs prb (root (red prb))) (labelling (black prb) ba) pc' \<and> sat pc')" 
    using base by (auto simp add : free_from_def out_arcs_def)

    ultimately
    show ?thesis 
    using base 1 2
          finite_labels_RB.base_SP_from2[of prb]
          no_feasible_arc_from_init_imp[of "confs prb (root (red prb))" "black prb"] 
    by (simp add : finite_labels_RB_def)
  qed



next

  (* Cas de l'ajout d'arc par exécution symbolique. *)
    
  case (se_step prb rg' ra pc prb')

  show ?case
  unfolding subset_iff
  proof (intro allI impI)

    (* Soient rl un sommet de la nouvelle partie rouge et bas un chemin faisable de la partie noire 
       issu de (fst rl). 

       On distingue dans un premier temps le cas où rl est un sommet de l'ancienne partie rouge 
       de celui où rl est la cible du nouvel arc.*)

    fix bas

    assume "bas \<in> feasible_subpaths_from (black prb') (confs prb' rl) (fst rl)"

    have   "rl \<in> red_locations prb \<or> rl = tgt ra"
    and    "rl \<in> red_locations prb'"
    using se_step by (auto simp add : extends_vertices)

    show "bas  \<in> SP_from prb' rl"
    using `rl \<in> red_locations prb \<or> rl = tgt ra`
    proof (elim disjE)

      (* si rl est un sommet de la partie rouge, alors on peut utiliser l'IH pour déduire que bas
         est un chemin rouge-noir de l'ancien rb_lts. *)
      assume "rl \<in> red_locations prb"

      moreover
      hence  "rl \<noteq> tgt ra" using se_step by (auto simp add : extendsE2)

      ultimately
      have  "bas \<in> SP_from prb rl"
      using se_step `bas \<in> feasible_subpaths_from (black prb') (confs prb' rl) (fst rl)`
      by fastforce
      
      (* On a alors deux cas possibles :
         1) bas est entièrement représentable dans l'ancienne partie rouge,
         2) bas peut se décomposer en un préfixe maximal dans l'ancienne partie rouge suivi d'un 
            suffixe noir issu de la fringe de prb. *)
      thus ?thesis 
      apply (subst (asm) SP_from_def)
      unfolding Un_iff image_def Bex_def mem_Collect_eq
      proof (elim disjE exE conjE)

        (* dans le premier cas, alors bas est également entièrement représentable dans la 
           nouvelle partie rouge, cqfd. *)
        case (goal1 ras)

        thus ?thesis 
        using se_step sp_in_extends_w_subs[of "red prb" ra rg' rl ras _ "subs prb"]
        unfolding SP_from_def Un_iff subpath_from_def
        by (intro disjI1) force

      next

        (* Dans le second cas, on appelle ras1 le préfixe rouge, bas2 le suffixe noir et rl' le 
           sommet de la fringe de l'ancienne partie rouge en lequel termine ras1 et commence bas2.*)
        fix ras1 bas2 rl'

        assume 1 : "bas = ui_as ras1 @ bas2"
        and    2 : "rl' \<in> fringe prb"
        and    3 : "subpath (red prb) rl ras1 rl' (subs prb)"
        and    4 : "\<not> (\<exists>ras21 bas22. bas2 = ui_as ras21 @ bas22 
                               \<and> ras21 \<noteq> [] 
                               \<and> subpath_from (red prb) rl' ras21 (subs prb))"
        and    5 : "Graph.subpath_from (black prb) (fst rl') bas2"

        (* On distingue selong que le suffixe noir est vide ou non. *)
        show ?thesis
        proof (case_tac "bas2 = []")

          (* Si le suffixe noir est vide, alors bas se résume à son préfixe rouge, lequel est 
             est également un sous-chemin de la nouvelle partie rouge. *)
          assume "bas2 = []"

          thus ?thesis 
          using se_step
                `bas = ui_as ras1 @ bas2`
                `subpath (red prb) rl ras1 rl' (subs prb)`
                sp_in_extends_w_subs[of "red prb" ra rg' rl ras1 rl' "subs prb"]
          unfolding SP_from_def Un_iff subpath_from_def image_def Bex_def mem_Collect_eq
          by (simp del : split_paired_Ex) blast

        next

          (* Sinon, bas2 se décompose en (ba # bas2'). *)

          assume "bas2 \<noteq> []"

          then obtain ba bas2' where "bas2 = ba # bas2'" by (simp add : neq_Nil_conv) blast

          (* On doit maintenant distinguer deux cas :
             1) le cas où le préfixe rouge ras1 termine en la source du nouvel arc,
             2) le cas où ras1 termine dans un sommet quelconque de la fringe de l'ancienne partie 
                rouge. *)
          show ?thesis 
          proof (case_tac "rl' = src ra")

            (* Supposons que ras1 termine en la source du nouvel arc. *)
            assume "rl' = src ra"

            (* On distingue les cas où le nouvel arc représente ou non ba. *)
            show ?thesis 
            proof (case_tac "ui_arc ra = ba")

              (* Supposons que le nouvel arc représente ba *)
              assume "ui_arc ra = ba"

              (* Puisque ra représente ba, bas "passe par" le nouvel arc ra et en particulier 
                 par la cible de cet arc. 

                 On va alors distinguer selon deux cas :
                 
                 1) dans la partie noire, aucun arc n'est issu du sommet noir représenté par la 
                    cible de ce nouvel arc ou bien il existe un arc libre depuis cette cible dans 
                    la nouvelle partie rouge.

                    Dans ce premier cas, la cible du nouvel arc fait partie de la fringe de 
                    la nouvelle partie rouge, et on peut conclure. Le préfixe recherché est 
                    ras1@[ra] et le suffixe bas2'.

                  2) dans la partie noire, il existe des arcs issus du sommet noir représenté 
                     par la cible du nouvel arc mais il n'existe aucun arc libre issu de cette 
                     cible dans la nouvelle partie rouge.

                     Dans ce cas, on a forcément que bas2' est vide (sinon son premier arc 
                     serait faisable et, étant non représenté depuis la cible du nouvel arc 
                     dans la nouvelle partie rouge, y serait nécessairement libre, contradiction). 
                     
                     On obtient alors que bas est entièrement représentable dans la nouvelle partie 
                     rouge, ce qui nous permet de conclure. *)
              show ?thesis 
              proof (case_tac "out_arcs (black prb') (fst (tgt ra)) = {} 
                               \<or> (\<exists> ba. free_from prb' (tgt ra) ba)")

                (* Supposons que dans la partie noire, aucun arc n'est issu du sommet noir 
                   représenté par la cible du nouvel arc ou bien il existe un arc libre depuis 
                   cette cible dans la nouvelle partie rouge. *)

                case goal1

                show ?thesis 
                unfolding SP_from_def Un_iff Bex_def mem_Collect_eq
                apply (intro disjI2) 
                apply (rule_tac ?x="ras1@[ra]" in exI)
                apply (rule_tac ?x="bas2'" in exI)
                proof (intro conjI)
                
                  show "bas = ui_as (ras1 @ [ra]) @ bas2'" 
                  using 1 `bas2 = ba # bas2'` `ui_arc ra = ba` by simp

                next

                  case goal2 show ?case
                  proof (rule_tac ?x="tgt ra" in exI, intro conjI)

                    show "tgt ra \<in> fringe prb'" 
                    using se_step goal1 seE_fringe1[OF se_step(1,3)] seE_fringe3[OF se_step(1,3)]
                    by auto

                  next

                    have "subpath (red prb') rl ras1 rl' (subs prb')" 
                    using se_step 3 by (auto simp add : sp_in_extends_w_subs)
                    
                    moreover
                    have "ra \<in> arcs (red prb')" using se_step by (auto simp add : extends_arcs)

                    ultimately
                    show "subpath (red prb') rl (ras1 @ [ra]) (tgt ra) (subs prb')" 
                    using `rl' = src ra` rb_sp_append_one[OF RB.se_step[OF se_step(1,3)]]
                    by auto

                  next

                    case goal3 
                    show ?case
                    proof (intro notI, elim exE conjE)

                      fix ras21 bas22
                      
                      assume "bas2' = ui_as ras21 @ bas22"
                      and    "ras21 \<noteq> []"
                      and    "subpath_from (red prb') (tgt ra) ras21 (subs prb')"

                      then obtain ra' ras21' where "ras21 = ra'#ras21'" 
                      by (simp add : neq_Nil_conv) blast

                      have "ra' \<in> out_arcs (red prb') (tgt ra)" 
                      proof -

                        have "tgt ra = src ra' \<or> (tgt ra, src ra') \<in> (subs prb')"
                        and  "ra' \<in> arcs (red prb')"
                        using `subpath_from (red prb') (tgt ra) ras21 (subs prb')`
                              `ras21 = ra'# ras21'`
                              rb_sp_Cons[OF RB.se_step[OF se_step(1,3)]]
                        unfolding subpath_from_def by auto

                        moreover
                        have "tgt ra \<notin> subsumees (subs prb')"
                        proof (intro notI)
                          assume "tgt ra \<in> subsumees (subs prb')"
                          
                          hence  "tgt ra \<in> red_locations prb"
                          using se_step subs_sub_rel_of[OF se_step(1)]
                          by (auto simp add : sub_rel_of_def subsumees_def)
                          
                          thus False using se_step by (auto simp add : extendsE2)
                        qed

                        ultimately
                        show ?thesis by (force simp add : out_arcs_def subsumees_def)

                      qed

                      thus False using se_step by (auto simp add : extends_tgt_out_arcs)
                    qed

                  next

                    show "Graph.subpath_from (black prb') (fst (tgt ra)) bas2'" 
                    using se_step 5 `bas2 = ba # bas2'` `ui_arc ra = ba`
                    unfolding Graph.subpath_from_def by (auto simp add : Graph.sp_Cons)
                    
                  qed
                qed

              next

                (* Supposons que, dans la partie noire, il existe des arcs issus du sommet noir 
                   représenté par la cible du nouvel arc mais il n'existe aucun arc libre issu de 
                   cette cible dans la nouvelle partie rouge. *)
                case goal2

                (* Montrons que bas2' est vide, par contradiction *)
                have "bas2' = []" 
                proof (rule ccontr)

                  (* Supposons que bas2' soit non-vide et appelons ba' son premier arc et bas2'' 
                     le suffixe *)
                  assume "bas2' \<noteq> []"

                  then obtain ba' bas2'' where "bas2' = ba' # bas2''" 
                  by (simp add : neq_Nil_conv) blast

                  (* montrons que ba' est libre depuis la cible du nouvel arc. *)
                  have "free_from prb' (tgt ra) ba'"
                  proof -

                    (* ba' est bien un arc de la partie noire issu du sommet représenté par 
                       la cible du nouvel arc, puisque bas2 est issu de ce sommet. *)
                    have "ba' \<in> out_arcs (black prb') (fst (tgt ra))"
                    using se_step 5 `bas2 = ba # bas2'` `ui_arc ra = ba` `bas2' = ba' # bas2''` 
                    by (auto simp add : Graph.subpath_from_def Graph.sp_Cons out_arcs_def)

                    (* de plus, ba' ne peut être représenté depuis la cible de ra dans la nouvelle 
                       partie rouge. *)
                    moreover
                    have "(\<forall>ra\<in>out_arcs (red prb') (tgt ra). ui_arc ra \<noteq> ba')" 
                    using se_step by (auto simp add : extends_tgt_out_arcs)

                    (* montrons finalement que ba' est faisable depuis la cible de ra *)
                    moreover
                    have "\<exists>pc'. se (confs prb' (tgt ra)) (labelling (black prb') ba') pc' \<and> sat pc'"
                    proof -

                      (* En effet, puisque la cible du nouvel arc est bien un sommet de la nouvelle 
                         partie rouge, alors il existe une configuration pc' obtenue par ES de ba' 
                         depuis la configuration en (tgt ra). *)
                      have "tgt ra \<in> red_locations prb'" 
                      using se_step by (auto simp add : extends_vertices)

                      then obtain pc' where "se (confs prb' (tgt ra)) (labelling (black prb') ba') pc'"
                      using se_step RB.se_step[OF se_step(1,3)]
                          finite_labels_RB.ex_se_succ[of prb' "tgt ra" "labelling (black prb') ba'"]
                      unfolding finite_labels_RB_def finite_labels_lts_def by blast

                      (* montrons que cette configuration est satisfiable. En utilisant le fait 
                         que bas est un chemin noir faisable et par la relation liant les 
                         configurations de la partie rouge d'un RB_lts. *)
                      moreover
                      have "sat pc'"
                      proof - 

                        obtain pc1 pc2 pc3 pc4
                        where a : "se_star (confs prb' rl) (trace (ui_as ras1) (labelling (black prb'))) pc1"
                        and   b : "se pc1 (labelling (black prb') ba) pc2"
                        and   c : "se pc2 (labelling (black prb') ba') pc3"
                        and   d : "se_star pc3 (trace bas2'' (labelling (black prb'))) pc4"
                        and   e : "sat pc4"
                        using `bas \<in> feasible_subpaths_from (black prb') (confs prb' rl) (fst rl)`
                              `bas = ui_as ras1 @ bas2` `bas2 = ba # bas2'` `bas2' = ba' # bas2''`
                        unfolding feasible_subpath_from_def feasible_subpath_def feasible_def
                        by (simp add : se_star_append se_star_Cons)  blast

                        have "pc2 \<sqsubseteq>\<^sub>p\<^sub>c confs prb' (tgt ra)"
                        proof -

                          have "subpath (red prb') rl (ras1 @ [ra]) (tgt ra) (subs prb')"
                          proof - 
                            have "subpath (red prb') rl ras1 rl' (subs prb')" 
                            using se_step 3 by (auto simp add : sp_in_extends_w_subs)
                    
                            moreover
                            have "ra \<in> arcs (red prb')" using se_step by (auto simp add : extends_arcs)

                            ultimately
                            show ?thesis 
                            using `rl' = src ra` rb_sp_append_one[OF RB.se_step[OF se_step(1,3)]]
                            by auto
                          qed

                          moreover
                          have "se_star (confs prb' rl) 
                                        (trace (ui_as (ras1 @ [ra])) (labelling (black prb'))) 
                                        pc2"
                          using a b `ui_arc ra = ba` by (simp add : se_star_append_one) blast

                          ultimately
                          show ?thesis 
                          using se_step RB.se_step[OF se_step(1,3)] finite_labels_RB.SE_rel
                          unfolding finite_labels_RB_def by fast
                          
                        qed

                        hence "pc3 \<sqsubseteq>\<^sub>p\<^sub>c pc'"
                        using `se (confs prb' (tgt ra)) (labelling (black prb') ba') pc'`
                              `se pc2 (labelling (black prb') ba') pc3`
                        by (simp add : se_mono_for_sub)

                        moreover
                        have "sat pc3" using d e by (simp add : se_star_sat_imp_sat se_sat_imp_sat)

                        ultimately
                        show ?thesis by (simp add : sat_sub_by_sat)
                      qed

                      (* donc ba' est faisable depuis la cible du nouvel arc, cqfd. *)
                      ultimately
                      show ?thesis by blast
                    qed
                    
                    (* on obtient finalement que ba' est libre depuis tgt ra, cqfd. *)
                    ultimately
                    show ?thesis by (simp add : free_from_def)
                  qed
                  
                  (* or il n'existe aucun arc libre issu de tgt ra, par hypothèse. cqfd. *)
                  thus False using goal2 by auto
                qed

                (* Puisque bas2' est vide, alors bas est entièrement représentable dans la nouvelle 
                   partie rouge, ce qui en fait un chemin rouge-noir.*)
                show ?thesis 
                unfolding SP_from_def Un_iff image_def Bex_def mem_Collect_eq
                proof (intro disjI1, rule_tac ?x="ras1@[ra]" in exI, intro conjI)

                  have "subpath (red prb') rl ras1 rl' (subs prb')" 
                  using se_step 3 by (auto simp add : sp_in_extends_w_subs)
                    
                  moreover
                  have "ra \<in> arcs (red prb')" using se_step by (auto simp add : extends_arcs)

                  ultimately
                  have "subpath (red prb') rl (ras1 @ [ra]) (tgt ra) (subs prb')" 
                  using `rl' = src ra` rb_sp_append_one[OF RB.se_step[OF se_step(1,3)]]
                  by auto

                  thus "subpath_from (red prb') rl (ras1 @ [ra]) (subs prb')" 
                  unfolding subpath_from_def by blast

                next

                  show "bas = ui_as (ras1 @ [ra])" 
                  using 1 `bas2 = ba # bas2'` `ui_arc ra = ba` `bas2' = []` by simp
                  
                qed
              qed

            next

              (* Supposons que ba ne soit pas l'arc représenté par ba. 

                 Dans ce cas, bas étant faisable depuis rl, bas2 est faisable depuis rl' et ba
                 est faisable depuis rl'. Or ba n'est pas représenté depuis rl', donc ba 
                 est libre depuis rl'.

                 rl' est donc dans la fringe de la nouvelle partie rouge. 

                 On conclue en montrant que ras1 est bien un préfixe maximal de bas menant en rl' 
                 dans la nouvelle partie rouge menant en rl', bas2 étant toujours un suffixe noir 
                 de bas issu de rl'. *)
            
              assume "ui_arc ra \<noteq> ba"

              show ?thesis
              unfolding SP_from_def Un_iff Bex_def mem_Collect_eq
              proof (intro disjI2, rule_tac ?x="ras1" in exI, rule_tac ?x="bas2" in exI, intro conjI)

                show "bas = ui_as ras1 @ bas2" using `bas = ui_as ras1 @ bas2` by assumption

              next

                (* On montre que rl' est bien dans la fringe de la nouvelle partie rouge. *)
                case goal2
                
                (* On montre d'abord que ba est libre depuis rl'. *)
                have "free_from prb' rl' ba"
                proof -

                  (* ba est bien un arc issu du sommet noir représenté par rl' *)
                  have "ba \<in> out_arcs (black prb') (fst rl')" 
                  using se_step 5 `bas2 = ba # bas2'`
                  by (simp add  : Graph.subpath_from_def sp_Cons out_arcs_def)
                  
                  (* de plus ba n'est pas représenté depuis rl' dans la nouvelle partie rouge. *)
                  moreover
                  have "(\<forall>ra\<in>out_arcs (red prb') rl'. ui_arc ra \<noteq> ba)" 
                  proof (rule ccontr, simp only : Ball_def not_all not_imp not_not, elim exE conjE)

                    (* En effet supposons qu'il existe un arc rouge ra' issu de rl' et 
                       représentant ba. *)
                    fix ra' 

                    assume "ra' \<in> out_arcs (red prb') rl'"
                    and    "ui_arc ra' = ba"

                    (* dans ce cas ra' est également un arc issu de rl' dans l'ancienne partie 
                       rouge puisque ra' représente ba, qui n'est pas l'arc représenté par le 
                       nouvel arc. *)
                    hence "ra' \<in> out_arcs (red prb) rl'"
                    using se_step `rl' = src ra` `ui_arc ra \<noteq> ba`
                    by (auto simp add : extends_src_out_arcs)

                    (* dans ce cas, [ra'] est un sous-chemin de l'ancienne partie rouge issu de 
                       rl'. Donc il existe un préfixe de bas2 qui est représentable dans 
                       l'ancienne partie rouge, ce qui vient contredire le fait ras1 est 
                       maximal. *)
                    hence "subpath_from (red prb) rl' [ra'] (subs prb)" 
                    using rb_sp_one[OF se_step(1)] unfolding subpath_from_def out_arcs_def by blast
                    
                    (* cqfd *)
                    thus False 
                    using 4 `bas2 = ba # bas2'` `ui_arc ra' = ba`
                    by (elim notE, rule_tac ?x="[ra']" in exI, rule_tac ?x="bas2'" in exI) simp
                  qed

                  (* Montrons de plus que ba est faisable depuis rl'. *)
                  moreover
                  have "\<exists>pc'. se (confs prb' rl') (labelling (black prb') ba) pc' \<and> sat pc'"
                  proof -

                    (* rl' étant un sommet rouge du nouveau rb_lts, il existe une configuration
                       pc' obtenue depuis (confs prb' rl') par ES de ba. *)
                    have "rl' \<in> red_locations prb'" 
                    using se_step `rl' = src ra` by (auto simp add : extends_vertices extendsE1)

                    then obtain pc' where "se (confs prb' rl') (labelling (black prb') ba) pc'"
                    using se_step `rl' = src ra`
                          RB.se_step[OF se_step(1,3)]
                          finite_labels_RB.ex_se_succ[of prb' rl' "labelling (black prb') ba"]
                    unfolding finite_labels_RB_def finite_labels_lts_def
                    by simp blast

                    (* On montre que pc' est satisfiable en se servant du fait que bas est 
                       un chemin faisable noir. *)
                    moreover
                    have "sat pc'"
                    proof -

                      (* Soient pc1 pc2 pc3 des configurations telles que : 
                         - pc1 est obtenue par ES de la trace de ras1 depuis la conf. en rl, 
                         - pc2 est obtenue par ES de ba depuis pc1, 
                         - pc3 est obtenue par ES de bas2' depuis pc2,
                         - pc3 est satisfiable. *)
                      obtain pc1 pc2 pc3
                      where "se_star (confs prb' rl) (trace (ui_as ras1) (labelling (black prb'))) pc1"
                      and   "se pc1 (labelling (black prb') ba) pc2"
                      and   "se_star pc2 (trace bas2' (labelling (black prb'))) pc3"
                      and   "sat pc3"
                      using `bas \<in> feasible_subpaths_from (black prb') (confs prb' rl) (fst rl)` 
                            1 `bas2 = ba # bas2'`
                      unfolding feasible_subpath_from_def feasible_subpath_def feasible_def
                      by (simp add : se_star_append se_star_Cons) blast

                      (* Donc pc2 est également satisfiable. *)
                      hence "sat pc2" by (simp add : se_star_sat_imp_sat se_sat_imp_sat)

                      (* De plus, ras1 est chemin de la nouvelle partie rouge menant de rl à rl'. *)
                      moreover
                      have  "subpath (red prb') rl ras1 rl' (subs prb')"
                      using `subpath (red prb)  rl ras1 rl' (subs prb)`
                      using se_step by (auto simp add : sp_in_extends_w_subs)

                      (* Donc pc1 est subsumée par la configuration en rl', par la relation 
                         existant entre les configurations de la partie rouge. *)
                      hence "pc1 \<sqsubseteq>\<^sub>p\<^sub>c confs prb' rl'" 
                      using se_step
                            `se_star (confs prb' rl) (trace (ui_as ras1) (labelling (black prb'))) pc1`
                            RB.se_step[OF se_step(1,3)]
                            finite_labels_RB.SE_rel[of prb' rl ras1 rl' pc1]
                      by (simp add : finite_labels_RB_def)

                      (* Donc pc2 est subsumée par pc', par monotonie de l'ES pour la 
                         subsomption. *)
                      hence "pc2 \<sqsubseteq>\<^sub>p\<^sub>c pc'" 
                      using `se (confs prb' rl') (labelling (black prb') ba) pc'`
                            `se pc1 (labelling (black prb') ba) pc2`
                      by (simp add : se_mono_for_sub)
                      
                      (* Or puisque pc2 est satisfiable, pc' l'est également, par def. de la 
                         subsomption. *)
                      ultimately
                      show ?thesis  by (simp add : sat_sub_by_sat)
                    qed

                    (* donc ba est bien faisable depuis rl' *)
                    ultimately
                    show ?thesis by blast
                  qed

                  (* donc ba est libre depuis rl' dans la nouvelle partie rouge. *)
                  ultimately
                  show ?thesis by (simp add : free_from_def)
                qed

                (* puisque ba est libre depuis rl' dans la nouvelle partie rouge, rl' 
                   fait bien partie de sa fringe. *)
                hence "rl' \<in> fringe prb'"
                using se_step `rl' \<in> fringe prb` `rl' = src ra`
                      seE_fringe1[OF se_step(1,3)] seE_fringe2[OF se_step(1,3)] 
                by auto
                
                (* de plus ras1, étant un chemin de l'ancienne partie rouge, est bien un chemin de 
                   la nouvelle partie rouge. *)
                moreover
                have "subpath (red prb') rl ras1 rl' (subs prb')" 
                using se_step 3 by (auto simp add : sp_in_extends_w_subs)

                (* Montrons que ras1 est maximal *)
                moreover
                have "\<not> (\<exists>ras21 bas22. bas2 = ui_as ras21 @ bas22 
                                     \<and> ras21 \<noteq> [] 
                                     \<and> subpath_from (red prb') rl' ras21 (subs prb'))"
                proof (intro notI, elim exE conjE)

                  (* supposons qu'il existe un préfixe rouge non-vide ras21 de bas2 *)
                  fix ras21 bas22

                  assume "bas2 = ui_as ras21 @ bas22"
                  and    "ras21 \<noteq> []"
                  and    "subpath_from (red prb') rl' ras21 (subs prb')"

                  (* soit ra' le premier arc de ce préfixe. *)
                  then obtain ra' ras21' where "ras21 = ra' # ras21'" 
                  by (simp add : neq_Nil_conv) blast

                  (* ra' n'est pas ra, car ra' représente ba (contrairement à ra) *)
                  have "ra' \<noteq> ra" 
                  proof -
                    have "ui_arc ra' = ba"
                    using `bas2 = ba # bas2'` `bas2 = ui_as ras21 @ bas22` `ras21 = ra' # ras21'`
                    by simp

                    thus ?thesis using `ui_arc ra \<noteq> ba` by auto
                  qed
                  
                  (* de plus, ra' est un arc de la nouvelle partie rouge issu de rl' *)
                  moreover
                  have "ra' \<in> out_arcs (red prb') rl'"
                  proof -
                    have "rl' = src ra' \<or> (rl', src ra') \<in> subs prb'" 
                    and  "ra' \<in> arcs (red prb')"
                    using `subpath_from (red prb') rl' ras21 (subs prb')`
                          `ras21 = ra' # ras21'`
                          rb_sp_Cons[OF RB.se_step[OF se_step(1,3)]]
                    unfolding subpath_from_def by auto
              
                    moreover
                    have "(rl', src ra') \<notin> subs prb'" 
                    using `rl' \<in> fringe prb'` by (force simp add : fringe_def subsumees_def)

                    ultimately
                    show ?thesis by (simp add : out_arcs_def)
                  qed

                  (* donc ra' est également un arc de l'ancienne partie rouge issu de rl' *)
                  ultimately
                  have "ra' \<in> out_arcs (red prb) rl'" 
                  using se_step `rl' = src ra` by (auto simp add : extends_src_out_arcs)

                  (* donc [ra'] consistue un préfixe rouge de bas2 *)
                  hence "subpath_from (red prb) rl' [ra'] (subs prb)"
                  using rb_sp_one[OF se_step(1)] unfolding subpath_from_def out_arcs_def by fast

                  (* ce qui contredit le fait ras21 est maximal, cqfd. *)
                  thus False 
                  using 4 `bas2 = ui_as ras21 @ bas22` `ras21 = ra' # ras21'`
                  by (elim notE, 
                      rule_tac ?x="[ra']" in exI, 
                      rule_tac ?x="ui_as ras21'@bas22" in exI) simp
                qed

                ultimately
                show ?case using se_step 2 3 4 5 by (simp del : split_paired_Ex) blast
              qed
            qed

          next

            (* Supposons que le préfixe rouge ne termine pas en la source du nouvel arc, 
               mais en un sommet quelconque (rl') de la fringe de l'ancienne partie rouge. 

               Dans ce cas, rl' n'a pas été "impactée" par l'ajout du nouvel arc et fait donc 
               également partie de la fringe de la nouvelle partie rouge. 

               De plus, ras1 est bien un préfixe maximal de bas entierement représentable dans 
               la nouvelle partie rouge menant en rl', et bas2 est bien un suffixe de bas 
               issu de (fst rl').

               Donc bas est bien un chemin rouge-noir du nouveau rb_lts. *)

            assume "rl' \<noteq> src ra"
            
            (* En particulier, rl' ne peut être la cible du nouvel arc. *)
            have   "rl' \<noteq> tgt ra" 
            proof -
              have "rl' \<in> red_locations prb" using `rl' \<in> fringe prb` by (simp add : fringe_def)

              thus ?thesis using se_step by (auto simp add : extendsE2)
            qed

            show ?thesis 
            unfolding SP_from_def Un_iff Bex_def mem_Collect_eq
            proof (intro disjI2, rule_tac ?x="ras1" in exI, rule_tac ?x="bas2" in exI, intro conjI)
            
              show "bas = ui_as ras1 @ bas2" using `bas = ui_as ras1 @ bas2` by assumption

            next

              case goal2
              show ?case
              proof (rule_tac ?x="rl'" in exI, intro conjI)

                (* puisque rl' est dans la fringe de l'ancienne partie rouge et n'est ni la source 
                   ni la cible du nouvel arc, rl' est également dans la fringe de la nouvelle 
                   partie rouge. *)
                show  "rl' \<in> fringe prb'" 
                using se_step `rl' \<in> fringe prb` `rl' \<noteq> src ra` `rl' \<noteq> tgt ra`
                      seE_fringe1[of prb ra rg' pc prb']
                      seE_fringe2[of prb ra rg' pc prb']
                      seE_fringe3[of prb ra rg' pc prb']
                      seE_fringe4[of prb ra rg' pc prb']
                by blast

              next

                (* de plus, puisque ras1 est entierement représentable de rl à rl' dans l'ancienne 
                   partie rouge, il l'est également dans la nouvelle. *)
                have "rl' \<in> red_locations prb" 
                using `subpath (red prb) rl ras1 rl' (subs prb)` 
                      sub_rel_of.lst_of_sp_is_vert[OF subs_sub_rel_of[OF se_step(1)]]
                by fast

                thus "subpath (red prb') rl ras1 rl' (subs prb')" 
                using se_step `rl \<in> red_locations prb` `subpath (red prb) rl ras1 rl' (subs prb)`
                by (auto simp add : sp_in_extends_w_subs)

              next

                (* On montre que ras1 est maximal dans la nouvelle partie rouge, ie. il n'existe 
                   aucun préfixe de bas2 représentable dans la nouvelle partie rouge. *)
                case goal3 
                show ?case 
                proof (intro notI, elim exE conjE)
                
                  (* supposons qu'il existe un tel préfixe, et soit ras21 sa représentation *)
                  fix ras21 bas22

                  assume "bas2 = ui_as ras21 @ bas22"
                  and    "ras21 \<noteq> []"
                  and    "subpath_from (red prb') rl' ras21 (subs prb')"

                  (* ras21 étant non vide, on appele ra' son premier arc *)
                  moreover
                  then obtain ra' ras21' where "ras21 = ra' # ras21'" 
                  by (simp add : neq_Nil_conv) blast

                  (* montrons que [ra'] est représentable depuis rl' dans l'ancienne partie rouge. *)
                  moreover
                  have "subpath_from (red prb) rl' [ra'] (subs prb)"
                  proof -

                    (* puisque ras21 est un sous-chemin de la nouvelle partie rouge issu de rl',
                       son premier arc est bien un arc de la nouvelle partie rouge. *)
                    have "rl' = src ra' \<or> (rl',src ra') \<in> subs prb"
                    and  "ra' \<in> arcs (red prb')"
                    using se_step `subpath_from (red prb') rl' ras21 (subs prb')` `ras21 = ra' # ras21'`
                          rb_sp_Cons[OF RB.se_step[OF se_step(1,3)]] 
                    by (auto simp add : subpath_from_def)
                    
                    (* De plus, la source de ce premier arc est bien rl', car rl' est dans la fringe 
                       de prb' et ne peut être subsumée. *)
                    have "rl' = src ra'" 
                    proof -
                      have "rl' \<notin> subsumees (subs prb)" 
                      using `rl' \<in> fringe prb` by (simp add : fringe_def)

                      thus ?thesis 
                      using `rl' = src ra' \<or> (rl',src ra') \<in> subs prb` 
                      by (force simp add : subsumees_def)
                    qed

                    (* on obtient de plus que ra' est un arc de l'ancienne partie rouge, puisque 
                       rl' n'est ni la source ni la cible du nouvel arc. *)
                    moreover
                    have "ra' \<in> arcs (red prb)" 
                    proof -

                      have "ra' \<in> out_arcs (red prb') rl'"
                      using `ra' \<in> arcs (red prb')` `rl' = src ra'` by (simp add : out_arcs_def)

                      hence "ra' \<in> out_arcs (red prb) rl'"
                      using se_step `rl' \<noteq> src ra` `rl' \<noteq> tgt ra` 
                      by (auto simp add : extends_not_end_out_arcs)

                      thus ?thesis by (simp add : out_arcs_def)
                    qed

                    (* donc ra' est un arc issu de rl' dans l'ancienne partie rouge, donc [ra'] est 
                       bien un sous-chemin issu de rl' dans l'ancienne partie rouge. *)
                    ultimately
                    show ?thesis 
                    using rb_sp_one[OF se_step(1)] unfolding subpath_from_def by blast
                  qed
                  
                  (* donc il existe un préfixe de bas2 non-vide représentable par [ra'] dans 
                     l'ancienne partie rouge, ce qui avait été supposé faux précédemment, cqfd. *)
                  ultimately
                  show False using 4 unfolding not_ex
                  by (drule_tac ?x="[ra']" in spec, drule_tac ?x="ui_as ras21'@bas22" in spec) simp
                qed
                
              next

                show "Graph.subpath_from (black prb') (fst rl') bas2" 
                using se_step `Graph.subpath_from (black prb) (fst rl') bas2` by simp

              qed
            qed
          qed
        qed
      qed

    next

      (* Si rl est la cible du nouvel arc, alors on distingue selon que le sous-chemin faisable 
         noir considéré (bas) est vide ou non. *)

      assume "rl = tgt ra"

      show ?thesis 
      proof (case_tac "bas = []")

        (* si bas est vide, alors il est représentable dans la nouvelle partie rouge, cqfd. *)

        assume "bas = []"

        thus ?thesis 
        using `rl \<in> red_locations prb'` Nil_sp[of "red prb'" rl _ "subs prb'"]
        unfolding SP_from_def Un_iff subpath_from_def image_def Bex_def mem_Collect_eq
        by blast

      next

        (* Supposons que bas est non-vide. Dans ce cas, le seul préfixe de bas dans la nouvelle 
           partie rouge est la liste vide, car dans la nouvelle partie rouge aucun arc n'est 
           issu de rl (qui est la cible du nouvel arc). 

           De plus, rl est dans la fringe de prb', car aucun arc n'en est issu. Or puisque bas 
           est faisable, ba l'est égalemenent depuis rl. ba est donc libre depuis rl.

           On peut donc exhiber un préfixe rouge maximal de bas menant à la fringe de la nouvelle 
           partie rouge ([]), ainsi qu'un suffixe noir de bas issu prolongeant ce préfixe (bas).

           Appelons ba le premier arc de bas et bas' la suite de ce sous-chemin. *)

        assume "bas \<noteq> []"

        then obtain ba bas' where "bas = ba # bas'" by (simp only : neq_Nil_conv) blast

        show ?thesis 
        unfolding SP_from_def Un_iff mem_Collect_eq
        proof (intro disjI2, rule_tac ?x="[]" in exI, rule_tac ?x="bas" in exI, intro conjI)

          show "bas = ui_as [] @ bas" by simp

        next

          (* On montre que rl est dans la fringe de prb' *)
          case goal2 
          show ?case 
          unfolding Bex_def
          proof (rule_tac ?x="rl" in exI, intro conjI)

            case goal1 

            (* rl ne peut être subsumée dans la nouvelle partie rouge, car rl = tgt ra, qui vient 
               d'être ajouté par SE. *)
            have "rl \<notin> subsumees (subs prb')"
            proof (intro notI)
              assume "rl \<in> subsumees (subs prb')"

              hence  "tgt ra \<in> red_locations prb"
              using se_step subs_sub_rel_of[OF se_step(1)] `rl = tgt ra`
              unfolding sub_rel_of_def subsumees_conv by simp blast

              thus False using se_step by (auto simp add : extendsE2)
            qed

            (* de plus, ba, le premier arc de bas2, est libre depuis rl. En effet, il est
               issu de (fst rl) et n'est pas représenté depuis rl dans la nouvelle partie rouge.
               De plus son ES depuis la configuration en rl donne forcément une conf. sat 
               puisque bas est faisable. *)
            moreover
            have "free_from prb' rl ba"
            unfolding free_from_def
            proof (intro conjI)
              show "ba \<in> out_arcs (black prb') (fst rl)" 
              using `bas \<in> feasible_subpaths_from (black prb') (confs prb' rl) (fst rl)`
                    `bas = ba # bas'`
              unfolding feasible_subpath_from_def feasible_subpath_def
              by (auto simp add : Graph.sp_Cons out_arcs_def)    
            next
              show "\<forall>ra\<in>out_arcs (red prb') rl. ui_arc ra \<noteq> ba" 
              using se_step `rl = tgt ra` by (auto simp add : extends_tgt_out_arcs)
            next
              obtain pc1 pc2 where "se (confs prb' rl) (labelling (black prb') ba) pc1"
                             and   "se_star pc1 (trace bas' (labelling (black prb'))) pc2"
                             and   "sat pc2"
              using `bas \<in> feasible_subpaths_from (black prb') (confs prb' rl) (fst rl)`
                    `bas = ba # bas'`
              unfolding feasible_subpath_from_def feasible_subpath_def feasible_def
              by (simp add : se_star_Cons) blast

              moreover
              hence "sat pc1" by (simp add : se_star_sat_imp_sat)

              ultimately
              show "\<exists>pc'. se (confs prb' rl) (labelling (black prb') ba) pc' \<and> PreConf.sat pc'"
              by blast     
            qed

            ultimately
            show ?case using `rl \<in> red_locations prb'` by (auto simp add : fringe_def)

          next

            (* la séquence vide est bien un sous-chemin de la nouvelle partie rouge issu de rl,
               puisque rl est la cible du nouvel arc. *)
            show "subpath (red prb') rl [] rl (subs prb')" 
            using `rl \<in> red_locations prb'` by (simp add : Nil_sp)

          next

            (* on montre qu'il n'existe pas de préfixe non-vide de bas qui serait représentable 
               dans la nouvelle partie rouge *)
            case goal3 
            show ?case 
            proof (intro notI, elim exE conjE)

              fix ras21 bas22
              
              (* supposons qu'un tel préfixe existe *)
              assume "bas = ui_as ras21 @ bas22"
              and    "ras21 \<noteq> []"
              and    "subpath_from (red prb') rl ras21 (subs prb')"

              (* ce préfixe est vide puisqu'il est issu de rl, qui est la cible du nouvel arc *)
              moreover
              hence "ras21 = []" 
              using se_step `rl = tgt ra` 
                    subs_sub_rel_of sub_rel_of.sp_from_tgt_in_extendsion_is_Nil_w_subs
              by (force simp add : subpath_from_def)
              
              (* contradiction *)
              ultimately
              show False by (elim notE)
            qed

          next

            (* finalement, bas étant un sous-chemin faisable noir issu de (fst rl), bas est en 
               particulier un sous-chemin noir issu de (fs rl) *)
            
            show "Graph.subpath_from (black prb') (fst rl) bas" 
            using `bas \<in> feasible_subpaths_from (black prb') (confs prb' rl) (fst rl)`
            by (simp add : feasible_subpath_from_conv)

          qed
        qed
      qed
    qed
  qed

next

  case (subsum_step prb sub subrel prb')

  show ?case
  unfolding subset_iff
  proof (intro allI impI)

    (* Soit bas un sous-chemin faisable noir issu de (fst rl) *)
    fix bas

    assume "bas \<in> feasible_subpaths_from (black prb') (confs prb' rl) (fst rl)"

    (* par IH, bas est donc un chemin rouge/noir de prb *)
    hence "bas \<in> SP_from prb rl" using subsum_step by simp fast

    thus   "bas \<in> SP_from prb' rl" 
    apply (subst (asm) SP_from_def)
    unfolding mem_Collect_eq Un_iff Bex_def
    proof (elim disjE exE conjE)

      (* supposons que bas soit entierement representable dans la partie rouge de prb *) 
      assume "bas \<in> ui_as ` subpaths_from (red prb) rl (subs prb)"
      
      (* dans ce cas il l'est egalement dans la partie rouge de prb' *)
      thus "bas \<in> SP_from prb' rl" 
      unfolding SP_from_def
      using subsum_step SubExt.sp_from_in_extends[of "red prb" "subs prb" sub subrel rl]
      by auto
      
    next

      (* supposons maintenant que bas possede un prefixe maximal rouge ras1 dans prb
         menant au sommet rouge rl1 faisant partie de la fringe de prb et un suffixe noir 
         bas2 *)
      fix ras1 bas2 rl1
      
      assume 1 : "bas = ui_as ras1 @ bas2"
      and    2 : "rl1 \<in> fringe prb"
      and    3 : "subpath (red prb) rl ras1 rl1 (subs prb)"
      and    4 : "\<not> (\<exists> ras21 bas22. bas2 = ui_as ras21 @ bas22 
                                  \<and> ras21 \<noteq> [] 
                                  \<and> subpath_from (red prb) rl1 ras21 (subs prb))"
      and    5 : "Graph.subpath_from (black prb) (fst rl1) bas2"

      hence  6 : "subpath (red prb') rl ras1 rl1 (subs prb')" 
      using subsum_step SubExt.sp_in_extends[of "red prb" "subs prb" sub subrel rl]
      by auto

      show "bas \<in> SP_from prb' rl"
      proof (case_tac "rl1 = subsumee sub")

        (* supposons que rl1 soit le subsume de la nouvelle subsomption *)
        assume "rl1 = subsumee sub"

        (* dans ce cas le suffixe noir bas2 est un chemin rouge/noir du RB de départ issu du 
           subsumeur de la nouvelle subsomption. *)
        have  "bas2 \<in> SP_from prb (subsumer sub)"
        proof -

          have "subsumer sub \<in> red_locations prb" 
          using subsum_step by (auto simp add : SubExt.extends_def)

          (* en effet, montrons que bas2 est un sous-chemin noir faisable depuis (subsumer sub) *)
          moreover
          have "bas2 \<in> feasible_subpaths_from (black prb) (confs prb (subsumer sub)) (fst (subsumer sub))"
          proof -

            (* bas2 etant un sous-chemin noir issu de (fst rl1), bas2 est egalement un sous-chemin 
               noir issu de (fst (subsumer sub)), car rl1 = (subsumee sub) 
               et fst (subsumer sub) = fst (subsumee sub). *)
            have "Graph.subpath_from (black prb) (fst (subsumer sub)) bas2"
            using subsum_step 5 `rl1 = subsumee sub` 
                  subs_to_same_BL2[OF RB.subsum_step[OF subsum_step(1,3)]]
                  SubExt.extendsE1[of "red prb" "subs prb"]
            by (auto simp add : SubExt.extendsE1)

            (* De plus montrons que bas2 est faisable depuis (subsumer sub) *)
            moreover
            have "feasible (confs prb (subsumer sub)) (trace bas2 (labelling (black prb)))" 
            proof -

              (* Il existe bien une configuration pc obtenue par SE de bas2 depuis (subsumer sub) 
                 grace a finite_labels_RB. *)
              have "\<exists>pc. se_star (confs prb (subsumer sub)) (trace bas2 (labelling (black prb))) pc"
              using subsum_step(1,3,5,6,7) `subsumer sub \<in> red_locations prb`
                    finite_labels_RB.ex_se_star_succ [of prb]
              by (simp add : finite_labels_RB_def finite_labels_lts_def finite_labels_def)

              then obtain pc 
              where "se_star (confs prb (subsumer sub)) (trace bas2 (labelling (black prb))) pc"
              by blast

              (* montrons que pc est sat *)
              moreover
              have "sat pc" 
              proof -
                
                (* Puisque bas est un sous-chemin noir faisable issu de (fst rl) et de la 
                   forme "ui_as ras1 @ bas2", on obtient deux sommets noirs, bl1 et bl2, et deux 
                   configurations pc1 et pc2 tels que :
                   - (ui_as ras1) est un sous-chemin noir menant de (fst rl) à bl1, son ES 
                     depuis rl donne pc1,
                   - bas2 est un sous-chemin noir menant de bl1 à bl2, son ES depuis pc1 donne 
                     pc2,
                   - pc2 est sat. *)
                obtain bl1 bl2 pc1 pc2 
                where "Graph.subpath (black prb) (fst rl) (ui_as ras1) bl1"
                and   "Graph.subpath (black prb) bl1 bas2 bl2"
                and   "se_star (confs prb rl) (trace (ui_as ras1) (labelling (black prb))) pc1"
                and   "se_star pc1 (trace bas2 (labelling (black prb))) pc2"
                and   "sat pc2"
                using subsum_step `bas \<in> feasible_subpaths_from (black prb') (confs prb' rl) (fst rl)` 1
                unfolding feasible_subpath_from_def feasible_subpath_def feasible_def mem_Collect_eq
                by (simp add : se_star_append Graph.sp_append) blast

                (* On obtient par SE_rel que pc1 est subsumee par la conf en rl1, ie le nouveau 
                   subsumee *)
                moreover
                hence "pc1 \<sqsubseteq>\<^sub>p\<^sub>c (confs prb rl1)"
                using subsum_step 3 by (simp add : finite_labels_RB_def finite_labels_RB.SE_rel)

                (* donc pc1 est egalement subsumee par le nouveau subsumer *)
                hence "pc1 \<sqsubseteq>\<^sub>p\<^sub>c (confs prb (subsumer sub))"
                using subsum_step `rl1 = subsumee sub` subsums_trans by fast
 
                (* donc pc2 est subsumee par pc, par monotonie de la SE pour la subsomption. *)
                hence "pc2 \<sqsubseteq>\<^sub>p\<^sub>c pc"
                using `se_star (confs prb (subsumer sub)) (trace bas2 (labelling (black prb))) pc`
                      `se_star pc1 (trace bas2 (labelling (black prb))) pc2`
                by (simp add : se_star_mono_for_sub)

                (* or puisque pc2 est sat, pc est egalement sat *)
                ultimately
                show ?thesis by (simp add : sat_sub_by_sat)
              qed

              ultimately
              show ?thesis unfolding feasible_def by blast
            qed

            ultimately
            show ?thesis 
            by (simp add : feasible_subpath_from_def feasible_subpath_def Graph.subpath_from_def)
          qed

          (* On peut donc utiliser l'hypothese d'induction de subsum_step *)
          ultimately
          show ?thesis using subsum_step(2,3,5,6,7) by fastforce
        qed

        (*On montre que bas est un sous-chemin rouge-noir par induction (vers l'arrière) sur bas2*)
        thus ?thesis 
        using `bas \<in> feasible_subpaths_from (black prb') (confs prb' rl) (fst rl)` 1 3 5 6
        proof (induct bas2 arbitrary : bas ras1 rule : rev_induct)
        
          (* si bas2 est vide, alors bas est entierement representable dans la nouvelle partie
             rouge par son préfixe ras1, cqfd par def de SP_from. *)
          case (goal1 bas ras1) thus ?case 
          using `subpath (red prb') rl ras1 rl1 (subs prb')`
          unfolding SP_from_def Un_iff subpath_from_def by blast

        next

          (* On suppose que le suffixe noir est de la forme "bas2 @ [ba]".

             Dans ce cas, bas est de la forme : "ui_as ras1 @ bas2 @ [ba]". 

             On va montrer par IH que (ui_as ras1 @ bas2) est un sous-chemin rouge-noir 
             du nouveau RB. On montrera ensuite qu'en ajoutant ba à (ui_as ras1 @ bas2),
             on obtient encore un sous-chemin rouge-noir du nouveau RB.
             
             Puisque bas est un sous-chemin faisable noir issu de (fst rl), on obtient trois 
             sommets rouges bl1 bl2 bl3 et trois configurations pc1 pc2 pc3 tels que :
             - ui_as ras1 est un sous-chemin noir allant de (fst rl) a bl1, son ES depuis 
               (confs prb' rl) donne pc1,
             - bas2 est un sous-chemin noir allant de bl1 à bl2, son ES depuis pc1 donne pc2,
             - [ba] est un sous-chemin noir allant de bl3 à bl2, son ES depuis pc2 donne pc3, 
             - pc3 est sat. *)

          case (goal2 ba bas2 bas ras1)

          obtain bl1 bl2 bl3 pc1 pc2 pc3 
          where "Graph.subpath (black prb) (fst rl)  (ui_as ras1) bl1"
          and   "Graph.subpath (black prb) bl1 bas2 bl2"
          and   "Graph.subpath (black prb) bl2 [ba] bl3"
          and   "se_star (confs prb' rl) (trace (ui_as ras1) (labelling (black prb))) pc1"
          and   "se_star pc1 (trace bas2 (labelling (black prb))) pc2"
          and   "se pc2 (labelling (black prb) ba) pc3"
          and   "sat pc3"
          using subsum_step goal2(3,4)
          unfolding mem_Collect_eq feasible_subpath_from_def feasible_subpath_def feasible_def
          by (simp add : Graph.sp_append se_star_append se_star_one) blast

          (* Or puisque ras1 est un sous-chemin menant de rl à rl1 dans l'ancienne partie rouge,
             (ui_as ras1) est egalement un sous-chemin noir allant de (fst rl) à (fst rl1), donc 
             bl1 = (fst rl1)  *)
          have "bl1 = fst rl1" 
          proof -

            have "Graph.subpath (black prb) (fst rl) (ui_as ras1) (fst rl1)"
            using `subpath (red prb) rl ras1 rl1 (subs prb)`  
                  rsp_is_bsp[OF subsum_step(1)] 
            by fast

            thus ?thesis 
            using `Graph.subpath (black prb) (fst rl)  (ui_as ras1) bl1` 
            by (simp add : sp_same_src_imp_same_tgt)
          qed
          
          (* De plus, on a bl2 et bl3 qui sont resp. la source et la cible de ba. *)
          have "bl2 = src ba" and "bl3 = tgt ba" 
          using `Graph.subpath (black prb) bl2 [ba] bl3` by (auto simp add : Graph.sp_one) 

          (* Puisque bas2@[ba] est un sous-chemin rouge-noir issu de (subsumer sub), on a
            egalement que bas2 est un sous-chemin rouge-noir issu de (subsumer sub). *)
          have "bas2 \<in> SP_from prb (subsumer sub)"
          using subsum_step(1) goal2(2) by (simp add : SP_from_append_oneE)

          (* de plus, montrons que (ui_as ras1 @ bas2) est un sous-chemin noir faisable 
             depuis rl *)
          moreover
          have "ui_as ras1 @ bas2 \<in> feasible_subpaths_from (black prb') (confs prb' rl) (fst rl)" 
          proof -

            (* Par concatenation de (ui_as ras1) et bas2, on obtient que 
               (ui_as ras1 @ bas2 est un sous-chemin noir issu de rl.  *)
            have "Graph.subpath_from (black prb') (fst rl) (ui_as ras1 @ bas2)"
            using subsum_step `Graph.subpath (black prb) (fst rl)  (ui_as ras1) bl1`
                  `Graph.subpath (black prb) bl1 bas2 bl2`
            by (simp add : Graph.subpath_from_def Graph.sp_append) blast

            (* montrons que (ui_as ras1 @ bas2) est faisable depuis rl. *) 
            moreover
            have "feasible (confs prb' rl) (trace (ui_as ras1 @ bas2) (labelling (black prb')))"
            proof -

              (* par concatenation de (ui_as ras1) et bas2, on obtient que l'ES 
                 de (ui_as ras1 @ bas2) depuis rl donne pc2 *)
              have "se_star (confs prb' rl) (trace (ui_as ras1 @ bas2) (labelling (black prb'))) pc2"
              using subsum_step 
                    `se_star (confs prb' rl) (trace (ui_as ras1) (labelling (black prb))) pc1`
                    `se_star pc1 (trace bas2 (labelling (black prb))) pc2`
              by (simp add : se_star_append) blast

              (* or pc2 est sat, car pc3 est sat et obtenue par ES depuis pc2 *)
              moreover
              have "sat pc2" 
              using `se pc2 (labelling (black prb) ba) pc3` `sat pc3` by (simp add : se_sat_imp_sat)

              ultimately
              show ?thesis by (simp add : feasible_def) blast
            qed

            ultimately
            show ?thesis 
            by (simp add : feasible_subpath_from_def feasible_subpath_def Graph.subpath_from_def) 
          qed

          (* de plus, bas2 est un sous-chemin noir faisable issu de rl1 *) 
          moreover
          have  "Graph.subpath_from (black prb) (fst rl1) bas2"
          using `Graph.subpath_from (black prb) (fst rl1) (bas2 @ [ba])`
          using  Graph.sp_append_one[of "black prb"] unfolding Graph.subpath_from_def by blast

          (* donc on peut utiliser l'IH courante pour montrer que (ui_as ras1 @ bas2) est un 
             sous-chemin rouge-noir de la nouvelle partie rouge *)
          ultimately
          have "ui_as ras1 @ bas2 \<in> SP_from prb' rl" using goal2 by auto


          (* On montre alors que la concatenation de ba a (ui_as ras1 @ bas2) est un 
             sous-chemin rouge-noir du nouveau RB *)
          thus ?case 
          apply (subst (asm) SP_from_def)
          unfolding Un_iff Bex_def mem_Collect_eq
          proof (elim disjE exE conjE)

            (* supposons que (ui_as ras1 @ bas2) soit entierement representable dans la nouvelle 
               partie rouge. On va ensuite montrer que : 
               
               1) si ba est represente par un arc rouge depuis rl2 dans la nouvelle partie rouge, 
                  alors (ui_as ras1 @ bas2 @ [ba]) est egalement entierement represente dans 
                  la nouvelle partie rouge, cqfd, 
               
               2) si ba n'est pas represente dans la nouvelle partie rouge depuis rl2, alors :

                  - soit rl2 n'est pas subsumee. Dans ce cas, puisque bas est un sous-chemin noir 
                    faisable, alors ba est libre depuis rl2 et rl2 fait partie de la fringe 
                    de la nouvelle partie rouge, cqfd

                  - soit rl2 est subsume. Dans ce cas, on applique alors le même raisonnement à  
                    partir du subsumeur de rl2 : on se ramene alors au cas précédent puisque ce 
                    subsumeur n'est pas subsumé. *)

            case goal1 

            (* on appelle ras le sous-chemin rouge representant (ui_as ras1 @ bas2) depuis rl *)
            then obtain ras rl2 where "ui_as ras1 @ bas2 = ui_as ras"
                                and   "subpath (red prb') rl ras rl2 (subs prb')"
            unfolding subpath_from_def by blast

            (* montrons que rl2 est une version indicee de la source de ba *)
            have "src ba = fst rl2"
            proof -

              (* en effet, puisque ras est un sous-chemin de la nouvelle partie rouge 
                 menant de rl a rl2, (ui_as ras) est un sous-chemin noir 
                 menant de (fst rl) a (fst rl2). *)
              have "Graph.subpath (black prb) (fst rl) (ui_as ras) (fst rl2)" 
              using subsum_step `subpath (red prb') rl ras rl2 (subs prb')` 
                    rsp_is_bsp[OF RB.subsum_step[OF subsum_step(1,3)]]
              by clarsimp

              (* Or (ui_as ras1 @ bas2) (ie. ui_as ras) est egalement un sous-chemin noir allant 
                 de (fst rl) a bl2. *)
              moreover
              have "Graph.subpath (black prb) (fst rl) (ui_as ras1 @ bas2) bl2"
              using `Graph.subpath (black prb) (fst rl)  (ui_as ras1) bl1`
                    `Graph.subpath (black prb) bl1 bas2 bl2`
              by (simp add : Graph.sp_append) blast

              (* or bl2 = src ba, cqfd. *)
              ultimately
              show ?thesis
              using `ui_as ras1 @ bas2 = ui_as ras` `bl2 = src ba` by (simp add : sp_same_src_imp_same_tgt)
            qed
          

            (* On distingue selon que ba est represente depuis rl2 ou non *)
            show ?thesis 
            proof (case_tac "\<exists> ra \<in> out_arcs (red prb') rl2. ba = ui_arc ra")

              (* si c'est le cas, alors (ui_as ras1 @ bas2 @ [ba]) est un sous-chemin issu de rl 
                 dans la nouvelle partie rouge, cqfd *)
              case goal1 
                
              then obtain ra where "ra \<in> out_arcs (red prb') rl2"
                             and   "ba = ui_arc ra"
              by blast

              hence "subpath (red prb') rl2 [ra] (tgt ra) (subs prb')" 
              by (auto simp add : out_arcs_def arc_sp_eq)
                
              hence "subpath (red prb') rl (ras@[ra]) (tgt ra) (subs prb')"
              using wf_sub_rel.sp_appendI1[OF subs_wf_sub_rel[OF RB.subsum_step[OF subsum_step(1,3)]]]
                    `subpath (red prb') rl ras rl2 (subs prb')`
              by fast

              thus ?thesis 
              using `bas = ui_as ras1 @ bas2 @ [ba]` `ui_as ras1 @ bas2 = ui_as ras` `ba = ui_arc ra`
              unfolding SP_from_def subpath_from_def Un_iff image_def Bex_def mem_Collect_eq
              apply (intro disjI1)
              apply (rule_tac ?x="ras@[ra]" in exI)
              apply (intro conjI)
                apply (rule_tac ?x="tgt ra" in exI)
                apply simp_all
              done
            
            next

              (* sinon, on distingue selon que rl2 est subsumee ou non. *)
              case goal2

              show ?thesis
              proof (case_tac "rl2 \<in> subsumees (subs prb')")


                (* supposons que rl2 est subsumée dans la nouvelle partie rouge. *)
                assume "rl2 \<in> subsumees (subs prb')"

                (* on appelle arl2 le subsumeur de rl2. *)
                then obtain arl2 where "(rl2,arl2) \<in> (subs prb')" unfolding subsumees_conv by blast

                (* puisque ras mène en rl2, ras mène également en arl2 *)
                have "subpath (red prb') rl ras arl2 (subs prb')" 
                using `subpath (red prb') rl ras rl2 (subs prb')` `(rl2,arl2) \<in> (subs prb')` 
                      sub_rel_of.sp_append_sub[OF sub_rel_of_subsumE[OF subsum_step(1,3)]]
                by simp

                (* on distingue selon que ba est représenté ou non depuis arl2. *)
                show ?thesis
                proof (case_tac "\<exists> ra \<in> out_arcs (red prb') arl2. ba = ui_arc ra")

                  (* supposons que ba soit représenté depuis arl2. Dans ce cas, 
                     (ui_as ras1 @ bas2 @ [ba]) est un sous-chemin issu de rl dans la nouvelle 
                     partie rouge, cqfd. *)
                  assume "\<exists> ra \<in> out_arcs (red prb') arl2. ba = ui_arc ra"

                  (* soit ra l'arc représentant ba depuis arl2 *)
                  then obtain ra where "ra \<in> out_arcs (red prb') arl2"
                                 and   "ba = ui_arc ra"
                  by blast

                  (* alors [ra] est un sous-chemin de la nouvelle partie rouge issu de rl2. *)
                  hence "subpath (red prb') rl2 [ra] (tgt ra) (subs prb')" 
                  proof -
                    have "subpath (red prb') arl2 [ra] (tgt ra) (subs prb')"
                    using `ra \<in> out_arcs (red prb') arl2` 
                          rb_sp_one[OF RB.subsum_step[OF subsum_step(1,3)], of arl2 ra "tgt ra"]
                    by (auto simp add : out_arcs_def)
                    
                    thus ?thesis 
                    using `(rl2,arl2) \<in> (subs prb')` `ra \<in> out_arcs (red prb') arl2`
                          rb_sp_one[OF RB.subsum_step[OF subsum_step(1,3)], of rl2 ra "tgt ra"]
                    by (auto simp add : out_arcs_def)
                  qed

                  (* donc ras@[ra] est un sous-chemin de la nouvelle partie rouge issu de rl *)
                  hence "subpath (red prb') rl (ras@[ra]) (tgt ra) (subs prb')"
                  using wf_sub_rel.sp_appendI1[OF subs_wf_sub_rel[OF RB.subsum_step[OF subsum_step(1,3)]]]
                        `subpath (red prb') rl ras rl2 (subs prb')`
                  by fast

                  (* cqfd. *)
                  thus ?thesis 
                  using `bas = ui_as ras1 @ bas2 @ [ba]` `ui_as ras1 @ bas2 = ui_as ras` `ba = ui_arc ra`
                  unfolding SP_from_def subpath_from_def Un_iff image_def Bex_def mem_Collect_eq
                  apply (intro disjI1)
                  apply (rule_tac ?x="ras@[ra]" in exI)
                  apply (intro conjI)
                    apply (rule_tac ?x="tgt ra" in exI)
                      apply simp_all
                  done

                next

                  (* supposons que ba ne soit pas représenté depuis arl2. Dans ce cas, ba
                     est libre depuis arl2, donc arl2 est dans la fringe de la nouvelle partie 
                     rouge. 
                      
                     On peut alors conclure : le préfixe recherché est ras et le suffixe [ba] *)
                  assume "\<not> (\<exists> ra \<in> out_arcs (red prb') arl2. ba = ui_arc ra)"

                  have "arl2 \<in> red_locations prb'" 
                  using sub_rel_of_subsumE[OF subsum_step(1,3)]
                        `(rl2,arl2) \<in> (subs prb')`
                  unfolding sub_rel_of_def subsumers_conv by blast

                  moreover
                  have "arl2 \<notin> subsumees (subs prb')"
                  proof (intro notI)
                    assume "arl2 \<in> subsumees (subs prb')"

                    moreover
                    have   "arl2 \<in> subsumers (subs prb')" 
                    using `(rl2,arl2) \<in> (subs prb')` unfolding subsumers_conv by blast

                    ultimately 
                    show False 
                    using subs_wf_sub_rel_in_subsumE[OF subsum_step(1,3)]
                    by (auto simp add : wf_sub_rel_def)
                  qed

                  moreover
                  have "free_from prb' arl2 ba" 
                  proof -

                    have "ba \<in> out_arcs (black prb') (fst arl2)"
                    proof -
                      have "ba \<in> arcs (black prb')" 
                      using subsum_step `Graph.subpath (black prb) bl2 [ba] bl3` 
                      by (simp add : Graph.sp_one out_arcs_def)

                      moreover
                      have "fst rl2 = fst arl2" 
                      using `(rl2,arl2) \<in> (subs prb')`
                        subs_to_same_BL2[OF RB.subsum_step[OF subsum_step(1,3)]]
                      by fastforce

                      hence "src ba = fst arl2" using `bl2 = src ba` `src ba = fst rl2` by simp
                      
                      ultimately
                      show ?thesis by (simp add : out_arcs_def)
                    qed

                    moreover
                    have "\<forall>ra\<in>out_arcs (red prb') arl2. ui_arc ra \<noteq> ba"
                    using `\<not> (\<exists> ra \<in> out_arcs (red prb') arl2. ba = ui_arc ra)` by auto

                    moreover
                    have "\<exists>pc. se (confs prb' arl2) (labelling (black prb') ba) pc \<and> sat pc"
                    proof -
                    
                      have "finite_labels_RB prb'"
                      using subsum_step by (auto simp add : finite_labels_RB_def)
                      
                      hence "finite_label (labelling (black prb') ba)"
                      by (auto simp add : finite_labels_RB_def finite_labels_lts_def)

                      then obtain pc where "se (confs prb' arl2) (labelling (black prb') ba) pc" 
                      using finite_labels_RB.ex_se_succ
                            [OF `finite_labels_RB prb'` 
                                RB.subsum_step[OF subsum_step(1,3)]
                                `arl2 \<in> red_locations prb'`]
                      by blast       

                      moreover
                      have "sat pc"
                      proof -

                        have "ui_as ras @ [ba] \<in> feasible_subpaths_from (black prb') (confs prb' rl) (fst rl)"
                        using `bas \<in> feasible_subpaths_from (black prb') (confs prb' rl) (fst rl)`
                             `bas = ui_as ras1 @ bas2 @ [ba]` `ui_as ras1 @ bas2 = ui_as ras`
                        by clarify (erule subst, simp)
                    
                        then obtain pc1 pc2 
                        where "se_star (confs prb' rl) (trace (ui_as ras) (labelling (black prb'))) pc1"
                        and   "se pc1 (labelling (black prb') ba) pc2"
                        and   "sat pc2"
                        unfolding feasible_subpath_from_def feasible_subpath_def feasible_def
                        by (simp add : se_star_append_one) blast
                        
                        moreover
                        hence "pc1 \<sqsubseteq>\<^sub>p\<^sub>c (confs prb' arl2)"
                        using finite_labels_RB.SE_rel
                              [OF `finite_labels_RB prb'`
                                  RB.subsum_step[OF subsum_step(1,3)]
                                  `subpath (red prb') rl ras arl2 (subs prb')`]
                        by simp

                        ultimately
                        have "pc2 \<sqsubseteq>\<^sub>p\<^sub>c pc"
                        using `se (confs prb' arl2) (labelling (black prb') ba) pc` 
                        by (simp add : se_mono_for_sub)
                  
                        thus ?thesis using `sat pc2` by (simp add : sat_sub_by_sat)
                      qed

                      ultimately
                      show ?thesis by blast
                    qed

                    ultimately
                    show ?thesis by (simp add : free_from_def)
                  qed

                  ultimately
                  have "arl2 \<in> fringe prb'" by (simp add : fringe_def) blast

                  moreover
                  have "\<not> (\<exists>ras21 bas22. [ba] = ui_as ras21 @ bas22 
                                       \<and> ras21 \<noteq> [] 
                                       \<and> subpath_from (red prb') arl2 ras21 (subs prb'))"
                  proof (intro notI, elim exE conjE)

                    fix ras21 bas22
                    
                    assume "[ba] = ui_as ras21 @ bas22"
                    and    "ras21 \<noteq> []"
                    and    "subpath_from (red prb') arl2 ras21 (subs prb')"

                    hence "[ba] = ui_as ras21" by (simp only : neq_Nil_conv) auto

                    then obtain ra where "ras21 = [ra]" and "ba = ui_arc ra" by blast

                    moreover
                    have "ra \<in> out_arcs (red prb') arl2" 
                    proof -

                      have "ra \<in> arcs (red prb')" 
                      using `subpath_from (red prb') arl2 ras21 (subs prb')` `ras21 = [ra]`
                      by (auto simp add : subpath_from_def subpath_def)

                      moreover
                      have "arl2 = src ra" 
                      proof -
                      
                        have "arl2 = src ra \<or> (arl2,src ra) \<in> (subs prb')"
                        using `subpath_from (red prb') arl2 ras21 (subs prb')` `ras21 = [ra]`
                               rb_sp_one[OF RB.subsum_step[OF subsum_step(1,3)]]
                        by (simp add : subpath_from_def)

                        thus ?thesis 
                        using `arl2 \<notin> subsumees (subs prb')` unfolding subsumees_conv by blast
                      qed
                      
                      ultimately
                      show ?thesis by (simp add : out_arcs_def)
                    qed

                    ultimately
                    show False 
                    using `\<not> (\<exists> ra \<in> out_arcs (red prb') arl2. ba = ui_arc ra)` by auto
                  qed

                  moreover
                  have "fst arl2 = fst rl2" 
                  using `(rl2,arl2) \<in> (subs prb')`
                        subs_to_same_BL2[OF RB.subsum_step[OF subsum_step(1,3)]]
                  by fastforce

                  hence "Graph.subpath_from (black prb') (fst arl2) [ba]" 
                  using subsum_step
                      `Graph.subpath (black prb) bl2 [ba] bl3` `bl2 = src ba` `src ba = fst rl2`
                  by (simp add : Graph.subpath_from_def) blast

                  ultimately
                  show ?thesis
                  using `bas = ui_as ras1 @ bas2 @ [ba]` `ui_as ras1 @ bas2 = ui_as ras`
                        `subpath (red prb') rl ras arl2 (subs prb')`
                  unfolding SP_from_def Un_iff Bex_def mem_Collect_eq
                  apply (intro disjI2)
                  apply (rule_tac ?x="ras"  in exI)
                  apply (rule_tac ?x="[ba]" in exI)
                  apply (intro conjI)  
                    apply simp
                    apply (rule_tac ?x="arl2" in exI, simp)
                  done
                qed

              next

                (* si rl2 n'est pas subsumee *)
                assume "rl2 \<notin> subsumees (subs prb')"

                moreover
                have   "rl2 \<in> red_locations prb'" 
                using `subpath (red prb') rl ras rl2 (subs prb')`
                      sub_rel_of.lst_of_sp_is_vert 
                      subs_sub_rel_of[OF RB.subsum_step[OF subsum_step(1,3)]]
                by fast

                (* montrons que ba est libre depuis rl2 *)
                moreover
                have "free_from prb' rl2 ba"
                unfolding free_from_def
                proof (intro conjI)
                
                  show "ba \<in> out_arcs (black prb') (fst rl2)" 
                  using subsum_step 
                        `Graph.subpath_from (black prb) (fst rl1) (bas2 @ [ba])` `src ba = fst rl2`
                  by (auto simp add : Graph.subpath_from_def Graph.subpath_def out_arcs_def)

                next

                  show "\<forall>ra\<in>out_arcs (red prb') rl2. ui_arc ra \<noteq> ba" using goal2 by auto

                next
   
                  (* montrons que ba est faisable depuis rl2 *)

                  have "finite_labels_RB prb'"
                  using subsum_step by (auto simp add : finite_labels_RB_def)
                  
                  hence "finite_label (labelling (black prb') ba)"
                  by (auto simp add : finite_labels_RB_def finite_labels_lts_def)

                  (* d'apres finite_labels_RB, il existe bien une conf. pc obtenue depuis rl2 
                     par ES de ba. *)
                  then obtain pc where "se (confs prb' rl2) (labelling (black prb') ba) pc"
                  using finite_labels_RB.ex_se_succ
                        [OF `finite_labels_RB prb'` 
                             RB.subsum_step[OF subsum_step(1,3)]
                             `rl2 \<in> red_locations prb'`]
                  by blast                 
                  
                  (* montrons que pc est sat. *)
                  moreover
                  have "sat pc" 
                  proof -

                    have "ui_as ras @ [ba] \<in> feasible_subpaths_from (black prb') (confs prb' rl) (fst rl)"
                    using `bas \<in> feasible_subpaths_from (black prb') (confs prb' rl) (fst rl)`
                          `bas = ui_as ras1 @ bas2 @ [ba]` `ui_as ras1 @ bas2 = ui_as ras`
                    by clarify (erule subst, simp)
                    
                    then obtain pc1 pc2 
                    where "se_star (confs prb' rl) (trace (ui_as ras) (labelling (black prb'))) pc1"
                    and   "se pc1 (labelling (black prb') ba) pc2"
                    and   "sat pc2"
                    unfolding feasible_subpath_from_def feasible_subpath_def feasible_def
                    by (simp add : se_star_append_one) blast

                    moreover
                    hence "pc1 \<sqsubseteq>\<^sub>p\<^sub>c (confs prb' rl2)"
                    using finite_labels_RB.SE_rel
                          [OF `finite_labels_RB prb'`
                              RB.subsum_step[OF subsum_step(1,3)]
                              `subpath (red prb') rl ras rl2 (subs prb')`]
                    by simp

                    ultimately
                    have "pc2 \<sqsubseteq>\<^sub>p\<^sub>c pc"
                    using `se (confs prb' rl2) (labelling (black prb') ba) pc` 
                    by (simp add : se_mono_for_sub)
                  
                    thus ?thesis using `sat pc2` by (simp add : sat_sub_by_sat)
                  qed

                  ultimately
                  show "\<exists>pc . se (confs prb' rl2) (labelling (black prb') ba) pc \<and> sat pc"
                  by blast
                qed

                (* donc rl2 est dans la fringe de prb' *)
                ultimately
                have "rl2 \<in> fringe prb'" by (simp add : fringe_def) blast

                (* on montre que ras est maximal (ie pas de prefixe rouge de [ba]) *)
                moreover
                have "\<not> (\<exists>ras21 bas22. [ba] = ui_as ras21 @ bas22 
                                     \<and> ras21 \<noteq> [] 
                                     \<and> subpath_from (red prb') rl2 ras21 (subs prb'))"
                proof (intro notI, elim exE conjE)

                  fix ras21 bas22
                  assume "[ba] = ui_as ras21 @ bas22"
                  and    "ras21 \<noteq> []"
                  and    "subpath_from (red prb') rl2 ras21 (subs prb')"

                  hence "[ba] = ui_as ras21" by (simp only : neq_Nil_conv) auto

                  then obtain ra where "ras21 = [ra]" and "ba = ui_arc ra" by blast

                  moreover
                  have "ra \<in> out_arcs (red prb') rl2" 
                  proof -

                    have "ra \<in> arcs (red prb')" 
                    using `subpath_from (red prb') rl2 ras21 (subs prb')` `ras21 = [ra]`
                    by (auto simp add : subpath_from_def subpath_def)

                    moreover
                    have "rl2 = src ra" 
                    proof -
                      
                      have "rl2 = src ra \<or> (rl2,src ra) \<in> (subs prb')"
                      using `subpath_from (red prb') rl2 ras21 (subs prb')` `ras21 = [ra]`
                            rb_sp_one[OF RB.subsum_step[OF subsum_step(1,3)]]
                      by (simp add : subpath_from_def)

                      thus ?thesis 
                      using `rl2 \<notin> subsumees (subs prb')`
                      unfolding subsumees_conv by blast
                    qed

                    ultimately
                    show ?thesis by (simp add : out_arcs_def)
                  qed

                  ultimately
                  show False using goal2 by auto
                qed

                (* de plus [ba] est bien sous-chemin noir issu de (fst rl2) *)
                moreover
                have "Graph.subpath_from (black prb') (fst rl2) [ba]"
                using subsum_step
                      `Graph.subpath (black prb) bl2 [ba] bl3` `bl2 = src ba` `src ba = fst rl2`
                by (simp add : Graph.subpath_from_def) blast
                
                (* donc bas est bien un sous-chemin rouge-noir du nouveau RB. *)
                ultimately
                show ?thesis 
                using `bas = ui_as ras1 @ bas2 @ [ba]` `ui_as ras1 @ bas2 = ui_as ras`
                      `subpath (red prb') rl ras rl2 (subs prb')`
                unfolding SP_from_def Un_iff Bex_def mem_Collect_eq
                apply (intro disjI2)
                apply (rule_tac ?x="ras"  in exI)
                apply (rule_tac ?x="[ba]" in exI)
                apply (intro conjI)
                  apply simp
                  apply blast
                done
              qed
            qed


          next

            (* Supposons que ras se décompose en un préfixe ras1' maximal dans la nouvelle
               partie rouge, issu de rl et menant en rl2 qui fait partie de la fringe de cette 
               nouvelle partie rouge, et un suffixe noir bas2' issu de rl2. 

               On doit alors distinguer selon que bas2' est vide ou non. La distinction est 
               nécessaire car, si bas2' est vide, le fait qu'il n'existe aucun préfixe rouge 
               non-vide de bas2' est évident et n'implique pas qu'il n'existe aucun préfixe rouge 
               non-vide de bas2'@[ba], ce qui empêche, dans ce cas, de montrer que le préfixe 
               rouge est maximal. *)

            case (goal2 ras1' bas2' rl2)

            thus ?case
            proof (case_tac "bas2' = []")

              (* Supposons que bas2' soit vide. Dans ce cas, (ui_as ras1 @ bas2) est entièrement 
                 représentable dans la nouvelle partie rouge. On conclut selon que ba est 
                 représenté ou non dans la nouvelle partie rouge. *)
              assume "bas2' = []"

              (* alors (ui_as ras1' @ [ba]) est un chemin noir issu de rl et terminant en bl3. *)
              have "Graph.subpath (black prb') (fst rl) (ui_as ras1' @ [ba]) bl3"
              proof -

                have "Graph.subpath (black prb') (fst rl) ((ui_as ras1 @ bas2) @ [ba]) bl3"
                using subsum_step `bas = ui_as ras1 @ bas2 @ [ba]`
                    `Graph.subpath (black prb) (fst rl)  (ui_as ras1) bl1`
                    `Graph.subpath (black prb) bl1 bas2 bl2`
                    `Graph.subpath (black prb) bl2 [ba] bl3`
                by (auto simp add : Graph.sp_append)

                hence "Graph.subpath (black prb') (fst rl) ((ui_as ras1' @ bas2') @ [ba]) bl3"
                using `bas = ui_as ras1 @ bas2 @ [ba]` `ui_as ras1 @ bas2 = ui_as ras1' @ bas2'`
                by simp

                thus ?thesis 
                using `ui_as ras1 @ bas2 = ui_as ras1' @ bas2'` `bas2' = []`
                by simp
              qed

              (* donc : 
                 - (ui_as ras1') est un chemin noir issu de rl et menant en la source de ba
                 - ba est un arc de la partie noire,
                 - la cible de ba est bl3. *)
              hence "Graph.subpath (black prb') (fst rl) (ui_as ras1') (src ba)"
              and  "ba \<in> arcs (black prb')"
              and  "tgt ba = bl3"
              by (simp_all add : sp_append_one)

              (* Montrons que la source de ba est représentée rl2. *)
              have "fst rl2 = src ba"
              proof -

                have "Graph.subpath (black prb') (fst rl) (ui_as ras1') (fst rl2)"
                using `subpath (red prb') rl ras1' rl2 (subs prb')`
                      rsp_is_bsp[OF RB.subsum_step[OF subsum_step(1,3)]]
                by fast

                thus ?thesis 
                using `Graph.subpath (black prb') (fst rl) (ui_as ras1') (src ba)`
                by (simp add : sp_same_src_imp_same_tgt)
              qed

              show ?thesis 
              proof (case_tac "\<exists> ra \<in> out_arcs (red prb') rl2. ba = ui_arc ra")

                (* Supposons que ba est représenté par un arc ra depuis rl2 dans la nouvelle 
                   partie rouge. *)
                assume "\<exists> ra \<in> out_arcs (red prb') rl2. ba = ui_arc ra"

                then obtain ra where "ra \<in> out_arcs (red prb') rl2"
                               and   "ba = ui_arc ra"
                by blast

                (* [ra] est alors un sous-chemin issu de rl2 dans la nouvelle partie rouge. *)
                hence "subpath (red prb') rl2 [ra] (tgt ra) (subs prb')" 
                by (auto simp add : out_arcs_def arc_sp_eq)
                
                (* dans ce cas, ras1'@[ra] est un chemin de la nouvelle partie rouge issu 
                   de rl *)
                hence "subpath (red prb') rl (ras1'@[ra]) (tgt ra) (subs prb')"
                using wf_sub_rel.sp_appendI1[OF subs_wf_sub_rel[OF RB.subsum_step[OF subsum_step(1,3)]]]
                      `subpath (red prb') rl ras1' rl2 (subs prb')`
                by fast

                (* or ras1'@[ra] représente entièrement bas, cqfd. *)
                thus ?thesis 
                using `bas = ui_as ras1 @ bas2 @ [ba]` `ui_as ras1 @ bas2 = ui_as ras1' @ bas2'` 
                      `ba = ui_arc ra` `bas2' = []`
                unfolding SP_from_def subpath_from_def Un_iff image_def Bex_def mem_Collect_eq
                apply (intro disjI1)
                apply (rule_tac ?x="ras1'@[ra]" in exI)
                apply (intro conjI)
                  apply (rule_tac ?x="tgt ra" in exI)
                  apply simp_all
                done

              next {

                (* supposons maintenant que ba ne soit pas représenté depuis rl2 dans la 
                   nouvelle partie rouge. *)

                assume "\<not> (\<exists>ra\<in>out_arcs (red prb') rl2. ba = ui_arc ra)"

                (* alors il n'existe aucun préfixe non-vide de [ba] représentant [ba] 
                   dans la nouvelle partie rouge. Donc ras1' est un préfixe rouge maximal de bas. *)
                have "\<not> (\<exists>ras21 bas22. [ba] = ui_as ras21 @ bas22 
                                     \<and> ras21 \<noteq> [] 
                                     \<and> subpath_from (red prb') rl2 ras21 (subs prb'))"
                proof (intro notI, elim exE conjE)

                  fix ras21 bas22

                  assume "[ba] = ui_as ras21 @ bas22"
                  and    "ras21 \<noteq> []"
                  and    "subpath_from (red prb') rl2 ras21 (subs prb')"

                  then obtain ra where "ras21 = [ra]"
                                 and   "ba = ui_arc ra" 
                  by (auto simp add : neq_Nil_conv)

                  moreover
                  have "ra \<in> out_arcs (red prb') rl2"
                  proof -

                    have "ra \<in> arcs (red prb')"
                    using `subpath_from (red prb') rl2 ras21 (subs prb')` `ras21 = [ra]`
                    by (simp add : subpath_from_def subpath_def)
                    
                    moreover
                    have "src ra = rl2" 
                    proof -
                      have "rl2 = src ra \<or> (rl2,src ra) \<in> subs prb'"
                      using `subpath_from (red prb') rl2 ras21 (subs prb')` `ras21 = [ra]`
                            rb_sp_one[OF RB.subsum_step[OF subsum_step(1,3)]]
                      unfolding subpath_from_def by blast

                      thus ?thesis 
                      using `rl2 \<in> fringe prb'` unfolding fringe_def subsumees_conv by blast
                    qed

                    ultimately
                    show ?thesis by (simp add : out_arcs_def)
                  qed

                  ultimately
                  show False 
                  using `\<not> (\<exists>ra\<in>out_arcs (red prb') rl2. ba = ui_arc ra)` by auto
                qed

                (* de plus, [ba] est bien un suffixe noir de bas issu de rl2. *)
                moreover
                have "Graph.subpath_from (black prb') (fst rl2) [ba]" 
                using `ba \<in> arcs (black prb')` `fst rl2 = src ba`
                unfolding Graph.subpath_from_def using Graph.arc_sp_eq
                by (rule_tac ?x="tgt ba" in exI)  (simp add : Graph.arc_sp_eq)
                
                (* on conclut : le préfixe rouge est ras1', le suffixe noir [ba], rl2 étant 
                   le sommet de la fringe. *)
                ultimately
                show ?thesis 
                using `bas = ui_as ras1 @ bas2 @ [ba]` `ui_as ras1 @ bas2 = ui_as ras1' @ bas2'` 
                       `rl2 \<in> fringe prb'` `subpath (red prb') rl ras1' rl2 (subs prb')`
                       `bas2' = []` 
                unfolding SP_from_def Un_iff Bex_def mem_Collect_eq
                apply (intro disjI2)
                apply (rule_tac ?x="ras1'" in exI)
                apply (rule_tac ?x="[ba]" in exI)
                apply (intro conjI)
                  apply simp
                  apply (rule_tac ?x="rl2" in exI)
                  apply simp
                done
              } qed

            next

              (* supposons que bas2' est non-vide. Dans ce cas, on peut se servir du fait qu'il 
                 n'existe aucun préfixe rouge non-vide de bas2' pour montrer qu'il n'en existe 
                 aucun pour bas2'@[ba]. 

                 On conclut en prenant ras1' comme préfixe rouge et bas2'@[ba] comme suffixe noir,
                 rl2 étant dans la fringe de la nouvelle partie rouge. *)

              assume "bas2' \<noteq> []"

              then obtain bas2'' ba' where "bas2' = ba' # bas2''" 
              using neq_Nil_conv[of bas2'] by blast

              have "\<not> (\<exists>ras21' bas22'. bas2' @ [ba] = ui_as ras21' @ bas22' 
                                      \<and> ras21' \<noteq> [] 
                                      \<and> subpath_from (red prb') rl2 ras21' (subs prb'))"
              proof (intro notI, elim exE conjE)

                fix ras21' bas22'

                assume "bas2' @ [ba] = ui_as ras21' @ bas22'"
                and    "ras21' \<noteq> []" 
                and    "subpath_from (red prb') rl2 ras21' (subs prb')"

                moreover
                then obtain ra' ras21'' where "ras21' = ra' # ras21''" 
                using neq_Nil_conv[of ras21'] by blast

                ultimately
                have "ba' = ui_arc ra'" using `bas2' = ba' # bas2''` by auto

                moreover
                have "subpath_from (red prb') rl2 [ra'] (subs prb')" 
                using `subpath_from (red prb') rl2 ras21' (subs prb')` `ras21' = ra' # ras21''`
                      rb_sp_Cons[OF RB.subsum_step[OF subsum_step(1,3)]]
                      rb_sp_one [OF RB.subsum_step[OF subsum_step(1,3)]]
                unfolding subpath_from_def by blast

                ultimately
                have "\<exists>ras21 bas22. bas2' = ui_as ras21 @ bas22 
                                   \<and> ras21 \<noteq> [] 
                                   \<and> subpath_from (red prb') rl2 ras21 (subs prb')"
                using `bas2' = ba' # bas2''`
                apply (rule_tac ?x="[ra']"  in exI)
                apply (rule_tac ?x="bas2''" in exI)
                apply simp
                done

                thus False using goal2 by (elim notE)
              qed

              moreover
              have "Graph.subpath_from (black prb') (fst rl2) (bas2' @ [ba])"
              proof -

                have "Graph.subpath (black prb') (fst rl) ((ui_as ras1 @ bas2) @ [ba]) bl3"
                using subsum_step `bas = ui_as ras1 @ bas2 @ [ba]`
                      `Graph.subpath (black prb) (fst rl)  (ui_as ras1) bl1`
                      `Graph.subpath (black prb) bl1 bas2 bl2`
                      `Graph.subpath (black prb) bl2 [ba] bl3`
                by (auto simp add : Graph.sp_append)

                hence "Graph.subpath (black prb') (fst rl) ((ui_as ras1' @ bas2') @ [ba]) bl3"
                using `bas = ui_as ras1 @ bas2 @ [ba]` `ui_as ras1 @ bas2 = ui_as ras1' @ bas2'`
                by simp
                
                then obtain bl2' where "Graph.subpath (black prb') (fst rl) (ui_as ras1') bl2'"
                                 and   "Graph.subpath (black prb') bl2' (bas2'@[ba]) bl3"
                by (simp only : Graph.sp_append) blast
                
                moreover
                have "Graph.subpath (black prb') (fst rl) (ui_as ras1') (fst rl2)" 
                using `subpath (red prb') rl ras1' rl2 (subs prb')`
                      rsp_is_bsp[OF RB.subsum_step[OF subsum_step(1,3)]]
                by fast

                ultimately
                have "bl2' = fst rl2" by (simp add : sp_same_src_imp_same_tgt)

                thus ?thesis 
                using `Graph.subpath (black prb') bl2' (bas2'@[ba]) bl3`
                by (simp add : Graph.subpath_from_def) blast
              qed
              
              ultimately
              show ?thesis 
              using `bas = ui_as ras1 @ bas2 @ [ba]` goal2 
              unfolding SP_from_def Un_iff Bex_def mem_Collect_eq
              apply (intro disjI2)
              apply (rule_tac ?x="ras1'" in exI)
              apply (rule_tac ?x="bas2'@[ba]" in exI)
              apply (intro conjI)
                apply simp
                apply (rule_tac ?x="rl2" in exI) 
                apply simp
              done
            qed
          qed
        qed

      next

        assume "rl1 \<noteq> subsumee sub"

        (* dans ce cas rl1 egalement dans la fringe de prb' *)
        hence  "rl1 \<in> fringe prb'"
        using  `rl1 \<in> fringe prb` subsumE_fringe[OF subsum_step(3)] by simp
               
        (* montrons de plus que, dans ce cas, bas2 ne possede pas non plus de 
           prefixe rouge non-vide dans prb' (sinon il aurait egalement possede 
           un tel prefixe dans prb), rendant ainsi ras1 maximal dans la partie 
           rouge de prb'. *)
        moreover
        have "\<not> (\<exists> bas21 bas22. bas2 = ui_as bas21 @ bas22 
                        \<and> bas21 \<noteq> [] 
                        \<and> subpath_from (red prb') rl1 bas21 (subs prb'))"
        proof (intro notI, elim exE conjE)

          (* supposons que bas2 possede un prefixe rouge non-vide ras21 dans  prb' *)
          fix ras21 bas22

          assume "bas2 = ui_as ras21 @ bas22"
          and    "ras21 \<noteq> []"
          and    "subpath_from (red prb') rl1 ras21 (subs prb')"

          (* on appelle ra le premier arc (rouge) de ras21 *)
          then obtain ra bas21' rl2 where "ras21 = ra # bas21'"
                                    and   "subpath (red prb') rl1 ras21 rl2 (subs prb')"
          using neq_Nil_conv[of ras21] unfolding subpath_from_def by blast

          (* montrons que cet arc constitue a lui seul un sous-chemin de (la partie rouge de) 
             prb menant de rl1 à sa cible. *)
          hence "subpath_from (red prb) rl1 [ra] (subs prb)"
          proof -

            (* montrons que rl1 est la source de ra *)
            have "rl1 = src ra"
            proof -

              (* puisque bas2 est un sous-chemin rouge issu de rl1 et commençant par ra,
                 on a rl1 = src ra ou bien il existe une subsomption dans prb' 
                 liant rl1 a src ra *)

              have "rl1 = src ra \<or> (rl1, src ra) \<in> (subs prb')"
              using `ras21 = ra # bas21'`
                    `subpath (red prb') rl1 ras21 rl2 (subs prb')`
                    rb_sp_Cons[OF RB.subsum_step[OF subsum_step(1,3)]]
              by simp
              
              thus ?thesis 
              proof (elim disjE)
                (* dans le premier cas, on a bien rl1 = src ra *)
                case goal1 thus ?thesis by simp
              next

               (* dans ce cas, rl1 est directement lie à src ra, car il ne peut exister 
                  de chaine de subsomption dans un RB, or on sait que rl1 n'est pas subsume 
                  dans prb : contradiction *)

                case goal2
                
                hence "(rl1,src ra) \<in> (subs prb)"
                using subsum_step `rl1 \<noteq> subsumee sub`
                      SubExt.extendsE2[of "red prb" "subs prb" sub subrel]
                by auto

                thus ?thesis using 2 unfolding fringe_def subsumees_def by force
              qed
            qed

            (* de plus, ra est un arc rouge (puisqu'il fait partie d'un sous-chemin rouge
               de prb', il est un arc rouge de prb', qui possede les memes arcs rouges 
               que prb) *)
            moreover
            have "ra \<in> arcs (red prb)" 
            using subsum_step
                  `ras21 = ra # bas21'`
                  `subpath (red prb') rl1 ras21 rl2 (subs prb')`
                  subs_sub_rel_of[OF RB.subsum_step[OF subsum_step(1,3)]]
            by (auto simp add : sub_rel_of.sp_Cons)

            (* donc ra est bien un sous-chemin de prb menant de rl1 a sa cible *)
            ultimately
            show ?thesis
            using arc_sp_eq[of "red prb" ra "subs prb"]
            unfolding subpath_from_def by blast
          qed

          (* donc bas2 possede un prefixe rouge dans prb : [ra] *)
          hence "\<exists> bas21 bas22. bas2 = ui_as bas21 @ bas22 
                              \<and> bas21 \<noteq> [] 
                              \<and> subpath_from (red prb) rl1 bas21 (subs prb)"
          apply (rule_tac ?x="[ra]" in exI)
          apply (rule_tac ?x="ui_as bas21' @ bas22" in exI)
          apply (intro conjI)
            apply (simp add : `bas2 = ui_as ras21 @ bas22` `ras21 = ra # bas21'`)
            apply simp
            apply assumption
          done
        
          (* ce qui contredit le fait que ras1 soit maximal *)
          thus False using 4 by (elim notE)
        qed

        (* de plus, bas2 est bien un sous-chemin issu de (fst rl1) dans la partie noire 
           de prb', puisqu'il l'est dans celle de prb et que ces deux parties noires sont 
           en realite identiques *)
        moreover
        have "Graph.subpath_from (black prb') (fst rl1) bas2" 
        using subsum_step `Graph.subpath_from (black prb) (fst rl1) bas2` by simp

        (* donc bas est bien un chemin rouge/noir de prb' *)
        ultimately
        show  "bas \<in> SP_from prb' rl"
        using subsum_step `bas = ui_as ras1 @ bas2` `subpath (red prb') rl ras1 rl1 (subs prb')`
        unfolding SP_from_def Un_iff Bex_def mem_Collect_eq
        apply (intro disjI2)
        apply (rule_tac ?x="ras1" in exI)
        apply (rule_tac ?x="bas2" in exI)
        apply (intro conjI)
          apply simp
          apply (rule_tac ?x="rl1" in exI)
          apply simp
        done

      qed
    qed
  qed

next

  case (abstract_step prb rl e prb' rl')

  show ?case
  unfolding subset_iff
  proof (intro allI impI)

    fix bas

    assume "bas \<in> feasible_subpaths_from (black prb') (confs prb' rl') (fst rl')"

    show "bas \<in> SP_from prb' rl'"
    proof (case_tac "rl' = rl")

      assume "rl' = rl"

      show ?thesis 
      proof (case_tac "bas = []")

        assume "bas = []"

        moreover
        have  "subpath_from (red prb') rl' [] (subs prb')" 
        using abstract_step unfolding subpath_from_def Nil_sp by blast

        ultimately
        show ?thesis by (simp only : SP_from_def) blast

      next

        assume "bas \<noteq> []"

        then obtain ba bas' where "bas = ba # bas'" 
        using `bas \<noteq> []` by (simp add : neq_Nil_conv) blast

        show ?thesis 
        unfolding SP_from_def Un_iff mem_Collect_eq Bex_def
        apply (intro disjI2)
        apply (rule_tac ?x="[]" in exI)
        apply (rule_tac ?x="bas" in exI)
        proof (intro conjI)
          
          show "bas = ui_as [] @ bas" by simp

        next

          case goal2
          show ?case
          proof (rule_tac ?x="rl'" in exI, intro conjI)

            have "\<exists> ba. free_from prb' rl' ba"
            proof -

              have "ba \<in> out_arcs (black prb') (fst rl')" 
              using `bas \<in> feasible_subpaths_from (black prb') (confs prb' rl') (fst rl')`
                    `bas = ba # bas'`
              unfolding feasible_subpath_from_def feasible_subpath_def
              by (auto simp add : Graph.sp_Cons out_arcs_def)

              moreover  
              have "\<forall> ra \<in> out_arcs (red prb') rl'. ui_arc ra \<noteq> ba" 
              using abstract_step `rl' = rl` by auto

              moreover
              have "\<exists> pc. se (confs prb' rl') (labelling (black prb') ba) pc \<and> sat pc" 
              proof -

                obtain pc1 pc2 
                where "se (confs prb' rl') (labelling (black prb') ba) pc1"
                and   "se_star pc1 (trace bas' (labelling (black prb'))) pc2"
                and   "sat pc2"
                using `bas \<in> feasible_subpaths_from (black prb') (confs prb' rl') (fst rl')`
                      `bas = ba # bas'`
                unfolding feasible_subpath_from_def feasible_subpath_def feasible_def
                by (simp add : se_star_Cons) blast

                moreover
                hence "sat pc1" by (simp add : se_star_sat_imp_sat)

                ultimately
                show ?thesis by blast
              qed

              ultimately
              show ?thesis by (simp add : free_from_def) blast
            qed

            thus "rl' \<in> fringe prb'" 
            using abstract_step `rl' = rl`
                  finite_labels_RB.abstractE_fringe1[of prb rl prb' e]
                  finite_labels_RB.abstractE_fringe2[of prb rl prb' e]  
            unfolding finite_labels_RB_def by auto

          next

            show "subpath (red prb') rl' [] rl' (subs prb')" 
            using abstract_step by (simp add : Nil_sp)

          next

            case goal3 
            show ?case 
            proof (intro notI, elim exE conjE)

              fix ras21 bas22

              assume "bas = ui_as ras21 @ bas22"
              and    "ras21 \<noteq> []"
              and    "subpath_from (red prb') rl' ras21 (subs prb')"

              moreover
              then obtain ra ras21' where "ras21 = ra # ras21'"
              by (simp add : neq_Nil_conv) blast

              ultimately
              have "ui_arc ra = ba" using `bas = ba # bas'` by simp

              have "out_arcs (red prb') rl' \<noteq> {}"
              proof -

                have "rl' = src ra \<or> (rl', src ra) \<in> subs prb'" 
                and  "ra \<in> arcs (red prb')"
                using abstract_step
                      `subpath_from (red prb') rl' ras21 (subs prb')` `ras21 = ra # ras21'`
                      rb_sp_Cons[OF RB.abstract_step[OF abstract_step(1,3)]]
                by (auto simp add : subpath_from_def)

                moreover
                have "rl' \<notin> subsumees (subs prb')" using abstract_step `rl' = rl` by simp

                ultimately
                show ?thesis by (force simp add : out_arcs_def subsumees_def)
              qed

              thus False using abstract_step `rl' = rl` by auto

            qed

          next
            
            show "Graph.subpath_from (black prb') (fst rl') bas"
            using `bas \<in> feasible_subpaths_from (black prb') (confs prb' rl') (fst rl')`
            unfolding feasible_subpath_from_def feasible_subpath_def Graph.subpath_from_def
            by blast

          qed
        qed
      qed

    next
      
      assume "rl' \<noteq> rl" 

      hence  "bas \<in> SP_from prb rl'" 
      using abstract_step  `bas \<in> feasible_subpaths_from (black prb') (confs prb' rl') (fst rl')`
      by fastforce

      thus ?thesis 
      apply (subst (asm) SP_from_def)
      apply (simp only : Un_iff image_def Bex_def mem_Collect_eq)
      proof (elim disjE exE conjE)

        fix ras

        assume "subpath_from (red prb) rl' ras (subs prb)"
        and    "bas = ui_as ras"

        thus ?thesis using abstract_step by (simp add : SP_from_def)

      next
          
        case (goal2 ras1 bas2 rl'')
 
        moreover
        hence "rl'' \<in> fringe prb'" 
        using abstract_step 
              finite_labels_RB.abstractE_fringe1[of prb rl prb' e]
              finite_labels_RB.abstractE_fringe2[of prb rl prb' e]
              finite_labels_RB.abstractE_fringe3[of prb rl prb' e]
        unfolding finite_labels_RB_def by auto

        ultimately
        show ?thesis using abstract_step by (simp add : SP_from_def) blast
      qed
    qed
  qed
qed






end
